/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <string.h>
#include "udm_common.h"
#include "udm_utils.h"
#include "udm_searchtool.h"
#include "udm_coords.h"

/* Check if x is a valid body byte */
#define IS_BODY(x) ((((int) (unsigned char) (x)) ^ 0x80) < 0x40)

/*
  Scan one coord.
  Return the number of bytes scanned.
*/
size_t
udm_coord_get(size_t *pwc, const unsigned char *s, const unsigned char *e)
{
  unsigned char c;
  
  if (s >= e)
    return 0;
  
  c= s[0];
  if (c < 0x80) 
  {
    *pwc = c;
    return 1;
  } 
  else if (c < 0xc2)
  {
    *pwc= 0;
    return 1; /* Caller will skip one byte */
  }
  else if (c < 0xe0) 
  {
    if (s+2 > e) /* We need 2 characters */ 
      return 0;
    
    if (!((s[1] ^ 0x80) < 0x40))
      return 0;
    
    *pwc = ((size_t) (c & 0x1f) << 6) | (size_t) (s[1] ^ 0x80);
    return 2;
  } 
  else if (c < 0xf0) 
  {
    if (s+3 > e) /* We need 3 characters */
      return 0;
    
    if (!((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40 && (c >= 0xe1 || s[1] >= 0xa0)))
      return 0;
    
    *pwc = ((size_t) (c & 0x0f) << 12)   | 
           ((size_t) (s[1] ^ 0x80) << 6) | 
            (size_t) (s[2] ^ 0x80);
    
    return 3;
  } 
  else if (c < 0xf8)
  {
    if (s + 4 > e) /* Need 4 bytes */
      return 0;
    if (!((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40 && (s[3] ^ 0x80) < 0x40 && (c >= 0xf1 || s[1] >= 0x90)))
      return 0;
    
    *pwc= ((size_t) (c & 0x07) << 18)    |
          ((size_t) (s[1] ^ 0x80) << 12) |
          ((size_t) (s[2] ^ 0x80) << 6)  |
          (size_t)  (s[3] ^ 0x80);
    return 4;
  }
  else if (c < 0xfc)
  {
    if (s + 5 > e) /* We need 5 characters */
      return 0;

    if (!IS_BODY(s[1]) || !IS_BODY(s[2]) || !IS_BODY(s[3]) || !IS_BODY(s[4]) ||
        !(c >= 0xf9 || s[1] >= 0x88))
      return 0;
    *pwc= ((size_t) (c & 0x03) << 24) |
           ((size_t) (s[1] ^ 0x80) << 18) |
           ((size_t) (s[2] ^ 0x80) << 12) |
           ((size_t) (s[3] ^ 0x80) << 6) |
           ((size_t) (s[4] ^ 0x80));
    return 5;
  }
  else if (c < 0xfe)
  {
    if (s + 6 > e) /* We need 6 characters */
      return 0;

    if (!IS_BODY(s[1]) || !IS_BODY(s[2]) || !IS_BODY(s[3]) || !IS_BODY(s[4]) ||
        !IS_BODY(s[5]) || !(c >= 0xfd || s[1] >= 0x84))
      return 0;

    *pwc=   ((size_t) (c & 0x01) << 30)
          | ((size_t) (s[1] ^ 0x80) << 24)
          | ((size_t) (s[2] ^ 0x80) << 18)
          | ((size_t) (s[3] ^ 0x80) << 12)
          | ((size_t) (s[4] ^ 0x80) << 6)
          | ((size_t) (s[5] ^ 0x80));
    return 6;
  }
  return 0;
}



/* Inlining make "gcc -O3" produce a warning */
size_t
udm_coord_put(size_t wc, unsigned char *r, unsigned char *e)
{
  int count;
  
  if (r >= e)
    return 0;
  
  if (wc < 0x80) 
    count= 1;
  else if (wc < 0x800) 
    count= 2;
  else if (wc < 0x10000) 
    count= 3;
  else if (wc < 0x200000)
    count= 4;
  else if (wc < 0x4000000)
    count= 5;
  else if (wc <= 0x7fffffff)
    count= 6;
  else return 0;
  
  /* 
    e is a character after the string r, not the last character of it.
    Because of it (r+count > e), not (r+count-1 >e )
   */
  if (r + count > e) 
    return 0;
  
  switch (count)
  { 
    /* Fall through all cases */
    case 6: r[5]= (unsigned char) (0x80 | (wc & 0x3f)); wc = wc >> 6; wc |= 0x4000000;
    case 5: r[4]= (unsigned char) (0x80 | (wc & 0x3f)); wc = wc >> 6; wc |= 0x200000;
    case 4: r[3]= (unsigned char) (0x80 | (wc & 0x3f)); wc = wc >> 6; wc |= 0x10000;
    case 3: r[2]= (unsigned char) (0x80 | (wc & 0x3f)); wc = wc >> 6; wc |= 0x800;
    case 2: r[1]= (unsigned char) (0x80 | (wc & 0x3f)); wc = wc >> 6; wc |= 0xc0;
    case 1: r[0]= (unsigned char) wc;
  }
  return count;
}



/*
 Position and length encoding components.
 Based on UTF-8 encoding:
  
 1. 0-7F         [00-7F]                      0xxxxxxx
 2. 80-7FF       [C2-DF][80-BF]               110yyyxx 10xxxxxx
 3. 8000-FFFF    [E0-EF][80-BF][80-BF]        1110yyyy 10yyyyxx 10xxxxxx
 4. 10000-10FFFF [F0-F4][80-BF][80-BF][80-BF] 11110zzz 10zzyyyy 10yyyyxx 10xxxxxx

Unused overlong sequences:
2. [C0-C1][?]                   1100000x 10xxxxxx
3. [E0-E0][80-9F][80-BF]        11100000 100xxxxx 10xxxxxx
4. 11110000 1000xxxx 10xxxxxx 10xxxxxx
5. 11111000 10000xxx 10xxxxxx 10xxxxxx 10xxxxxx
6. 11111100 100000xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx


Used bytes:
 [00-7F] 0xxxxxxx Single byte sequence
 [80-BF] 10xxxxxx Second, third and fourth byte in a multi-byte sequence
 [C2-DF] 110xxxxx Start of a two-byte sequence
 [E0-EF] 1110xxxx Start of a three-byte sequence
 [F0-F4] 11110xxx Start of a four-byte sequence
         11110000
         11110001
         11110010
         11110011
         11110100
 
Unused bytes:
 [C0-C1] Overlong encoding: start of a 2-byte sequence,    codepoint <= 127
 [F5-F7] 11110xxx Restricted by RFC 3629: start of 4-byte sequence, codepoint > 10FFFF
         11110101
         11110110
         11110111
         
 [F8-FB] 111110xx Restricted by RFC 3629: start of 5-byte sequence
 [FC-FD] 1111110x Restricted by RFC 3629: start of 6-byte sequence
 [FE-FF] 1111111x Invalid


Example statistics for DBMode=blob.
==================================
Collection: Apache manual.
Documents: 1253     = select count(*) from bdicti;
Index size: 6158768 =  select sum(length(coords)) from bdict where word not like '##%';
- URL_ID size: 31% = 1970652 = 492663 times x 4bytes-
- Coord size:  69% = 4188116 bytes
  492082 coords/lengths use 1 byte.
  581 coords/lengths use 2 bytes.

Top 16 words:
mysql> select word, secno, length(coords) from bdict where word not like '##%' order by 3 desc limit 16;
+--------+-------+----------------+
| word   | secno | length(coords) |
+--------+-------+----------------+
| the    |     1 |        121499  | 
| to     |     1 |         57876  | 
| a      |     1 |         46039  | 
| of     |     1 |         41643  | 
| is     |     1 |         40950  | 
| and    |     1 |         36484  | 
| in     |     1 |         35399  | 
| for    |     1 |         33173  | 
| this   |     1 |         31080  | 
| server |     1 |         30422  | 
| be     |     1 |         29393  | 
| apache |     1 |         29022  | 
| that   |     1 |         24360  | 
| it     |     1 |         23114  | 
| on     |     1 |         22491  | 
| are    |     1 |         22449  | 
+--------+-------+----------------+
16 rows in set (0.02 sec)

16   top words use 10% disk space, 17-96 bytes for word coords per document.
51   top words use 20% disk space, 10-96
110  top words use 30% disk space, 7-96
202  top words use 40% disk space, 4-96
350  top words use 50% disk space, 2.7-96
574  top words use 60% disk space
937  top words use 70% disk space
1593 top words use 80% disk space
3081 top words use 90% disk space

The same collection with Deflate=yes
====================================
Index size: 2680858

Top 16 words:
+--------+-------+----------------+
| word   | secno | length(coords) |
+--------+-------+----------------+
| the    |     1 |          25949 | 
| to     |     1 |          13272 | 
| a      |     1 |          10341 | 
| of     |     1 |           9938 | 
| is     |     1 |           9835 | 
| and    |     1 |           9323 | 
| in     |     1 |           9087 | 
| for    |     1 |           8709 | 
| this   |     1 |           8486 | 
| server |     1 |           8048 | 
| be     |     1 |           8024 | 
| apache |     1 |           7672 | 
| that   |     1 |           7019 | 
| it     |     1 |           6816 | 
| by     |     1 |           6760 | 
| are    |     1 |           6692 | 
+--------+-------+----------------+

35 top words use 10% disk space
97 top words use 20% disk space

The same collection, 1 byte per URL_ID (delta)
4680779

The same collation, 1 byte per URL_ID (delta), Deflate=yes
2289516

*/


/*
  A faster version, without range checking
  The incoming string must have enough space
  to scan full miltibyte sequence.
*/
size_t
udm_coord_get_quick(size_t *pwc, const unsigned char *s)
{
  unsigned char c;
  
  c= s[0];
  if (c < 0x80) 
  {
    *pwc = c;
    return 1;
  } 
  else if (c < 0xc2) 
  {
    *pwc= 0;
    return 1; /* The caller will skip one byte */
  }
  else if (c < 0xe0) 
  {
    size_t s1;
    if (!((s1= (s[1] ^ 0x80)) < 0x40))
      return 0;
    *pwc= (((size_t) (c & 0x1f)) << 6) | s1;
    return 2;
  } 
  else if (c < 0xf0) 
  {
    if (!((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40 && (c >= 0xe1 || s[1] >= 0xa0)))
      return 0;
    
    *pwc = ((size_t) (c & 0x0f) << 12)   | 
           ((size_t) (s[1] ^ 0x80) << 6) | 
            (size_t) (s[2] ^ 0x80);
    
    return 3;
  } 
  else if (c < 0xf8)
  {
    if (!((s[1] ^ 0x80) < 0x40 && (s[2] ^ 0x80) < 0x40 && (s[3] ^ 0x80) < 0x40 && (c >= 0xf1 || s[1] >= 0x90)))
      return 0;
    
    *pwc= ((size_t) (c & 0x07) << 18)    |
          ((size_t) (s[1] ^ 0x80) << 12) |
          ((size_t) (s[2] ^ 0x80) << 6)  |
          (size_t)  (s[3] ^ 0x80);
    return 4;
  }

  return 0;
}


/*
  Skip the given amount of coords, accumulating sum.
*/

const unsigned char *
udm_coord_sum(size_t *sump,
              const unsigned char *s,
              const unsigned char *e,
              size_t n)
{
  size_t sum= 0, nbytes, tmp;
  
  /* String must have at least "n" bytes */
  if (s + (n * 4) > e)
    goto mb;
  
  /* Quickly sum leading 7bit numbers */
  for (; *s < 128 && n; n--)
    sum+= *s++;

  for ( ; n; n--)
  {
    if (!(nbytes= udm_coord_get_quick(&tmp, s)))
    {
      *sump= sum;
      return e;
    }
    s+= nbytes;
    sum+= tmp;
  }
  *sump= sum;
  return s;
  
mb:
  for ( ; n; n--)
  {
    if (!(nbytes= udm_coord_get(&tmp, s, e)))
    {
      *sump= sum;
      return e;
    }
    s+= nbytes;
    sum+= tmp;
  }
  *sump= sum;
  return s;
}


/*
  Skip the given anount of coords in the given string
  s - beginning of the string
  e - end of the string (last byte + 1)
  n - how many characters to skip
  RETURNS
    position after n characters, or end of the string (if shorter)
*/
const unsigned char *
udm_coord_skip(const unsigned char *s, const unsigned char *e, size_t n)
{
#if defined(__i386__) && !defined(_WIN64)
  {
    const unsigned char *e4= e - 4;
    for ( ; n > 4 && s < e4; n-= 4, s+= 4)
    {
      if (*((const uint4*)s) & 0x80808080)
        break;
    }
  }
#endif

  for ( ; n && s < e; n--)
  {
    if (*s++ < 0x80)
    {
      /* Single byte character, nothing to do */
    }
    else
    {
      unsigned char c= s[-1];
      if (c < 0xc2) /* Error */
        return e;
      else if (c < 0xe0) /* Two-byte sequence */
      {
        s++;
      }
      else if (c < 0xf0) /* Three-byte sequence */
      {
        s+= 2;
      }
      else if (c < 0xf8) /* Four-byte sequence */
      {
        s+= 3;
      }
      else
      {
        /* TODO: longer sequences */
        return e;
      }
      if (s > e)
        return e;
    }
  }
  return s;
}



/*
  Note, it can acutally be up to 6 in some cases (e.g. in poprank records),
  but not in UdmPackedCoordsToUnpackedCoords context.
*/
#define COORD_MAX_LEN 4



/*
  New VS Old statistics. Collection km.ru. Query q=1+2+3
  
  New:
    UdmPackedCoordsToUnpackedCoords called 547748 times
  
  Old:
    UdmPackedCoordsToUnpackedCoords called 2834091 times
*/


/*
  Unpack coordinates.
  s     - start of packed string
  e     - end of packed string
  nrecs - how many coordinates to unpack
  C     - pointer to initial coordinate, increment positions relatively to it.
  Coord - Where to out unpacked coordinates. Must have enough space allocated.
  **end - where unpacking stopped
  
  Returns - pointer after the last coordinate unpacked.
*/
UDM_COORD2 *
UdmPackedCoordsToUnpackedCoords(const unsigned char *s,
                                const unsigned char *e,
                                size_t nrecs,
                                UDM_COORD2 *C,
                                UDM_COORD2 *Coord,
                                const unsigned char **end)
{
  if (s + nrecs * COORD_MAX_LEN < e)
  {
    /* Not the last chunk: unpack without range check */
    for ( ; nrecs > 0 ; nrecs--)
    {
      if (*s < 128)
      {
        C->pos+= *s++;
      }
      else
      {
        size_t crd, nbytes;
        nbytes= udm_coord_get_quick(&crd, s);
        if (!nbytes) break;
        s+= nbytes;
        C->pos+= crd;
      }
      *Coord++= *C;
    }
  }
  else
  {
    /* Possibly last chunk: unpack with range check */
    for ( ; nrecs > 0 ; nrecs--)
    {
      if (s < e && *s < 128)
      {
        C->pos+= *s++;
      }
      else
      {
        size_t crd, nbytes;
        nbytes= udm_coord_get(&crd, s, e);
        if (!nbytes) break;
        s+= nbytes;
        C->pos+= crd;
      }
      *Coord++= *C;
    }
  }
  *end= s;
  return Coord;
}

udm_rc_t
UdmAddOneCoord(UDM_URLCRDLIST *CoordList, const UDM_URL_CRD *Coord)
{
  if (CoordList->ncoords == CoordList->acoords)
  {
    UDM_URL_CRD *tmp;
    size_t newsize= CoordList->acoords ? CoordList->acoords * 2 : 1024;
    if (!(tmp= (UDM_URL_CRD*) UdmRealloc(CoordList->Coords, newsize * sizeof(UDM_URL_CRD))))
      return UDM_ERROR;
    CoordList->Coords= tmp;
    CoordList->acoords= newsize;
  }
  CoordList->Coords[CoordList->ncoords]= *Coord;
  CoordList->ncoords++;
  return UDM_OK;
}


udm_rc_t
UdmCoordListMultiUnpack(UDM_URLCRDLIST *CoordList,
                             UDM_URL_CRD *CoordTemplate,
                             const unsigned char *intag, size_t lintag,
                             udm_bool_t save_section_size)
{
  const unsigned char *s, *e;
  size_t last= 0;
  size_t ncoords0= CoordList->ncoords;
  
  for (s= intag, e= intag + lintag; e > s; )
  {
    UDM_URL_CRD *Crd= &CoordList->Coords[CoordList->ncoords];
    size_t crd;
    size_t nbytes= udm_coord_get(&crd, s, e);
    if (!nbytes) break;
    last+= crd;
    Crd->urlid_coord.url_id= CoordTemplate->urlid_coord.url_id;
    Crd->urlid_coord.coord.pos= last;
    Crd->urlid_coord.coord.secno= CoordTemplate->urlid_coord.coord.secno;
    Crd->num= CoordTemplate->num;
    Crd->seclen= 0;
    CoordList->ncoords++;
    s+= nbytes;
  }
  
  if (save_section_size)
  {
    uint4 seclen= CoordList->Coords[--CoordList->ncoords].urlid_coord.coord.pos;
    for ( ; ncoords0 < CoordList->ncoords; ncoords0++)
      CoordList->Coords[ncoords0].seclen= seclen;
  }
  return UDM_OK;
}


static int
cmpurlid(UDM_URL_CRD *s1, UDM_URL_CRD *s2)
{
  if (s1->urlid_coord.url_id > s2->urlid_coord.url_id) return(1);
  if (s1->urlid_coord.url_id < s2->urlid_coord.url_id) return(-1);
  if (s1->urlid_coord.coord.secno > s2->urlid_coord.coord.secno) return 1;
  if (s1->urlid_coord.coord.secno < s2->urlid_coord.coord.secno) return -1;
  return (int) s1->urlid_coord.coord.pos - (int) s2->urlid_coord.coord.pos;
}


void
UdmURLCRDListSortByURLThenSecnoThenPos(UDM_URLCRDLIST *L)
{
  if(L->Coords && L->ncoords)
    UdmSort((void*)L->Coords, L->ncoords,
            sizeof(*L->Coords),(udm_qsort_cmp)cmpurlid);
}


static void
UdmURLCRDListMergeMultiWord(UDM_URLCRDLIST *Phrase, size_t wordnum, size_t nparts)
{
  UDM_URL_CRD *To= Phrase->Coords;
  UDM_URL_CRD *End= Phrase->Coords + Phrase->ncoords;
  UDM_URL_CRD *From= Phrase->Coords + nparts - 1;
  UDM_URL_CRD *Prev= Phrase->Coords + nparts - 2;
  
#if 0
  fprintf(stderr, "merge: wordnum=%d nparts=%d ncoords=%d\n",
          wordnum, nparts, Phrase->ncoords);
#endif
  
  if (nparts < 2) /* If one part, keep Phrase unchanged */
    return;
  
  if (Phrase->ncoords < nparts) /* Nothing found */
  {
    Phrase->ncoords= 0;
    return;
  }
  
  for ( ; From < End ; From++, Prev++)
  {
    if (Prev->urlid_coord.url_id == From->urlid_coord.url_id)
    {
      UDM_COORD coord= From->urlid_coord.coord;
      size_t num= From->num;
      if (coord.pos == Prev->urlid_coord.coord.pos + 1 &&
          coord.secno == Prev->urlid_coord.coord.secno  &&
          num == Prev->num + 1)
      {
        size_t i, nmatches;
        for (nmatches= 2, i= 2; i < nparts; i++)
        {
          if (From[-i].urlid_coord.url_id != From->urlid_coord.url_id         ||
              From[-i].urlid_coord.coord.secno != coord.secno     ||
              From[-i].urlid_coord.coord.pos != (coord.pos - i)   ||
              From[-i].num != (num - i))
            break;
            
          nmatches++;
        }
        if (nmatches == nparts)
        {
          To->urlid_coord.url_id= From->urlid_coord.url_id;
          To->urlid_coord.coord.pos= coord.pos - nparts + 1;
          To->urlid_coord.coord.secno= coord.secno;
          To->num= wordnum;
          To++;
        }
      }
    }
  }
  
  Phrase->ncoords= To - Phrase->Coords;
}


/*
  Convert URLCRDList to SectionList.
  Only single word is supported.
  Crd->num must be the same for all coords.
*/
static udm_rc_t
UdmURLCRDListToSearchSectionList(UDM_SEARCHSECTIONLIST *SectionList,
                                 UDM_URLCRDLIST *CoordList,
                                 udm_wordnum_t wordnum,
                                 udm_wordnum_t order)
{
  size_t ncoords= CoordList->ncoords;
  UDM_URL_CRD *CrdFrom, *CrdCurr;
  UDM_URL_CRD *CrdLast= CoordList->Coords + ncoords;
  UDM_COORD2 *Coord;
  UDM_SEARCHSECTION *Section;
  
  UdmSearchSectionListAlloc(SectionList, CoordList->ncoords, CoordList->ncoords);

#if 0
  UdmCoordListPrint(CoordList);
#endif
  
  if (!CoordList->ncoords)
    return UDM_OK;
  Coord= SectionList->Coord;
  Section= SectionList->Section;
  
  for (CrdFrom= CoordList->Coords; CrdFrom < CrdLast; CrdFrom= CrdCurr)
  {
    Section->PackedCoord= NULL; /* Not packed */
    Section->Coord= Coord;
    Section->secno= CrdFrom->urlid_coord.coord.secno;
    Section->wordnum= CrdFrom->num;
    UDM_ASSERT(CrdFrom->num == wordnum);
    Section->order= order; /* args->WWList->Word[CrdFrom->num].order; */
    for (CrdCurr= CrdFrom;
         CrdCurr < CrdLast &&
         CrdCurr->urlid_coord.url_id == CrdFrom->urlid_coord.url_id &&
         CrdCurr->urlid_coord.coord.secno == CrdFrom->urlid_coord.coord.secno ;
         CrdCurr++)
    {
      Coord->pos= CrdCurr->urlid_coord.coord.pos;
      Coord->order= Section->order;
      Section->maxpos= CrdCurr->urlid_coord.coord.pos;
      Coord++;
    }
    Section->url_id= CrdFrom->urlid_coord.url_id;
    Section->ncoords= CrdCurr - CrdFrom;
    Section->seclen= CrdFrom->seclen;
    Section->minpos= Section->Coord->pos;
    Section++;
  }
  SectionList->ncoords= Coord - SectionList->Coord;
  SectionList->nsections= Section - SectionList->Section;
  
#if 0
  UdmSectionListPrint(SectionList);
#endif
  
  return UDM_OK;
}


/*
  When we're searching for a multiword part,
  it is possible that
    wordnum >= args->WWList->nwords
  In this case "order" is not important.
  It will be later set to the correct value
  in UdmMultiWordAdd().
  
  In other cases "wordnum < args->WWList->nwords" should be true.
  TODO: add an assert about that.
*/
udm_rc_t
UdmURLCRDListListAddWithSort2(UDM_SEARCHSECTIONLISTLIST *SectionListList,
                              UDM_WIDEWORDLIST *WWList,
                              UDM_WIDEWORD *Word,
                              UDM_URLCRDLIST *CoordList)
{
  UDM_SEARCHSECTIONLIST SectionList;
  udm_wordnum_t wordnum= Word->Param.order;
  udm_wordnum_t order= wordnum < WWList->nwords ?
                       WWList->Word[wordnum].Param.order :
                       0 /* possible when in a multiword */;
  UdmURLCRDListToSearchSectionList(&SectionList, CoordList, wordnum, order);
  UdmSearchSectionListListAdd(SectionListList, &SectionList);
  UDM_FREE(CoordList->Coords);
  return UDM_OK;
}



/************** multi word routines **************************/

static udm_rc_t
UdmSectionListListToURLCRDList(UDM_URLCRDLIST *List,
                               UDM_SEARCHSECTIONLISTLIST *SrcList)
{
  size_t list, ncoords= 0, nsections= 0;
  UDM_URL_CRD *Crd;
  for (list=0; list < SrcList->nitems; list++)
  {
    /*UdmSectionListPrint(&SrcList->Item[list]);*/
    ncoords+= SrcList->Item[list].ncoords;
    nsections+= SrcList->Item[list].nsections;
  }
  bzero((void*) List, sizeof(*List));
  if (ncoords &&
      !(List->Coords= (UDM_URL_CRD*) UdmMalloc(ncoords * sizeof(UDM_URL_CRD))))
    return UDM_ERROR;
  for (Crd= List->Coords, list=0; list < SrcList->nitems; list++)
  {
    size_t sec;
    UDM_SEARCHSECTIONLIST *SectionList= &SrcList->Item[list];
    for (sec=0; sec < SectionList->nsections; sec++)
    {
      UDM_SEARCHSECTION *Section= &SectionList->Section[sec];
      size_t coord;
      for (coord= 0; coord < Section->ncoords; coord++)
      {
        UDM_COORD2 *Coord2= &Section->Coord[coord];
        Crd->urlid_coord.url_id= Section->url_id;
        Crd->seclen= Section->seclen;
        Crd->urlid_coord.coord.pos= Coord2->pos;
        Crd->num= Section->wordnum;
        Crd->urlid_coord.coord.secno= Section->secno;
        Crd++;
      }
    }
  }
  UDM_ASSERT(ncoords == Crd - List->Coords);
  List->ncoords= ncoords;
  /*UdmCoordListPrint(List);*/

  return UDM_OK;
}



static udm_rc_t
UdmMultiWordMergeCoords(UDM_URLCRDLIST *Phrase, 
                        UDM_SEARCHSECTIONLISTLIST *SrcList,
                        size_t orig_wordnum,
                        size_t nparts)
{
  UdmSectionListListToURLCRDList(Phrase, SrcList);
  UdmURLCRDListSortByURLThenSecnoThenPos(Phrase);
  UdmURLCRDListMergeMultiWord(Phrase, orig_wordnum, nparts);
  /*UdmCoordListPrint(Phrase);*/
  return UDM_OK;
}


udm_rc_t
UdmMultiWordAdd(UDM_SEARCHSECTIONLISTLIST *SectionListList,
                UDM_WIDEWORDLIST *WWList,
                UDM_WIDEWORD *Word,
                UDM_URLID_LIST *Limit,
                UDM_SEARCHSECTIONLISTLIST *OriginalSectionListList,
                size_t orig_wordnum, size_t nparts)
{
  UDM_URLCRDLIST Phrase;
  UDM_SEARCHSECTIONLIST SectionList;
  bzero((void*) &Phrase, sizeof(Phrase));
  UdmMultiWordMergeCoords(&Phrase, SectionListList, orig_wordnum, nparts);
  UDM_ASSERT(orig_wordnum < WWList->nwords);
  if (Limit->nurls)
    UdmApplyFastLimit(&Phrase, Limit);
  if (Phrase.ncoords)
  {
    udm_wordnum_t order= WWList->Word[orig_wordnum].Param.order;
    UdmURLCRDListToSearchSectionList(&SectionList, &Phrase, orig_wordnum, order);
    UdmSearchSectionListListAdd(OriginalSectionListList, &SectionList);
  }
  UDM_FREE(Phrase.Coords);
  Word->Param.count= Phrase.ncoords;
  return UDM_OK;
}
