/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_db.h"
#include "udm_db_int.h"
#include "udm_hash.h"
#include "udm_log.h"
#include "udm_vars.h"
#include "udm_word.h"
#include "udm_coords.h"
#include "udm_parsehtml.h"
#include "udm_indexcache.h"

#ifdef HAVE_SQL

/********************* MultiCache stuff *************************/

typedef struct {
  char *word;
  size_t nintags;
  uint4 *intags;
} UDM_MULTI_CACHE_WORD;


#define COORD_MAX_LEN 4

/*
  Encode intag:
  0 - hex encoding
  1 - hex encoding with '0x' prefix
  2 - PgSQL bytea encoding
*/
static char*
UdmMultiCachePutIntagUsingEncoding(UDM_AGENT *A, UDM_MULTI_CACHE_WORD *cache,
                                   UDM_DB *db, char type)
{
  size_t c;
  size_t len;
  char *_;
  unsigned char buf[COORD_MAX_LEN];
  unsigned char *bufend = buf + sizeof(buf);
  size_t nbytes;
  size_t last;
  size_t multiply= (type == 2) ? 3*5 : 3*2;
  size_t alloced_length;

  if (! cache->nintags) return(NULL);

  /* Allocate maximum possible length. */
  _ = (char*) UdmMalloc((alloced_length= cache->nintags * multiply + 3));
  if (! _) return(NULL);

  if (type == 1)
  {
    strcpy(_, "0x");
    len= 2;
  }
  else
  {
    *_ = 0;
    len = 0;
  }

  last = 0;
  for (c = 0; c < cache->nintags; c++)
  {
    if (cache->intags[c] < last)
    {
      UdmFree(_);
      return(NULL);
    }

    nbytes = udm_coord_put(cache->intags[c] - last, buf, bufend);
    if (! nbytes)
    {
      UdmFree(_);
      return(NULL);
    }

    if (type == 2)
    {
      len+= UdmDBSQLBinEscStr(A, db, _ + len, alloced_length, (char*) buf, nbytes);
      _[len]= '\0'; /* Check if really necessary - possibly  already done */
    }
    else
    {
      len+= UdmHexEncode(_ + len, (char*) buf, nbytes);
    }

    last = cache->intags[c];
  }

  return(_);
}


static char*
UdmMultiCachePutIntagUsingRaw(UDM_MULTI_CACHE_WORD *cache)
{
  size_t c;
  size_t len = 0;
  char *_;
  unsigned char buf[COORD_MAX_LEN];
  unsigned char *bufend = buf + sizeof(buf);
  size_t nbytes;
  size_t last;

  if (! cache->nintags) return(NULL);

  /* Allocate maximum possible length. */
  _= (char*) UdmMalloc(cache->nintags * COORD_MAX_LEN + 1);
  if (! _) return(NULL);

  last = 0;
  for (c = 0; c < cache->nintags; c++)
  {
    if (cache->intags[c] < last)
    {
      UdmFree(_);
      return(NULL);
    }

    nbytes = udm_coord_put(cache->intags[c] - last, buf, bufend);
    if (! nbytes)
    {
      UdmFree(_);
      return(NULL);
    }

    memcpy(_ + len, buf, nbytes);
    len += nbytes;

    last = cache->intags[c];
  }
  _[len] = 0;

  return(_);
}

#define UDM_THREADINFO(A,s,m)	if(A->Conf->ThreadInfo)A->Conf->ThreadInfo(A,s,m)


static udm_rc_t
UdmWordCacheWriteUsingBind(UDM_AGENT *A, UDM_DB *db, int seed, urlid_t url_id,
                           int secno, const char *word,
                           char *coord, size_t coord_len)
{
  udm_rc_t rc;
  int BindType= UDM_SQLTYPE_LONGVARBINARY;
  char qbuf[128];
  
  if (UdmSQLDBType(db) == UDM_DB_IBASE)
    BindType= UDM_SQLTYPE_VARCHAR;
  
  /*
    DBType   Data type        Native ODBC   BindType
    -------  ---------------- ------ ----   
    Cache    VARCHAR(15000)   n/a    n/a    LONGVARBINARY
    DB2      BLOB             n/a    n/a    LONGVARBINARY
    Ibase    VARCHAR(32000)   passed passed VARCHAR
    Mimer    VARCHAR(15000)   passed passed LONGVARBINARY
    MSSQL    VARBINARY(4000)  n/a    n/a    
    MySQL    BLOB             passed passed LONGVARBINARY
    Oracle   RAW(2000)        passed passed LONGVARBINARY
    PgSQL    ByteA            passed failed 
    SQLite   BLOB             passed n/a    LONGVARBINARY
    SQLite3  BLOB             passed n/a    LONGVARBINARY
    Sybase   VARBINARY(4000)  n/a    n/a    LONGVARBINARY
    Virtuoso VARBINARY(4000)  n/a    n/a    LONGVARBINARY
  */
  
  udm_snprintf(qbuf, sizeof(qbuf),
               "INSERT INTO dict%02X "
               "(url_id,secno,word,coords) "
               "VALUES (%d,%d,'%s',%s)",
               seed, url_id, secno, word,
               UdmDBSQLParamPlaceHolder(db, 1));
  if (UDM_OK != (rc= UdmDBSQLPrepare(A, db, qbuf)))
    return rc;

  if (UDM_OK != (rc= UdmDBSQLBindParameter(A, db, 1, coord, coord_len, BindType)))
    return rc;

  if(UDM_OK != (rc= UdmDBSQLExecute(A, db)) ||
     UDM_OK != (rc= UdmDBSQLStmtFree(A, db)))
    return rc;
  
  return rc;
}


udm_rc_t
UdmWordCacheWrite (UDM_AGENT *Indexer, UDM_DB *db, size_t limit)
{
  size_t i;
  size_t in_num, in_limit= UdmVarListFindInt(&Indexer->Conf->Vars, "URLSelectCacheSize", URL_DELETE_CACHE);
  int LastLocked = 0;
  udm_rc_t rc;
  UDM_WORD_CACHE *cache= &UdmSQLDB(db)->WordCache;
  UDM_DSTR buf, qbuf;
  UDM_MULTI_CACHE_WORD mintag;
  size_t aintags = 0;
  char arg[128];

  if (cache->nbytes <= limit) return(UDM_OK);
  UdmLog(Indexer, UDM_LOG_ERROR,
         "Writing words (%d words, %d bytes%s).",
         (int) cache->nwords, (int) cache->nbytes, limit ? "" : ", final");

  UDM_THREADINFO(Indexer, "Starting tnx", "");
  if (UDM_OK != (rc= UdmDBSQLBegin(Indexer, db)))
  {
    UdmWordCacheFree(cache);
    return(rc);
  }

  UdmDSTRInit(&buf, 8192);
  UdmDSTRInit(&qbuf, 8192);
  mintag.intags = NULL;
  mintag.word = NULL;
  
  /* Sort IN list */
  if (cache->nurls)
    UdmSort(cache->urls, cache->nurls, sizeof(urlid_t), (udm_qsort_cmp)UdmCmpURLID);

  /* Remove expired words */
  for (i= 0, in_num= 0; i < cache->nurls; i++)
  {
    if (UdmDSTRLength(&buf))
      UdmDSTRAppend(&buf, ",", 1);
    UdmDSTRAppendf(&buf, "%d", cache->urls[i]);

    /* Don't allow too long IN list */
    if (++in_num >= in_limit || (i + 1) == cache->nurls)
    {
      int dictno;
      for (dictno= 0; dictno <= MULTI_DICTS; dictno++)
      {
        udm_snprintf(arg, sizeof(arg), "dict%02X", dictno);
        UDM_THREADINFO(Indexer, "Deleting", arg);
        UdmDSTRReset(&qbuf);
        UdmDSTRAppendf(&qbuf, "DELETE FROM dict%02X WHERE url_id IN (%s)",
                       dictno, UdmDSTRPtr(&buf));
        if(UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf))))
          goto unlock_UdmStoreWordsMulti;
      }
      UdmDSTRReset(&buf);
      UdmDSTRReset(&qbuf);
      in_num= 0;
    }
  }

  UdmDSTRReset(&buf);
  UdmDSTRReset(&qbuf);

  if (cache->nwords)
    UdmWordCacheSort(cache);

  for (i = 0; i < cache->nwords;)
  {
    UDM_WORD_CACHE_WORD *cword = &cache->words[i];
    unsigned char seed = cword->seed;
    char *word = cword->word;
    urlid_t url_id = cword->urlid_coord.url_id;
    udm_pos_t secno = cword->urlid_coord.coord.secno;

    mintag.nintags = 0;

    do
    {
      if (mintag.nintags == aintags)
      {
        uint4 *itmp;
        itmp= (uint4*) UdmRealloc(mintag.intags, (aintags + 256) * sizeof(uint4));
	if (! itmp)
	{
	  goto unlock_UdmStoreWordsMulti;
	}
	mintag.intags = itmp;
	aintags += 256;
      }
      mintag.intags[mintag.nintags]= cword->urlid_coord.coord.pos;
      mintag.nintags++;

      i++;
      if (i == cache->nwords) break;
      cword = &cache->words[i];
    }
    while (seed == cword->seed &&
           url_id == cword->urlid_coord.url_id &&
           secno == cword->urlid_coord.coord.secno &&
           ! strcmp(word, cword->word));

    udm_snprintf(arg, sizeof(arg), "dict%02X", seed);
    UDM_THREADINFO(Indexer, "Writing", arg);

    if (UdmSQLDBFlags(db) & UDM_SQL_HAVE_BIND_BINARY)
    {
      char *intag= UdmMultiCachePutIntagUsingRaw(&mintag);
      if (!intag)
        continue;

      if (UDM_OK != (rc= UdmWordCacheWriteUsingBind(Indexer, db, seed, url_id,
                           secno, word, intag, strlen(intag))))
        goto unlock_UdmStoreWordsMulti;
      UdmFree(intag);
    }
    else if (UdmSQLDBType(db) == UDM_DB_MYSQL)
    {
      char *intag= UdmMultiCachePutIntagUsingEncoding(Indexer, &mintag, db, 1);
      if (! intag) continue;

      if (UdmDSTRLength(&buf))
      {
        UdmDSTRAppendf(&buf, ",(%d, %d, '%s', %s)",
	               url_id, secno, word, intag);
      } else {
        UdmDSTRAppendf(&buf, "INSERT INTO dict%02X (url_id,secno,word,coords) VALUES(%d,%d,'%s',%s)",
                       seed, url_id, secno, word, intag);
      }
      UdmFree(intag);
      if (seed != cword->seed || i == cache->nwords)
      {
        if (LastLocked <= seed)
	{
          if (LastLocked)
            UdmDBSQLQuery(Indexer, db, NULL, "UNLOCK TABLES");
	  LastLocked= seed;
	  UdmDSTRAppendf(&qbuf, "LOCK TABLES dict%02X WRITE", LastLocked);
          for (LastLocked++; LastLocked <= MULTI_DICTS; LastLocked++)
          {
	    if (LastLocked - seed == 0x10) break;
            UdmDSTRAppendf(&qbuf, ",dict%02X WRITE", LastLocked);
          }
          UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&qbuf));
	  UdmDSTRReset(&qbuf);
	}

        if (UdmDSTRLength(&buf))
        {
          if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&buf))))
            goto unlock_UdmStoreWordsMulti;
	  UdmDSTRReset(&buf);
        }
      }
    }
    else
    {
      const char *quot, *x, *castb, *caste;
      char *intag;
      
      if (UdmSQLDBType(db) == UDM_DB_ORACLE8 ||
          UdmSQLDBType(db) == UDM_DB_DB2 ||
          UdmSQLDBType(db) == UDM_DB_MONETDB)
        intag= UdmMultiCachePutIntagUsingEncoding(Indexer, &mintag, db, 0);
      else if (UdmSQLDBType(db) == UDM_DB_MSSQL ||
               UdmSQLDBType(db) == UDM_DB_SYBASE ||
               UdmSQLDBType(db) == UDM_DB_ACCESS)
        intag= UdmMultiCachePutIntagUsingEncoding(Indexer, &mintag, db, 1);
      else if (UdmSQLDBType(db) == UDM_DB_PGSQL)
        intag= UdmMultiCachePutIntagUsingEncoding(Indexer, &mintag, db, 2);
      else
      {
        if ((intag= UdmMultiCachePutIntagUsingRaw(&mintag)))
        {
          char *tmp= UdmDBSQLEscStrAlloc(Indexer, db, intag, strlen(intag)); /* Data */
          UdmFree(intag);
          intag= tmp;
        }
      }
      if (! intag) continue;

      if (UdmSQLDBType(db) == UDM_DB_MSSQL ||
          UdmSQLDBType(db) == UDM_DB_SYBASE ||
          UdmSQLDBType(db) == UDM_DB_ACCESS)
        quot="";
      else
        quot="'";

      if (UdmSQLDBType(db) == UDM_DB_DB2)
      {
        x="X";
        castb="CAST(";
        caste=" AS BLOB)";
      }
      else
      {
        x="";
        castb="";
        caste="";
      }

      UdmDSTRAppendf(&buf,
        "INSERT INTO dict%02X (url_id,secno,word,coords) VALUES(%d,%d,'%s',%s%s%s%s%s%s)",
        seed, url_id, secno, word, castb,x,quot,intag,quot,caste);
      UdmFree(intag);
      if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, UdmDSTRPtr(&buf))))
        goto unlock_UdmStoreWordsMulti;
      UdmDSTRReset(&buf);
    }
  }

unlock_UdmStoreWordsMulti:
  UDM_FREE(mintag.intags);
  UdmDSTRFree(&buf);
  UdmDSTRFree(&qbuf);
  UDM_THREADINFO(Indexer, "Committing tnx", "");

  if (LastLocked && rc == UDM_OK)
    rc= UdmDBSQLQuery(Indexer, db, NULL, "UNLOCK TABLES");

  if (rc == UDM_OK)
    rc=UdmDBSQLCommit(Indexer, db);

  UdmWordCacheFree(&UdmSQLDB(db)->WordCache);
  UdmLog(Indexer, UDM_LOG_ERROR, "The words are written successfully.%s", limit ? "" : " (final)");
  return(rc);
}


static udm_rc_t
UdmStoreWordsMulti(UDM_AGENT *Indexer, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  udm_rc_t rc= UDM_OK;
  urlid_t  url_id= UdmVarListFindInt(&Doc->Sections, "ID", 0);
  unsigned char PrevStatus= UdmVarListFindInt(&Doc->Sections, "PrevStatus", 0) ? 1 : 0;
  UDM_WORDLIST Words;

  UDM_LOCK_CHECK_OWNER(Indexer, UDM_LOCK_DB);

  UdmWordListInit(&Words);

  if (UDM_OK != (rc= UdmPrepareWords(Indexer, Doc, &Words)))
    goto ret;

  /*
    Don't need to delete words here.
    DBMode=multi deletes words from multiple
    documents in a portion, when doing UdmWordCacheWrite()
  */

  if (UdmVarListFindBool(&Indexer->Conf->Vars, "SaveSectionSize", UDM_TRUE))
  {
    if (UDM_OK!= (rc= UdmWordListSaveSectionSize(&Words)))
      goto ret;
  }
  
  if (PrevStatus)
    UdmWordCacheAddURL(&UdmSQLDB(db)->WordCache, url_id);

  rc= UdmWordCacheAddWordList(&UdmSQLDB(db)->WordCache, &Words, url_id);

ret:
  UdmWordListFree(&Words);
  return rc;
}


static udm_rc_t
UdmMultiAddCoords(UDM_QUERY *Query, UDM_FINDWORD_ARGS *args, UDM_SQLRES *SQLRes)
{
  size_t i;
  size_t numrows;
  UDM_URLCRDLIST CoordList;
  UDM_URL_CRD CoordTemplate;

  bzero((void*)&CoordList, sizeof(CoordList));
  bzero((void*)&CoordTemplate, sizeof(CoordTemplate));
  CoordTemplate.num= args->Word.Param.order;
  
  numrows= UdmSQLNumRows(SQLRes);

  for (i= 0; i < numrows; i++)
  {
    size_t tmp = UdmSQLLen(SQLRes, i, 2);
    if (tmp) CoordList.acoords += tmp;
    else CoordList.acoords += strlen(UdmSQLValue(SQLRes, i, 2));
  }
  CoordList.Coords= (UDM_URL_CRD*)UdmMalloc((CoordList.acoords) * sizeof(UDM_URL_CRD));

  for (i = 0; i < numrows; i++)
  {
    size_t lintag= UdmSQLLen(SQLRes, i, 2);
    const unsigned char *intag= (const unsigned char *)UdmSQLValue(SQLRes, i, 2);
    
    CoordTemplate.urlid_coord.url_id= UDM_ATOI(UdmSQLValue(SQLRes, i, 0));
    CoordTemplate.urlid_coord.coord.secno= UDM_ATOI(UdmSQLValue(SQLRes, i, 1));
    
    if (!lintag)
      lintag= strlen((const char *)intag); /* FIXME: Check UdmSQLLen */
    
    if (args->query_param.wf[CoordTemplate.urlid_coord.coord.secno])
    {
      UdmCoordListMultiUnpack(&CoordList, &CoordTemplate, 
                              intag, lintag,
                              args->query_param.SaveSectionSize);
    }
  }
  
  if (args->urls.nurls)
    UdmApplyFastLimit(&CoordList, &args->urls);
  if (CoordList.ncoords)
  {
    /*
      We have to sort here, because DBMode=multi
      returns data unsorted.
    */
    UdmURLCRDListSortByURLThenSecnoThenPos(&CoordList);
    UdmURLCRDListListAddWithSort2(&args->SearchSectionListList,
                                  &Query->Res.WWList,
                                  &args->Word, &CoordList);
  }
  else
  {
    UdmFree(CoordList.Coords);
  }
  args->Word.Param.count= CoordList.ncoords;
  return UDM_OK;
}


static udm_rc_t
UdmFindWordMulti(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                 UDM_FINDWORD_ARGS *args)
{
  char qbuf[4096], secno[64]= "";
  UDM_SQLRES SQLRes;
  int tnum, tmin, tmax;
  udm_timer_t ticks;
  udm_rc_t rc;

  if (args->Word.Param.match_mode != UDM_MATCH_FULL)
  {
    /* This is for substring search!  */
    /* In Multi mode: we have to scan */
    /* almost all tables except those */
    /* with to short words            */
      
    tmin= 0;
    tmax= MULTI_DICTS;
  }
  else
  {
    tmin= tmax= UdmStrHash32(args->Word.Word.str) & MULTI_DICTS;
  }

  if (args->Word.Param.secno)
    udm_snprintf(secno, sizeof(secno),
                 " AND dict.secno=%d", (int) args->Word.Param.secno);

  for(tnum= tmin; tnum <= tmax; tnum++)
  {
    if (*args->where)
    {
      udm_snprintf(qbuf, sizeof(qbuf) - 1,"\
SELECT dict.url_id,dict.secno,dict.coords \
FROM dict%02X dict, url%s \
WHERE dict.%s \
AND url.rec_id=dict.url_id AND %s%s",
        tnum, UdmSQLDBQueryFrom(Query), args->cmparg, args->where, secno);
    }
    else
    {
      udm_snprintf(qbuf, sizeof(qbuf) - 1,
                   "SELECT url_id,secno,coords FROM dict%02X dict WHERE %s%s",
                   tnum, args->cmparg, secno);
    }

    if (UDM_OK != (rc= UdmDBSQLQuery(A, db, &SQLRes, qbuf)))
      return rc;

    UdmLog(A, UDM_LOG_DEBUG, "Start UdmMultiAddCoords");
    ticks= UdmStartTimer();
    UdmMultiAddCoords(Query, args, &SQLRes);
    UdmLog(A, UDM_LOG_DEBUG, "Stop UdmMultiAddCoords\t%.2f", UdmStopTimer(&ticks));
    UdmSQLFree(&SQLRes);
  }
  return UDM_OK;
}


static udm_rc_t
UdmDeleteWordsFromURLMulti(UDM_AGENT *Indexer, UDM_DB *db, urlid_t url_id)
{
  int i;
  char qbuf[512];
  for(i= 0; i <= MULTI_DICTS; i++)
  {
    udm_rc_t rc;
    udm_snprintf(qbuf, sizeof(qbuf),
                 "DELETE FROM dict%02X WHERE url_id=%d", i, url_id);
    if (UDM_OK != (rc= UdmDBSQLQuery(Indexer, db, NULL, qbuf)))
      return rc;
  }
  return UDM_OK;
}


static udm_rc_t
UdmTruncateDictMulti(UDM_AGENT *Indexer,UDM_DB *db)
{
  udm_rc_t rc= UDM_OK;
  int i;
  for(i= 0 ; i <= MULTI_DICTS; i++)
  {
    char tablename[32];
    sprintf(tablename, "dict%02X", i);
    if (UDM_OK != (rc= UdmDBSQLTableTruncateOrDelete(Indexer, db, tablename)))
      return rc;
  }
  return UDM_OK;
}


static udm_rc_t
UdmWordStatCreateMulti(UDM_AGENT *A, UDM_DB *db)
{
  int i;
  
  for (i=0; i <= 0xFF; i++)
  {
    udm_rc_t rc;
    char qbuf[128];
    UdmLog(A, UDM_LOG_EXTRA, "Processing table %02X", i);

    sprintf(qbuf, "SELECT word, count(*) FROM dict%02X GROUP BY word", i);
    if (UDM_OK != (rc= UdmWordStatQuery(A, db, qbuf)))
      return rc;
  }
  return UDM_OK;
}



static udm_rc_t
UdmDumpWordInfoOneDocMulti(UDM_AGENT *A, UDM_DB *db, UDM_DOCUMENT *Doc)
{
  return UDM_OK;
}


static udm_rc_t
UdmMultiInitSearch(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query,
                   UDM_FINDWORD_ARGS *args)
{
  return UDM_OK;
}


static udm_rc_t
UdmQueryActionMulti(UDM_AGENT *A, UDM_DB *db, UDM_QUERY *Query, udm_querycmd_t cmd)
{
  switch (cmd)
  {
    case UDM_QUERYCMD_CLEAR:    return UdmTruncateDictMulti(A, db);
    case UDM_QUERYCMD_WORDSTAT: return UdmWordStatCreateMulti(A, db);
    case UDM_QUERYCMD_INDEX:
      return udm_dbmode_handler_blob.QueryAction(A, db, Query, cmd);
    default: break;
  }
  return UDM_NOTARGET;
}


const UDM_DBMODE_HANDLER udm_dbmode_handler_multi=
{
  "multi",
  UdmStoreWordsMulti,
  UdmQueryActionMulti,
  UdmDeleteWordsFromURLMulti,
  UdmFindWordMulti,
  UdmDumpWordInfoOneDocMulti,
  UdmMultiInitSearch,
};

#endif /* HAVE_SQL */
