/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>

#include "udm_common.h"
#include "udm_doc.h"
#include "udm_utils.h"
#include "udm_doc.h"
#include "udm_word.h"
#include "udm_hrefs.h"
#include "udm_guesser.h"
#include "udm_vars.h"
#include "udm_textlist.h"
#include "udm_parsehtml.h"
#include "udm_xmalloc.h"
#include "udm_url.h"
#include "udm_http.h"
#include "udm_searchtool.h"
#include "udm_indexer.h"
#include "udm_env.h"


UDM_API(void)
UdmDocInit(UDM_DOCUMENT *Doc)
{
  bzero((void*)Doc, sizeof(UDM_DOCUMENT));
  Doc->Spider.read_timeout= UDM_READ_TIMEOUT;
  Doc->Spider.doc_timeout= UDM_DOC_TIMEOUT;
  Doc->Spider.net_error_delay_time= UDM_DEFAULT_NET_ERROR_DELAY_TIME;
  Doc->connp.connp= &Doc->connp2;
  UdmURLInit(&Doc->CurURL);
}


UDM_API(void)
UdmDocFree(UDM_DOCUMENT *Result)
{
  if(!Result)return;
  UdmHTTPBufFree(&Result->Buf);
  UDM_FREE(Result->connp.hostname);
  UDM_FREE(Result->connp.user);
  UDM_FREE(Result->connp.pass);
  
  UdmHrefListFree(&Result->Hrefs);
  UdmVarListFree(&Result->RequestHeaders);
  UdmVarListFree(&Result->Sections);
  UdmTextListFree(&Result->TextList);
  
  UdmURLFree(&Result->CurURL);

  bzero((void*)Result, sizeof(UDM_DOCUMENT));
}


/* Aproximation for memory used by a document, e.g. in the Document cache */
size_t
UdmDocUsedMemory(const UDM_DOCUMENT *Doc)
{
  size_t mem= sizeof(UDM_DOCUMENT);
  mem+= UdmVarListUsedMemory(&Doc->Sections);
  mem+= (sizeof(UDM_HREF) + 35) * Doc->Hrefs.nhrefs;
  return mem;
}


/************************************************************************/

#include "udm_server.h"
#include "udm_log.h"

static udm_rc_t
UdmDocParse(UDM_AGENT *Indexer, UDM_DOCUMENT *Doc)
{
  const char *content_type= UdmVarListFindStr(&Doc->Sections,
                                              "Cached.Content-Type", "text/plain");
  udm_content_type_t ct= UdmContentTypeByName(content_type);
  if (ct == UDM_CONTENT_TYPE_UNKNOWN)
  {
    UdmLog(Indexer, UDM_LOG_WARN,
           "Warning: URLID=%d: Unknown Content-Type: %s",
            UdmVarListFindInt(&Doc->Sections, "ID", 0), content_type);
    return UDM_ERROR;
  }
  return UdmDocInternalParserExec(Indexer, Doc, ct);
}


static udm_rc_t
UdmSectionAppend(const UDM_ENV *Env, UDM_CONV *cnv,
                 UDM_VAR *Sec, const UDM_TEXTITEM *Item)
{
  if (Sec)
  {
    size_t maxlen= UdmEnvSectionMaxLength(Env, UdmVarSecno(Sec));
    if (UdmVarStr(Sec) && UdmVarStr(Sec)[0])
      UdmVarAppendStrn(Sec, UDM_CSTR_WITH_LEN(" "));
    if (UdmVarFlags(Sec) & UDM_VARFLAG_RAW)
    {
      UdmVarAppendStrn(Sec, Item->str, strlen(Item->str));
    }
    else
    {
      UdmVarAppendConv(Sec, maxlen, cnv, UDM_RECODE_HTML,
                       Item->str, strlen(Item->str));
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmMakeSections(const UDM_ENV *Env, UDM_CONV *cnv,
                UDM_VARLIST *Vars,
                const UDM_TEXTLIST *TextList,
                udm_bool_t raw)
{
  size_t i;
  for (i= 0; i < TextList->nitems; i++)
  {
    UDM_TEXTITEM *Item= &TextList->Item[i];
    if (!(Item->Param.flags & UDM_TEXTLIST_FLAG_SKIP_ADD_SECTION))
    {
      UdmSectionAppend(Env, cnv,
                       UdmVarListFindVar(Vars, Item->section_name), Item);
      if (raw)
      {
        char name[128];
        udm_snprintf(name, sizeof(name), "Raw.%s", Item->section_name);
        UdmSectionAppend(Env, cnv, UdmVarListFindVar(Vars, name), Item);
      }
    }
  }
  return UDM_OK;
}


udm_rc_t
UdmDocToConstWordList(UDM_AGENT *Indexer,
                      UDM_DOCUMENT *Doc,
                      UDM_CONSTWORDLIST *CWL,
                      UDM_CHARSET *doccs, int cnvflags, int all_section_flags,
                      udm_timer_t *timer_parse,
                      udm_timer_t *timer_words)
{
  udm_timer_t ticks= UdmStartTimer();

  UdmSpiderParamInit(&Doc->Spider);
  Doc->Spider.robots.follow= UDM_FALSE;
  UdmVarListReplaceLst(&Doc->Sections, &Indexer->Conf->Sections, NULL, "*");

  if (!Doc->Buf.buf)
    return UDM_OK;

  UdmDocParse(Indexer, Doc);
  *timer_parse+= UdmStartTimer() - ticks;

  ticks= UdmStartTimer();
  if (Doc->Spider.robots.index)
  {
    UDM_CHARSET *metacs= UdmVarListFindCharset(&Doc->Sections, "Strong-Meta-Charset", doccs);
    const char *seg= UdmVarListFindStr(&Indexer->Conf->Vars, "Segmenter", NULL);
    const char *lang= UdmVarListFindStr(&Doc->Sections, "Content-Language",
                                        UdmVarListFindStr(&Doc->Sections,
                                                         "Meta-Language", NULL));
    int segmenter;
    /* TODO34: revise segmenter activation logic */
    if ((seg || lang) && (segmenter= UdmUniSegmenterFind(Indexer->Conf, lang, seg)))
    {
      UdmTextListSegment(Indexer, &Doc->TextList, metacs, segmenter);
    }
    if (all_section_flags & UDM_VARFLAG_USERDEF)
    {
      UDM_CONST_STR content;
      UDM_VAR *HTTPContent= UdmVarListFindVar(&Doc->Sections, "HTTP.Content");
      UDM_VAR *HTTPLocalCharsetContent= UdmVarListFindVar(&Doc->Sections, "HTTP.LocalCharsetContent");
      if ((HTTPContent||HTTPLocalCharsetContent) &&
          !UdmHTTPBufContentToConstStr(&Doc->Buf, &content) && content.length)
      {
        if (HTTPContent)
          UdmVarSetStrn(HTTPContent, content.str, content.length);
        if (HTTPLocalCharsetContent)
        {
          UDM_CONV cnv;
          UdmConvInit(&cnv, metacs, Doc->lcs);
          UdmVarSetConv(HTTPLocalCharsetContent, &cnv, UDM_RECODE_HTML_NONASCII,
                        content.str, content.length);
        }
      }
      UdmUserSectionsParse(Indexer, &Indexer->Conf->SectionMatch, Doc, metacs);
      UdmUserSectionsParse(Indexer, &Indexer->Conf->SectionGsrMatch, Doc, metacs);
    }
    if (all_section_flags & UDM_VARFLAG_DECIMAL)
    {
      size_t i;
      for (i= 0; i < Doc->TextList.nitems; i++)
      {
        UDM_TEXTITEM *Item= &Doc->TextList.Item[i];
        const UDM_VAR *Section= UdmVarListFindBySecno(&Indexer->Conf->Sections, Item->Param.secno);
        if (Section && (UdmVarFlags(Section) & UDM_VARFLAG_DECIMAL))
          UdmTextItemNormalizeDecimal(Item);
      }
    }
    if (Indexer->Conf->SectionFilters.nitems)
    {
      UDM_CONV cnv;
      udm_method_t method;
      char reason[1024]= "none";
      UdmConvInit(&cnv, metacs, Doc->lcs);
      /*UdmVarListPrint(&Doc->Sections, stderr);*/
      UdmMakeSections(Indexer->Conf, &cnv, &Doc->Sections, &Doc->TextList,
                      (all_section_flags & UDM_VARFLAG_RAW) ?
                      UDM_TRUE : UDM_FALSE);
      method= UdmSectionFilterListFindMethod(&Indexer->Conf->SectionFilters,
                                             Doc, reason, sizeof(reason));
      switch (method)
      {
        case UDM_METHOD_NOINDEX:
          goto end;
        case UDM_METHOD_GET:
        case UDM_METHOD_UNKNOWN:
        case UDM_METHOD_DISALLOW:
        case UDM_METHOD_HEAD:
        case UDM_METHOD_HREFONLY:
        case UDM_METHOD_CHECKMP3:
        case UDM_METHOD_CHECKMP3ONLY:
        case UDM_METHOD_VISITLATER:
        case UDM_METHOD_INDEX:
        case UDM_METHOD_IMPORTONLY:
          break;
      }
    }
    /* TODO34: compare results from UdmPrepareWords and UdmTextListToConstWordList  */
    /* TODO34: check rc */
    UdmTextListToConstWordList(&Doc->TextList,
                               Indexer->Conf->unidata, metacs, cnvflags, CWL);
  }
end:
  *timer_words+= UdmStartTimer() - ticks;

  return UDM_OK;
}
