/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include <udm_config.h>

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#ifdef   HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_ZLIB
#include <zlib.h>
#endif

#include "udm_common.h"
#include "udm_store.h"
#include "udm_utils.h"
#include "udm_log.h"
#include "udm_vars.h"
#include "udm_parsehtml.h"
#include "udm_unicode.h"
#include "udm_unidata.h"
#include "udm_searchtool.h"
#include "udm_guesser.h"
#include "udm_proto.h"
#include "udm_parser.h"
#include "udm_http.h"
#include "udm_textlist.h"
#include "udm_match.h"
#include "udm_wild.h"
#include "udm_env.h"


typedef struct
{
  size_t offset;
  size_t length;
  uint4 quality;
} UDM_SNIPPETCHUNK;


typedef struct
{
  size_t nitems;
  size_t mitems;
  UDM_SNIPPETCHUNK *Item;
} UDM_SNIPPETCHUNKLIST;


static void
UdmSnippetChunkListInit(UDM_SNIPPETCHUNKLIST *list)
{
  bzero((void*) list, sizeof(*list));
}


static void
UdmSnippetChunkListFree(UDM_SNIPPETCHUNKLIST *list)
{
  UdmFree((void*) list->Item);
}


static void
UdmSnippetChunkListReset(UDM_SNIPPETCHUNKLIST *list)
{
  list->nitems= 0;
}


static udm_rc_t
UdmSnippetChunkListRealloc(UDM_SNIPPETCHUNKLIST *list)
{
  if (list->nitems < list->nitems)
    return UDM_OK;
  if (!(list->Item= UdmRealloc(list->Item,
                                (list->mitems + 256) * sizeof(list->Item[0]))))
    return UDM_ERROR;
  list->mitems+= 256;
  return UDM_OK;
}


static udm_rc_t
UdmSnippetChunkListAdd(UDM_SNIPPETCHUNKLIST *list, const UDM_SNIPPETCHUNK *Item)
{
  if (UDM_OK != UdmSnippetChunkListRealloc(list))
    return UDM_ERROR;
  list->Item[list->nitems++]= Item[0];
  return UDM_OK;
}


typedef struct
{
  /*UDM_DSTR body_native;*/
  UDM_DSTR body_unicode;
  UDM_SNIPPETCHUNKLIST chunks;
  size_t best_chunk;
} UDM_SNIPPET;


static udm_rc_t
UdmSnippetInit(UDM_SNIPPET *snippet)
{
  snippet->best_chunk= 0;
  UdmSnippetChunkListInit(&snippet->chunks);
  return UdmDSTRInit(&snippet->body_unicode, 4096);
}


static void
UdmSnippetFree(UDM_SNIPPET *snippet)
{
  UdmSnippetChunkListFree(&snippet->chunks);
  UdmDSTRFree(&snippet->body_unicode);
}


static void UdmSnippetReset(UDM_SNIPPET *snippet)
{
  snippet->best_chunk= 0;
  UdmSnippetChunkListReset(&snippet->chunks);
  UdmDSTRReset(&snippet->body_unicode);
}


static udm_rc_t
UdmSnippetAddChunk(UDM_SNIPPET *snippet, UDM_SNIPPETCHUNK *chunk)
{
  if (UDM_OK != UdmSnippetChunkListAdd(&snippet->chunks, chunk))
    return UDM_ERROR;
  if (snippet->chunks.Item[snippet->best_chunk].quality < chunk->quality)
    snippet->best_chunk= snippet->chunks.nitems - 1;
  return UDM_OK;
}


/* Put space character into the end of snippet->body_unicode */
static size_t UdmSnippetAppendSpaceUnicode(UDM_SNIPPET *snippet)
{
  static int i20= 0x20;
  return UdmDSTRAppend(&snippet->body_unicode, (const char *) &i20, sizeof(i20));
}


typedef struct
{
  UDM_MATCH_PART fragment;
  int quality;
} UDM_EXCERPT_FRAGMENT_QUALITY;


#define NQUALITY 256
typedef struct
{
  size_t nitems;
  UDM_EXCERPT_FRAGMENT_QUALITY Item[NQUALITY];
} UDM_EXCERPT_FRAGMENT_QUALITY_LIST;


static void
UdmExcerptFragmentQualityListInit(UDM_EXCERPT_FRAGMENT_QUALITY_LIST *List)
{
  List->nitems= 0;
}
static udm_rc_t
UdmExcerptFragmentQualityListAdd(UDM_EXCERPT_FRAGMENT_QUALITY_LIST *List,
                                 UDM_MATCH_PART *fragment, int quality)
{
  UDM_EXCERPT_FRAGMENT_QUALITY *Item;
  if (List->nitems >= NQUALITY)
    return UDM_ERROR;
  Item= &List->Item[List->nitems];
  Item->fragment= *fragment;
  Item->quality= quality;
  List->nitems++;
  return UDM_OK;
}
static UDM_EXCERPT_FRAGMENT_QUALITY *
UdmExcerptFragmentQualityListFind(UDM_EXCERPT_FRAGMENT_QUALITY_LIST *List,
                                  size_t offset)
{
  size_t i;
  for (i= 0; i < List->nitems; i++)
  {
    UDM_EXCERPT_FRAGMENT_QUALITY *Item= &List->Item[i];
    if (Item->fragment.beg < (int) offset &&
        Item->fragment.end > (int) offset)
      return Item;
  }
  return NULL;
}


typedef struct udm_excerpt_helper_st
{
  UDM_AGENT *Agent;
  UDM_QUERY *Query;
  UDM_DOCUMENT *Doc;
  UDM_CONST_STR content;
  UDM_SNIPPET snippet;
  UDM_DSTR body;
  UDM_HIGHLIGHT_CONV hconv_bcs;
  UDM_HIGHLIGHT_CONV hconv_uni;
  UDM_EXCERPT_FRAGMENT_QUALITY_LIST Quality;
} UDM_EXCERPT_HELPER;


#define DEFAULT_FRAGMENT_QUALITY 255

static size_t
UdmSnippetHlAppendUnicode(UDM_EXCERPT_HELPER *helper,
                          UDM_SNIPPET *dst,
                          int fragment_quality,
                          const char * src, size_t length)
{
  UDM_SNIPPETCHUNK_STAT chunk_stat;
  UDM_SNIPPETCHUNK chunk;
  size_t res_length;

  if (!fragment_quality)
    return 0;

  chunk.offset= UdmDSTRLength(&dst->body_unicode);
  res_length= UdmHlConvertExtWithConv(helper->Agent,
                                      &dst->body_unicode, &chunk_stat,
                                      &helper->Query->Res.WWList,
                                      src, length,
                                      &helper->hconv_uni);
  if (chunk_stat.mask) /* Some query words were found in the chunk */
  {
    chunk.length= UdmDSTRLength(&dst->body_unicode) - chunk.offset;
    /*printf("HERE: quality=%d-%d\n", chunk_stat.quality, fragment_quality);*/
    if ((chunk.quality= chunk_stat.quality * fragment_quality / 255))
    {
      if (UDM_OK != UdmSnippetAddChunk(dst, &chunk))
        return 0;
    }
  }
  return res_length;
}


static void
UdmExcerptHelperInitConverters(UDM_EXCERPT_HELPER *param, UDM_CHARSET *dcs)
{
  UdmExcerptConvInitFromEnv(&param->hconv_bcs,
                            param->Agent->Conf->bcs,
                            dcs,
                            param->Agent->Conf->bcs,
                            param->Agent->Conf);
  UdmExcerptConvInitFromEnv(&param->hconv_uni,
                            param->Agent->Conf->bcs,
                            dcs,
                            &udm_charset_sys_int,
                            param->Agent->Conf);
}


static udm_rc_t
UdmExcerptHelperInit(UDM_EXCERPT_HELPER *param,
                     UDM_AGENT *Agent,
                     UDM_QUERY *Query,
                     UDM_DOCUMENT *Doc,
                     UDM_CHARSET *dcs)
{
  bzero(param, sizeof(*param));
  param->Agent= Agent;
  param->Query= Query;
  param->Doc= Doc;    
  if (UDM_OK != UdmDSTRInit(&param->body, 1024))
    return UDM_ERROR;
  if (UDM_OK != UdmSnippetInit(&param->snippet))
  {
    UdmDSTRFree(&param->body);
    return UDM_ERROR;
  }
  UdmExcerptHelperInitConverters(param, dcs);
  return UDM_OK;
}


static void
UdmExcerptHelperFree(UDM_EXCERPT_HELPER *param)
{
  UdmDSTRFree(&param->body);
  UdmSnippetFree(&param->snippet);
}


static udm_rc_t
UdmGetExcerptSourceFromXMLContentTextHandler(UDM_HTML_PARSER *parser)
{
  UDM_EXCERPT_HELPER *param= (UDM_EXCERPT_HELPER *) parser->user_data;
  if (!parser->state.script && !parser->state.comment && !parser->state.style)
    UdmDSTRAppend(&param->body, parser->lasttok.str, parser->lasttok.length);
  else
    UdmDSTRAppend(&param->body, " ", 1);
  return UDM_OK;
}


static udm_rc_t
UdmGetExcerptSourceFromXMLContentTagCommentHandler(UDM_HTML_PARSER *parser)
{
  UDM_EXCERPT_HELPER *param= (UDM_EXCERPT_HELPER *) parser->user_data;
  UdmDSTRAppend(&param->body, " ", 1);
  return UDM_OK;
}


static udm_rc_t
UdmGetExcerptSourceFromXMLContent(UDM_EXCERPT_HELPER *param,
                                  const char *src, size_t srclen)
{
  UDM_HTML_PARSER parser;
  UdmHTMLParserInit(&parser);
  UdmHTMLParserSetUserData(&parser, param);
  UdmHTMLParserSetTextHandler(&parser, UdmGetExcerptSourceFromXMLContentTextHandler);
  UdmHTMLParserSetTagHandler(&parser, UdmGetExcerptSourceFromXMLContentTagCommentHandler);
  UdmHTMLParserSetCommentHandler(&parser, UdmGetExcerptSourceFromXMLContentTagCommentHandler);
  return UdmHTMLParserExec(&parser, src, srclen);
}


static udm_rc_t
UdmGetExcerptSourceFromHTMLContentCommentHandler(UDM_HTML_PARSER *parser)
{
  UDM_EXCERPT_HELPER *param= (UDM_EXCERPT_HELPER *) parser->user_data;
  UdmSnippetAppendSpaceUnicode(&param->snippet);
  return UDM_OK;
}


static udm_rc_t
UdmGetExcerptSourceFromHTMLContentTagHandler(UDM_HTML_PARSER *parser)
{
  UDM_EXCERPT_HELPER *param= (UDM_EXCERPT_HELPER *) parser->user_data;
  if (!UdmConstStrNCaseCmp(&parser->tag.toks[0].name, UDM_CSTR_WITH_LEN("meta")))
  {
    UDM_CONST_STR *name= UdmHTMLTagFindAttrMetaName(&parser->tag);
    UDM_CONST_STR *content= UdmHTMLTagFindAttrByName(&parser->tag, UDM_CSTR_WITH_LEN("content"));
    if (name && content)
    {
      UDM_CONST_TEXTITEM Item;
      UDM_TEXT_PARAM TextParam;
      UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_HTML, 0);
      UdmConstTextItemInit(&Item);
      Item.text= *content;
      UdmConstStrTrim(&Item.text, " \r\n\t");
      if (Item.text.length)
      {
        char secname[128];
        size_t secname_length= udm_snprintf(secname, sizeof(secname),
                                            "meta.%.*s",
                                            (int) name->length, name->str);
        UdmConstStrSet(&Item.section_name, secname, secname_length);
        UdmTextListAddConst(&param->Doc->TextList, &Item, &TextParam);
      }
    }
  }
  UdmSnippetAppendSpaceUnicode(&param->snippet);
  return UDM_OK;
}


static udm_rc_t
UdmGetExcerptSourceFromHTMLContentTextHandler(UDM_HTML_PARSER *parser)
{
  UDM_EXCERPT_HELPER *param= (UDM_EXCERPT_HELPER *) parser->user_data;
  if (parser->state.title &!parser->state.comment)
  {
    UDM_TEXT_PARAM TextParam;
    UDM_CONST_TEXTITEM Item;
    UdmTextParamInit(&TextParam, UDM_TEXTLIST_FLAG_HTML, 0);
    Item.text= parser->lasttok;
    UdmConstStrTrim(&Item.text, " \r\n\t");
    if (Item.text.length)
    {
      UdmConstStrSet(&Item.section_name, UDM_CSTR_WITH_LEN("title"));
      UdmConstStrSet(&Item.href, UDM_CSTR_WITH_LEN(""));
      UdmTextListAddConst(&param->Doc->TextList, &Item, &TextParam);
    }
  }
  if (!parser->state.title && !parser->state.script &&
      !parser->state.comment && !parser->state.style)
  {
    size_t offset= parser->lasttok.str - param->content.str;
    UDM_EXCERPT_FRAGMENT_QUALITY *q= UdmExcerptFragmentQualityListFind(&param->Quality, offset);
    /*printf("HERE: q=%d\n", q ? q->quality : DEFAULT_FRAGMENT_QUALITY);*/
    UdmSnippetHlAppendUnicode(param, &param->snippet,
                              q ? q->quality : DEFAULT_FRAGMENT_QUALITY,
                              parser->lasttok.str,  parser->lasttok.length);
  }
  else
  {
    UdmSnippetAppendSpaceUnicode(&param->snippet);
  }
  return UDM_OK;
}


static udm_rc_t
UdmGetExcerptSourceFromHTMLContent(UDM_EXCERPT_HELPER *param,
                                   const char *src, size_t srclen)
{
  UDM_HTML_PARSER tag;
  UdmHTMLParserInit(&tag);
  UdmHTMLParserSetUserData(&tag, param);
  UdmHTMLParserSetTextHandler(&tag, UdmGetExcerptSourceFromHTMLContentTextHandler);
  UdmHTMLParserSetTagHandler(&tag, UdmGetExcerptSourceFromHTMLContentTagHandler);
  UdmHTMLParserSetCommentHandler(&tag, UdmGetExcerptSourceFromHTMLContentCommentHandler);
  return UdmHTMLParserExec(&tag, src, srclen);
}


static void
UdmTextListToVarList(const UDM_ENV *Env,
                     UDM_CONV *cnv, UDM_TEXTLIST *TextList, UDM_VARLIST *Vars)
{
  size_t i;
  for (i= 0; i < TextList->nitems; i++)
  {
    UDM_TEXTITEM *Item= &TextList->Item[i];
    UDM_VAR *Var= UdmVarListFindOrCreateStr(Vars, Item->section_name);
    if (Var && UdmVarValueDataType(Var) == UDM_VALUE_DATA_TYPE_STR)
    {
      size_t maxlen= UdmEnvSectionMaxLength(Env, UdmVarSecno(Var));
      if (UdmVarLength(Var))
        UdmVarAppendStrn(Var, " ", 1);
      UdmVarAppendConv(Var, maxlen, cnv, UDM_RECODE_HTML,
                       Item->str, strlen(Item->str));
    }
  }
}


/*
  Add sections collected from cached copy to sections
  loaded from the "urlinfo" table.
  At this point UdmConvert() was already called for helper->Res.
  So WWList and Doc->Sections are already in "bcs" charset.
*/
static udm_rc_t
UdmMixURLInfoSectionsAndCachedSections(UDM_EXCERPT_HELPER *helper)
{
  size_t i;
  UDM_CONV dcs_lcs;
  UDM_VARLIST Vars;
  UDM_HIGHLIGHT_CONV ec;
  UDM_VARLIST *Sections= &helper->Doc->Sections;
  UDM_CHARSET *bcs= helper->Agent->Conf->bcs;
  UDM_CHARSET *dcs= UdmVarListFindCharset(Sections, "Cached.Charset",
                                          UdmVarListFindCharset(Sections,
                                          "Charset",
                                          UdmVarListFindCharset(Sections,
                                          "Server-Charset", &udm_charset_latin1)));
  UdmConvInit(&dcs_lcs, dcs, bcs);
  UdmVarListInit(&Vars);
  UdmTextListToVarList(helper->Agent->Conf, &dcs_lcs,
                       &helper->Doc->TextList, &Vars);
  UdmExcerptConvInit(&ec, bcs, bcs, bcs,
                      helper->hconv_bcs.hlstop, helper->hconv_bcs.segmenter);
  UdmVarListHlConvert(helper->Agent, &helper->Query->Res.WWList, &Vars, &ec);
  /*
    Put variables collected from cached copy,
    but only those that are not previously loaded from "urlinfo".
  */
  for (i= 0; i < Vars.nvars; i++)
  {
    const UDM_VAR *v= UdmVarListFindConstByIndex(&Vars, i), *v1;
    if (!(v1= UdmVarListFind(Sections, UdmVarName(v))) ||
        (UdmVarValueDataType(v1) == UDM_VALUE_DATA_TYPE_STR &&
         UdmVarLength(v1) == 0))
      UdmVarListReplaceStr(Sections, UdmVarName(v), UdmVarStr(v));
  }
  UdmVarListFree(&Vars);
  return UDM_OK;
}


/*
  TODO34: remove this.
  Make excerpt before UdmConvert().
*/
static void
removehl(char *dst, size_t dstlen, const char *src)
{
  UDM_ASSERT(dstlen > 1);
  for ( ; dstlen > 1 && *src; dstlen--, src++)
  {
    if (*src != 0x02 && *src != 0x03)
      *dst++= *src;
  }
  *dst= '\0';
}


static udm_content_type_t
UdmCachedContentType(UDM_ENV *Env, UDM_DOCUMENT *Doc, UDM_CONST_STR *content)
{
  const char *cached_content_type= UdmVarListFindStr(&Doc->Sections, "Cached.Content-Type", NULL);
  udm_content_type_t ct;
  char ctbuf[128]= "";
  if (cached_content_type)
    removehl(ctbuf, sizeof(ctbuf), cached_content_type);
  if (!cached_content_type || !(ct= UdmContentTypeByName(ctbuf)))
  {
    udm_content_type_t deftype= UdmVarListFindContentType(&Env->Vars, "DefaultContentType", UDM_CONTENT_TYPE_UNKNOWN);
    ct= UdmGuessContentType(content->str, content->length, deftype);
  }
  return ct;
}


static udm_rc_t
UdmExcerptFragmentQualityAddWild(UDM_EXCERPT_HELPER *helper,
                                 const UDM_CONST_STR *pattern,
                                 const UDM_CONST_STR *content,
                                 int quality)
{ 
  static UDM_WILD_PARAM param=
  {
    '?',  /* one  */
    '*',  /* many */
    0     /* eol  */
  };
  UDM_MATCH_PART Part;
  if (UdmWildCmpSubstr(content->str, content->str + content->length,
                       pattern->str, pattern->str + pattern->length,
                       &param, &Part))
    return UDM_ERROR;
  return UdmExcerptFragmentQualityListAdd(&helper->Quality, &Part, quality);
}


static udm_rc_t
UdmExcerptFragmentAdd(UDM_EXCERPT_HELPER *helper,
                      UDM_EXCERPT_FRAGMENT *Fragment,
                      const UDM_CONST_STR *content)
{
  switch (Fragment->Match.Param.match_mode)
  {
  case UDM_MATCH_WILD:
    {
      UDM_CONST_STR pattern;
      UdmConstStrSet(&pattern, UdmMatchPatternConstStr(&Fragment->Match),
                               UdmMatchPatternLength(&Fragment->Match));
      return UdmExcerptFragmentQualityAddWild(helper, &pattern, content, Fragment->quality);
    }
    break;
  case UDM_MATCH_REGEX:
    {
      UDM_MATCH_PART Part[10];
      if (UdmMatchExec(&Fragment->Match, content->str, content->length, NULL, 10, Part))
        return UDM_ERROR;
      return UdmExcerptFragmentQualityListAdd(&helper->Quality, &Part[0], Fragment->quality);
    }
    break;
  case UDM_MATCH_FULL:
  case UDM_MATCH_BEGIN:
  case UDM_MATCH_SUBSTR:
  case UDM_MATCH_END:
  case UDM_MATCH_SUBNET:
  case UDM_MATCH_NUMERIC_LT:
  case UDM_MATCH_NUMERIC_GT:
  case UDM_MATCH_RANGE:
    break;
  }
  return UDM_OK;
}


static void
UdmExcerptFragmentQualityInit(UDM_EXCERPT_HELPER *helper,
                              const UDM_CONST_STR *content)
{
  size_t i;
  UDM_VARLIST *Vars= &helper->Agent->Conf->Vars;
  UdmExcerptFragmentQualityListInit(&helper->Quality);
  for (i= 0; i < Vars->nvars; i++)
  {
    UDM_VAR *Var= UdmVarListFindByIndex(Vars, i);
    if (UdmVarValueDataType(Var) == UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT)
    {
      UDM_EXCERPT_FRAGMENT *Specific= (UDM_EXCERPT_FRAGMENT *) UdmVarDataPtr(Var);
      UdmExcerptFragmentAdd(helper, Specific, content);
    }
  }
}


static size_t
UdmGetExcerptSourceFromContent(UDM_EXCERPT_HELPER *helper)
{
  UdmSnippetReset(&helper->snippet);
  if (UdmHTTPBufContentToConstStr(&helper->Doc->Buf, &helper->content))
    return 0;

  UdmExcerptFragmentQualityInit(helper, &helper->content);

  switch (UdmCachedContentType(helper->Agent->Conf, helper->Doc, &helper->content))
  {
    case UDM_CONTENT_TYPE_TEXT_XML:
    {
      UdmGetExcerptSourceFromXMLContent(helper, helper->content.str,
                                                helper->content.length);
      UdmSnippetHlAppendUnicode(helper, &helper->snippet,
                                DEFAULT_FRAGMENT_QUALITY,
                                UdmDSTRPtr(&helper->body),
                                UdmDSTRLength(&helper->body));
      break;
    }

    case UDM_CONTENT_TYPE_TEXT_HTML:
    {
      UdmGetExcerptSourceFromHTMLContent(helper, helper->content.str,
                                                 helper->content.length);
      UdmMixURLInfoSectionsAndCachedSections(helper);
      break;
    }

    case UDM_CONTENT_TYPE_MESSAGE_RFC822:
    {
      /*
        We need a new UDM_EXCERPT_HELPER because document charset
        may change afrer parsing the message.
      */
      UDM_EXCERPT_HELPER helper2;
      UDM_CHARSET *dcs;

      UdmMessageRFC822ExcerptSource(helper->Agent, helper->Query, helper->Doc,
                                    &helper->content, &helper->body);

      dcs= UdmVarListFindCharset(&helper->Doc->Sections, "Charset", &udm_charset_latin1);

      if (UDM_OK != UdmExcerptHelperInit(&helper2, helper->Agent,
                                         helper->Query, helper->Doc, dcs))
        return 0;

      UdmSnippetHlAppendUnicode(&helper2, &helper->snippet,
                                DEFAULT_FRAGMENT_QUALITY,
                                UdmDSTRPtr(&helper->body),
                                UdmDSTRLength(&helper->body));
      UdmMixURLInfoSectionsAndCachedSections(&helper2);
      UdmExcerptHelperFree(&helper2);
      break;
    }

    case UDM_CONTENT_TYPE_DOCX:
    {
      UDM_EXCERPT_HELPER helper2;
      UdmDOCXExcerptSource(helper->Agent, helper->Query,
                           helper->Doc, &helper->content, &helper->body);
      if (UDM_OK != UdmExcerptHelperInit(&helper2, helper->Agent,
                                         helper->Query, helper->Doc,
                                         &udm_charset_utf8))
        return 0;
      UdmSnippetHlAppendUnicode(&helper2, &helper->snippet,
                                DEFAULT_FRAGMENT_QUALITY,
                                UdmDSTRPtr(&helper->body),
                                UdmDSTRLength(&helper->body));
      UdmExcerptHelperFree(&helper2);
      break;
    }

    case UDM_CONTENT_TYPE_HTDB:
    {
      UDM_EXCERPT_HELPER helper2;
      UdmHTDBExcerptSource(helper->Agent, helper->Query,
                           helper->Doc, &helper->content, &helper->body);
      if (UDM_OK != UdmExcerptHelperInit(&helper2, helper->Agent,
                                         helper->Query, helper->Doc,
                                         &udm_charset_utf8))
        return 0;
      UdmSnippetHlAppendUnicode(&helper2, &helper->snippet,
                                DEFAULT_FRAGMENT_QUALITY,
                                UdmDSTRPtr(&helper->body),
                                UdmDSTRLength(&helper->body));
      UdmExcerptHelperFree(&helper2);
      break;
    }

    case UDM_CONTENT_TYPE_TEXT_RTF:
    {
      UDM_CHARSET *dcs;
      UDM_EXCERPT_HELPER helper2;
      UdmRTFExcerptSource(helper->Agent, helper->Query,
                          helper->Doc, &helper->content, &helper->body);
      dcs= UdmVarListFindCharset(&helper->Doc->Sections, "Meta-Charset",
                                 &udm_charset_latin1);
      if (UDM_OK != UdmExcerptHelperInit(&helper2, helper->Agent,
                                         helper->Query, helper->Doc, dcs))
        return 0;
      UdmSnippetHlAppendUnicode(&helper2, &helper->snippet,
                                DEFAULT_FRAGMENT_QUALITY,
                                UdmDSTRPtr(&helper->body),
                                UdmDSTRLength(&helper->body));
      UdmExcerptHelperFree(&helper2);      
      break;
    }

    case UDM_CONTENT_TYPE_TEXT_PLAIN:
      UdmSnippetHlAppendUnicode(helper, &helper->snippet,
                                DEFAULT_FRAGMENT_QUALITY,
                                helper->content.str, helper->content.length);
      break;

    case UDM_CONTENT_TYPE_AUDIO_MPEG:
    case UDM_CONTENT_TYPE_UNKNOWN:
      break;
  }
  return UdmDSTRLength(&helper->snippet.body_unicode);
}


static size_t
UdmGetExcerptSourceCachedCopy(UDM_EXCERPT_HELPER *helper)
{
  if (UdmHTTPBufSize(&helper->Doc->Buf))
    return UdmGetExcerptSourceFromContent(helper);
  return 0;
}


static size_t
UdmGetExcerptSourceLocalCopy(UDM_EXCERPT_HELPER *helper)
{
  if (UdmVarListFindBool(&helper->Agent->Conf->Vars, "UseLocalCachedCopy", UDM_FALSE))
  {
    const char *url= UdmVarListFindStr(&helper->Doc->Sections, "url", NULL);
    udm_rc_t rc= url ? UdmGetURLSimple(helper->Agent, helper->Doc, url) : UDM_ERROR;
    if (0)
    {
      const char *cl= UdmVarListFindStr(&helper->Doc->Sections, "Content-Length", "?");
      const char *ct= UdmVarListFindStr(&helper->Doc->Sections, "Content-Type", "?");
      const char *ce= UdmVarListFindStr(&helper->Doc->Sections, "Content-Encoding", "?");
      fprintf(stderr, "rc=%d url=%s size=%d cl=%s ce=%s ct=%s\n",
              (int) rc, url, (int) UdmHTTPBufSize(&helper->Doc->Buf), cl, ce, ct);
    }
    return UdmGetExcerptSourceFromContent(helper);
  }
  return 0;
}


static size_t
UdmGetExcerptSourceBody(UDM_EXCERPT_HELPER *helper)
{
  const char *Source= UdmVarListFindStr(&helper->Doc->Sections, "body", NULL);
  UDM_CHARSET *sys_int= &udm_charset_sys_int;
  UDM_CONV conv;
  if (!Source || !helper->Agent->Conf->bcs)
    return 0;
  UdmConvInit(&conv, helper->Agent->Conf->bcs, sys_int);
  UdmDSTRAppendConv(&helper->snippet.body_unicode, &conv,
                    Source, strlen(Source), UDM_RECODE_HTML);
  return UdmDSTRLength(&helper->snippet.body_unicode);
}

static int dots[]= { 0x2E, 0x2E, 0x2E }; /* "..." */


static void
make_excerpt(const int *Source, size_t offset, size_t ul, UDM_DSTR *buf,
             size_t ExcerptPadding, size_t ExcerptSize)
{
  size_t i, prev_right= 0;

  /* Get excerpt */
  for (i= offset; i < ul; i++)
  {
    if (Source[i] == 2) /* Beginning of a highlighted area found */
    {
      size_t j, left, right;
      for (j= i + 1; j < ul; j++) /* Find the end of highighted area */
        if (Source[j] == 3)
          break;
      
      /*
        Move towards the beginning of the document
        and find where to break excerpt:
        - either on the closest space character
        - or put all "ExcerptPadding" characters if no space found.
          This is very likely in case of Asian texts with no spaces.
      */
      left= ExcerptPadding < i ? i - ExcerptPadding : 0;
      if (left < prev_right)
      {
        left= prev_right;
      }
      if (left)
      {
        size_t left0= left;
        for ( ; Source[left] != 0x20; left++)
        {
          if (left >= i) /* no space found */
          {
            left= left0;
            break;
          }
        }
      }
      
      /*
        Now move towards the end of the document and
        find the right excerpt border:
        - either on the closest space character
        - or after ExcerptSize characters
      */
      right= ExcerptPadding + j;
      if (right >= ul)
      {
        right= ul - 1;
      }
      else
      {
        size_t right0= right;
        for ( ; Source[right] != 0x20 ;  right--)
        {
          if (right <= j) /* No space found */
          {
            right= right0;
            /*
              Do not break excerpt in the middle of a highlighted word.
            */
            for ( ; j <  right; j++)
            {
              if (Source[j] == 0x02)
              {
                right= j - 1;
                break;
              }
            }
            break;
          }
        }
      }
      
      /*
        Check if there is a space for the new excerpt.
        We must stay under ExcerptSize limit.
      */
      if (ExcerptSize < UdmDSTRLength(buf) / sizeof(int) + right - left + 1)
        break;
      if (left != prev_right)
        UdmDSTRAppend(buf, (char *) dots, sizeof(dots));
      UdmDSTRAppend(buf, (const char *) &Source[left], (right - left + 1) * sizeof(int));
      i= right;

       /*
         On the next step we should not add dots if the new highlighted word
         immediately follows this word: <02>W1<03><02><W2>
         This is possible in case of "Segmenter cjk".
       */
      if (Source[right] == 0x03)
        right++;
      prev_right= right;
    }
  }
}


char *
UdmExcerptDoc(UDM_AGENT *Agent, UDM_QUERY *Query, UDM_DOCUMENT *Doc,
              size_t ExcerptSize, size_t ExcerptPadding)
{
  size_t ul, l, offset= 0;
  char *res= NULL;
  UDM_CONV conv;
  UDM_CHARSET *sys_int= &udm_charset_sys_int;
  UDM_CHARSET *bcs= Agent->Conf->bcs;
  UDM_DSTR buf;
  UDM_EXCERPT_HELPER helper;
  UDM_CHARSET *dcs= UdmVarListFindCharset(&Doc->Sections, "Cached.Charset",
                                          UdmVarListFindCharset(&Doc->Sections,
                                          "Charset",
                                          UdmVarListFindCharset(&Doc->Sections,
                                          "Server-Charset", &udm_charset_latin1)));
#ifdef EXTRA_DEBUG
  unsigned long ticks= UdmStartTimer();
#endif

  if (!bcs || UDM_OK != UdmExcerptHelperInit(&helper, Agent, Query, Doc, dcs))
    return NULL;

  /* Get excerpt source */
  if (!(UdmGetExcerptSourceLocalCopy(&helper)) &&
      !(UdmGetExcerptSourceCachedCopy(&helper)) &&
      !(UdmGetExcerptSourceBody(&helper)))
    goto ex;

/*
  fprintf(stderr, "HERE: nchunks=%d best=%d\n",
     (int) helper.snippet.chunks.nitems,
     (int) helper.snippet.best_chunk);

  {
    size_t i;
    int *dst=  (int*) helper.snippet.body_unicode.Val.str;
    size_t nchars_total= helper.snippet.body_unicode.Val.length / sizeof(int);
    for (i= 0; i < nchars_total; i++)
    {
      fprintf(stderr, "HERE: '%c' %02X\n", dst[i], dst[i]);
    }
  }
*/

  if (UdmVarListFindBool(&Agent->Conf->Vars, "NewExcerpt", UDM_TRUE) &&
      helper.snippet.chunks.nitems)
  {
    UDM_SNIPPETCHUNK *chunk= &helper.snippet.chunks.Item[helper.snippet.best_chunk];
    size_t nchars_total= UdmDSTRLength(&helper.snippet.body_unicode) / sizeof(int);
    size_t nchars_before= chunk->offset / sizeof(int);
    size_t nchars_after= nchars_total - nchars_before;
    int *dst= (int*) helper.snippet.body_unicode.Val.str;
    /*
    fprintf(stderr, "HERE: ofs/len=%d/%d bits=%llu\n",
           (int) chunk->offset, (int) chunk->length,
           chunk->found_mask);
    */

    /* Copy the part before the best chunk */
    if (nchars_before)
    {
      offset= UdmUniStrStripSpaces(dst, dst, nchars_before);
      if (offset > 0 && offset < nchars_before)
        dst[offset++]= ' ';
    }
    /* Copy the best chunk and the rest */
    ul= offset + UdmUniStrStripSpaces(dst + offset, dst + nchars_before,
                                             nchars_after);
  }
  else
  {
    ul= UdmUniStrStripSpaces((int*) helper.snippet.body_unicode.Val.str,
                             (int*) helper.snippet.body_unicode.Val.str,
                             UdmDSTRLength(&helper.snippet.body_unicode) / sizeof(int));
  }


#ifdef EXTRA_DEBUG
  UdmLog(Agent, UDM_LOG_ERROR, "GetSour: %.5f ul=%d", UdmStopTimer(&ticks), ul);
  ticks= UdmStartTimer();
#endif

  if (UDM_OK != UdmDSTRInit(&buf, 1024))
    goto ex;

  make_excerpt((const int*) UdmDSTRPtr(&helper.snippet.body_unicode),
                offset, ul, &buf,
               ExcerptPadding, ExcerptSize);
  if (!UdmDSTRLength(&buf))
  {
    ul= ExcerptSize > ul ? ul : ExcerptSize;
    UdmDSTRAppend(&buf, UdmDSTRPtr(&helper.snippet.body_unicode), ul * sizeof(int));
  }

  ul= UdmDSTRLength(&buf) / sizeof(int) * 20;
  if (!ul || !(res= (char*) UdmMalloc(ul))) 
    goto ex2;

  /* Convert Unicode to BrowserCharset */
  UdmConvInit(&conv, sys_int, bcs);
  l= UdmConvHTML(&conv, res, ul, UdmDSTRPtr(&buf), UdmDSTRLength(&buf));
  res[l]= 0;

  /*
    Remove trailing spaces to have equal results from cluster searches
    and from direct searches. Cluster searches trim spaces, as they use XML.
  */
  while (l > 0 && res[l - 1] == ' ')
  {
    l--;
    res[l]= '\0';
  }

#ifdef EXTRA_DEBUG
  UdmLog(Agent, UDM_LOG_ERROR, "Excerpt: %.5f l=%d", UdmStopTimer(&ticks), l);
#endif  

ex2:
  UdmDSTRFree(&buf);
ex:
  UdmExcerptHelperFree(&helper);
  return res;
}
