/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_spell.h"
#include "udm_hrefs.h"
#include "udm_utils.h"
#include "udm_xmalloc.h"
#include "udm_sgml.h"
#include "udm_url.h"
#include "udm_vars.h"

/* Max URLs in cache: 4K URLs will use about 200K of RAM         */
/* This should be a configurable parameter but we'll use 4K now  */

#define HSIZE		256	/* Length of buffer increment  TUNE */
#define RESORT_HREFS	256	/* Max length of unsorted part TUNE */


void
UdmHrefParamInit(UDM_HREFPARAM *H)
{
  bzero((void*)H, sizeof(*H));
}


void
UdmHrefInit(UDM_HREF * H)
{
  bzero((void*)H, sizeof(*H));
}


void
UdmHrefFree(UDM_HREF * H)
{
  if (H->url)
    UdmFree(H->url);
  UdmVarListFree(&H->HrefVars);
}


/* Function to sort URLs in alphabetic order */
static int
cmphrefs(const void * v1, const void * v2)
{
  return(strcmp(((const UDM_HREF*)v1)->url,((const UDM_HREF*)v2)->url));
}


static void
UdmHrefCopyParam(UDM_HREFPARAM *Dst, const UDM_HREFPARAM *Src)
{
  *Dst= *Src;
}


/*
  Find URL in the sorted part of HrefList
*/
static UDM_HREF *
UdmHrefListFindSorted(UDM_HREFLIST *HrefList, const char *ehref)
{
  int l, r, c;
  for (l=0, r= HrefList->shrefs - 1; l <= r; )
  {
    int res;
    c= (l + r) / 2;
    if (!(res= strcmp(HrefList->Href[c].url, ehref)))
      return &HrefList->Href[c];
    if (res < 0)
      l= c + 1;
    else
      r= c - 1;
  }
  return NULL;
}


/*
  Find URL in the unsorted part of HrefList
*/
static UDM_HREF *
UdmHrefListFindUnsorted(UDM_HREFLIST *HrefList, const char *ehref)
{
  int c;
  for(c= HrefList->shrefs; c < HrefList->nhrefs; c++)
  {
    if (!strcmp(HrefList->Href[c].url, ehref))
      return &HrefList->Href[c];
  }
  return NULL;
}


static udm_rc_t
UdmHrefListAddInternal(UDM_HREFLIST *HrefList,
                       const UDM_HREFPARAM *HrefParam,
                       const UDM_VARLIST *HrefVars,
                       const UDM_CONST_STR *url)
{
  char *ehref;
  UDM_HREF *dst;

  /* Don't add an empty link */
  if (!url->str || !(ehref= (char*) UdmMalloc(3 * url->length + 1)))
    return UDM_ERROR;

  UdmURLCanonize(url->str, ehref, 3 * url->length + 1);
  UdmSGMLUnescape(ehref);

  /* Find current URL in the sorted part, then in the unsorted part */
  if (!(HrefList->flags & UDM_HREFLIST_FLAG_NONUNIQ) &&
      ((dst= UdmHrefListFindSorted(HrefList, ehref)) ||
       (dst= UdmHrefListFindUnsorted(HrefList, ehref))))
  {
    UdmHrefCopyParam(&dst->Param, HrefParam);
    UdmVarListFree(&dst->HrefVars);
    UdmVarListInit(&dst->HrefVars);
    if (HrefVars)
      UdmVarListReplaceLst(&dst->HrefVars, HrefVars, NULL, "*");
    UDM_FREE(ehref);
    return UDM_OK;
  }

  if(HrefList->nhrefs >= HrefList->mhrefs)
  {
    HrefList->mhrefs+= HSIZE;
    HrefList->Href= (UDM_HREF *) UdmRealloc(HrefList->Href,
                                            HrefList->mhrefs * sizeof(UDM_HREF));
    if (!HrefList->Href) /* Fatal */
    {
      UdmHrefListInit(HrefList);
      return UDM_ERROR;
    }
  }

  dst= &HrefList->Href[HrefList->nhrefs];
  dst->url= (char*) UdmStrdup(ehref);
  UdmHrefCopyParam(&dst->Param, HrefParam);
  UdmVarListInit(&dst->HrefVars);
  if (HrefVars)
    UdmVarListReplaceLst(&dst->HrefVars, HrefVars, NULL, "*");
  HrefList->nhrefs++;

  /* Sort unsorted part */
  if (!(HrefList->flags & UDM_HREFLIST_FLAG_NONUNIQ) &&
      HrefList->nhrefs-HrefList->shrefs > RESORT_HREFS)
  {
    UdmSort(HrefList->Href, HrefList->nhrefs, sizeof(UDM_HREF), cmphrefs);
    /* Remember count of sorted URLs  */
    HrefList->shrefs= HrefList->nhrefs;
  }
  UDM_FREE(ehref);
  return UDM_OK;
}


udm_rc_t
UdmHrefListAddConstStr(UDM_HREFLIST *HrefList,
                       const UDM_HREFPARAM *H,
                       const UDM_CONST_STR *href)
{
  /*
  udm_rc_t rc;
  UDM_CONST_STR str;
  char *tmp= UdmConstStrDup(href);
  if (tmp)
    return UDM_ERROR;
  str.str= tmp;
  str.length= href->length;
  rc= UdmHrefListAddInternal(HrefList, H, NULL, &str);
  UdmFree(tmp);
  return rc;
  */
  return UdmHrefListAddInternal(HrefList, H, NULL, href);
}


udm_rc_t
UdmHrefListAddConstStrSGMLUnescape(UDM_HREFLIST *HrefList,
                                   const UDM_HREFPARAM *Template,
                                   const UDM_VARLIST *Vars,
                                   const UDM_CONST_STR *href)
{
  udm_rc_t rc;
  UDM_CONST_STR Str;
  char *tmp;
  if (!(tmp= UdmConstStrDup(href)))
    return UDM_ERROR;
  UdmSGMLUnescape(tmp);
  Str.str= tmp;
  Str.length= strlen(tmp);
  rc= UdmHrefListAddInternal(HrefList, Template, Vars, &Str);
  UdmFree(tmp);
  return rc;
}


UDM_API(udm_rc_t)
UdmHrefListAdd(UDM_HREFLIST * HrefList, const UDM_HREF *H)
{
  UDM_CONST_STR str;
  UdmConstStrSetStr(&str, H->url);
  return UdmHrefListAddInternal(HrefList, &H->Param, &H->HrefVars, &str);
}


udm_rc_t
UdmHrefListAddConst(UDM_HREFLIST * HrefList,
                    const UDM_HREFPARAM *H,
                    const char *href)
{
  UDM_CONST_STR str;
  UdmConstStrSetStr(&str, href);
  return UdmHrefListAddInternal(HrefList, H, NULL, &str);
}


UDM_API(void)
UdmHrefListFree(UDM_HREFLIST * HrefList)
{
  size_t i;
  for(i=0;i<HrefList->nhrefs;i++)
    UdmHrefFree(&HrefList->Href[i]);
  UDM_FREE(HrefList->Href);
  bzero((void*)HrefList, sizeof(*HrefList));
}


UDM_API(void)
UdmHrefListInit(UDM_HREFLIST * Hrefs)
{
  bzero((void*)Hrefs, sizeof(*Hrefs));
}


const char *
UdmLinkSourceStr(udm_link_source_t link_source)
{
  switch (link_source)
  {
    case UDM_LINK_SOURCE_UNKNOWN:         return "unk";
    case UDM_LINK_SOURCE_CMDLINE:         return "cline";
    case UDM_LINK_SOURCE_CONF:            return "conf";
    case UDM_LINK_SOURCE_A_HREF:          return "a";
    case UDM_LINK_SOURCE_FRAME_SRC:       return "frame";
    case UDM_LINK_SOURCE_IMG_SRC:         return "img";
    case UDM_LINK_SOURCE_REDIRECT:        return "redir";
    case UDM_LINK_SOURCE_META_REFRESH:    return "meta";
    case UDM_LINK_SOURCE_LINK_HREF:       return "link";
    case UDM_LINK_SOURCE_AREA_HREF:       return "area";
    case UDM_LINK_SOURCE_IFRAME_SRC:      return "iframe";
    case UDM_LINK_SOURCE_SCRIPT_SRC:      return "script";
    case UDM_LINK_SOURCE_HTDB:            return "htdb";
    case UDM_LINK_SOURCE_URLFILE:         return "ufile";
    case UDM_LINK_SOURCE_ROBOTS_SITEMAP:  return "robots";
    case UDM_LINK_SOURCE_XML:             return "xml";
    case UDM_LINK_SOURCE_DIR:             return "dir";
  }
  return "unknown";
}


udm_link_source_t
UdmLinkSourceByName(const char *str)
{
  if (!strcasecmp(str, "cline"))
    return UDM_LINK_SOURCE_CMDLINE;
  if (!strcasecmp(str, "conf"))
    return UDM_LINK_SOURCE_CONF;
  if (!strcasecmp(str, "a"))
    return UDM_LINK_SOURCE_A_HREF;
  if (!strcasecmp(str, "frame"))
    return UDM_LINK_SOURCE_FRAME_SRC;
  if (!strcasecmp(str, "img"))
    return UDM_LINK_SOURCE_IMG_SRC;
  if (!strcasecmp(str, "redir"))
    return UDM_LINK_SOURCE_REDIRECT;
  if (!strcasecmp(str, "meta"))
    return UDM_LINK_SOURCE_META_REFRESH;
  if (!strcasecmp(str, "link"))
    return UDM_LINK_SOURCE_LINK_HREF;
  if (!strcasecmp(str, "area"))
    return UDM_LINK_SOURCE_AREA_HREF;
  if (!strcasecmp(str, "iframe"))
    return UDM_LINK_SOURCE_IFRAME_SRC;
  if (!strcasecmp(str, "script"))
    return UDM_LINK_SOURCE_SCRIPT_SRC;
  if (!strcasecmp(str, "htdb"))
    return UDM_LINK_SOURCE_HTDB;
  if (!strcasecmp(str, "ufile"))
    return UDM_LINK_SOURCE_URLFILE;
  if (!strcasecmp(str, "robots"))
    return UDM_LINK_SOURCE_ROBOTS_SITEMAP;
  if (!strcasecmp(str, "xml"))
    return UDM_LINK_SOURCE_XML;
  return UDM_LINK_SOURCE_UNKNOWN;
}
