/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#ifdef WIN32
#include <process.h>
#endif

#ifdef HAVE_READLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif

#ifdef CHASEN
#include <chasen.h>
#endif

#ifdef MECAB
#include <mecab.h>
#endif

#include "udmsearch.h"
#include "udm_sqldbms.h"  /* TODO34: Remove this */
#include "udm_db_int.h"   /* TODO34: Remove this */
#include "udm_query.h"    /* TODO34: Remove this */
#include "udm_http.h"

/* This should be last include */
#ifdef DMALLOC
#include "dmalloc.h"
#endif

static int loglevel= UDM_LOG_INFO;
static char cname[1024]= "";

static UDM_ENV Conf;

extern unsigned int seconds;   /* To sleep between documents   */
extern int flags;              /* For indexer                  */
extern int total_threads;      /* Total threads number         */
extern int sleep_threads;      /* Number of sleepping threads  */
extern int max_index_time;
extern int maxthreads;
extern UDM_CRAWLER *ThreadCrawlers;
extern int thd_errors;

#ifdef HAVE_PTHREAD
static udm_thread_t *threads= NULL;

#ifdef WIN32
#include <time.h>
static struct tm*
localtime_r(const time_t *clock, struct tm *result)
{
  *result= *localtime(clock);
  return result;
}
#define strptime(x, y, z) (0)
#endif /* WIN32 */
#endif /* HAVE_PTHREAD */


/* CallBack function for log information */
#ifdef WIN32
void UdmShowInfo(UDM_AGENT* A, const char *state, const char* str)
{
  printf("%d %s %s\n", A ? A->handle : 0,state,str);
}
#else
extern UDM_API(void) UdmShowInfo(UDM_AGENT* A, const char *state, const char* str);
#endif /* WIN32 */


static const char *execsql= NULL;
static char *
sqlexecgets(UDM_IOHANDLER *iohandler, char *str, size_t len)
{
  UDM_SQLMON_PARAM *prm= (UDM_SQLMON_PARAM*) iohandler->user_data;
  if (!execsql)
    return NULL;
  udm_snprintf(str, len, "%s", execsql);
  prm->flags|= UDM_SQLMON_DONT_NEED_SEMICOLON;
  execsql= NULL;
  return str;
}


static char* sqlmongets(UDM_IOHANDLER *iohandler, char *str, size_t size)
{
  UDM_SQLMON_PARAM *prm= (UDM_SQLMON_PARAM*) iohandler->user_data;
#ifdef HAVE_READLINE
  if (isatty(0))
  {
     char prompt[]="SQL>";
     char *line= readline(prompt);
     if (!line)
       return 0;
     
     if (*line) add_history(line);
     /* We need "\n" at the end to make sqlmon work properly */
     udm_snprintf(str, size, "%s\n", line);
  }
  else
#endif
  {
    if (loglevel >= UDM_LOG_INFO)
      prm->iohandler.prompt(&prm->iohandler, UDM_MSG_PROMPT, "SQL>");
    if (!fgets(str, size, stdin))
      return 0;
  }
  return str;
}


static udm_rc_t
sqlmonprompt(UDM_IOHANDLER *iohandler, udm_msg_t msgtype, const char *msg)
{
  UDM_SQLMON_PARAM *prm= (UDM_SQLMON_PARAM*) iohandler->user_data;
  switch (msgtype)
  {
    case UDM_MSG_ERROR:
      if (prm->mode == udm_sqlmon_mode_batch)
        fprintf(stdout, "ERROR at line %d: %s\n", (int) prm->lineno + 1, msg); 
      else
        fprintf(stdout, "ERROR: %s\n", msg);
      break;
    case UDM_MSG_INFO:
      if (loglevel >= UDM_LOG_INFO)
        fprintf(stdout, "%s\n", msg);
      break;
    case UDM_MSG_COMMAND:
      if (loglevel >= UDM_LOG_INFO)
        fprintf(stdout, "'%s'\n", msg);
      break;
    case UDM_MSG_EOL:
      fprintf(stdout, "\n");
      break;
    case UDM_MSG_DATA:
    case UDM_MSG_PROMPT:
      fprintf(stdout, "%s", msg);
      break;
  }
  return UDM_OK;
}


static udm_rc_t
UdmIndCreate(UDM_AGENT *Agent)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmDBAction(Agent, UDM_DBCMD_CREATE)))
    UdmLog(Agent, UDM_LOG_ERROR, "Error: '%s'", UdmEnvErrMsg(Agent->Conf));
  return rc;
}


static udm_rc_t
UdmIndDrop(UDM_AGENT *Agent)
{
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmDBAction(Agent, UDM_DBCMD_DROP)))
    UdmLog(Agent, UDM_LOG_ERROR, "Error: '%s'", UdmEnvErrMsg(Agent->Conf));
  return rc;
}


static udm_rc_t
ShowStatistics(UDM_AGENT *Indexer, UDM_QUERY *Query)
{
  udm_rc_t rc;
  struct tm tm;
  const char *stat_time;
  char sbuf[32];
  size_t snum;
  UDM_STAT  Total;

  bzero((void*)&Total, sizeof(Total));
  Query->StatList.time= time(NULL);
  stat_time = UdmVarListFindStr(&Conf.Vars, "stat_time", "0");
  bzero(&tm, sizeof(tm));

#ifndef WIN32
  if (stat_time &&
      strlen(stat_time) >= 7 &&
      stat_time[4] == '-' &&
      (stat_time[7] == '-' || !stat_time[7]) &&
      (strptime(stat_time, "%Y-%m-%d %H:%M:%S", &tm) ||
       strptime(stat_time, "%Y-%m-%d %H:%M", &tm) ||
       strptime(stat_time, "%Y-%m-%d %H:%M", &tm) ||
       strptime(stat_time, "%Y-%m-%d %H", &tm) ||
       strptime(stat_time, "%Y-%m-%d", &tm) ||
       strptime(stat_time, "%Y-%m", &tm)))
  {
    Query->StatList.time = mktime(&tm);
  }
  else if (stat_time && (Query->StatList.time= Udm_dp2time_t(stat_time)) >= 0)
  {
    Query->StatList.time += time(NULL);
    localtime_r(&Query->StatList.time, &tm);
  }
  else
  {
    Query->StatList.time = time(NULL);
    localtime_r(&Query->StatList.time, &tm);
  }
#else
  {
    struct tm *tm1;
    Query->StatList.time= time(NULL);
    tm= *(tm1= localtime(&Query->StatList.time));
  }
#endif
  if (UDM_OK != (rc= UdmQueryAction(Indexer, Query, UDM_QUERYCMD_STATISTICS)))
  {
    UdmLog(Indexer, UDM_LOG_ERROR, "Error: '%s'", UdmEnvErrMsg(Indexer->Conf));
    goto ex;
  }

  strftime(sbuf, sizeof(sbuf), "%Y-%m-%d %H:%M:%S", &tm);
  printf("\n          Database statistics [%s]\n\n", sbuf);
  printf("%10s %10s %10s\n","Status","Expired","Total");
  printf("   -----------------------------\n");
  for (snum= 0; snum < Query->StatList.nstats; snum++)
  {
    UDM_STAT *S= &Query->StatList.Stat[snum];
    printf("%10d %10d %10d %s\n",S->status,S->expired,S->total,UdmHTTPErrMsg(S->status));
    Total.expired+=S->expired;
    Total.total+=S->total;
  }
  printf("   -----------------------------\n");
  printf("%10s %10d %10d\n","Total",Total.expired,Total.total);
  printf("\n");

ex:
  return rc;
}

/* CallBack Func for Referers*/
static void UdmRefProc(int code, const char *url, const char * ref)
{
     printf("%d %s %s\n",code,url,ref);
}

static udm_rc_t
ShowReferers(UDM_AGENT *Indexer, UDM_QUERY *Query)
{
  return UdmQueryAction(Indexer, Query, UDM_QUERYCMD_REFERERS);
}


#undef THINFO_TEST
#ifdef THINFO_TEST
/* CallBack function for Thread information */
void UdmShowThreadInfoProc(int handle,char *state, char* str)
{
  printf("%d %s %s\n",handle,state,str);
}
#endif


static int cmpgrp(const void *v1, const void *v2)
{
  int res;
  UDM_CHARSET *c1= (UDM_CHARSET*) v1;
  UDM_CHARSET *c2= (UDM_CHARSET*) v2;
  if ((res = strcasecmp(UdmCsGroup(c1), UdmCsGroup(c2))))
    return res;
  return strcasecmp(c1->name,c2->name);
}

static void display_charsets(FILE *file)
{
  UDM_CHARSET *cs=NULL;
  struct udm_cset_st c[100];
  size_t i=0;
  size_t n=0;
  int family=-1;
     
  for(cs=UdmGetCharSetByID(0) ; cs && cs->name ; cs++)
  {
    /* Skip not compiled charsets */
    if (cs->family != UDM_CHARSET_UNKNOWN)
      c[n++]=*cs;
  }
  fprintf(file,"\n%d charsets available:\n", (int) n);

  UdmSort(c,n,sizeof(UDM_CHARSET),&cmpgrp);
  for(i=0;i<n;i++)
  {
    if (family!=c[i].family)
    {
      fprintf(file, "\n%19s : ", UdmCsGroup(&c[i]));
      family=c[i].family;
    }
    fprintf(file,"%s ",c[i].name);
  }
  fprintf(file,"\n");
}


static int usage(int level, UDM_CMDLINE_OPT *options)
{
  FILE *file= stdout;
  fprintf(file, "\n");
  fprintf(file, "indexer from %s-%s-%s\n", PACKAGE, VERSION, UDM_DBTYPE);
  fprintf(file, "http://www.mnogosearch.org/ (C)1998-2015, LavTech Corp.\n");
  fprintf(file, "\n");
  fprintf(file, "Usage: indexer [OPTIONS]  [configfile]\n");

  UdmCmdLineOptionsPrint(options, stdout);
  
  fprintf(file, "\n");
  fprintf(file, "\n");
  fprintf(file, "Please post bug reports and suggestions at http://www.mnogosearch.org/bugs/\n");


  if (level>1)display_charsets(file);
  return(0);
}


/*
  Load indexer.conf and check if any DBAddr were given
*/
static udm_rc_t
UdmIndexerEnvLoad(UDM_AGENT *Indexer, const char *fname,int lflags)
{
  udm_rc_t rc;
  if (UDM_OK == (rc= UdmEnvLoad(Indexer, fname, lflags)))
  {
    if (Indexer->Conf->DBList.nitems == 0)
    {
      sprintf(Indexer->Conf->errstr, "Error: '%s': No required DBAddr commands were specified", fname);
      rc= UDM_ERROR;
    }
  }
  return rc;
}



/*
  Parse command line
*/
static int UdmARGC;
static char **UdmARGV;



typedef struct
{
  const char *url_filename;
  udm_indcmd_t cmd;
  udm_bool_t insert;
  udm_bool_t expire;
  udm_bool_t block;
  udm_bool_t help;
  udm_bool_t log2stderr;
  udm_bool_t have_loglevel;
  udm_bool_t warnings;

  int add_servers;
  int add_server_urls;
  int load_langmaps;
  int load_spells;
  int load_for_dump;

} UDM_INDEXER_OPTIONS;


static void
UdmIndexerOptionsInit(UDM_INDEXER_OPTIONS *io)
{
  io->cmd=    UDM_IND_CRAWL;
  io->insert= UDM_FALSE;
  io->expire= UDM_FALSE;
  io->block=  UDM_FALSE;
  io->help=   UDM_FALSE;
  io->url_filename= NULL;
  io->log2stderr= UDM_TRUE;
  io->have_loglevel= UDM_FALSE;

  io->add_servers=      UDM_FLAG_ADD_SERV;
  io->add_server_urls=  UDM_FLAG_ADD_SERVURL;
  io->load_langmaps=    UDM_FLAG_LOAD_LANGMAP;
  io->load_spells=      UDM_FLAG_SPELL;
  io->load_for_dump=    0;
  io->warnings= UDM_TRUE;
}


static void
UdmFeaturesPrint(const char *mask)
{
  UDM_VARLIST W;
  size_t i;
  UdmVarListInit(&W);
  UdmFeatures(&W);
  for (i= 0; i < W.nvars; i++)
  {
    const UDM_VAR *var= UdmVarListFindConstByIndex(&W, i);
    if (!UdmWildCaseCmp(UdmVarName(var), mask))
    {
      UDM_CONST_STR strbuf, *str= UdmVarGetConstStr(var, &strbuf);
      printf("%s:%.*s\n", UdmVarName(var), (int) str->length, str->str);
    }
  }
  UdmVarListFree(&W);
}

/* 
  Available new options:
  Capital letters:  B    GH JK M OP   T VXWXYZ
  Small   letters:           k             x z
  Digits:          123456789
*/
static int
UdmCmdLineHandleOption(void *user_data,
                       const UDM_CMDLINE_OPT *opt, const char *value)
{
  UDM_INDEXER_OPTIONS *io= (UDM_INDEXER_OPTIONS*) user_data;
  switch (opt->id)
  {
    case 'F':
    {
      UdmFeaturesPrint(value);
      exit(0);
    }
    case 'C':
      io->cmd= UDM_IND_DELETE;
      io->add_servers= io->load_langmaps= io->load_spells=0;
      break;
    case 'S':
      io->cmd= UDM_IND_STAT;
      io->add_servers= io->load_langmaps= io->load_spells= 0;
      break;
    case 'I':
      io->cmd= UDM_IND_REFERERS;
      io->add_servers= io->load_langmaps= io->load_spells= 0;
      break;
    case 'Q':
      io->cmd= UDM_IND_SQLMON;
      io->add_servers= io->load_langmaps= io->load_spells= 0;
      break;
    case UDM_IND_INDEX:
    case UDM_IND_CRAWL:
    case UDM_IND_CREATE:
    case UDM_IND_DROP:
    case UDM_IND_CHECKCONF:
    case UDM_IND_CONVERT:
    case UDM_IND_MULTI2BLOB:
    case UDM_IND_EXPORT:
    case UDM_IND_WRDSTAT:
    case UDM_IND_REWRITEURL:
    case UDM_IND_REWRITEPOP:
    case UDM_IND_HASHSPELL:
    case UDM_IND_DUMPSPELL:
    case UDM_IND_REWRITELIMITS:
    case UDM_IND_DUMPCONF:
    case UDM_IND_DUMPDATA:
    case UDM_IND_RESTOREDATA:
      io->cmd= (udm_indcmd_t) opt->id;
      break;
    case UDM_IND_EXECSQL:
      io->cmd= UDM_IND_SQLMON;
      UdmVarListReplaceStr(&Conf.Vars, "exec", value);
      io->add_servers= io->load_langmaps= io->load_spells= 0;
      break;
    case UDM_IND_SET0:
      UDM_ASSERT(opt->name && opt->name[0]);
      UdmVarListReplaceStr(&Conf.Vars, opt->name, value);
      break;
    case UDM_IND_SET:
      {
        const char *eq= strchr(value, '=');
        char name[80];
        if (eq && ((size_t)(eq - value)) < sizeof(name))
        {
          memcpy(name, value, eq - value);
          name[eq-value]= '\0';
          UdmVarListReplaceStr(&Conf.Vars, name, eq + 1);
        }
        else
          UdmVarListReplaceInt(&Conf.Vars, value, 1);
      }
      break;

    case 'q': 
      if (io->add_server_urls == 0) /* -q already given */
      {
        /*
          "indexer -qq" is given, do even faster start-up.
          Don't synchonize "Server" commands in indexer.conf
          with the "server" table content.
        */
        io->add_servers|= UDM_FLAG_DONT_ADD_TO_DB;
      }
      io->add_server_urls= 0;
      break;
    case 'l': io->log2stderr= UDM_FALSE; break;
    case 'a': io->expire= UDM_TRUE; break;
    case 'b': io->block= UDM_TRUE; break;
    case 'e': flags|=UDM_FLAG_SORT_EXPIRED;break;
    case 'o': flags|=UDM_FLAG_SORT_HOPS;break;
    case 'r': flags|=UDM_FLAG_DONTSORT_SEED; break;
    case 'm': flags|=UDM_FLAG_REINDEX;break;
    case 'n': Conf.url_number=atoi(value);break;
    case 'c': max_index_time=atoi(value);break;
    case 'v': loglevel= atoi(value); io->have_loglevel= UDM_TRUE; break;
    case 'p': seconds=atoi(value);break;
    case 't': UdmVarListAddStr(&Conf.Vars,"tag" , value);break;
    case 's': UdmVarListAddStr(&Conf.Vars, "status", value);break;
    case 'y': UdmVarListAddStr(&Conf.Vars,"type", value);break;
    case 'L': UdmVarListAddStr(&Conf.Vars,"lang", value);break;
    case 'u': UdmVarListAddStr(&Conf.Vars,"u"   , value);
      if (io->insert)
      {
        UDM_HREFPARAM HrefParam;
        UdmHrefParamInit(&HrefParam);
        HrefParam.link_source= UDM_LINK_SOURCE_CMDLINE;
        UdmHrefListAddConst(&Conf.Hrefs, &HrefParam, value);
      }
      break;
    case 'N':
      maxthreads=atoi(value);
      UdmVarListReplaceInt(&Conf.Vars, "CrawlerThreads", maxthreads);
      UdmVarListReplaceInt(&Conf.Vars, "IndexerThreads", maxthreads);
      break;
    case 'f': io->url_filename= value; break;
    case 'i': io->insert= UDM_TRUE; break;
    case 'w': io->warnings= UDM_FALSE; break;
    case 'j': UdmVarListAddStr(&Conf.Vars, "stat_time", value); break;
    case 'd': strncpy(cname, value, sizeof(cname));
      cname[sizeof(cname) - 1] = '\0';
      break;
    case 'D': UdmVarListAddStr(&Conf.Vars,"DBLimit"   , value); break;
    case '?':
    case 'h':
    default:
      io->help++;
  }
  return 0;
}


/*
  Available new options:
  Capital letters: AB    GH JK M OP   TUVXWXYZ
  Small   letters:           k             x z
  Digits:          123456789
*/
static UDM_CMDLINE_OPT udm_indexer_options[]=
{
  {-1, "",  UDM_OPT_TITLE,NULL, "\nCrawler options:"},
  {'a', "", UDM_OPT_BOOL, NULL, "Revisit all documents even if not expired (can be\n"
                                "limited using -t, -u, -s, -c, -y and -f options)"},
  {'m', "", UDM_OPT_BOOL, NULL, "Update expired documents even if not modified (can be\n"
                                "limited using -t, -u, -c, -s, -y and -f options)"},
  {'e', "", UDM_OPT_BOOL, NULL, "Visit 'most expired' (oldest) documents first"},
  {'o', "", UDM_OPT_BOOL, NULL, "Visit documents with less depth (hops value) first"},
  {'r', "", UDM_OPT_BOOL, NULL, "Do not try to reduce remote servers load by randomising\n"
                                "crawler queue order (faster, but less polite)"},
  {'n', "", UDM_OPT_INT,  NULL, "Visit only # documents and exit"},
  {'c', "", UDM_OPT_INT,  NULL, "Visit only # seconds and exit"},
  {'q', "", UDM_OPT_BOOL, NULL, "Quick startup (do not add Server URLs);  -qq even quicker"},
  {'b', "", UDM_OPT_BOOL, NULL, "Block starting more than one indexer instances"},
  {'i', "", UDM_OPT_BOOL, NULL, "Insert new URLs (URLs to insert must be given using -u or -f)"},
  {'p', "", UDM_OPT_INT,  NULL, "Sleep # seconds after downloading every URL"},
  {'w', "", UDM_OPT_BOOL, NULL, "Do not ask for confirmation when clearing documents\n"
                                "from the database (e.g.: indexer -Cw)"},
  {'N', "", UDM_OPT_INT,  NULL, "Run # threads (for crawling or indexing)"},


  {-1, "",  UDM_OPT_TITLE,NULL, "\nSubsection control options (can be combined):"},
  {'s', "", UDM_OPT_STR,  NULL, "Limit indexer to documents matching status (HTTP Status code)"},
  {'t', "", UDM_OPT_STR,  NULL, "Limit indexer to documents matching tag"},
  {'y', "", UDM_OPT_STR,  NULL, "Limit indexer to documents matching content-type"},
  {'L', "", UDM_OPT_STR,  NULL, "Limit indexer to documents matching language"},
  {'u', "", UDM_OPT_STR,  NULL, "Limit indexer to documents with URLs matching pattern\n"
                                "(supports SQL LIKE wildcards '%' and '_')"},
  {0, "seed",UDM_OPT_STR, NULL, "Limit indexer to documents with the given seed (0-255)"},
  {'D', "", UDM_OPT_STR,  NULL, "Work with the n-th database only (i.e. with the n-th DBAddr)"},
  {'f', "", UDM_OPT_STR,  NULL, "Read URLs to be visited/inserted/deleted from file (with -a\n"
                                "or -C option, supports SQL LIKE wildcard '%%'; has no effect\n"
                                "when combined with -m option)"},
  {-1,  "", UDM_OPT_TITLE,NULL,
  "  -f -            Use stdin instead of a file as an URL list"},


  {-1,  "", UDM_OPT_TITLE,NULL, "\nLogging options:"},
  {'l', "", UDM_OPT_BOOL, NULL, "Do not log to stdout/stderr"},
  {'v', "", UDM_OPT_INT,  NULL, "Verbose level (0-5)"},


  {-1, "",  UDM_OPT_TITLE,NULL, "\nMisc. options:"},
  {'F', "", UDM_OPT_STR,  NULL, "Print compile configuration and exit (e.g.: indexer -F '*')"},
  {'h',"help",UDM_OPT_BOOL, NULL,"Print help page and exit; -hh print more help"},
  {'?', "", UDM_OPT_BOOL, NULL, "Print help page and exit; -?? print more help"},
  {'d', "", UDM_OPT_STR,  NULL, "Use the given configuration file instead of indexer.conf"
#ifndef WIN32
                                "\nThis option is usefull when running indexer as an\n"
                                "interpreter, e.g.: #!/usr/local/sbin/indexer -d"
#endif
  },
  {'j', "", UDM_OPT_STR,  NULL, "Set current time for statistic (use with -S),\n"
                                "format: YYYY-MM[-DD[ HH[:MM[:SS]]]]\n"
                                "or time offset, e.g. 1d12h (see Period in indexer.conf)"},
  {UDM_IND_SET,          "set",          UDM_OPT_STR, NULL, "Set variable"},
  {-1, "",  UDM_OPT_TITLE,NULL, "\nCommands (can be used with subsection control options):"},
  {UDM_IND_CRAWL,        "crawl",        UDM_OPT_BOOL,NULL, "Crawl (default command)"},
  {UDM_IND_MULTI2BLOB,   "index",        UDM_OPT_BOOL,NULL, "Create search index"},
  {UDM_IND_WRDSTAT,      "wordstat",     UDM_OPT_BOOL,NULL, "Create statistics for misspelled word suggestions"},
  {UDM_IND_REWRITEURL,   "rewriteurl",   UDM_OPT_BOOL,NULL, "Rewrite URL data into the current search index"},
  {UDM_IND_REWRITELIMITS,"rewritelimits",UDM_OPT_BOOL,NULL, "Recreate all Limit, UserScore, UserOrder data"},
  {UDM_IND_REWRITEPOP,   "rewritepop",   UDM_OPT_BOOL,NULL, "Recreate popularity data"},
  {UDM_IND_DELETE, /*C*/  "delete",      UDM_OPT_BOOL,NULL, "Delete documents from the database"},
  {UDM_IND_STAT,   /*S*/ "statistics",   UDM_OPT_BOOL,NULL, "Print statistics and exit"},
  {UDM_IND_REFERERS,/*I*/ "referers",    UDM_OPT_BOOL,NULL, "Print referers and exit "},

  {-1, "",  UDM_OPT_TITLE,NULL, "\nOther commands:"},
#ifdef HAVE_SQL /* TODO34 */
  {UDM_IND_CREATE,       "create",       UDM_OPT_BOOL,NULL, "Create SQL table structure and exit"},
  {UDM_IND_DROP,         "drop",         UDM_OPT_BOOL,NULL, "Drop SQL table structure and exit"},
  {UDM_IND_SQLMON, /*Q*/ "sqlmon",       UDM_OPT_BOOL,NULL, "Run interactive SQL monitor"},
  {UDM_IND_EXECSQL,      "exec",         UDM_OPT_STR, NULL, "Execute SQL query"},
#endif
  {UDM_IND_CHECKCONF,    "checkconf",    UDM_OPT_BOOL,NULL, "Check configuration file for good syntax"},
  {UDM_IND_EXPORT,       "export",       UDM_OPT_BOOL,NULL, NULL}, /* TODO */
  {UDM_IND_HASHSPELL,    "hashspell",    UDM_OPT_BOOL,NULL, "Create hash files for the active Ispell dictionaries"},
  {UDM_IND_DUMPSPELL,    "dumpspell",    UDM_OPT_BOOL,NULL, "Dump Ispell data for use with SQLWordForms"},
  {UDM_IND_DUMPCONF,     "dumpconf",     UDM_OPT_BOOL,NULL, NULL},
  {UDM_IND_DUMPDATA,     "dumpdata",     UDM_OPT_BOOL,NULL, "Dump collected data using SQL statements"},
  {UDM_IND_RESTOREDATA,  "restoredata",  UDM_OPT_BOOL,NULL, "Load prevously dumped data (give a filename using -f)"},

  {UDM_IND_SET0, "fl", UDM_OPT_STR, NULL, NULL},
  {0,NULL,0,NULL,NULL}
};


static int
UdmParseCmdLine(UDM_INDEXER_OPTIONS *io,
                int argc, char **argv, size_t *noptions)
{
  return UdmCmdLineOptionsGet(io, argc, argv, udm_indexer_options,
                              UdmCmdLineHandleOption, noptions);
}


/*
static int
UdmReloadEnv(UDM_AGENT *Indexer)
{
  UDM_ENV   NewConf;
  int  rc;

  UdmLog(Indexer,UDM_LOG_ERROR,"Reloading config '%s'",cname);
  UdmEnvInit(&NewConf);
  UdmSetLockProc(&NewConf,UdmLockProc);
  UdmSetRefProc(&NewConf,UdmRefProc);

  Indexer->Conf = &NewConf;
  rc = UdmIndexerEnvLoad(Indexer, cname, add_servers + load_langmaps + UDM_FLAG_SPELL);
  Indexer->Conf = &Conf;
               
  if (rc!=UDM_OK)
  {
    UdmLog(Indexer,UDM_LOG_ERROR,"Can't load config: %s",UdmEnvErrMsg(&NewConf));
    UdmLog(Indexer,UDM_LOG_ERROR,"Continuing with old config");
    UdmEnvFree(&NewConf);
  }
  else
  {
    size_t noptions;
    UdmEnvFree(&Conf);
    Conf=NewConf;
    UdmParseCmdLine(UdmARGC, UdmARGV, &noptions);
#ifndef WIN32
    UdmOpenLog("indexer", &Conf, log2stderr);
#endif
  }
  return UDM_OK;
}
*/  


static size_t
create_shared_info(UDM_AGENT *A, char *str, size_t len)
{
  UDM_ENV *Env= A->Conf;
  size_t res= udm_snprintf(str, len,
                          "Hrefs: %d,"
                          "Targets: %d,"
                          "Cookies: %d,"
                          "Robots: %d,"
                          "Hosts: %d,"
                          "IPs: %d",
                          (int) Env->Hrefs.nhrefs,
                          (int) Env->Targets.num_rows,
                          (int) Env->Cookies.nvars,
                          (int) Env->Robots.nrobots,
                          (int) Env->Hosts.nhost_addr,
                          (int) Env->InAddr.nitems);
  return res;
}


static udm_rc_t
httpd_client_handler(int client, UDM_AGENT *A)
{
  char request[4096];
  char response[1024];
  char speed_info[128]= "";
  char shared_info[128]= "";
  ssize_t nrecv;
  size_t i, len, total_docs= 0, total_sec= 0;
  udm_uint8 total_bytes= 0;
  time_t now= time(0);
  
  nrecv= recv(client, request, sizeof(request), 0);
  UdmLog(A, UDM_LOG_ERROR, "Received request len=%d", (int) nrecv);
  udm_snprintf(response, sizeof(response) - 1,
               "HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n");
  UdmSend(client, response, strlen(response), 0);
  len= sprintf(response,
               "Threads:"
               "<table border=1 cellspacing=1 cellpadding=1>\n"
               "<tr><th>ID</th>"
               "<th>Docs</th>"
               "<th>Size</th>"
               "<th>Task</th>"
               "<th>Time</th>"
               "<th>Param</th>"
               "<th>Extra</th></tr>");
  UdmSend(client, response, len, 0);
  for (i= 0; i < (size_t) maxthreads; i++)
  {
    UDM_AGENT *Tmp= &ThreadCrawlers[i].Agent;
    char mutex_owned_info[64]= "", *mi;
    size_t sec;
    udm_mutexno_t mutex;
    
    for (mutex= 0, mi= mutex_owned_info;
         mutex < UDM_LOCK_MAX;
         mutex++)
    {
      mi+= UdmMutexStatePrint(mi, sizeof(mutex_owned_info) - (mi - mutex_owned_info),
                              Tmp, mutex);
    }
    
    len= sprintf(response,
                 "<tr><td>%d</td>"
                 "<td align=right>%d</td>"
                 "<td align=right>%llu</td>"
                 "<td>%s</td>"
                 "<td align=right>%d</td>"
                 "<td>%s&nbsp;</td>"
                 "<td>%s%s&nbsp;</td></tr>\n",
                 Tmp->handle,
                 (int) Tmp->ndocs,
                 (unsigned long long) Tmp->nbytes,
                 Tmp->State.task,
                 (int) (now - Tmp->State.start_time),
                 UDM_NULL2EMPTY(Tmp->State.param),
                 UDM_NULL2EMPTY(Tmp->State.extra),
                 mutex_owned_info);
    UdmSend(client, response, len, 0);
    total_docs+= Tmp->ndocs;
    total_bytes+= Tmp->nbytes;
    sec= (size_t) (now - Tmp->start_time);
    if (sec > total_sec)
      total_sec= sec;
  }
  if (total_sec)
  {
    udm_snprintf(speed_info, sizeof(speed_info) - 1,
                 "%d seconds, %d docs/sec, %d bytes/sec",
                 (int) total_sec,
                 (int) (total_docs / total_sec),
                 (int) (total_bytes / total_sec));
  }
  
  len= sprintf(response,
                 "<tr><td>&nbsp;</td>"
                 "<td align=right>%d</td>"
                 "<td align=right>%llu</td>"
                 "<td>&nbsp;</td>"
                 "<td align=right>&nbsp;</td>"
                 "<td>%s&nbsp;</td>"
                 "<td>&nbsp;</td></tr>\n",
                 (int) total_docs,
                 (unsigned long long) total_bytes,
                 speed_info);
  UdmSend(client, response, len, 0);
  len= sprintf(response, "</table>\n");
  UdmSend(client, response, len, 0);
  len= create_shared_info(A, shared_info, sizeof(shared_info));
  UdmSend(client, shared_info, len, 0);
  return UDM_OK;
}


#ifdef  WIN32
unsigned int __stdcall thread_main_httpd(void *arg)
#else
static void* thread_main_httpd(void *arg)
#endif
{
  UDM_AGENT *A= (UDM_AGENT*) arg;
#ifndef WIN32
  UdmStartHTTPD(A, httpd_client_handler);
#endif
  return 0;
}



static char pidname[1024];
static char time_pid[100];

static void exitproc(void)
{
  unlink(pidname);
}


static char * time_pid_info(void)
{
  struct tm * tim;
  time_t t;
  t= time(NULL);
  tim= localtime(&t);
  strftime(time_pid,sizeof(time_pid),"%a %d %H:%M:%S",tim);
  sprintf(time_pid+strlen(time_pid)," [%d]",(int)getpid());
  return(time_pid);
}


static void UdmWSAStartup(void)
{
#ifdef WIN32
  WSADATA wsaData;
  if (WSAStartup(0x101,&wsaData)!=0)
  {
    fprintf(stderr,"WSAStartup() error %d\n",WSAGetLastError);
    exit(1);
  }
#endif
}

static void UdmWSACleanup(void)
{
#ifdef WIN32
  WSACleanup();
#endif
  return;
}

static int UdmConfirm(const char *msg)
{
  char str[5];
  printf("%s",msg);
  return (fgets(str,sizeof(str),stdin) && !strncmp(str,"YES",3));
}


static udm_rc_t
UdmIndDelete(UDM_AGENT *A, UDM_QUERY *Query, const UDM_INDEXER_OPTIONS *io)
{
  int clear_confirmed=1;
  udm_rc_t rc= UDM_OK;
  if (io->warnings)
  {
    size_t i;
    printf("You are going to delete content from the database(s):\n");
    for (i = 0; i < Conf.DBList.nitems; i++)
    {
      UDM_DB *db= &Conf.DBList.Item[i];
      char dbaddr[128];
      size_t nbytes;
      db->dbhandler->Info(db, dbaddr, sizeof(dbaddr), &nbytes, UDM_DBINFO_ADDR);
      printf("%s\n", dbaddr);
    }
    clear_confirmed=UdmConfirm("Are you sure?(YES/no)");
  }
  
  if (clear_confirmed)
  {
    if (io->url_filename)
    {
      rc= UdmURLFile(A, io->url_filename, UDM_URL_FILE_CLEAR);
    }
    else
    {
      rc= UdmQueryAction(A, Query, UDM_QUERYCMD_CLEAR);
    }
  }
  else
  {
    printf("Canceled\n");
  }
  if (rc != UDM_OK)
  {
    fflush(stdout);
    UdmLog(A, UDM_LOG_ERROR, "Error: '%s'", UdmEnvErrMsg(A->Conf));
  }
  return rc;
}

static void
UdmPrintCumulativeStatistics(UDM_AGENT *Indexer,
                             UDM_CRAWLER *Item, size_t nitems, time_t sec)
{
  size_t i;
  unsigned int ndocs= 0;
  unsigned long long nbytes= 0;
  double M= 0.0, K= 0.0;

  for (i= 0; i < nitems; i++)
  {
    ndocs+= Item[i].Agent.ndocs;
    nbytes+= Item[i].Agent.nbytes;
  }
  if (sec > 0)
  {
    /* Convert to int64 - conversion from uint64 to double doesn't work on windows */
    M= ((udm_timer_t) nbytes) / 1048576.0 / sec;
    if (M < 1.0) K= ((udm_timer_t) nbytes) / 1024.0 / sec;
  }
  UdmLog(Indexer, UDM_LOG_ERROR,
         "Done (%d seconds, %u documents, %llu bytes, %5.2f %cbytes/sec.)",
         (int) sec, (unsigned int) ndocs, (unsigned long long) nbytes,
         (M < 1.0) ? K : M, (M < 1.0) ? 'K' : 'M' );
}


static udm_rc_t
UdmCrawl(UDM_AGENT *A, UDM_QUERY *Query)
{
  UDM_AGENT httpd_agent;
  size_t nbytes;
  const UDM_VAR *Listen;
  time_t start_time= time(0);

  maxthreads= UdmVarListFindInt(&A->Conf->Vars, "CrawlerThreads", 1);
  nbytes= maxthreads * sizeof(UDM_CRAWLER);
  ThreadCrawlers= (UDM_CRAWLER*) UdmMalloc(nbytes);
  bzero((void*) ThreadCrawlers, nbytes);
  UdmAgentInit(&httpd_agent, A->Conf, 0);
  
  if ((Listen= UdmVarListFind(&A->Conf->Vars, "Listen")))
  {
#ifdef HAVE_PTHREAD
    if (UdmVarStr(Listen) && UdmVarStr(Listen)[0])
    {
      udm_thread_t httpd_thread;
      UdmThreadCreate(&httpd_thread, thread_main_httpd, &httpd_agent);
    }
    else
#endif
    {
      UdmLog(A, UDM_LOG_ERROR, "Not starting HTTPD");
    }
  }
  
  
#ifdef HAVE_PTHREAD
  {
    int i;
    for (i= 0; i < maxthreads; i++)
    {
      UdmAgentInit(&ThreadCrawlers[i].Agent, A->Conf, i + 1);
      ThreadCrawlers[i].Query= Query;
      ThreadCrawlers[i].Agent.flags= flags;
    }
  }
  
  {
    int i;
    nbytes= maxthreads * sizeof(udm_thread_t);
    threads= (udm_thread_t*) UdmMalloc(nbytes);
    bzero((void*) threads, nbytes);
    for(i= 0; i < maxthreads; i++)
    {
      if (seconds) UDMSLEEP(seconds);
      
      UdmThreadCreate(&threads[i], &UdmCrawlerMain, &ThreadCrawlers[i]);

      UDM_GETLOCK(A, UDM_LOCK_THREAD);
      total_threads= i + 1;
      UDM_RELEASELOCK(A,UDM_LOCK_THREAD);
    }
#ifndef WIN32
    for (i = 0; i < maxthreads; i++)
      pthread_join(threads[i], NULL);
#else
    while(1)
    {
      int num;
      UDM_GETLOCK(A,UDM_LOCK_THREAD);
      num=total_threads;
      UDM_RELEASELOCK(A,UDM_LOCK_THREAD);
      if (!num)break;
      UDMSLEEP(1);
    }
#endif
    UDM_FREE(threads);
  }
#else
  A->handle = 1;
  UdmCrawlerMain(A);
#endif

  if (maxthreads > 1)
    UdmPrintCumulativeStatistics(A, ThreadCrawlers, maxthreads,
                                 time(0) - start_time);
  UdmAgentFree(&httpd_agent);
  UDM_FREE(ThreadCrawlers);
  return thd_errors ? UDM_ERROR : UDM_OK;
}


static int
UdmDumpDocument(UDM_AGENT *A, UDM_DOCUMENT *D)
{
  size_t i;
  printf("  <item>\n");
  /*
  printf("    <link>%s</link>\n", UdmVarListFindStr(&D->Sections, "URL", ""));
  printf("    <id>%s</id>\n", UdmVarListFindStr(&D->Sections, "ID", ""));
  printf("    <content-length>%s</content-length>\n", UdmVarListFindStr(&D->Sections, "Content-Length", ""));
  printf("    <status>%s</status>\n", UdmVarListFindStr(&D->Sections, "Status", ""));
  printf("    <hops>%s</hops>\n", UdmVarListFindStr(&D->Sections, "Hops", ""));
  printf("    <crc32>%s</crc32>\n", UdmVarListFindStr(&D->Sections, "crc32", ""));
  printf("    <modified>%s</modified>\n", UdmVarListFindStr(&D->Sections, "Last-Modified", ""));
  */
  for (i= 0; i < D->Sections.nvars; i++)
  {
    const UDM_VAR *S= UdmVarListFindConstByIndex(&D->Sections, i);
    UDM_CONST_STR strbuf, *str= UdmVarGetConstStr(S, &strbuf);
    printf("    <%s>%.*s</%s>\n",
           UdmVarName(S), (int) str->length, str->str, UdmVarName(S));
  }
  printf("  </item>\n");
  return 0;
}


static udm_rc_t
UdmDumpDocuments(UDM_AGENT *A, UDM_QUERY *Query)
{
  udm_rc_t rc;
  A->Conf->DumpDoc= UdmDumpDocument;
  
  /* printf("<rss><channel>\n"); */
  if (UDM_OK != (rc= UdmQueryAction(A, Query, UDM_QUERYCMD_DUMPDATA)))
    return rc;
  /* printf("</channel></rss>\n"); */
  return UDM_OK;
}


#include <sys/stat.h>
/*
  TODO for dump/restore:
  - Escape bad characters to entities on dump, unescape on restore.
  - Huge file size
  - Subsection control
  - Database number limit
  - Insert into database
*/
static udm_rc_t
UdmRestoreDocuments(UDM_AGENT *A, const UDM_INDEXER_OPTIONS *io)
{
  udm_rc_t rc= UDM_OK;
  UDM_DSTR dstr;
  struct stat sb;
  int fd;
  UDM_CHARSET *cs;
  
  if (!io->url_filename)
  {
    fprintf(stderr, "Required option -f ommitted\n");
    return UDM_ERROR;
  }
  fprintf(stderr, "Restoring data from '%s'\n", io->url_filename);
  
  if (stat(io->url_filename, &sb))
  {
    fprintf(stderr, "Can't stat '%s'\n", io->url_filename);
    return UDM_ERROR;
  }
  UdmDSTRInit(&dstr, 1024);
  if ((fd= open(io->url_filename, O_RDONLY|UDM_BINARY)) <= 0)
  {
    rc= UDM_ERROR;
    fprintf(stderr, "Can't open '%s'", io->url_filename);
    goto ex;
  }

  if (UDM_OK != (rc= UdmDSTRReadFile(&dstr, fd, sb.st_size)))
  {
    rc= UDM_ERROR;
#ifndef WIN32
    fprintf(stderr, "Failed to read %lld bytes from '%s'\n",
            (long long) sb.st_size, io->url_filename);
#endif
    goto ex; 
  }
  
  cs= &udm_charset_latin1;
  rc= UdmQueryFromXML(A, NULL, UdmDSTRPtr(&dstr), UdmDSTRLength(&dstr), cs);
ex:
  UdmDSTRFree(&dstr);
  return rc;
}


static udm_rc_t
UdmIndHashSpell(UDM_AGENT *Agent)
{
  UDM_ENV *Env= Agent->Conf;
  char errmsg[256];
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmSpellListListLoad(&Env->Spells, errmsg, sizeof(errmsg)))||
      UDM_OK != (rc= UdmSpellListListWriteHash(&Env->Spells, errmsg, sizeof(errmsg))))
  {
    fprintf(stderr, "error: %s\n", errmsg);
  }
  return rc;
}


static udm_rc_t
UdmIndDumpSpell(UDM_AGENT *Agent)
{
  char errmsg[256];
  int spflags= UDM_SPELL_NOPREFIX;
  UDM_ENV *Env= Agent->Conf;
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmSpellListListLoad(&Env->Spells,
                                          errmsg, sizeof(errmsg)))||
      UDM_OK != (rc= UdmAffixListListLoad(&Env->Affixes, spflags,
                                          errmsg, sizeof(errmsg)))||
      UDM_OK != (rc= UdmSpellDump(&Env->Spells, &Env->Affixes,
                                          errmsg, sizeof(errmsg))))
  {
    fprintf(stderr, "error: %s\n", errmsg);
  }
  return rc;
}


static udm_rc_t
UdmIndSQLMonitor(UDM_AGENT *Agent, UDM_INDEXER_OPTIONS *io)
{
  udm_rc_t rc;
  UDM_SQLMON_PARAM prm;
  bzero((void*)&prm,sizeof(prm));
  prm.flags= loglevel > 0 ? UDM_SQLMON_DISPLAY_FIELDS : 0;
  execsql= UdmVarListFindStr(&Conf.Vars, "exec", NULL);
  prm.iohandler.gets= execsql ? sqlexecgets : sqlmongets;
  prm.iohandler.prompt= sqlmonprompt;
  prm.iohandler.user_data= &prm;
#ifdef HAVE_READLINE
  if (isatty(0))
    prm.mode= udm_sqlmon_mode_interactive;
#endif
  rc= UdmSQLMonitor(Agent, &Conf, &prm);
  if (prm.iohandler.gets == sqlmongets)
  {
    /* Display "\n" after the last ">SQL" prompt */
    sqlmonprompt(&prm.iohandler, UDM_MSG_EOL, "");
  }
  return rc;
}


static udm_rc_t
UdmIndBlock(void)
{
  int pid_fd;
  char pidbuf[128];
  /* Check that another instance isn't running and create PID file. */
  const char *vardir= UdmVarListFindStr(&Conf.Vars,"VarDir",UDM_VAR_DIR);
  sprintf(pidname,"%s/%s", vardir ,"indexer.pid");
  pid_fd = open(pidname,O_CREAT|O_EXCL|O_WRONLY,0644);
  if (pid_fd < 0)
  {
    fprintf(stderr,"%s Can't create '%s': %s\n", time_pid_info(), pidname, strerror(errno));
    if (errno == EEXIST)
    {
      fprintf(stderr,"It seems that another indexer is already running!\n");
      fprintf(stderr,"Remove '%s' if it is not true.\n",pidname);
    }
    return UDM_ERROR;
  }
  udm_snprintf(pidbuf, sizeof(pidbuf), "%d\n", (int)getpid());
  write(pid_fd, &pidbuf, strlen(pidbuf));
#ifdef HAVE_ATEXIT
  atexit(&exitproc);
#endif
  return UDM_OK;
}


static udm_rc_t
UdmIndInsertFromFile(UDM_AGENT *Agent, UDM_INDEXER_OPTIONS *io)
{
  udm_rc_t rc= UDM_OK;
  if (strcmp(io->url_filename,"-"))
  {
    /* Make sure all URLs to be inserted are OK */
    if (UDM_OK != (rc= UdmURLFile(Agent, io->url_filename, UDM_URL_FILE_PARSE)))
    {
      UdmLog(Agent, UDM_LOG_ERROR,"Error: Invalid URL in '%s'", io->url_filename);
      goto ex;
    }
  }
  if (UDM_OK != (rc= UdmURLFile(Agent, io->url_filename, UDM_URL_FILE_INSERT)))
  {
    UdmLog(Agent, UDM_LOG_ERROR,"Error: '%s'", UdmEnvErrMsg(Agent->Conf));
    goto ex;
  }
ex:
  return rc;
}


static udm_rc_t
UdmIndExpire(UDM_AGENT *Agent, UDM_QUERY *Query, const UDM_INDEXER_OPTIONS *io)
{
  udm_rc_t rc= io->url_filename ?
               UdmURLFile(Agent, io->url_filename, UDM_URL_FILE_REINDEX) :
               UdmQueryAction(Agent, Query, UDM_QUERYCMD_EXPIRE);
  return rc;
}


static udm_rc_t
UdmIndCrawl(UDM_AGENT *Agent, UDM_QUERY *Query, const UDM_INDEXER_OPTIONS *io)
{
  udm_rc_t rc;
  if (io->block && UDM_OK != (rc= UdmIndBlock()))
    goto ex;
  UdmLog(Agent, UDM_LOG_WARN, "indexer from %s-%s-%s started with '%s'", PACKAGE, VERSION, UDM_DBTYPE, cname);
  UdmSigHandlersInit(Agent);

  if (io->expire)
  {
    if (UDM_OK != (rc= UdmIndExpire(Agent, Query, io)))
      goto ex;
  }

  if (UDM_OK != (rc= UdmCrawl(Agent, Query)))
    goto ex;

ex:
  return rc;
}


int main(int argc, char **argv)
{
  UDM_INDEXER_OPTIONS io;
  UDM_AGENT Main;
  UDM_QUERY Query;
  udm_rc_t rc= UDM_OK;
  char  *REQUEST_METHOD= getenv("REQUEST_METHOD");
  FILE  *logfile= REQUEST_METHOD ? stdout : stderr;
  size_t noptions;
#ifdef CHASEN
  char  *chasen_argv[] = { "chasen", "-b", "-f", "-F", "%m ", NULL };
  chasen_getopt_argv(chasen_argv, NULL);
#endif

  UdmIndexerOptionsInit(&io);

  if (REQUEST_METHOD)
    printf("Content-Type: text/plain\r\n\r\n");

  UdmWSAStartup();

  UdmInit(); /* Initialize library */

  UdmInitMutexes();
  UdmEnvInit(&Conf);
  UdmVarListAddEnviron(&Conf.Vars,"ENV");
  UdmSetLockProc(&Conf,UdmLockProc);

#if defined(HAVE_PTHREAD) && !defined(WIN32)
  Conf.THDHandler.ThreadCreate= UdmThreadCreate;
  Conf.THDHandler.ThreadJoin= UdmThreadJoin;
#endif

  UdmSetRefProc(&Conf,UdmRefProc);
#ifdef THINFO_TEST
  UdmSetThreadProc(&Conf,UdmShowThreadInfoProc);
#endif
  UdmAgentInit(&Main,&Conf,0);
  UdmQueryInit(&Query);

  UdmARGC= argc;
  UdmARGV= argv;

  if (UdmParseCmdLine(&io, UdmARGC, UdmARGV, &noptions))
    goto ex;

  if (io.cmd == UDM_IND_INDEX)
  {
    fprintf(stderr, "\nWARNING: \"indexer -Eindex\" is deprecated. "
                    "Use \"indexer -Ecrawl\" instead!\n\n");
    io.cmd= UDM_IND_CRAWL;
  }
  
  if (io.cmd == UDM_IND_AMBIGUOUS)
  {
    fprintf(stderr, "Ambiguous indexer command in -E\n");
    io.help++;
  }
  
  if (io.cmd == UDM_IND_UNKNOWN)
  {
    fprintf(stderr, "Unknown indexer command in -E\n");
    io.help++;
  }
  
  if (io.cmd == UDM_IND_DUMPCONF)
  {
    io.load_for_dump|= UDM_FLAG_DONT_ADD_TO_DB;
    io.load_langmaps= 0;
    io.load_spells= 0;
  }
  else if (io.cmd != UDM_IND_CRAWL)
  {
    if (io.cmd != UDM_IND_MULTI2BLOB && /* TODO34: rename this to UDM_IND_INDEX */
        io.cmd != UDM_IND_REWRITEPOP) 
      io.add_servers= 0;
    else
    {
      io.add_servers|= UDM_FLAG_DONT_ADD_TO_DB;
      io.add_server_urls= 0;
    }
    io.load_langmaps= 0;
    if (io.cmd != UDM_IND_HASHSPELL && io.cmd  != UDM_IND_DUMPSPELL)
      io.load_spells= 0;
  }

  flags|= io.add_servers;
  flags |= io.add_server_urls;
  Main.flags= flags;

  argc-= noptions;
  argv+= noptions;

  if (argc > 1 || io.help)
  {
    usage(io.help, udm_indexer_options);
    rc= UDM_ERROR;
    goto ex;
  }
  
  if (!*cname)
  {
    if (argc == 1)
    {
      strncpy(cname,argv[0],sizeof(cname));
      cname[sizeof(cname)-1]='\0';
    }
    else
    {
      const char *cd=UdmVarListFindStr(&Conf.Vars,"ConfDir",UDM_CONF_DIR);
      udm_snprintf(cname,sizeof(cname),"%s%s%s",cd,UDMSLASHSTR,"indexer.conf");
      cname[sizeof(cname)-1]='\0';
    }
  }
  
  
  if (UDM_OK != (rc= UdmIndexerEnvLoad(&Main, cname,
                                       io.add_servers + io.load_langmaps +
                                       io.load_spells +
                                       io.add_server_urls + io.load_for_dump)))
  {
    fprintf(logfile, "%s\n", UdmEnvErrMsg(&Conf));
    goto ex;
  }


#ifdef HAVE_LOCALE_H
  /*
    Needed for Mimer to make non-ascii Latin1 characters work.
    The locale should typically be set to en_US.iso88591
  */
  if (UdmVarListFind(&Conf.Vars, "Locale"))
    setlocale(LC_ALL, UdmVarListFindStr(&Conf.Vars, "Locale", "C"));
#endif

  if (io.url_filename && strcmp(io.url_filename,"-"))
  {
    /* Make sure URL file is readable if not STDIN */
    FILE *url_file;
    if (!(url_file= fopen(io.url_filename,"r")))
    {
      UdmLog(&Main, UDM_LOG_ERROR,
             "Error: can't open url file '%s': %s",
             io.url_filename, strerror(errno));
      goto ex;
    }
    fclose(url_file);
  }
  
  
  
  if (io.cmd == UDM_IND_DUMPCONF)
  {
    if (UDM_OK != (rc= UdmEnvSave(&Main, "-", 0)))
      fprintf(logfile, "%s\n", UdmEnvErrMsg(&Conf));
    goto ex;
  }
  
  
  if (io.cmd == UDM_IND_CHECKCONF)
  {
    rc= 0;
    goto ex;
  }
  
  
  UdmEnvSetLogLevel(Main.Conf, io.have_loglevel ? loglevel :
                    UdmVarListFindInt(&Main.Conf->Vars, "LogLevel", UDM_LOG_INFO));
#ifndef WIN32
  UdmOpenLog("indexer", &Conf, io.log2stderr);
#endif

  if (io.insert && io.url_filename)
  {
    if (UDM_OK != UdmIndInsertFromFile(&Main, &io))
    {
      UdmLog(&Main,UDM_LOG_ERROR,"Error: '%s'",UdmEnvErrMsg(Main.Conf));
      goto ex;
    }
  }

  switch (io.cmd)
  {
    case UDM_IND_HASHSPELL:     rc= UdmIndHashSpell(&Main);     goto ex;
    case UDM_IND_DUMPSPELL:     rc= UdmIndDumpSpell(&Main);     goto ex;
    case UDM_IND_DUMPDATA:      rc= UdmDumpDocuments(&Main, &Query); goto ex;
    case UDM_IND_RESTOREDATA:   rc= UdmRestoreDocuments(&Main, &io); goto ex;
    case UDM_IND_SQLMON:        rc= UdmIndSQLMonitor(&Main, &io);    goto ex;
    case UDM_IND_EXECSQL:       rc= UdmIndSQLMonitor(&Main, &io);    goto ex;
    case UDM_IND_MULTI2BLOB:    rc= UdmQueryAction(&Main, &Query, UDM_QUERYCMD_INDEX);  goto ex;
    case UDM_IND_EXPORT:        rc= UdmQueryAction(&Main, &Query, UDM_QUERYCMD_EXPORT); goto ex;
    case UDM_IND_WRDSTAT:       rc= UdmQueryAction(&Main, &Query, UDM_QUERYCMD_WORDSTAT); goto ex;
    case UDM_IND_REWRITEURL:    rc= UdmQueryAction(&Main, &Query, UDM_QUERYCMD_REWRITE_URLDATA);    goto ex;
    case UDM_IND_REWRITELIMITS: rc= UdmQueryAction(&Main, &Query, UDM_QUERYCMD_REWRITE_LIMITS);     goto ex;
    case UDM_IND_REWRITEPOP:    rc= UdmQueryAction(&Main, &Query, UDM_QUERYCMD_REWRITE_POPULARITY); goto ex;
    case UDM_IND_STAT:          rc= ShowStatistics(&Main, &Query); goto ex;
    case UDM_IND_REFERERS:      rc= ShowReferers(&Main, &Query);goto ex;
    case UDM_IND_CREATE:        rc= UdmIndCreate(&Main);        goto ex;
    case UDM_IND_DROP:          rc= UdmIndDrop(&Main);          goto ex;
    case UDM_IND_DELETE:        rc= UdmIndDelete(&Main, &Query, &io);   goto ex;
    case UDM_IND_CRAWL:         rc= UdmIndCrawl(&Main, &Query, &io);    goto ex;
    case UDM_IND_INDEX:         rc= UdmIndCrawl(&Main, &Query, &io);    goto ex;

    case UDM_IND_AMBIGUOUS:
    case UDM_IND_UNKNOWN:
    case UDM_IND_CHECKCONF:
    case UDM_IND_CONVERT:
    case UDM_IND_DUMPCONF:
    case UDM_IND_SET:
    case UDM_IND_SET0:
      break;
  }
  
  
ex:
  total_threads= 0;
  UdmAgentFree(&Main);
  UdmQueryFree(&Query);
  UdmEnvFree(&Conf);
  UdmDestroyMutexes();
  UdmWSACleanup();
#ifndef HAVE_ATEXIT
  exitproc();
#endif
  return rc == UDM_OK ? 0 : 1;
}
