/*
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#include <math.h>

#include "udm_common.h"
#include "udm_template.h"
#include "udm_server.h"
#include "udm_env.h"
#include "udm_conf.h"
#include "udm_db.h"
#include "udm_http.h"
#include "udm_parsehtml.h"
#include "udm_host.h"
#include "udm_contentencoding.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_wild.h"
#include "udm_searchtool.h"
#include "udm_doc.h"
#include "udm_match.h"
#include "udm_sgml.h"
#include "udm_proto.h"
#include "udm_ctype.h"
#include "udm_lex.h"
#include "udm_prog.h"


/*
TODO:
- Add C++ keywords
- Add size_t
- Add sizeof
- Refactor cout, make << an operator in ostream.
- Change functions _argv() and _argc() to paramenters argv and argc
- More efficient data alignment: see UdmValueHandlerAlignedDataSize()
- Add a warning: "Suggest parentheses around &&" when mixing || and &&
- Restore a warning: see ExpressionHasEffect()
- Create function call arguments on stack rather than on heap.
- Allow function calls as arguments to other function calls.
- maximum execution time
- string comparison: if (x < "yyy")
- string concatenation with operator+

TODO-TEMPLATE:
- DetectClones: copy from msearch-test/include/search.htm
- doccount is 0 if there is only one search word
- qcache and "missing" do not work together
- mark stopwords (check origin)
- don't display suggestions (origin=6) in word statistics
- "Strict mode ... too few result" does not get displayed
   if Strict found 0 results
- display query in browser charset, e.g. res.property_html("q") ?
- Remove DateFormat, implement strftime() instead
- HTML-aware string::left(). See UdmHtmlStrLeft().
- ExcerptSize and ExcerptPadding should probably be parameters to
  property(), or to a new method.
*/

#define UDM_CMDNAME(x) {UDM_CSTR_WITH_LEN(x)}
#define UDM_CMD1_VOID(name, exec) {UDM_CMDNAME(name), 1, exec}
#define UDM_CMD2_VOID(name, exec) {UDM_CMDNAME(name), 2, exec}
#define UDM_CMD1_IREG0(name, exec) {UDM_CMDNAME(name), 1, exec}


typedef enum
{
  UDM_ASSIGNMENT_MUL,
  UDM_ASSIGNMENT_DIV,
  UDM_ASSIGNMENT_REM,
  UDM_ASSIGNMENT_ADD,
  UDM_ASSIGNMENT_SUB,
  UDM_ASSIGNMENT_AND,
  UDM_ASSIGNMENT_OR,
  UDM_ASSIGNMENT_XOR,
  UDM_ASSIGNMENT_LSHIFT,
  UDM_ASSIGNMENT_RSHIFT
} udm_special_assignment_t;


typedef enum
{
  UDM_UNARY_OPERATOR_PREFIX_INC,
  UDM_UNARY_OPERATOR_PREFIX_DEC,
  UDM_UNARY_OPERATOR_POSTFIX_INC,
  UDM_UNARY_OPERATOR_POSTFIX_DEC
} udm_unary_operator_t;


typedef enum
{
  UDM_EXPRESSION_IDENT,
  UDM_EXPRESSION_CONST,
  UDM_EXPRESSION_OBJECT,      /* The value can be in a temp var   */
  UDM_EXPRESSION_COMPUTED     /* The value is in registers        */
} udm_expression_t;


typedef struct
{
  UDM_PROG_VAR Var;
  udm_expression_t type;
  char buf[2];
  udm_offset_and_length_t textdata_addr;
} UDM_PROG_PRIMARY_EXPRESSION;


typedef struct
{
  UDM_PROG_PRIMARY_EXPRESSION args[UDM_PROG_MAXARG];
  size_t nargs;
} UDM_PROG_FUNCTION_CALL_ARGUMENTS;


typedef struct
{
  UDM_PROG_PRIMARY_EXPRESSION primary_expr;
  UDM_PROG_FUNCTION_CALL_ARGUMENTS arglist;
} UDM_PROG_EXPRESSION;


static const UDM_VALUE_HANDLER *
UdmProgVarHandler(const UDM_PROG_VAR *Var)
{
  return Var->value.handler;
}


static void
UdmProgVarSetHandler(UDM_PROG_VAR *Var, const UDM_VALUE_HANDLER *handler)
{
  Var->value.handler= handler;
}


static const UDM_VALUE_HANDLER *
UdmPrimaryExprHandler(const UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  return expr->Var.value.handler;
}


static const UDM_VALUE_HANDLER *
UdmExprHandler(const UDM_PROG_EXPRESSION *expr)
{
  return expr->primary_expr.Var.value.handler;
}


static void
UdmExprSetHandler(UDM_PROG_EXPRESSION *expr,
                  const UDM_VALUE_HANDLER *handler)
{
  expr->primary_expr.Var.value.handler= handler;
}


/************************************************************/

static udm_bool_t
UdmGenerateOp(UDM_PROG_COMPILER *compiler, udm_tmpl_cmd_t cmd)
{
  return UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg, cmd);
}


static void
UdmProgItemArgInitFromLiteral(UDM_PROG_ARG *arg,
                              const UDM_CONST_TOKEN *str,
                              const udm_offset_and_length_t *textdata)
{
  arg->textdata_addr= *textdata;
}


static void
UdmProgItemArgInitFromLiteralAndHandler(UDM_PROG_ARG *arg,
                                        const UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  arg->handler= UdmPrimaryExprHandler(expr);
  arg->textdata_addr= expr->textdata_addr;
}


static void
UdmProgItemArgInitFromVar(UDM_PROG_ARG *arg, const UDM_PROG_VAR *Var)
{
  arg->stack_offset= Var->value.stack_offset;
  arg->handler= Var->value.handler;
}


static udm_rc_t
UdmProgAddJmpWithComment(UDM_PROG *prog,
                         udm_tmpl_cmd_t cmdnum,
                         int addr, const char *comment)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmdnum= cmdnum;
  i.comment= strdup(comment);
  i.cmdarg.jump= addr;
  return UdmProgAdd(prog, &i);
}


static udm_bool_t
UdmGenerateOpOffset(UDM_PROG_COMPILER *compiler,
                    udm_tmpl_cmd_t cmdnum,
                    size_t offset)
{
  UDM_PROG_ITEM it;
  UdmProgItemInit(&it);
  it.cmdnum= cmdnum;
  it.cmdarg.offset= offset;
  return UDM_OK == UdmProgAdd(compiler->prg, &it);
}


/*********************************************************************/

static const UDM_PROG_ITEM *
UdmProgExecutorStateCurrentCmd(const UDM_PROG_EXECUTOR_STATE *p)
{
  return p->curr;
}


static void *
UdmProgExecutorStateStackAddr0(UDM_PROG_EXECUTOR_STATE *st,
                               const UDM_PROG_ITEM *it)
{
  return st->stack.Val.str + it->item_args.stack_offset;
}


static void
Template_mov_dregn_var0(UDM_PROG_EXECUTOR_STATE *T, size_t dregno)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  void *Data= UdmProgExecutorStateStackAddr0(T, it);
  it->item_args.handler->GetDouble(Data, &T->reg.d[dregno]);
}


static void
Template_mov_iregn_var0(UDM_PROG_EXECUTOR_STATE *T, size_t iregno)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  void *Data= UdmProgExecutorStateStackAddr0(T, it);
  it->item_args.handler->GetInt(Data, &T->reg.i[iregno]);
}


/* Variable initialization */

/*TODO34: destructor is never called for declared variables!!! */
static void
Template_construct_var(UDM_PROG_EXECUTOR_STATE *T)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  const UDM_VALUE_HANDLER *ha= it->item_args.handler;
  void *Data= UdmProgExecutorStateStackAddr0(T, it);
  UDM_ASSERT(it->item_args.stack_offset + ha->DataSize() < T->stack.size_alloced);
  if (UDM_OK != ha->Constructor(Data, NULL, 0))
    T->fatal_error= UDM_TRUE;
}


static void
Template_set_var_from_ireg0(UDM_PROG_EXECUTOR_STATE *T)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  void *Data= UdmProgExecutorStateStackAddr0(T, it);
  const UDM_VALUE_HANDLER *ha= it->item_args.handler;
  ha->SetInt(Data, T->reg.i[0]);
}


static void
Template_set_var_from_dreg0(UDM_PROG_EXECUTOR_STATE *T)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  void *Data= UdmProgExecutorStateStackAddr0(T, it);
  const UDM_VALUE_HANDLER *ha= it->item_args.handler;
  ha->SetDouble(Data, T->reg.d[0]);
}


static void
Template_set_var_from_text0(UDM_PROG_EXECUTOR_STATE *T)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  void *Data= UdmProgExecutorStateStackAddr0(T, it);
  const UDM_VALUE_HANDLER *ha= it->item_args.handler;
  ha->SetStrn(Data,
              T->prog->textdata.Val.str + it->item_args.textdata_addr.offset,
              it->item_args.textdata_addr.length);
}


static void
Template_set_var_from_cs0_z0(UDM_PROG_EXECUTOR_STATE *T)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  void *Dst= UdmProgExecutorStateStackAddr0(T, it);
  if (UDM_OK != it->item_args.handler->SetStrn(Dst, T->reg.cs0, T->reg.z0))
    T->fatal_error= UDM_TRUE;
}


static void
Template_get_var_to_cs0_z0(UDM_PROG_EXECUTOR_STATE *T)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  void *Src= UdmProgExecutorStateStackAddr0(T, it);
  UDM_CONST_STR str;
  it->item_args.handler->GetConstStr(Src, &str);
  T->reg.cs0= str.str;
  T->reg.z0= str.length;
}


static void
Template_get_textdata_to_cs0_z0(UDM_PROG_EXECUTOR_STATE *T)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  T->reg.cs0= T->prog->textdata.Val.str + it->item_args.textdata_addr.offset;
  T->reg.z0= it->item_args.textdata_addr.length;
}


static void
Template_destruct_var(UDM_PROG_EXECUTOR_STATE *T)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  void *Data= UdmProgExecutorStateStackAddr0(T, it);
  it->item_args.handler->Destructor(Data);
}


static void
Template_mov_ireg0_var0(UDM_PROG_EXECUTOR_STATE *T)
{
  Template_mov_iregn_var0(T, 0);
}


static void
Template_mov_ireg1_var0(UDM_PROG_EXECUTOR_STATE *T)
{
  Template_mov_iregn_var0(T, 1);
}


static void
Template_mov_dreg0_var0(UDM_PROG_EXECUTOR_STATE *T)
{
  Template_mov_dregn_var0(T, 0);
}


static void
Template_mov_dreg1_var0(UDM_PROG_EXECUTOR_STATE *T)
{
  Template_mov_dregn_var0(T, 1);
}


#define UDM_CMDNAME(x) {UDM_CSTR_WITH_LEN(x)}


#define UDM_CMD1_VOID(name, exec) {UDM_CMDNAME(name), 1, exec}
#define UDM_CMD2_VOID(name, exec) {UDM_CMDNAME(name), 2, exec}
#define UDM_CMD3_VOID(name, exec) {UDM_CMDNAME(name), 3, exec}

#define UDM_CMD0_IREG0(name, exec) {UDM_CMDNAME(name), 0, exec}
#define UDM_CMD0_IREG1(name, exec) {UDM_CMDNAME(name), 0, exec}
#define UDM_CMD0_DREG0(name, exec) {UDM_CMDNAME(name), 0, exec}
#define UDM_CMD0_DREG1(name, exec) {UDM_CMDNAME(name), 0, exec}

#define UDM_CMD1_IREG0(name, exec) {UDM_CMDNAME(name), 1, exec}
#define UDM_CMD1_IREG1(name, exec) {UDM_CMDNAME(name), 1, exec}
#define UDM_CMD1_DREG0(name, exec) {UDM_CMDNAME(name), 1, exec}
#define UDM_CMD1_DREG1(name, exec) {UDM_CMDNAME(name), 1, exec}

#define UDM_CMD2_IREG0(name, exec) {UDM_CMDNAME(name), 2, exec}



/* Registers and variables */
static UDM_PROG_CMD
udm_tmpl_mov_ireg0_var0= UDM_CMD1_IREG0("GetInt i0 v0", Template_mov_ireg0_var0),
udm_tmpl_mov_ireg1_var0= UDM_CMD1_IREG1("GetInt i1 v0", Template_mov_ireg1_var0);

static UDM_PROG_CMD
udm_tmpl_mov_dreg0_var0= UDM_CMD1_DREG0("GetDbl d0 v0", Template_mov_dreg0_var0),
udm_tmpl_mov_dreg1_var0= UDM_CMD1_DREG0("GetDbl d1 v0", Template_mov_dreg1_var0);

static UDM_PROG_CMD
udm_tmpl_get_var_to_cs0_z0= UDM_CMD1_DREG0("GetStr cs0/z0 v0", Template_get_var_to_cs0_z0);

static UDM_PROG_CMD
udm_tmpl_textdata_to_cs0_z0= UDM_CMD1_DREG0("GetStr cs0/z0 textdata", Template_get_textdata_to_cs0_z0);

/* Registers and values */

static UDM_PROG_CMD
udm_tmpl_set_var_from_text0= UDM_CMD2_VOID("SetStr v0 text0", Template_set_var_from_text0),
udm_tmpl_set_var_from_ireg0=  UDM_CMD2_VOID("SetInt v0 i0", Template_set_var_from_ireg0),
udm_tmpl_set_var_from_dreg0=  UDM_CMD2_VOID("SetDbl v0 d0", Template_set_var_from_dreg0),
udm_tmpl_set_var_from_cs0_z0=   UDM_CMD2_VOID("SetStr v0 cs0/z0", Template_set_var_from_cs0_z0),
udm_tmpl_construct_var= UDM_CMD2_VOID("construct", Template_construct_var),
udm_tmpl_destruct_var= UDM_CMD2_VOID("destruct", Template_destruct_var);

/*********************************************************************/

static void
Exec_method_make_arg(UDM_PROG_EXECUTOR_STATE *T, size_t argno)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  const UDM_VALUE_HANDLER *handler= it->item_args.handler;
  T->vreg[argno].Data= UdmMalloc(handler->DataSize());
  handler->Constructor(T->vreg[argno].Data, NULL, 0);
  handler->SetStrn(T->vreg[argno].Data,
                   T->prog->textdata.Val.str + it->item_args.textdata_addr.offset,
                   it->item_args.textdata_addr.length);
}


static void
Exec_method_free_arg(UDM_PROG_EXECUTOR_STATE *T, size_t argno)
{
  const UDM_PROG_ITEM *it= UdmProgExecutorStateCurrentCmd(T);
  const UDM_VALUE_HANDLER *handler= it->item_args.handler;
  handler->Destructor(T->vreg[argno].Data);
  UdmFree(T->vreg[argno].Data);
  T->vreg[argno].Data= NULL;
}

/**** arg0 ***/

static void
Exec_method_make_arg0(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_make_arg(T, 0);
}

static void
Exec_method_free_arg0(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_free_arg(T, 0);
}


static void
Exec_method_clear_arg0(UDM_PROG_EXECUTOR_STATE *T)
{
  T->vreg[0].Data= NULL;
}


/*** arg1 ***/

static void
Exec_method_make_arg1(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_make_arg(T, 1);
}

static void
Exec_method_free_arg1(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_free_arg(T, 1);
}

static void
Exec_method_clear_arg1(UDM_PROG_EXECUTOR_STATE *T)
{
  T->vreg[1].Data= NULL;
}


/*** arg2 ***/

static void
Exec_method_make_arg2(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_make_arg(T, 2);
}

static void
Exec_method_free_arg2(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_free_arg(T, 2);
}

static void
Exec_method_clear_arg2(UDM_PROG_EXECUTOR_STATE *T)
{
  T->vreg[2].Data= NULL;
}

/*** arg3 ***/

static void
Exec_method_make_arg3(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_make_arg(T, 3);
}

static void
Exec_method_free_arg3(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_free_arg(T, 3);
}

static void
Exec_method_clear_arg3(UDM_PROG_EXECUTOR_STATE *T)
{
  T->vreg[3].Data= NULL;
}

/*** arg4 ***/

static void
Exec_method_make_arg4(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_make_arg(T, 4);
}

static void
Exec_method_free_arg4(UDM_PROG_EXECUTOR_STATE *T)
{
  Exec_method_free_arg(T, 4);
}

static void
Exec_method_clear_arg4(UDM_PROG_EXECUTOR_STATE *T)
{
  T->vreg[4].Data= NULL;
}

/****/

static UDM_PROG_CMD
udm_method_clear_arg0=  UDM_CMD1_VOID("mov  v0,0", Exec_method_clear_arg0),
udm_method_clear_arg1=  UDM_CMD1_VOID("mov  v1,0", Exec_method_clear_arg1),
udm_method_clear_arg2=  UDM_CMD1_VOID("mov  v2,0", Exec_method_clear_arg2),
udm_method_clear_arg3=  UDM_CMD1_VOID("mov  v3,0", Exec_method_clear_arg3),
udm_method_clear_arg4=  UDM_CMD1_VOID("mov  v4,0", Exec_method_clear_arg4);


static UDM_PROG_CMD
udm_method_make_arg0=   UDM_CMD1_VOID("MAKE-ARG0", Exec_method_make_arg0),
udm_method_free_arg0=   UDM_CMD1_VOID("FREE-ARG0", Exec_method_free_arg0);


static UDM_PROG_CMD
udm_method_make_arg1=   UDM_CMD1_VOID("MAKE-ARG1", Exec_method_make_arg1),
udm_method_free_arg1=   UDM_CMD1_VOID("FREE-ARG1", Exec_method_free_arg1);


static UDM_PROG_CMD
udm_method_make_arg2=   UDM_CMD1_VOID("MAKE-ARG2", Exec_method_make_arg2),
udm_method_free_arg2=   UDM_CMD1_VOID("FREE-ARG2", Exec_method_free_arg2);


static UDM_PROG_CMD
udm_method_make_arg3=   UDM_CMD1_VOID("MAKE-ARG3", Exec_method_make_arg3),
udm_method_free_arg3=   UDM_CMD1_VOID("FREE-ARG3", Exec_method_free_arg3);


static UDM_PROG_CMD
udm_method_make_arg4=   UDM_CMD1_VOID("MAKE-ARG4", Exec_method_make_arg4),
udm_method_free_arg4=   UDM_CMD1_VOID("FREE-ARG4", Exec_method_free_arg4);


/*********************************************************************/

static void
Template_cout_ireg0(UDM_PROG_EXECUTOR_STATE *T)
{
  fprintf(T->globals.stdout, "%d", T->reg.i[0]);
} 


static void
Template_cout_ireg0_char(UDM_PROG_EXECUTOR_STATE *T)
{
  char ch= T->reg.i[0];
  (void) fwrite(&ch, 1, 1, T->globals.stdout);
  
} 


static void
Template_cout_dreg0(UDM_PROG_EXECUTOR_STATE *T)
{
  fprintf(T->globals.stdout, "%f", T->reg.d[0]);
} 


static void
Template_cout_cs0_z0(UDM_PROG_EXECUTOR_STATE *T)
{
  fwrite(T->reg.cs0, T->reg.z0, 1, T->globals.stdout);
} 


/* New echo functions */
static UDM_PROG_CMD
udm_tmpl_cout_ireg0=          UDM_CMD1_VOID("cout i0", Template_cout_ireg0),
udm_tmpl_cout_ireg0_char=     UDM_CMD1_VOID("cout i0 (char)", Template_cout_ireg0_char),
udm_tmpl_cout_dreg0=          UDM_CMD1_VOID("cout d0", Template_cout_dreg0),
udm_tmpl_cout_cs0_z0=          UDM_CMD1_VOID("cout   cs0/z0", Template_cout_cs0_z0);


/*********************************************************************/


static void
UdmPIScanLexToken(UDM_LEX_SCANNER *p, UDM_LEX_TOKEN *a)
{
beg:
  UdmLexScannerSkipSpaces(p);
  a->token.str= UdmLexScannerCur(p);
  if (UdmLexScannerEOF(p))
  {
    a->type= UDM_LEX_EOF;
    a->token.end= a->token.str;
  }
  else if (!UdmLexScannerScanCComment(p, a))
  {
    goto beg;
  }
  else if (!UdmLexScannerScanCOperator(p, a))
  {
  }
  else if (!UdmLexScannerScanIdentifier(p, a))
  {
  }
  else if (!UdmLexScannerScanPunctuation(p, a))
  {
  }
  else if (!UdmLexScannerScanEscapedString(p, a))
  {
  }
  else if (!UdmLexScannerScanChar(p, a))
  {
  }
  else if (!UdmLexScannerScanUnsignedNumber(p, a))
  {
  }
  else
  {
    a->type= UDM_LEX_UNKNOWN;
  }

#if 0
  fprintf(stderr, "LEX=%s='%.*s'\n",
          UdmLex2str(a->type), (int) (a->token.end-a->token.str), a->token.str);
#endif
}


static void
UdmPIParserInit(UDM_PROG_COMPILER *compiler, const char *src, size_t length)
{
  UdmLexScannerInit(&compiler->scanner, src, length);
  UdmProgVarListListInit(&compiler->Vars2);
  compiler->errstr[0]= '\0';
  compiler->current_loop_continue= 0;
}


static void
UdmPIParserFree(UDM_PROG_COMPILER *compiler)
{
  UdmProgVarListListFree(&compiler->Vars2);
}


static size_t
UdmProgCompilerLineno(UDM_PROG_COMPILER *compiler)
{
  const char *s;
  size_t lineno= compiler->lineno;
  for (s= compiler->start; s < compiler->scanner.token.token.str; s++)
  {
    if (*s == '\n')
      lineno++;
  }
  return lineno + 1;
}


static size_t
UdmProgCompilerNextVarStackOffset(UDM_PROG_COMPILER *compiler)
{
  UDM_PROG_VARLIST *Vars= &compiler->Vars2.Item[compiler->Vars2.nitems - 1];
  UDM_ASSERT(compiler->Vars2.nitems);
  return Vars->next_stack_offset;
}


static udm_rc_t
UdmProgCompilerNextVarStackOffsetIncrement(UDM_PROG_COMPILER *compiler,
                                           size_t inc)
{
  UDM_PROG_VARLIST *Vars= &compiler->Vars2.Item[compiler->Vars2.nitems - 1];
  Vars->next_stack_offset+= inc;
  return UDM_OK;
}


static udm_bool_t
UdmProgCompilerEnterBlock(UDM_PROG_COMPILER *compiler)
{
  UDM_PROG_VARLIST tmp;
  UdmProgVarListInit(&tmp);
  tmp.next_stack_offset= compiler->Vars2.nitems ?
                         UdmProgCompilerNextVarStackOffset(compiler) : 0;
  if (UDM_OK != UdmProgVarListListAdd(&compiler->Vars2, &tmp))
    return UDM_FALSE;
  return UDM_TRUE;
}


/**
  Generate destructors for all variables in the list,
  in reverse creation order.
*/
static udm_bool_t
UdmProgCompilerGenerateVarListDestructors(UDM_PROG_COMPILER *compiler,
                                          UDM_PROG_VARLIST *Vars)
{
  size_t i;
  for (i= 0; i < Vars->nitems; i++)
  {
    const UDM_PROG_VAR2 *Var= &Vars->Item[Vars->nitems - i - 1];
    UDM_PROG_ITEM it;
    UdmProgItemInit(&it);
    it.item_args.stack_offset= Var->value.stack_offset;
    it.item_args.handler= Var->value.handler;
    it.cmd= &udm_tmpl_destruct_var;
    if (compiler->generate_debug_info)
      it.comment= UdmStrdup(Var->name);
    if (UDM_OK != UdmProgAdd(compiler->prg, &it))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmProgCompilerGenerateVarListListDestructors(UDM_PROG_COMPILER *compiler)
{
  size_t i;
  /*
    Generate destructors for all variables on all levels,
    to avoid memory leaks.
  */
  for (i= compiler->Vars2.nitems; i > 0; i--)
  {
    UDM_PROG_VARLIST *Vars= &compiler->Vars2.Item[i - 1];
    if (!UdmProgCompilerGenerateVarListDestructors(compiler, Vars))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmProgCompilerLeaveBlock(UDM_PROG_COMPILER *compiler)
{
  UDM_PROG_VARLIST *Vars= &compiler->Vars2.Item[compiler->Vars2.nitems - 1];
  UDM_ASSERT(compiler->Vars2.nitems);
  if (!UdmProgCompilerGenerateVarListDestructors(compiler, Vars))
    return UDM_FALSE;
  UdmProgVarListFree(Vars);
  compiler->Vars2.nitems--;
  return UDM_TRUE;
}


static udm_bool_t
UdmProgCompilerRedeclarationError(UDM_PROG_COMPILER *compiler,
                                  const char *name)
{
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%d: ERROR: Redeclaration of '%s'",
               (int) UdmProgCompilerLineno(compiler), name);
  return UDM_FALSE;
}


typedef struct
{
  udm_unary_operator_t op;
  UDM_VALUE_HANDLER *ha;
} UDM_UNARY_OPERATOR;


static UDM_UNARY_OPERATOR unary_operators[]=
{
  {UDM_UNARY_OPERATOR_PREFIX_INC, &UdmValueHandlerInt},
  {UDM_UNARY_OPERATOR_PREFIX_DEC, &UdmValueHandlerInt},
  {UDM_UNARY_OPERATOR_POSTFIX_INC, &UdmValueHandlerInt},
  {UDM_UNARY_OPERATOR_POSTFIX_DEC, &UdmValueHandlerInt},

  {UDM_UNARY_OPERATOR_PREFIX_INC,  &UdmValueHandlerChar},
  {UDM_UNARY_OPERATOR_PREFIX_DEC,  &UdmValueHandlerChar},
  {UDM_UNARY_OPERATOR_POSTFIX_INC, &UdmValueHandlerChar},
  {UDM_UNARY_OPERATOR_POSTFIX_DEC, &UdmValueHandlerChar},

  {UDM_UNARY_OPERATOR_PREFIX_INC,  &UdmValueHandlerDouble},
  {UDM_UNARY_OPERATOR_PREFIX_DEC,  &UdmValueHandlerDouble},
  {UDM_UNARY_OPERATOR_POSTFIX_INC, &UdmValueHandlerDouble},
  {UDM_UNARY_OPERATOR_POSTFIX_DEC, &UdmValueHandlerDouble},

  {UDM_UNARY_OPERATOR_PREFIX_INC, NULL}
};


static const UDM_UNARY_OPERATOR *
UdmUnaryOperatorFind(const UDM_UNARY_OPERATOR *ops,
                     udm_unary_operator_t op,
                     const UDM_VALUE_HANDLER *h1)
{
  for (  ; ops->ha; ops++)
  {
    if (ops->ha == h1 && ops->op == op)
      return ops;
  }
  return NULL;
}


static const UDM_UNARY_OPERATOR *
UdmUnaryOperatorFindWithError(UDM_PROG_COMPILER *compiler,
                              const UDM_UNARY_OPERATOR *ops,
                              udm_unary_operator_t op,
                              const UDM_VALUE_HANDLER *ha,
                              const char *name)
{
  const UDM_UNARY_OPERATOR *res;
  if ((res= UdmUnaryOperatorFind(ops, op, ha)))
    return res;
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%d: no operator '%s' declared for '%s'",
               (int) UdmProgCompilerLineno(compiler),
               name, ha->type_name);
  return NULL;
}


typedef struct
{
  const UDM_VALUE_HANDLER *h1;
  const UDM_VALUE_HANDLER *h2;
  const UDM_VALUE_HANDLER *res; /* Aggregate type for h1 and h2 */
} UDM_DYADIC_OPERATOR;


static UDM_DYADIC_OPERATOR assignment_operators[]=
{
  {&UdmValueHandlerDouble,    &UdmValueHandlerDouble,    NULL},
  {&UdmValueHandlerDouble,    &UdmValueHandlerInt,       NULL},
  {&UdmValueHandlerDouble,    &UdmValueHandlerChar,      NULL},
  {&UdmValueHandlerInt,       &UdmValueHandlerDouble,    NULL},
  {&UdmValueHandlerInt,       &UdmValueHandlerInt,       NULL},
  {&UdmValueHandlerInt,       &UdmValueHandlerChar,      NULL},
  {&UdmValueHandlerChar,      &UdmValueHandlerDouble,    NULL},
  {&UdmValueHandlerChar,      &UdmValueHandlerInt,       NULL},
  {&UdmValueHandlerChar,      &UdmValueHandlerChar,      NULL},
  {&UdmValueHandlerSimple,    &UdmValueHandlerSimple,    NULL},
  {&UdmValueHandlerQueryWord, &UdmValueHandlerQueryWord, NULL},
  {NULL, NULL}
};


static const UDM_DYADIC_OPERATOR *
UdmDyadicOperatorFind(UDM_DYADIC_OPERATOR *ops,
                      const UDM_VALUE_HANDLER *h1,
                      const UDM_VALUE_HANDLER *h2)
{
  const UDM_DYADIC_OPERATOR *op;
  for (op= ops; op->h1; op++)
  {
    if (op->h1 == h1 && op->h2 == h2)
      return op;
  }
  return NULL;
}


static udm_bool_t
UdmCompilerErrorInvalidOperands(UDM_PROG_COMPILER *compiler,
                                const UDM_VALUE_HANDLER *h1,
                                const UDM_VALUE_HANDLER *h2,
                                const char *name)
{
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%d: Invalid operands of types for '%s' and '%s' to operator '%s'",
               (int) UdmProgCompilerLineno(compiler),
               h1->type_name, h2->type_name, name);
  return UDM_FALSE;
}


static const UDM_DYADIC_OPERATOR *
UdmDyadicOperatorFindWithError(UDM_PROG_COMPILER *compiler,
                               UDM_DYADIC_OPERATOR *ops,
                               const UDM_VALUE_HANDLER *h1,
                               const UDM_VALUE_HANDLER *h2,
                               const char *name)
{
  const UDM_DYADIC_OPERATOR *op;
  if ((op= UdmDyadicOperatorFind(ops, h1, h2)))
    return op;
  UdmCompilerErrorInvalidOperands(compiler, h1, h2, name);
  return NULL;
}


static udm_bool_t
UdmProgCompilerError(const UDM_PROG_COMPILER *compiler)
{
  return compiler->errstr[0] != '\0';
}


static void
UdmProgCompilerScan(UDM_PROG_COMPILER *compiler)
{
  UdmPIScanLexToken(&compiler->scanner, &compiler->scanner.token);
}


static udm_lex_t
UdmProgCompilerLastTokenType(const UDM_PROG_COMPILER *compiler)
{
  return compiler->scanner.token.type;
}


static udm_bool_t
ParseTerm(UDM_PROG_COMPILER *compiler, udm_lex_t term)
{
  if (UdmProgCompilerError(compiler))
    return UDM_FALSE;
  if (UdmProgCompilerLastTokenType(compiler) != term)
    return UDM_FALSE;
  UdmProgCompilerScan(compiler);
  return UDM_TRUE;
}


static udm_bool_t
ParseTermOrError(UDM_PROG_COMPILER *compiler, udm_lex_t term)
{
  if (UdmProgCompilerError(compiler))
    return UDM_FALSE;
  if (UdmProgCompilerLastTokenType(compiler) != term)
  {
    udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
                 "%d: ERROR: Expected %s got %s",
                 (int) UdmProgCompilerLineno(compiler),
                 UdmLex2str(term),
                 UdmLex2str(UdmProgCompilerLastTokenType(compiler)));
    return UDM_FALSE;
  }
  UdmProgCompilerScan(compiler);
  return UDM_TRUE;
}


static udm_bool_t
UdmProgCompilerErrorMultiCharacter(UDM_PROG_COMPILER *compiler)
{
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%d: error: multi-character character constant",
               (int) UdmProgCompilerLineno(compiler));
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerGenericError(UDM_PROG_COMPILER *compiler, const char *msg)
{
  if (!UdmProgCompilerError(compiler))
    udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
                 "%d: ERROR: %s",
                 (int) UdmProgCompilerLineno(compiler), msg);
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerExpectedRule(UDM_PROG_COMPILER *compiler, const char *expected)
{
  if (!UdmProgCompilerError(compiler))
    udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
                 "%d: ERROR: Expected %s got %s",
                 (int) UdmProgCompilerLineno(compiler), expected,
                 UdmLex2str(UdmProgCompilerLastTokenType(compiler)));
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerNotSupported(UDM_PROG_COMPILER *compiler, const char *msg)
{
  if (!UdmProgCompilerError(compiler))
    udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
                 "%d: ERROR: %s is not supported",
                 (int) UdmProgCompilerLineno(compiler), msg);
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerExponentHasNoDigits(UDM_PROG_COMPILER *compiler)
{
  if (!UdmProgCompilerError(compiler))
    udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
                 "%d: ERROR: exponent has no digits",
                 (int) UdmProgCompilerLineno(compiler));
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerExponentInvalidSuffix(UDM_PROG_COMPILER *compiler,
                                     const UDM_CONST_TOKEN *exponent)
{
  if (!UdmProgCompilerError(compiler))
    udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
                 "%d: ERROR: invalid suffix \"%.*s\" on floating exponent",
                 (int) UdmProgCompilerLineno(compiler),
                 (int) (exponent->end - exponent->str), exponent->str);
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerUnknownIdentifierError(UDM_PROG_COMPILER *compiler,
                                      const UDM_CONST_TOKEN *name)
{
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%d: ERROR: Unknown identifier '%.*s'",
               (int) UdmProgCompilerLineno(compiler),
               (int) (name->end - name->str), name->str);
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerBadDataTypeForArg(UDM_PROG_COMPILER *compiler, size_t argno,
                                 const UDM_VALUE_HANDLER *expected,
                                 const UDM_VALUE_HANDLER *actual)
{
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%d: ERROR: Bad data type for argument #%d: expected '%s', got '%s'",
               (int) UdmProgCompilerLineno(compiler),
               (int) argno, expected->type_name, actual->type_name);
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerBadDataTypeForOperator2(UDM_PROG_COMPILER *compiler,
                                       const char *name,
                                       const UDM_VALUE_HANDLER *arg1,
                                       const UDM_VALUE_HANDLER *arg2)
{
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%d: ERROR: Invalid operands to binary operator '%s' "
               "(have '%s' and '%s')",
               (int) UdmProgCompilerLineno(compiler),
               name, arg1->type_name, arg2->type_name);
  return UDM_FALSE;
}


typedef struct
{
  UDM_CONST_STR name;
  UDM_VALUE_HANDLER *handler;
} UDM_TYPE_SPECIFIER;

#define STD_TYPE_SPECIFIER(x, h) {{UDM_CSTR_WITH_LEN(x)}, h}

static UDM_TYPE_SPECIFIER std_type_specifiers[]=
{
  STD_TYPE_SPECIFIER("string", &UdmValueHandlerSimple),
  STD_TYPE_SPECIFIER("ENV", &UdmValueHandlerEnv),
  STD_TYPE_SPECIFIER("RESULT", &UdmValueHandlerResult),
  STD_TYPE_SPECIFIER("DOCUMENT", &UdmValueHandlerDocument),
  STD_TYPE_SPECIFIER("QUERYWORD", &UdmValueHandlerQueryWord),
  {{0, 0}}
};

/*
<type-specifier> ::= void
                   | char
                   | short
                   | int
                   | long
                   | float
                   | double
                   | signed
                   | unsigned
                   | <struct-or-union-specifier>
                   | <enum-specifier>
                   | <typedef-name>
*/
static const UDM_VALUE_HANDLER *
FindTypeByName(const UDM_CONST_TOKEN *name)
{
  UDM_TYPE_SPECIFIER *t;
  for (t= std_type_specifiers; t->name.str; t++)
  {
    size_t name_length= name->end - name->str;
    if (name_length == t->name.length &&
        !memcmp(name->str, t->name.str, name_length))
      return t->handler;
  }
  return NULL;
}


static udm_bool_t
ParseTypeSpecifier(UDM_PROG_COMPILER *compiler,
                   const UDM_VALUE_HANDLER **ha)
{
  if (ParseTerm(compiler, UDM_LEX_CHAR))
  {
    ha[0]= &UdmValueHandlerChar;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_INT))
  {
    ha[0]= &UdmValueHandlerInt;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_DOUBLE))
  {
    ha[0]= &UdmValueHandlerDouble;
    return UDM_TRUE;
  }
  if (UdmProgCompilerLastTokenType(compiler) != UDM_LEX_IDENT)
    return UDM_FALSE;
  if (!(ha[0]= FindTypeByName(&compiler->scanner.token.token)))
    return UDM_FALSE;
  return ParseTerm(compiler, UDM_LEX_IDENT);
}


/**
  This needs to do look-ahead for one more step, to distinguish between:
    x= (ident);            // a parenthesized expression
    x= (type) expression;  // a type cast
*/
static udm_bool_t
TestParenthesizedTypeSpecifier(UDM_PROG_COMPILER *compiler)
{
  udm_bool_t rc;
  UDM_LEX_SCANNER scanner;
  const UDM_VALUE_HANDLER *ha;
  if (UdmProgCompilerLastTokenType(compiler) != UDM_LEX_LP)
    return UDM_FALSE;
  scanner= compiler->scanner;
  if (!ParseTerm(compiler, UDM_LEX_LP))
  {
    UDM_ASSERT(0);
    return UDM_FALSE;  
  }
  rc= ParseTypeSpecifier(compiler, &ha) && ParseTerm(compiler, UDM_LEX_RP);
  compiler->scanner= scanner; /* Rollback to the initial state */
  return rc;
}


/*
<declaration-specifier> ::= <storage-class-specifier>
                          | <type-specifier>
                          | <type-qualifier>
*/
static udm_bool_t
ParseDeclarationSpecifier(UDM_PROG_COMPILER *compiler,
                          const UDM_VALUE_HANDLER **ha)
{
  return ParseTypeSpecifier(compiler, ha);
}


static void
UdmDSTRAppendArgTypes(UDM_DSTR *dstr,
                      const UDM_PROG_FUNCTION_CALL_ARGUMENTS *arglist)
{
  size_t i;
  for (i= 0; i <  arglist->nargs; i++)
  {
    if (i > 0)
      UdmDSTRAppend(dstr, ",", 1);
    UdmDSTRAppendSTR(dstr, UdmPrimaryExprHandler(&arglist->args[i])->type_name);
  }
}


static udm_bool_t
UdmProgCompilerErrorUnknownFunction(UDM_PROG_COMPILER *compiler,
                                    const UDM_CONST_TOKEN *name,
                                    const UDM_PROG_FUNCTION_CALL_ARGUMENTS *arglist)
{
  UDM_DSTR dstr;
  UdmDSTRInit(&dstr, 128);
  UdmDSTRAppendf(&dstr, "%d: ERROR: unknown function: %.*s(",
                 (int) UdmProgCompilerLineno(compiler),
                 (int) UdmConstTokenLength(name), UdmConstTokenStr(name));
  UdmDSTRAppendArgTypes(&dstr, arglist);
  UdmDSTRAppend(&dstr, ")", 1);
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr), "%.*s",
               (int) UdmDSTRLength(&dstr), UdmDSTRPtr(&dstr));
  UdmDSTRFree(&dstr);
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerUnknownMethodError(UDM_PROG_COMPILER *compiler,
                                  const UDM_VALUE_HANDLER *handler,
                                  const UDM_CONST_TOKEN *name,
                                  const UDM_PROG_FUNCTION_CALL_ARGUMENTS *args)
{
  size_t i;
  UDM_DSTR dstr;
  UdmDSTRInit(&dstr, 128);
  UdmDSTRAppendf(&dstr, "%d: ERROR: %s has no method '%.*s(",
               (int) UdmProgCompilerLineno(compiler),
               handler->type_name, (int) UdmConstTokenLength(name),
               UdmConstTokenStr(name));
  for (i= 0; i < args->nargs; i++)
  {
    UdmDSTRAppendf(&dstr, "%s%s&", i > 0 ? "," : "",
                   UdmPrimaryExprHandler(&args->args[i])->type_name);
  }
  UdmDSTRAppend(&dstr, ")'", 2);
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%.*s", (int) UdmDSTRLength(&dstr), UdmDSTRPtr(&dstr));
  UdmDSTRFree(&dstr);
  return UDM_FALSE;
}


static udm_bool_t
UdmProgCompilerVoidValueIsNotIgnored(UDM_PROG_COMPILER *compiler)
{
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "%d: ERROR: void value is not ignored as it ought to be",
               (int) UdmProgCompilerLineno(compiler));
  return UDM_FALSE;
}


static udm_bool_t
UdmProgGenerateVariableItemWithComment(UDM_PROG_COMPILER *compiler,
                                       const UDM_PROG_VAR *Var,
                                       UDM_PROG_ITEM *it)
{
  UdmProgItemArgInitFromVar(&it->item_args, Var);
  if (compiler->generate_debug_info)
    it->comment= UdmConstTokenStrDup(&Var->name);
  return UdmProgAdd(compiler->prg, it) == UDM_OK;
}


static udm_bool_t
UdmGenerateVariableOp(UDM_PROG_COMPILER *compiler,
                      const UDM_PROG_VAR *Var,
                      udm_tmpl_cmd_t op)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmdnum= op;
  return UdmProgGenerateVariableItemWithComment(compiler, Var, &i);
}


static udm_bool_t
UdmGenerateVariableOperation(UDM_PROG_COMPILER *compiler,
                             const UDM_PROG_VAR *Var,
                             UDM_PROG_CMD *cmd)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmd= cmd;
  return UdmProgGenerateVariableItemWithComment(compiler, Var, &i);
}


static udm_bool_t
UdmGenerateConstructor(UDM_PROG_COMPILER *c, const UDM_PROG_VAR *Var)
{
  return UdmGenerateVariableOperation(c, Var, &udm_tmpl_construct_var);
}


static udm_bool_t
UdmGenerateVariableOperationToIReg0(UDM_PROG_COMPILER *compiler,
                                    const UDM_PROG_VAR *Var)
{
  if (UdmProgVarHandler(Var)->datatype == UDM_VALUE_DATA_TYPE_INT)
  {
    UDM_PROG_ITEM it;
    UdmProgItemInit(&it);
    it.cmdnum= UDM_PROG_MOV_IREG0_INT_PTR_STACK_OFFSET;
    return UdmProgGenerateVariableItemWithComment(compiler, Var, &it);
  }
  return UdmGenerateVariableOperation(compiler, Var, &udm_tmpl_mov_ireg0_var0);
}


static udm_bool_t
UdmGenerateVariableOperationToDReg0(UDM_PROG_COMPILER *compiler,
                                    const UDM_PROG_VAR *Var)
{
  if (UdmProgVarHandler(Var)->datatype == UDM_VALUE_DATA_TYPE_DOUBLE)
  {
    UDM_PROG_ITEM it;
    UdmProgItemInit(&it);
    it.cmdnum= UDM_PROG_MOV_DREG0_DBL_PTR_STACK_OFFSET;
    return UdmProgGenerateVariableItemWithComment(compiler, Var, &it);
  }
  return UdmGenerateVariableOperation(compiler, Var, &udm_tmpl_mov_dreg0_var0);
}


static udm_bool_t
UdmGenerateVariableOperationToIReg1(UDM_PROG_COMPILER *compiler,
                                    const UDM_PROG_VAR *Var)
{
  if (UdmProgVarHandler(Var)->datatype == UDM_VALUE_DATA_TYPE_INT)
  {
    UDM_PROG_ITEM it;
    UdmProgItemInit(&it);
    it.cmdnum= UDM_PROG_MOV_IREG1_INT_PTR_STACK_OFFSET;
    return UdmProgGenerateVariableItemWithComment(compiler, Var, &it);
  }
  return UdmGenerateVariableOperation(compiler, Var, &udm_tmpl_mov_ireg1_var0);
}


static udm_bool_t
UdmGenerateVariableOperationToDReg1(UDM_PROG_COMPILER *compiler,
                                    const UDM_PROG_VAR *Var)
{
  if (UdmProgVarHandler(Var)->datatype == UDM_VALUE_DATA_TYPE_DOUBLE)
  {
    UDM_PROG_ITEM it;
    UdmProgItemInit(&it);
    it.cmdnum= UDM_PROG_MOV_DREG1_DBL_PTR_STACK_OFFSET;
    return UdmProgGenerateVariableItemWithComment(compiler, Var, &it);
  }
  return UdmGenerateVariableOperation(compiler, Var, &udm_tmpl_mov_dreg1_var0);
}


static udm_bool_t
UdmGenerateBooleanTest(UDM_PROG_COMPILER *compiler,
                       const UDM_VALUE_HANDLER *handler,
                       udm_bool_t positive)
{
  if (!handler)
    return UdmProgCompilerVoidValueIsNotIgnored(compiler);
  switch (handler->datatype)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    break;
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    return UdmProgCompilerNotSupported(compiler, "boolean test for data type");
  }

  if (handler == &UdmValueHandlerInt)
  {
    if (UDM_OK != UdmProgAddArg0SimpleOp(compiler->prg,
                                         UDM_PROG_MOV_IREG1_0))
      return UDM_FALSE;
    if (UDM_OK != UdmProgAddArg0SimpleOp(compiler->prg,
                                          positive ?
                                          UDM_PROG_CMP_IREG0_IREG1_NE:
                                          UDM_PROG_CMP_IREG0_IREG1_EQ))
      return UDM_FALSE;
  }
  else
  {
    if (UDM_OK != UdmProgAddArg0SimpleOp(compiler->prg,
                                         UDM_PROG_MOV_DREG1_0))
      return UDM_FALSE;
    if (UDM_OK != UdmProgAddArg0SimpleOp(compiler->prg,
                                         positive ?
                                         UDM_PROG_CMP_DREG0_DREG1_NE :
                                         UDM_PROG_CMP_DREG0_DREG1_EQ))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*
  Store the value of (!expr) into ireg0.
*/
static udm_bool_t
UdmGenerateIsNotTrueTest(UDM_PROG_COMPILER *compiler,
                         const UDM_VALUE_HANDLER *handler)
{
  return UdmGenerateBooleanTest(compiler, handler, UDM_FALSE);
}


static udm_bool_t
UdmGenerateIsTrueTest(UDM_PROG_COMPILER *compiler,
                      const UDM_VALUE_HANDLER *handler)
{
  return UdmGenerateBooleanTest(compiler, handler, UDM_TRUE);
}



static udm_bool_t
UdmGenerateSwapNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                              const UDM_VALUE_HANDLER *handler)
{
  return UdmGenerateOp(compiler,
                       handler->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE ?
                       UDM_PROG_SWP_DREG0_DREG1:
                       UDM_PROG_SWP_IREG0_IREG1);
}


static udm_bool_t
UdmGenerateAddNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                             const UDM_VALUE_HANDLER **handler,
                             udm_lex_t op)
{
  UDM_ASSERT(op == UDM_LEX_PLUS || op == UDM_LEX_MINUS);
  switch (handler[0]->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
    return UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg,
                                            op == UDM_LEX_PLUS ?
                                            UDM_PROG_ADD_IREG0_IREG1 :
                                            UDM_PROG_SUB_IREG0_IREG1);

  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg,
                                            op == UDM_LEX_PLUS ?
                                            UDM_PROG_ADD_DREG0_DREG1:
                                            UDM_PROG_SUB_DREG0_DREG1);
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateMulNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                             const UDM_VALUE_HANDLER **handler,
                             udm_lex_t op)
{
  UDM_ASSERT(op == UDM_LEX_ASTERISK || op == UDM_LEX_SLASH || op == UDM_LEX_PERCENT);
  switch (handler[0]->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateOp(compiler, op == UDM_LEX_ASTERISK ?
                                   UDM_PROG_MUL_IREG0_IREG1 :
                                   op == UDM_LEX_SLASH ?
                                   UDM_PROG_DIV_IREG0_IREG1 :
                                   UDM_PROG_REM_IREG0_IREG1);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    if (op == UDM_LEX_PERCENT)
      return UdmProgCompilerNotSupported(compiler, "binary operator % for the `double` data type");
    return UdmGenerateOp(compiler, op == UDM_LEX_ASTERISK ?
                                   UDM_PROG_MUL_DREG0_DREG1 :
                                   UDM_PROG_DIV_DREG0_DREG1);
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateShiftNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                             const UDM_VALUE_HANDLER **handler,
                             udm_lex_t op)
{
  UDM_ASSERT(op == UDM_LEX_LSHIFT || op == UDM_LEX_RSHIFT);
  switch (handler[0]->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateOp(compiler, op == UDM_LEX_LSHIFT ?
                                   UDM_PROG_SHL_IREG0_IREG1 :
                                   UDM_PROG_SHR_IREG0_IREG1);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmProgCompilerNotSupported(compiler, "<< and >> for this data type");
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateRelNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                             const UDM_VALUE_HANDLER **handler,
                             udm_lex_t op)
{
  UDM_ASSERT(op == UDM_LEX_LE || op == UDM_LEX_LT ||
             op == UDM_LEX_GT || op == UDM_LEX_GE);
  switch (handler[0]->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
    switch (op)
    {
    case UDM_LEX_LE: return UdmGenerateOp(compiler, UDM_PROG_CMP_IREG0_IREG1_LE);
    case UDM_LEX_LT: return UdmGenerateOp(compiler, UDM_PROG_CMP_IREG0_IREG1_LT);
    case UDM_LEX_GT: return UdmGenerateOp(compiler, UDM_PROG_CMP_IREG0_IREG1_GT);
    case UDM_LEX_GE: return UdmGenerateOp(compiler, UDM_PROG_CMP_IREG0_IREG1_GE);
    default:
      break;
    }
    break;
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    *handler= &UdmValueHandlerInt;
    switch (op)
    {
    case UDM_LEX_LE: return UdmGenerateOp(compiler, UDM_PROG_CMP_DREG0_DREG1_LE);
    case UDM_LEX_LT: return UdmGenerateOp(compiler, UDM_PROG_CMP_DREG0_DREG1_LT);
    case UDM_LEX_GT: return UdmGenerateOp(compiler, UDM_PROG_CMP_DREG0_DREG1_GT);
    case UDM_LEX_GE: return UdmGenerateOp(compiler, UDM_PROG_CMP_DREG0_DREG1_GE);
    default:
      break;
    }
    break;

  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateEqNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                            const UDM_VALUE_HANDLER **handler,
                            udm_lex_t op)
{
  UDM_ASSERT(op == UDM_LEX_EQ_EQ || op == UDM_LEX_NOT_EQ);
  switch (handler[0]->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
    switch (op)
    {
    case UDM_LEX_EQ_EQ:  return UdmGenerateOp(compiler, UDM_PROG_CMP_IREG0_IREG1_EQ);
    case UDM_LEX_NOT_EQ: return UdmGenerateOp(compiler, UDM_PROG_CMP_IREG0_IREG1_NE);
    default:
      break;
    }
    break;
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    *handler= &UdmValueHandlerInt;
    switch (op)
    {
    case UDM_LEX_EQ_EQ:  return UdmGenerateOp(compiler, UDM_PROG_CMP_DREG0_DREG1_EQ);
    case UDM_LEX_NOT_EQ: return UdmGenerateOp(compiler, UDM_PROG_CMP_DREG0_DREG1_NE);
    default:
      break;
    }
    break;
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateAndNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                             const UDM_VALUE_HANDLER **handler,
                             udm_lex_t op)
{
  UDM_ASSERT(op == UDM_LEX_AMPERSAND);
  switch (handler[0]->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateOp(compiler, UDM_PROG_AND_IREG0_IREG1);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmProgCompilerNotSupported(compiler, "binary & for this data type");
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateXorNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                             const UDM_VALUE_HANDLER **handler,
                             udm_lex_t op)
{
  UDM_ASSERT(op == UDM_LEX_CARET);
  switch (handler[0]->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateOp(compiler, UDM_PROG_XOR_IREG0_IREG1);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmProgCompilerNotSupported(compiler, "binary ^ for this data type");
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateOrNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                            const UDM_VALUE_HANDLER **handler,
                            udm_lex_t op)
{
  UDM_ASSERT(op == UDM_LEX_VBAR);
  switch (handler[0]->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateOp(compiler, UDM_PROG_OR_IREG0_IREG1);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmProgCompilerNotSupported(compiler, "binary | for this data type");
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateDiadicOperationNativeReg0Reg1(UDM_PROG_COMPILER *compiler,
                                         const UDM_VALUE_HANDLER **handler,
                                         udm_lex_t op)
{
  switch (op)
  {
  case UDM_LEX_VBAR:
    return UdmGenerateOrNativeReg0Reg1(compiler, handler, op);
  case UDM_LEX_CARET:
    return UdmGenerateXorNativeReg0Reg1(compiler, handler, op);
  case UDM_LEX_AMPERSAND:
    return UdmGenerateAndNativeReg0Reg1(compiler, handler, op);
  case UDM_LEX_EQ_EQ:
  case UDM_LEX_NOT_EQ:
    return UdmGenerateEqNativeReg0Reg1(compiler, handler, op);
  case UDM_LEX_LE:
  case UDM_LEX_GE:
  case UDM_LEX_LT:
  case UDM_LEX_GT:
    return UdmGenerateRelNativeReg0Reg1(compiler, handler, op);
  case UDM_LEX_ASTERISK:
  case UDM_LEX_SLASH:
  case UDM_LEX_PERCENT:
    return UdmGenerateMulNativeReg0Reg1(compiler, handler, op);
  case UDM_LEX_PLUS:
  case UDM_LEX_MINUS:
    return UdmGenerateAddNativeReg0Reg1(compiler, handler, op);
  case UDM_LEX_LSHIFT:
  case UDM_LEX_RSHIFT:
    return UdmGenerateShiftNativeReg0Reg1(compiler, handler, op);
  default:
    UDM_ASSERT(0);
  }
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateVarToNativeReg0(UDM_PROG_COMPILER *compiler,
                           const UDM_PROG_VAR *src)
{
  switch (UdmProgVarHandler(src)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateVariableOperationToIReg0(compiler, src);
  case UDM_VALUE_DATA_TYPE_CHAR:
    return UdmGenerateVariableOperation(compiler, src, &udm_tmpl_mov_ireg0_var0);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmGenerateVariableOperationToDReg0(compiler, src);
  default:
    return UdmGenerateVariableOperation(compiler, src, &udm_tmpl_get_var_to_cs0_z0);
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateVarToNativeReg1(UDM_PROG_COMPILER *compiler,
                           const UDM_PROG_VAR *src)
{
  switch (UdmProgVarHandler(src)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateVariableOperationToIReg1(compiler, src);
  case UDM_VALUE_DATA_TYPE_CHAR:
    return UdmGenerateVariableOperation(compiler, src, &udm_tmpl_mov_ireg1_var0);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmGenerateVariableOperationToDReg1(compiler, src);
  default:
    return UdmProgCompilerNotSupported(compiler, "Variable of a complex type in this context");
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateNativeReg0ToVar(UDM_PROG_COMPILER *compiler,
                           const UDM_PROG_VAR *Var)
{
  switch (UdmProgVarHandler(Var)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
  case UDM_VALUE_DATA_TYPE_CHAR:
    return UdmGenerateVariableOperation(compiler, Var, &udm_tmpl_set_var_from_ireg0);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmGenerateVariableOperation(compiler, Var, &udm_tmpl_set_var_from_dreg0);
  default:
    UDM_ASSERT(0);
    return UDM_FALSE;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateLiteralToNativeRegNInt(UDM_PROG_COMPILER *compiler,
                                  const UDM_PROG_PRIMARY_EXPRESSION *expr,
                                  size_t regno)
{
  udm_tmpl_cmd_t cmd= !regno ? UDM_PROG_MOV_IREG0_INT : UDM_PROG_MOV_IREG1_INT;
  int tmp= udm_strntoi(expr->Var.name.str, expr->Var.name.end - expr->Var.name.str);
  if (UDM_OK != UdmProgAddArg1SimpleOpInt(compiler->prg, cmd, tmp))
    return UDM_FALSE;
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateLiteralToNativeRegNDbl(UDM_PROG_COMPILER *compiler,
                                  const UDM_PROG_PRIMARY_EXPRESSION *expr,
                                  size_t regno)
{
  udm_tmpl_cmd_t cmd= !regno ? UDM_PROG_MOV_DREG0_DBL : UDM_PROG_MOV_DREG1_DBL;
  double tmp= udm_strntod(expr->Var.name.str, expr->Var.name.end - expr->Var.name.str);
  if (UDM_OK != UdmProgAddArg1SimpleOpDouble(compiler->prg, cmd, tmp))
    return UDM_FALSE;
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateLiteralToNativeRegNChar(UDM_PROG_COMPILER *compiler,
                                   const UDM_PROG_PRIMARY_EXPRESSION *expr,
                                   size_t regno)
{
  udm_tmpl_cmd_t cmd= !regno ? UDM_PROG_MOV_IREG0_INT : UDM_PROG_MOV_IREG1_INT;
  return UDM_OK == UdmProgAddArg1SimpleOpInt(compiler->prg,
                                             cmd, expr->Var.name.str[0]);
}


static udm_bool_t
UdmGenerateTextSegmentTo_cs0_z0(UDM_PROG_COMPILER *compiler,
                                udm_offset_and_length_t addr)
{
  UDM_PROG_ITEM i;
  UdmProgItemInit(&i);
  i.cmd= &udm_tmpl_textdata_to_cs0_z0;
  i.item_args.textdata_addr= addr;
  return UDM_OK == UdmProgAdd(compiler->prg, &i);
}


static udm_bool_t
UdmGenerateMovNativeReg0ToReg1(UDM_PROG_COMPILER *compiler,
                               const UDM_VALUE_HANDLER *handler)
{
  switch (handler->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_INT:
    return UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_IREG1_IREG0);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_DREG1_DREG0);
  default:
    UDM_ASSERT(0);
  }
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateClearNativeReg0(UDM_PROG_COMPILER *compiler,
                           const UDM_VALUE_HANDLER *handler)
{
  switch (handler->native_reg_type)
  {
  case UDM_VALUE_DATA_TYPE_INT:
    return UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_IREG0_0);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_DREG0_0);
  default:
    UDM_ASSERT(0);
  }
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateLiteralToNativeRegN(UDM_PROG_COMPILER *compiler,
                               const UDM_PROG_PRIMARY_EXPRESSION *expr,
                               size_t regno)
{
  switch (UdmPrimaryExprHandler(expr)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmGenerateLiteralToNativeRegNDbl(compiler, expr, regno);
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateLiteralToNativeRegNInt(compiler, expr, regno);
  case UDM_VALUE_DATA_TYPE_CHAR:
    return UdmGenerateLiteralToNativeRegNChar(compiler, expr, regno);
  case UDM_VALUE_DATA_TYPE_STR:
    UDM_ASSERT(regno == 0);
    return UdmGenerateTextSegmentTo_cs0_z0(compiler, expr->textdata_addr);
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateExpressionToNativeReg0(UDM_PROG_COMPILER *compiler,
                                  UDM_PROG_EXPRESSION *expr,
                                  const UDM_PROG_VAR *func_result)
{
  switch (expr->primary_expr.type)
  {
    case UDM_EXPRESSION_CONST:
      expr->primary_expr.type= UDM_EXPRESSION_COMPUTED;
      return UdmGenerateLiteralToNativeRegN(compiler, &expr->primary_expr, 0);
    case UDM_EXPRESSION_IDENT:
      expr->primary_expr.type= UDM_EXPRESSION_COMPUTED;
      return UdmGenerateVarToNativeReg0(compiler, &expr->primary_expr.Var);
    case UDM_EXPRESSION_OBJECT:
      expr->primary_expr.type= UDM_EXPRESSION_COMPUTED;
      if (func_result && UdmProgVarHandler(func_result))
        return UdmGenerateVarToNativeReg0(compiler, func_result);
      else
        return UDM_TRUE; /* Simple type: char, int, double */
    case UDM_EXPRESSION_COMPUTED:
      return UDM_TRUE;
  }
  UDM_ASSERT(0);
  return UdmProgCompilerNotSupported(compiler, "cout with postfix expression of this type");
}


static udm_bool_t
UdmGenerateNativeReg0TypeConversion(UDM_PROG_COMPILER *compiler,
                                    const UDM_VALUE_HANDLER *dst,
                                    const UDM_VALUE_HANDLER *src)
{
  if (dst->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE &&
      src->native_reg_type == UDM_VALUE_DATA_TYPE_INT)
  {
    if (UDM_OK != UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_DREG0_IREG0))
      return UDM_FALSE;
  }
  else if (dst->native_reg_type == UDM_VALUE_DATA_TYPE_INT &&
           src->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE)
  {
    if (UDM_OK != UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_IREG0_DREG0))
      return UDM_FALSE;
  }
  return UDM_TRUE; /* No conversion needed */
}


static udm_bool_t
UdmGenerateNativeReg1TypeConversion(UDM_PROG_COMPILER *compiler,
                                    const UDM_VALUE_HANDLER *dst,
                                    const UDM_VALUE_HANDLER *src)
{
  if (dst->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE &&
      src->native_reg_type == UDM_VALUE_DATA_TYPE_INT)
  {
    if (UDM_OK != UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_DREG1_IREG1))
      return UDM_FALSE;
  }
  else if (dst->native_reg_type == UDM_VALUE_DATA_TYPE_INT &&
           src->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE)
  {
    if (UDM_OK != UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_IREG1_DREG1))
      return UDM_FALSE;
  }
  return UDM_TRUE; /* No conversion needed */
}


static size_t
UdmValueHandlerAlignedDataSize(const UDM_VALUE_HANDLER *handler)
{
  size_t data_size= handler->DataSize();
  if (data_size < sizeof(void*))
    data_size= sizeof(void*);
  return data_size;
}


static udm_bool_t
UdmGenerateDeclareNonInitializedVariableWithName(UDM_PROG_COMPILER *compiler,
                                                 UDM_PROG_VARLIST *Vars,
                                                 const UDM_PROG_VAR *PIVar,
                                                 const char *name)
{
  size_t data_size;
  UDM_ASSERT(compiler->Vars2.nitems);
  if (UdmProgVarListFind(Vars, name))
  {
    UDM_ASSERT(name[0] != '/'); /* Not a temporary variable */
    return UdmProgCompilerRedeclarationError(compiler, name);
  }
  /*fprintf(stderr, "stackaddr=%d name=%s type=%s\n",
           (int) UdmProgCompilerNextVarStackOffset(compiler),
           name, PIVar->handler->type_name);*/

  data_size= UdmValueHandlerAlignedDataSize(UdmProgVarHandler(PIVar));
  if (UDM_OK != UdmProgCompilerNextVarStackOffsetIncrement(compiler, data_size) ||
      UDM_OK != UdmProgVarListAdd(Vars, name, &PIVar->value))
    return UDM_FALSE;
  return UDM_TRUE;
}


/**
  Create a named or a temporary variable on the last stack level.
*/
static udm_bool_t
UdmGenerateDeclareNonInitializedVariable(UDM_PROG_COMPILER *compiler,
                                         const UDM_PROG_VAR *PIVar)
{
  char name[128];
  size_t name_length= UdmConstTokenLength(&PIVar->name);
  UDM_PROG_VARLIST *Vars= &compiler->Vars2.Item[compiler->Vars2.nitems - 1];
  if (name_length)
    udm_snprintf(name, sizeof(name), "%.*s",
                 (int) (PIVar->name.end - PIVar->name.str), PIVar->name.str);
  else
    udm_snprintf(name, sizeof(name), "/tmp_%03X", (int) Vars->nitems);
  return UdmGenerateDeclareNonInitializedVariableWithName(compiler, Vars,
                                                          PIVar, name);
}


static const UDM_PROG_VAR2 *
UdmProgCompilerFindVariable(UDM_PROG_COMPILER *compiler,
                            UDM_CONST_TOKEN *str)
{
  size_t i;
  char name[128]; /* TODO34: find by name+length */
  udm_snprintf(name, sizeof(name), "%.*s",
               (int) (str->end - str->str), str->str);
  for (i= compiler->Vars2.nitems; i > 0; i--)
  {
    const UDM_PROG_VAR2 *Var= UdmProgVarListFind(&compiler->Vars2.Item[i - 1],
                                                 name);
    if (Var)
      return Var;
  }
  return NULL;
}


static udm_bool_t
ParseIdentifier(UDM_PROG_COMPILER *compiler, UDM_CONST_TOKEN *ident)
{
  if (UdmProgCompilerLastTokenType(compiler) == UDM_LEX_IDENT)
    *ident= compiler->scanner.token.token;
  return ParseTerm(compiler, UDM_LEX_IDENT);
}


static udm_bool_t
ParseUnsignedNumber(UDM_PROG_COMPILER *compiler, UDM_PROG_VAR *Var)
{
  UDM_CONST_TOKEN exponent;
  if (UdmProgCompilerLastTokenType(compiler) != UDM_LEX_UINT)
    return UDM_FALSE;
  Var->name= compiler->scanner.token.token;
  UdmProgVarSetHandler(Var, &UdmValueHandlerInt);
  if (!ParseTerm(compiler, UDM_LEX_UINT))
  {
    UDM_ASSERT(0);
    return UDM_FALSE;
  }
  if (ParseTerm(compiler, UDM_LEX_DOT))
  {
    UdmProgVarSetHandler(Var, &UdmValueHandlerDouble);
    if (UdmProgCompilerLastTokenType(compiler) == UDM_LEX_UINT)
    {
      Var->name.end= compiler->scanner.token.token.end;
      if (!ParseTerm(compiler, UDM_LEX_UINT))
      {
        UDM_ASSERT(0);
        return UDM_FALSE;
      }
    }
  }
  if (UdmProgCompilerLastTokenType(compiler) != UDM_LEX_IDENT)
    return UDM_TRUE;
  UdmProgVarSetHandler(Var, &UdmValueHandlerDouble);
  exponent= compiler->scanner.token.token;
  UDM_ASSERT(exponent.end - exponent.str > 0);
  if (exponent.str[0] != 'E' && exponent.str[0] != 'e')
    return UDM_TRUE;
  if (!ParseTerm(compiler, UDM_LEX_IDENT))
  {
    UDM_ASSERT(0);
    return UDM_FALSE;
  }
  if (exponent.end - exponent.str > 1) /* e123 was scanned as UDM_LEX_IDENT*/
  {
    const char *e;
    if (!udm_isdigit(exponent.str[1]))
      return UdmProgCompilerExponentHasNoDigits(compiler);
    for (e= exponent.str + 2; e < exponent.end; e++)
    {
      if (!udm_isdigit(*e))
      {
        exponent.str= e;
        return UdmProgCompilerExponentInvalidSuffix(compiler, &exponent);
      }
    }
    Var->name.end= exponent.end;
    return UDM_TRUE;
  }
  if (!ParseTerm(compiler, UDM_LEX_MINUS) &&
      !ParseTerm(compiler, UDM_LEX_PLUS))
    return UdmProgCompilerExponentHasNoDigits(compiler); /* e.g. 123e; */
  if (UdmProgCompilerLastTokenType(compiler) != UDM_LEX_UINT)
    return UdmProgCompilerExponentHasNoDigits(compiler); /* e.g. 123e+; */
  Var->name.end= compiler->scanner.token.token.end;
  if (!ParseTerm(compiler, UDM_LEX_UINT))
  {
    UDM_ASSERT(0);
    return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
ParseString(UDM_PROG_COMPILER *compiler, UDM_CONST_TOKEN *string)
{
  if (UdmProgCompilerLastTokenType(compiler) != UDM_LEX_STRING &&
      UdmProgCompilerLastTokenType(compiler) != UDM_LEX_ESCAPED_STRING)
    return UDM_FALSE;
  *string= compiler->scanner.token.token;
  UdmProgCompilerScan(compiler);
  return UDM_TRUE;
}


static udm_bool_t
ParseChar(UDM_PROG_COMPILER *compiler, UDM_CONST_TOKEN *string)
{
  if (UdmProgCompilerLastTokenType(compiler) == UDM_LEX_CHAR_LITERAL)
  {
    *string= compiler->scanner.token.token;
    if (UdmConstTokenLength(string) != 1 &&
        string->str[0] != '\\')
      return UdmProgCompilerErrorMultiCharacter(compiler);
  }
  return ParseTerm(compiler, UDM_LEX_CHAR_LITERAL);
}


static udm_bool_t
ParseLCB(UDM_PROG_COMPILER *compiler)
{
  return ParseTerm(compiler, UDM_LEX_LCB);
}


static udm_bool_t
ParsePrimaryExpressionIdentifier(UDM_PROG_COMPILER *compiler,
                                 UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  if (ParseIdentifier(compiler, &expr->Var.name))
  {
    const UDM_PROG_VAR2 *Var= UdmProgCompilerFindVariable(compiler,
                                                    &expr->Var.name);
    if (Var)
      expr->Var.value= Var->value;
    else
      UDM_BZERO(&expr->Var.value, sizeof(expr->Var.value));
    expr->type= UDM_EXPRESSION_IDENT;
    return UDM_TRUE;
  }
  return UDM_FALSE;
}


static int
CharUnescape(int ch)
{
  switch (ch)
  {
  case 'n': return '\n';
  case 'r': return '\r';
  case 't': return '\t';
  case '0': return '\0';
  }
  return ch;
}


static udm_bool_t
UdmCreateTextSegmentEntry(UDM_PROG_COMPILER *compiler,
                         udm_offset_and_length_t *addr,
                         const char *src, size_t length,
                         udm_bool_t unescape,
                         udm_bool_t append)
{
  if (append)
  {
    UDM_ASSERT(UdmDSTRLength(&compiler->prg->textdata));
    UdmDSTRShrinkLast(&compiler->prg->textdata);
  }
  addr->offset= UdmDSTRLength(&compiler->prg->textdata);
  /*
  fprintf(stderr, "CREATE: pos=%d len=%d\n",
        (int) addr->offset,
        (int) length);
  */
  if (unescape)
  {
    const char *end;
    UdmDSTRReserve(&compiler->prg->textdata, length);
    for (end= src + length ; src < end; src++)
    {
      if (*src != '\\')
        UdmDSTRAppend(&compiler->prg->textdata, src, 1);
      else if (src + 1 < end)
      {
        char ch= CharUnescape(src[1]);
        UdmDSTRAppend(&compiler->prg->textdata, &ch, 1);
        src++;
      }
    }
  }
  else
    UdmDSTRAppend(&compiler->prg->textdata, src, length);
  addr->length= UdmDSTRLength(&compiler->prg->textdata) - addr->offset;
  UdmDSTRAppend(&compiler->prg->textdata, "\0", 1);
  return UDM_TRUE;
}


static udm_bool_t
UdmProgGenerateCoutStringFromTextSegment(UDM_PROG_COMPILER *compiler,
                                         udm_offset_and_length_t addr)
{
  return UdmGenerateTextSegmentTo_cs0_z0(compiler, addr) &&
         UDM_OK == UdmProgAddArg0Simple(compiler->prg, &udm_tmpl_cout_cs0_z0);
}


static udm_bool_t
UdmProgGenerateCoutText(UDM_PROG_COMPILER *compiler,
                        const char *str,
                        const char *strend,
                        udm_bool_t unescape)
{
  if (str && str != strend)
  {
    udm_offset_and_length_t addr;
    UdmCreateTextSegmentEntry(compiler, &addr, str, strend - str, unescape, UDM_FALSE);
    return UdmProgGenerateCoutStringFromTextSegment(compiler, addr);
  }
  return UDM_OK;
}


static udm_bool_t
ParsePrimaryExpressionString(UDM_PROG_COMPILER *compiler, UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  UDM_CONST_TOKEN chunk;
  udm_offset_and_length_t addr;
  if (!ParseString(compiler, &expr->Var.name))
    return UDM_FALSE;
  expr->type= UDM_EXPRESSION_CONST;
  UdmProgVarSetHandler(&expr->Var, &UdmValueHandlerSimple);
  if (!UdmCreateTextSegmentEntry(compiler, &expr->textdata_addr,
                                   expr->Var.name.str,
                                   expr->Var.name.end - expr->Var.name.str,
                                   UDM_TRUE, UDM_FALSE))
    return UDM_FALSE;
  while (ParseString(compiler, &chunk))
  {
    if (!UdmCreateTextSegmentEntry(compiler, &addr,
                                   chunk.str, chunk.end - chunk.str,
                                   UDM_TRUE, UDM_TRUE))
      return UDM_FALSE;
    expr->textdata_addr.length+= addr.length;
  }
  return UDM_TRUE;
}


static udm_bool_t
ParsePrimaryExpressionUnsignedNumber(UDM_PROG_COMPILER *compiler,
                                     UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  if (!ParseUnsignedNumber(compiler, &expr->Var))
    return UDM_FALSE;
  expr->type= UDM_EXPRESSION_CONST;
  return UdmCreateTextSegmentEntry(compiler, &expr->textdata_addr,
                                   expr->Var.name.str,
                                   expr->Var.name.end - expr->Var.name.str,
                                   UDM_FALSE, UDM_FALSE);
}


static void
CharToStr(char *dst, const UDM_CONST_TOKEN *src)
{
  UDM_ASSERT(UdmConstTokenLength(src) > 0);
  if (src->str[0] == '\\')
  {
    UDM_ASSERT(UdmConstTokenLength(src) > 1);
    dst[0]= CharUnescape(src->str[1]);
  }
  else
    dst[0]= src->str[0];
  dst[1]= '\0';
}


static udm_bool_t
ParsePrimaryExpressionChar(UDM_PROG_COMPILER *compiler,
                           UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  if (!ParseChar(compiler, &expr->Var.name))
    return UDM_FALSE;
  CharToStr(expr->buf, &expr->Var.name);
  expr->Var.name.str= expr->buf;
  expr->Var.name.end= expr->buf + 1;
  expr->type= UDM_EXPRESSION_CONST;
  UdmProgVarSetHandler(&expr->Var, &UdmValueHandlerChar);
  return UdmCreateTextSegmentEntry(compiler, &expr->textdata_addr,
                                   expr->Var.name.str, 1,
                                   UDM_FALSE, UDM_FALSE);
}


/*
 <primary-expression> ::= <identifier>
                        | <constant>
                        | <string>
                        | ( <expression> )
*/
static udm_bool_t
ParsePrimaryExpressionLiteral(UDM_PROG_COMPILER *compiler,
                              UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  if (ParsePrimaryExpressionString(compiler, expr))
    return UDM_TRUE;
  if (ParsePrimaryExpressionChar(compiler, expr))
    return UDM_TRUE;
  if (ParsePrimaryExpressionUnsignedNumber(compiler, expr))
    return UDM_TRUE;
  return UDM_FALSE;
}

/*
  TODO: move parsing of parenthesized expression to here.
*/
static udm_bool_t
ParsePrimaryExpression(UDM_PROG_COMPILER *compiler, UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  if (ParsePrimaryExpressionIdentifier(compiler, expr))
    return UDM_TRUE;
  if (ParsePrimaryExpressionLiteral(compiler, expr))
    return UDM_TRUE;
  return UDM_FALSE;
}


/*
  <function-call-argument> ::= <assignment-expression>
*/
static udm_bool_t
ParseFunctionCallArgument(UDM_PROG_COMPILER *compiler,
                          UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  if (ParsePrimaryExpressionIdentifier(compiler, expr))
  {
    if (!UdmPrimaryExprHandler(expr))
      return UdmProgCompilerUnknownIdentifierError(compiler, &expr->Var.name);
    return UDM_TRUE;
  }
  return ParsePrimaryExpressionLiteral(compiler, expr);
}


/*
<function-call-argument> ::= <assignment-expression>

<function-call-argument-list> ::=     <function-call-argument>
                                 {',' <function-call-argument>}*

<function-call-arguments> ::= '(' <function-call-argument-list>? ')'
                            | '(' ')'
*/
static udm_bool_t
ParseFunctionCallArguments(UDM_PROG_COMPILER *compiler,
                           UDM_PROG_FUNCTION_CALL_ARGUMENTS *func)
{
  if (!ParseTerm(compiler, UDM_LEX_LP))
    return UDM_FALSE;
  func->nargs= 0;
  if (!ParseFunctionCallArgument(compiler, &func->args[0]))
  {
    if (UdmProgCompilerError(compiler))
      return UDM_FALSE;
    goto rp;
  }
  func->nargs= 1;
  for ( ; ; )
  {
    if (!ParseTerm(compiler, UDM_LEX_COMMA))
      break;
    if (func->nargs >= UDM_PROG_MAXARG)
    {
      udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
                   "In <function-call-arguments>: Too many arguments");
      return UDM_FALSE;
    }
    func->nargs++;
    if (!ParseFunctionCallArgument(compiler, &func->args[func->nargs - 1]))
      return UdmProgCompilerExpectedRule(compiler, "<function-call-argument>");
  }
rp:
  return ParseTermOrError(compiler, UDM_LEX_RP);
}


/**
  Check if a function prototype is compatible with the actual paramenter.
  Returns UDM_TRUE  on failure (incompatible)
  Returns UDM_FALSE on success (compatible)
*/
static udm_bool_t
UdmFunctionPrototypeCheck(const UDM_FUNCTION_PROTOTYPE *prototype,
                          const UDM_PROG_FUNCTION_CALL_ARGUMENTS *arglist,
                          size_t offset)
{
  size_t i;
  for (i= 0; i < arglist->nargs; i++)
  {
    if (prototype->args[i + offset].handler !=
        UdmPrimaryExprHandler(&arglist->args[i]))
      return UDM_TRUE;
  }
  return UDM_FALSE;
}


static const UDM_FUNCTION *
UdmFunctionListFind(const UDM_FUNCTION *functions,
                    const char *name, size_t length,
                    const UDM_VALUE_HANDLER *handler_for_this,
                    const UDM_PROG_FUNCTION_CALL_ARGUMENTS *arglist,
                    size_t offset /* 0, or 1 in case of method */)
{
  const UDM_FUNCTION *method;
  for (method= functions; method->name.str; method++)
  {
    if (handler_for_this &&
        handler_for_this != method->prototype.args[0].handler)
      continue;
    if (method->prototype.nargs == arglist->nargs + offset &&
        method->name.length == length &&
        !memcmp(method->name.str, name, length) &&
        !UdmFunctionPrototypeCheck(&method->prototype, arglist, offset))
      return method;
  }
  return NULL;
}


static const UDM_FUNCTION *
UdmValueHandlerMethodFind(const UDM_VALUE_HANDLER *ha,
                          const char *name, size_t length,
                          const UDM_PROG_FUNCTION_CALL_ARGUMENTS *arglist)
{
  if (!ha->method)
    return NULL;
  return UdmFunctionListFind(ha->method, name, length, ha, arglist, 1);
}


typedef struct
{
  udm_tmpl_cmd_t create_ident;
  UDM_PROG_CMD *create_value;
  UDM_PROG_CMD *delete_value;
  UDM_PROG_CMD *clear_value;
} UDM_CALL_ARG_CMD;


#define UDM_MAX_FUNC_ARG 5

static UDM_CALL_ARG_CMD call_arg_cmd[UDM_MAX_FUNC_ARG]=
{
  {
    UDM_PROG_MOV_VREG0_STACK_OFFSET,
    &udm_method_make_arg0,
    &udm_method_free_arg0,
    &udm_method_clear_arg0
  },
  {
    UDM_PROG_MOV_VREG1_STACK_OFFSET,
    &udm_method_make_arg1,
    &udm_method_free_arg1,
    &udm_method_clear_arg1
  },
  {
    UDM_PROG_MOV_VREG2_STACK_OFFSET,
    &udm_method_make_arg2,
    &udm_method_free_arg2,
    &udm_method_clear_arg2
  },
  {
    UDM_PROG_MOV_VREG3_STACK_OFFSET,
    &udm_method_make_arg3,
    &udm_method_free_arg3,
    &udm_method_clear_arg3
  },
  {
    UDM_PROG_MOV_VREG4_STACK_OFFSET,
    &udm_method_make_arg4,
    &udm_method_free_arg4,
    &udm_method_clear_arg4
  }
};


/*
  Create a temporary value.
*/
static udm_bool_t
UdmGenerateFunctionCallArgCreate(UDM_PROG_COMPILER *compiler,
                                 const UDM_PROG_PRIMARY_EXPRESSION *arg,
                                 size_t offset)
{
  UDM_PROG_ITEM item;
  UdmProgItemInit(&item);
  if (arg->type == UDM_EXPRESSION_IDENT)
  {
    return UdmGenerateOpOffset(compiler, call_arg_cmd[offset].create_ident,
                               arg->Var.value.stack_offset);
  }
  else
  {
    item.cmd= call_arg_cmd[offset].create_value;
    UdmProgItemArgInitFromLiteralAndHandler(&item.item_args, arg);
  }
  return UDM_OK == UdmProgAdd(compiler->prg, &item);
}


/*
  Setup a pointer to a variable, or create a temporary value.
*/
static udm_bool_t
UdmGenerateFunctionCallArgSetup(UDM_PROG_COMPILER *compiler,
                                const UDM_FUNCTION *func,
                                const UDM_PROG_PRIMARY_EXPRESSION *arg,
                                size_t argno)
{
  if (arg->type == UDM_EXPRESSION_IDENT)
  {
    if (!UdmPrimaryExprHandler(arg))
      return UdmProgCompilerUnknownIdentifierError(compiler, &arg->Var.name);
  }
  if (func->prototype.args[argno].handler != UdmPrimaryExprHandler(arg))
  {
    return UdmProgCompilerBadDataTypeForArg(compiler, argno,
                                            func->prototype.args[argno].handler,
                                            UdmPrimaryExprHandler(arg));
  }
  return UdmGenerateFunctionCallArgCreate(compiler, arg, argno);
}


static udm_bool_t
UdmGenerateFunctionCallArgDelete(UDM_PROG_COMPILER *compiler,
                                 const UDM_PROG_PRIMARY_EXPRESSION *arg,
                                 size_t offset)
{
  UDM_PROG_ITEM item;
  UdmProgItemInit(&item);
  item.cmd= arg->type == UDM_EXPRESSION_IDENT ?
            call_arg_cmd[offset].clear_value :
            call_arg_cmd[offset].delete_value;
  if (arg->type != UDM_EXPRESSION_IDENT)
    item.item_args.handler= UdmPrimaryExprHandler(arg);
  /*item.item_args.stack_offset= arg->Var.value.stack_offset;*/
  return UDM_OK == UdmProgAdd(compiler->prg, &item);
}


/**
  Returns UDM_TRUE if a function returning a value
  of the handler requires an object.
  Returns UDM_FALSE if the function can simply return
  the value in registers.
*/
static udm_bool_t
UdmValueHandlerIsComplex(const UDM_VALUE_HANDLER *handler)
{
  switch (handler->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
  case UDM_VALUE_DATA_TYPE_DOUBLE:
  case UDM_VALUE_DATA_TYPE_CHAR:
    return UDM_FALSE;
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  return UDM_TRUE;
}


/**
  Initialize everything but name.
*/
static void
UdmProgStackVarInitAttributes(UDM_PROG_VAR *Var,
                    UDM_PROG_COMPILER *compiler,
                    const UDM_VALUE_HANDLER *handler)
{
  UdmProgVarSetHandler(Var, handler);
  Var->value.stack_offset= UdmProgCompilerNextVarStackOffset(compiler);
}


/**
  Construct a temporary variable on stack.
  Return its address on result->stack_offset.
*/
static udm_bool_t
UdmGenerateReserveNonInitializedTemporaryVariableOnStack(UDM_PROG_COMPILER *c,
                                              const UDM_VALUE_HANDLER *handler,
                                                        UDM_PROG_VAR *result)
{
  UDM_BZERO(&result->name, sizeof(result->name));
  UdmProgStackVarInitAttributes(result, c, handler);
  if (!UdmGenerateDeclareNonInitializedVariable(c, result))
    return UDM_FALSE;
  if (UDM_OK != UdmProgAddValueHandler(c->prg, UDM_PROG_MOV_HREG0_PTR,
                                       handler))
    return UDM_FALSE;
  if (!UdmGenerateOpOffset(c, UDM_PROG_MOV_SREG0_STACK_OFFSET,
                           result->value.stack_offset))
    return UDM_FALSE;
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateFunctionReturnValue(UDM_PROG_COMPILER *compiler,
                               const UDM_FUNCTION *func,
                               UDM_PROG_VAR *result,
                               const UDM_PROG_VAR *dstvar)
{
  /*
    TODO34: a better test for this. Perhaps a method in handler.
  */

  if (func->prototype.result.handler &&
      UdmValueHandlerIsComplex(func->prototype.result.handler))
  {
    if (!(!result ^ !dstvar))
    {
      /*
        Called from UdmParseConditionalExpression,
        with both result==NULL and dstvar==NULL.
      */
      /*UDM_ASSERT(0);*/
      return UdmProgCompilerNotSupported(compiler, "function call of this type in this context (e.g. in a conditional expression)");
    }
    UDM_ASSERT(!result ^ !dstvar); /* At least on of them should be set */
    if (dstvar) /* string x= str.upper(); */
    {
      if (func->prototype.result.handler != UdmProgVarHandler(dstvar))
        return UdmCompilerErrorInvalidOperands(compiler,
                                               func->prototype.result.handler,
                                               UdmProgVarHandler(dstvar), "=");
      if (UDM_OK != UdmProgAddValueHandler(compiler->prg, UDM_PROG_MOV_HREG0_PTR,
                                           func->prototype.result.handler))
        return UDM_FALSE;
      if (!UdmGenerateOpOffset(compiler, UDM_PROG_MOV_SREG0_STACK_OFFSET,
                               dstvar->value.stack_offset))
        return UDM_FALSE;
    }
    else if (result) /* Want complex return value on stack rather than to be alloced */
    {
      UDM_ASSERT(!UdmProgVarHandler(result)); /* Must be initialized */
      if (!UdmGenerateReserveNonInitializedTemporaryVariableOnStack(compiler,
                                              func->prototype.result.handler,
                                                                    result))
        return UDM_FALSE;      
    }
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateFunctionArgsSetup(UDM_PROG_COMPILER *compiler,
                             const UDM_FUNCTION *func,
                             const UDM_PROG_FUNCTION_CALL_ARGUMENTS *arglist,
                             size_t offset)
{
  size_t i;
  for (i= 0; i < arglist->nargs && i < UDM_MAX_FUNC_ARG; i++)
  {
    if (!UdmGenerateFunctionCallArgSetup(compiler, func,
                                         &arglist->args[i], i + offset))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateFunctionArgsDelete(UDM_PROG_COMPILER *compiler,
                              const UDM_FUNCTION *func,
                              const UDM_PROG_FUNCTION_CALL_ARGUMENTS *arglist,
                              size_t offset)
{
  size_t i;
  for (i= 0; i < arglist->nargs && i < UDM_MAX_FUNC_ARG; i++)
  {
    if (!UdmGenerateFunctionCallArgDelete(compiler,
                                          &arglist->args[i], i + offset))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateCallInstruction(UDM_PROG_COMPILER *compiler,
                           const UDM_VALUE_HANDLER *classhandler,
                           const UDM_FUNCTION *func)
{
  UDM_PROG_ITEM item;
  char comment[128];
  UdmProgItemInit(&item);
  item.cmdarg.func= func;
  item.cmdnum= UDM_PROG_CALL;
  udm_snprintf(comment, sizeof(comment), "%s::%.*s()",
               classhandler ?
               classhandler->type_name : "",
               (int) func->name.length, func->name.str);
  return UDM_OK == UdmProgAdd(compiler->prg, &item);
}


/*
  @param [OUT] result - non-NULL means that we want the result on stack.
                        A new stack variable is created and
                        is returned in result[0].
                        NULL means that we want the result on heap in (h0,s0).
  @param       dstvar - non-NULL if the function call is directly assigned
                        to a variable:
                          string x= str.upper();
                        A new variable is not created in this case,
                        neither on stack, nor on heap.
  
  Function call returns resuts as follows:
  - INT    is returned in i0
  - DOUBLE is returned in d0
  - Other datatypes are returned in the register pair (h0,s0),
    where h0 stores the pointer to the data type handler,
    and s0 stores the pointer to the data.
*/
static udm_bool_t
UdmGenerateNewFunctionCall(UDM_PROG_COMPILER *compiler,
                           const UDM_FUNCTION *func,
                           const UDM_PROG_VAR *thisptr,
                           const UDM_PROG_FUNCTION_CALL_ARGUMENTS *arglist,
                           UDM_PROG_VAR *result,
                           const UDM_PROG_VAR *dstvar)
{
  size_t offset= thisptr ? 1 : 0; /* 1 means method, 0 means function */
  if (!UdmGenerateFunctionReturnValue(compiler, func, result, dstvar))
    return UDM_FALSE;

  if (thisptr)
  {
    /* Pass "this" as the first arg */
    if (!UdmGenerateOpOffset(compiler, UDM_PROG_MOV_VREG0_STACK_OFFSET,
                             thisptr->value.stack_offset))
      return UDM_FALSE;
  }
  if (!UdmGenerateFunctionArgsSetup(compiler, func, arglist, offset))
    return UDM_FALSE;

  if (!UdmGenerateCallInstruction(compiler,
                                  thisptr ?
                                  UdmProgVarHandler(thisptr) : NULL, func))
    return UDM_FALSE;

  if (!UdmGenerateFunctionArgsDelete(compiler, func, arglist, offset))
    return UDM_FALSE;

  if (thisptr)
  {
    if (UDM_OK != UdmProgAddArg0Simple(compiler->prg, &udm_method_clear_arg0))
      return UDM_FALSE;
  }
  if (func->func == UdmFunction_exit)
  {
    if (!UdmProgCompilerGenerateVarListListDestructors(compiler))
      return UDM_FALSE;
    UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_EXIT);    
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateFunctionCall(UDM_PROG_COMPILER *compiler,
                        const UDM_CONST_TOKEN *name,
                        UDM_PROG_EXPRESSION *expr,
                        UDM_PROG_VAR *complex_func_result_on_stack,
                        const UDM_PROG_VAR *dstvar)
{
  const UDM_FUNCTION *func2;
  if (!(func2= UdmFunctionListFind(udm_builtin_functions,
                                  name->str, name->end - name->str,
                                  NULL, &expr->arglist, 0)))
    return UdmProgCompilerErrorUnknownFunction(compiler, name, &expr->arglist);
  if (!UdmGenerateNewFunctionCall(compiler, func2,
                                    NULL, &expr->arglist,
                                    complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  UdmExprSetHandler(expr, func2->prototype.result.handler);
  expr->primary_expr.type= UDM_EXPRESSION_OBJECT;
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateMethodCall(UDM_PROG_COMPILER *compiler,
                      const UDM_CONST_TOKEN *name,
                      UDM_PROG_EXPRESSION *expr,
                      UDM_PROG_VAR *complex_func_result_on_stack,
                      const UDM_PROG_VAR *dstvar)
{
  const UDM_FUNCTION *method;

  if (!(method= UdmValueHandlerMethodFind(UdmExprHandler(expr),
                                          name->str, name->end - name->str,
                                          &expr->arglist)))
    return UdmProgCompilerUnknownMethodError(compiler, UdmExprHandler(expr),
                                             name, &expr->arglist);
  if (!UdmGenerateNewFunctionCall(compiler, method,
                                  &expr->primary_expr.Var, &expr->arglist,
                                  complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  UdmExprSetHandler(expr, method->prototype.result.handler);
  expr->primary_expr.type= UDM_EXPRESSION_OBJECT;
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateStackVarIncrement(UDM_PROG_COMPILER *compiler, const UDM_PROG_VAR *Var)
{
  if (!UdmUnaryOperatorFindWithError(compiler, unary_operators,
                                    UDM_UNARY_OPERATOR_POSTFIX_INC,
                                    UdmProgVarHandler(Var), "++"))
    return UDM_FALSE;
  switch (UdmProgVarHandler(Var)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateVariableOp(compiler, Var, UDM_PROG_INC_INT_PTR_STACK_OFFSET);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmGenerateVariableOp(compiler, Var, UDM_PROG_INC_DBL_PTR_STACK_OFFSET);
  case UDM_VALUE_DATA_TYPE_CHAR:
    return UdmGenerateVariableOp(compiler, Var, UDM_PROG_INC_CHR_PTR_STACK_OFFSET);
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateStackVarDecrement(UDM_PROG_COMPILER *compiler, const UDM_PROG_VAR *Var)
{
  if (!UdmUnaryOperatorFindWithError(compiler, unary_operators,
                                     UDM_UNARY_OPERATOR_POSTFIX_DEC,
                                     UdmProgVarHandler(Var), "--"))
    return UDM_FALSE;
  switch (UdmProgVarHandler(Var)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
    return UdmGenerateVariableOp(compiler, Var, UDM_PROG_DEC_INT_PTR_STACK_OFFSET);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmGenerateVariableOp(compiler, Var, UDM_PROG_DEC_DBL_PTR_STACK_OFFSET);
  case UDM_VALUE_DATA_TYPE_CHAR:
    return UdmGenerateVariableOp(compiler, Var, UDM_PROG_DEC_CHR_PTR_STACK_OFFSET);
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateStackVarIncOrDec(UDM_PROG_COMPILER *compiler,
                            const UDM_PROG_VAR *Var,
                            udm_lex_t op)
{
  if (op == UDM_LEX_INC)
    return UdmGenerateStackVarIncrement(compiler, Var);
  if (op == UDM_LEX_DEC)
    return UdmGenerateStackVarDecrement(compiler, Var);
  UDM_ASSERT(0);
  return UDM_FALSE;
}


/*
<postfix-expression-part2> ::= '[' <expression> ']'
                             | <function-call-arguments>
                             | '.' <identifier>
                             | '->' <identifier>
                             | '++'
                             | '--'
*/
static udm_bool_t
ParsePostfixExpressionVarPart2(UDM_PROG_COMPILER *compiler,
                               UDM_PROG_EXPRESSION *expr)
{
  udm_lex_t cmd;
  if (ParseTerm(compiler, (cmd= UDM_LEX_INC)) ||
      ParseTerm(compiler, (cmd= UDM_LEX_DEC)))
  {
    expr->primary_expr.type= UDM_EXPRESSION_COMPUTED;
    return
      UdmGenerateVarToNativeReg0(compiler, &expr->primary_expr.Var) &&
      UdmGenerateStackVarIncOrDec(compiler, &expr->primary_expr.Var, cmd);
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmParseExpression(UDM_PROG_COMPILER *compiler,
                   UDM_PROG_EXPRESSION *expr,
                   UDM_PROG_VAR *complex_result_on_stack,
                   const UDM_PROG_VAR *dstvar);

static udm_bool_t
ParseParenthesizedExpression(UDM_PROG_COMPILER *compiler,
                             UDM_PROG_EXPRESSION *expr,
                             UDM_PROG_VAR *complex_func_result_on_stack,
                             const UDM_PROG_VAR *dstvar)
{
  if (TestParenthesizedTypeSpecifier(compiler))
    return UDM_FALSE; /* This is a type cast, e.g.: y= (int) x; */
  if (!ParseTerm(compiler, UDM_LEX_LP))
    return UDM_FALSE;
  if (!UdmParseExpression(compiler, expr,
                          complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  return ParseTermOrError(compiler, UDM_LEX_RP);
}


static udm_bool_t
UdmParseMemberAccess(UDM_PROG_COMPILER *compiler,
                     UDM_PROG_EXPRESSION *expr,
                     UDM_PROG_VAR *complex_func_result_on_stack,
                     const UDM_PROG_VAR *dstvar)
{
  UDM_CONST_TOKEN name;     
  if (ParseTerm(compiler, UDM_LEX_OPERATOR))
  {
    if (!ParseTermOrError(compiler, UDM_LEX_EQ))
      return UDM_FALSE;
    UdmConstTokenSet(&name, UDM_CSTR_WITH_LEN("operator="));
  }
  else if (!ParseIdentifier(compiler, &name))
    return UdmProgCompilerExpectedRule(compiler, "<method-name> or `operator`");
  if (!ParseFunctionCallArguments(compiler, &expr->arglist))
    return UDM_FALSE;
  return UdmGenerateMethodCall(compiler, &name, expr,
                               complex_func_result_on_stack, dstvar);
}


static void
UdmPrepareFunctionResultForMethodCall(UDM_PROG_EXPRESSION *expr,
                                      UDM_PROG_VAR *complex_func_result_on_stack)
{
    /* Preserve expr->primary_expr.Var.handler */
    UdmConstTokenSet(&expr->primary_expr.Var.name, NULL, 0);
    expr->primary_expr.type= UDM_EXPRESSION_IDENT;
    expr->primary_expr.Var.value.stack_offset=
      complex_func_result_on_stack->value.stack_offset;
    UDM_BZERO(complex_func_result_on_stack, sizeof(*complex_func_result_on_stack));

}


static udm_bool_t
UdmParseRecursiveMemberAccess(UDM_PROG_COMPILER *compiler,
                              UDM_PROG_EXPRESSION *expr,
                              UDM_PROG_VAR *complex_func_result_on_stack,
                              const UDM_PROG_VAR *dstvar)
{
  if (!UdmParseMemberAccess(compiler, expr,
                            complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  while (ParseTerm(compiler, UDM_LEX_DOT))
  {
    if (dstvar)
      return UdmProgCompilerNotSupported(compiler,
                                         "recursive member access in this context");
    UdmPrepareFunctionResultForMethodCall(expr, complex_func_result_on_stack);
    if (!UdmParseMemberAccess(compiler, expr,
                              complex_func_result_on_stack, dstvar))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}

/*

<postfix-expression> ::= <primary-expression> {<postfix-expression-part2>}*

*/
static udm_bool_t
ParsePostfixExpression(UDM_PROG_COMPILER *compiler,
                       UDM_PROG_EXPRESSION *expr,
                       UDM_PROG_VAR *complex_func_result_on_stack,
                       const UDM_PROG_VAR *dstvar)
{
  if (ParseParenthesizedExpression(compiler, expr,
                                   complex_func_result_on_stack, dstvar))
  {
    if (ParseTerm(compiler, UDM_LEX_DOT))
    {
      UdmPrepareFunctionResultForMethodCall(expr, complex_func_result_on_stack);
      return UdmParseRecursiveMemberAccess(compiler, expr,
                                           complex_func_result_on_stack, dstvar);
    }
    return UDM_TRUE;
  }
  if (!ParsePrimaryExpression(compiler, &expr->primary_expr))
    return UDM_FALSE;

  if (expr->primary_expr.type == UDM_EXPRESSION_IDENT)
  {
    if (!UdmExprHandler(expr))
    {
      if (!ParseFunctionCallArguments(compiler, &expr->arglist))
      {
        if (UdmProgCompilerError(compiler))
          return UDM_FALSE;
        return UdmProgCompilerUnknownIdentifierError(compiler, &expr->primary_expr.Var.name);
      }
      else
      {
        return UdmGenerateFunctionCall(compiler, &expr->primary_expr.Var.name, expr,
                                       complex_func_result_on_stack,
                                       dstvar);
      }
    }

    if (ParseTerm(compiler, UDM_LEX_DOT))
      return UdmParseRecursiveMemberAccess(compiler, expr,
                                           complex_func_result_on_stack, dstvar);
    return ParsePostfixExpressionVarPart2(compiler, expr);
  }
  return UDM_TRUE;
}


static udm_bool_t
ExpressionHasEffect(const UDM_PROG_EXPRESSION *expr)
{
  switch (expr->primary_expr.type)
  {
  case UDM_EXPRESSION_IDENT:
  case UDM_EXPRESSION_CONST:
    return UDM_FALSE;
  case UDM_EXPRESSION_OBJECT: 
  case UDM_EXPRESSION_COMPUTED:
    return UDM_TRUE;
  }
  return UDM_FALSE;
}


static udm_bool_t ParseUnaryExpression(UDM_PROG_COMPILER *compiler,
                                       UDM_PROG_EXPRESSION *expr,
                                       UDM_PROG_VAR *complex_func_result_on_stack,
                                       const UDM_PROG_VAR *dstvar);

static udm_bool_t
UdmParseCastExpression(UDM_PROG_COMPILER *compiler,
                       UDM_PROG_EXPRESSION *expr,
                       UDM_PROG_VAR *complex_func_result_on_stack,
                       const UDM_PROG_VAR *dstvar);

/*
<unary-operator> ::= &
                   | *
                   | +
                   | -
                   | ~
                   | !

<prefix-expression> ::= <unary-operator> [<cast>] <cast-expression>
                      | '++' <unary-expression>
                      | '--' <unary-expression>
                      | 'sizeof' <unary-expression>
                      | 'sizeof' <type-name>

<unary-expression> ::= <postfix-expression>
                     | <prefix-expression>
*/
static udm_bool_t
ParseUnaryExpressionInternal(UDM_PROG_COMPILER *compiler,
                             UDM_PROG_EXPRESSION *expr,
                             UDM_PROG_VAR *complex_func_result_on_stack,
                             const UDM_PROG_VAR *dstvar)
{
  if (ParseTerm(compiler, UDM_LEX_DEC))
  {
    if (!ParseUnaryExpression(compiler, expr,
                              complex_func_result_on_stack, dstvar))
      return UdmProgCompilerExpectedRule(compiler, "<unary-expression>");
    expr->primary_expr.type= UDM_EXPRESSION_COMPUTED;
    return UdmGenerateStackVarDecrement(compiler, &expr->primary_expr.Var) &&
           UdmGenerateVarToNativeReg0(compiler, &expr->primary_expr.Var);
  }
  if (ParseTerm(compiler, UDM_LEX_INC))
  {
    if (!ParseUnaryExpression(compiler, expr,
                              complex_func_result_on_stack, dstvar))
      return UdmProgCompilerExpectedRule(compiler, "<unary-expression>");
    expr->primary_expr.type=  UDM_EXPRESSION_COMPUTED;
    return UdmGenerateStackVarIncrement(compiler, &expr->primary_expr.Var) &&
           UdmGenerateVarToNativeReg0(compiler, &expr->primary_expr.Var);
  }
  if (!ParsePostfixExpression(compiler, expr,
                              complex_func_result_on_stack, dstvar))
    return UDM_FALSE;

  return UDM_TRUE;
}


static udm_bool_t
ParseUnaryExpression(UDM_PROG_COMPILER *compiler,
                     UDM_PROG_EXPRESSION *expr,
                     UDM_PROG_VAR *complex_func_result_on_stack,
                     const UDM_PROG_VAR *dstvar)
{
  if (ParseTerm(compiler, UDM_LEX_EXCLAM))
  {
    if (!UdmParseCastExpression(compiler, expr,
                                complex_func_result_on_stack, dstvar))
      return UDM_FALSE;
    if (!UdmGenerateExpressionToNativeReg0(compiler, expr, NULL) ||
        !UdmGenerateIsNotTrueTest(compiler, UdmExprHandler(expr)))
      return UDM_FALSE;
    UdmExprSetHandler(expr, &UdmValueHandlerInt);
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_TILDE))
  {
    if (!UdmParseCastExpression(compiler, expr,
                                complex_func_result_on_stack, dstvar))
      return UDM_FALSE;
    if (UdmExprHandler(expr)->native_reg_type != UDM_VALUE_DATA_TYPE_INT)
      return UdmProgCompilerNotSupported(compiler, "unary ~ for this data type");
    if (!UdmGenerateExpressionToNativeReg0(compiler, expr, NULL) ||
        !UdmGenerateOp(compiler, UDM_PROG_NOT_IREG0))
      return UDM_FALSE;
    UdmExprSetHandler(expr, &UdmValueHandlerInt);
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_MINUS))
  {
    if (!UdmParseCastExpression(compiler, expr,
                                complex_func_result_on_stack, dstvar))
      return UDM_FALSE;
    if (UdmValueHandlerIsComplex(UdmExprHandler(expr)))
      return UdmProgCompilerNotSupported(compiler, "unary '-' for this data type");
    if (!UdmGenerateExpressionToNativeReg0(compiler, expr, NULL))
      return UDM_FALSE;
    if (!UdmGenerateMovNativeReg0ToReg1(compiler, UdmExprHandler(expr)))
      return UDM_FALSE;
    if (!UdmGenerateClearNativeReg0(compiler, UdmExprHandler(expr)))
      return UDM_FALSE;

    if (!UdmGenerateAddNativeReg0Reg1(compiler,
                                      &expr->primary_expr.Var.value.handler,
                                      UDM_LEX_MINUS))
      return UDM_FALSE;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_PLUS))
  {
    if (!UdmParseCastExpression(compiler, expr,
                                complex_func_result_on_stack, dstvar))
      return UDM_FALSE;
    if (UdmValueHandlerIsComplex(UdmExprHandler(expr)))
      return UdmProgCompilerNotSupported(compiler, "unary '+' for this data type");
    return UDM_TRUE;
  }
  return ParseUnaryExpressionInternal(compiler, expr,
                                      complex_func_result_on_stack, dstvar);
}


/**
  <cast-expression> ::=   <unary-expression>
                        | ( <type-name> ) <cast-expression>
*/
static udm_bool_t
UdmParseCastExpression(UDM_PROG_COMPILER *compiler,
                       UDM_PROG_EXPRESSION *expr,
                       UDM_PROG_VAR *complex_func_result_on_stack,
                       const UDM_PROG_VAR *dstvar)
{
  const UDM_VALUE_HANDLER *ha;
  /*
    The cast operator starts from parenthesis, so does the
    parenthesized expression. This rule needs one more look-ahead.
    TestParenthesizedTypeSpecifier() makes one more look-ahead with rollback.
  */
  if (!TestParenthesizedTypeSpecifier(compiler))
    return ParseUnaryExpression(compiler, expr,
                                complex_func_result_on_stack, dstvar);

  if (!ParseTerm(compiler, UDM_LEX_LP) ||
      !ParseTypeSpecifier(compiler, &ha) ||
      !ParseTermOrError(compiler, UDM_LEX_RP))
    return UDM_FALSE;

  if (!UdmParseCastExpression(compiler, expr,
                              complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  if (UdmExprHandler(expr) == ha)
    return UDM_TRUE;

  if (UdmValueHandlerIsComplex(ha))
    return UdmProgCompilerNotSupported(compiler, "cast to this data type");
  if (UdmValueHandlerIsComplex(UdmExprHandler(expr)))
    return UdmProgCompilerNotSupported(compiler, "cast from this data type");
  if (!UdmGenerateExpressionToNativeReg0(compiler, expr, NULL) ||
      !UdmGenerateNativeReg0TypeConversion(compiler, ha, UdmExprHandler(expr)))
    return UDM_FALSE;
  UdmExprSetHandler(expr, ha);
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateSaveNativeReg0ToTempVariableOnStack(UDM_PROG_COMPILER *c,
                                               const UDM_VALUE_HANDLER *h,
                                               UDM_PROG_VAR *dst)
{
  return UdmGenerateReserveNonInitializedTemporaryVariableOnStack(c, h, dst) &&
         UdmGenerateConstructor(c, dst) &&
         UdmGenerateNativeReg0ToVar(c, dst);
}


static udm_bool_t
UdmGenerateAggregateAndConvertReg0Reg1(UDM_PROG_COMPILER *compiler,
                                       const UDM_VALUE_HANDLER *b,
                                       const UDM_VALUE_HANDLER *a,
                                       const UDM_VALUE_HANDLER **agg)
{
  *agg= a->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE ||
        b->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE ?
       &UdmValueHandlerDouble : &UdmValueHandlerInt;
  return UdmGenerateNativeReg1TypeConversion(compiler, *agg, a) &&
         UdmGenerateNativeReg0TypeConversion(compiler, *agg, b);
}


static udm_bool_t
ParseMultiplicativeExpression(UDM_PROG_COMPILER *compiler,
                              UDM_PROG_EXPRESSION *expr,
                              UDM_PROG_VAR *complex_func_result_on_stack,
                              const UDM_PROG_VAR *dstvar);
static udm_bool_t
ParseAdditiveExpression(UDM_PROG_COMPILER *compiler,
                        UDM_PROG_EXPRESSION *expr,
                        UDM_PROG_VAR *complex_func_result_on_stack,
                        const UDM_PROG_VAR *dstvar);

static udm_bool_t
UdmParseShiftExpression(UDM_PROG_COMPILER *compiler,
                        UDM_PROG_EXPRESSION *expr,
                        UDM_PROG_VAR *complex_func_result_on_stack,
                        const UDM_PROG_VAR *dstvar);


static udm_bool_t
UdmParseRelationalExpression(UDM_PROG_COMPILER *compiler,
                             UDM_PROG_EXPRESSION *expr,
                             UDM_PROG_VAR *complex_func_result_on_stack,
                             const UDM_PROG_VAR *dstvar);

static udm_bool_t
UdmParseEqualityExpression(UDM_PROG_COMPILER *compiler,
                           UDM_PROG_EXPRESSION *expr,
                           UDM_PROG_VAR *complex_func_result_on_stack,
                           const UDM_PROG_VAR *dstvar);


static udm_bool_t
UdmParseAndExpression(UDM_PROG_COMPILER *compiler,
                      UDM_PROG_EXPRESSION *expr,
                      UDM_PROG_VAR *complex_func_result_on_stack,
                      const UDM_PROG_VAR *dstvar);


static udm_bool_t
UdmParseOrExpression(UDM_PROG_COMPILER *compiler,
                     UDM_PROG_EXPRESSION *expr,
                     UDM_PROG_VAR *complex_func_result_on_stack,
                     const UDM_PROG_VAR *dstvar);


static udm_bool_t
UdmGenerateDiadicOperationPart2(UDM_PROG_COMPILER *compiler,
                                UDM_PROG_EXPRESSION *expr,
                                udm_lex_t op)
{
  UDM_PROG_EXPRESSION expr2;
  UDM_PROG_VAR tmp;
  const UDM_VALUE_HANDLER *agg;

  if (!UdmExprHandler(expr))
    return UdmProgCompilerVoidValueIsNotIgnored(compiler);

  switch (UdmExprHandler(expr)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_INT:
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    break;
  case UDM_VALUE_DATA_TYPE_STR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    return UdmProgCompilerNotSupported(compiler, "Dyadic operation for this data type");
  }

  if (!UdmGenerateExpressionToNativeReg0(compiler, expr, NULL))
    return UDM_FALSE;

  /* Save the current value in reg0 to a temporary variable */
  if (!UdmProgCompilerEnterBlock(compiler) ||
      !UdmGenerateSaveNativeReg0ToTempVariableOnStack(compiler,
                                                      UdmExprHandler(expr),
                                                      &tmp))
    return UDM_FALSE;
  /* Parse the second operand and bring its value to reg0  */
  switch (op)
  {
  case UDM_LEX_VBAR:
    if (!UdmParseOrExpression(compiler, &expr2, NULL, NULL))
      return UDM_FALSE;
    break;
  case UDM_LEX_CARET:
    if (!UdmParseAndExpression(compiler, &expr2, NULL, NULL))
      return UDM_FALSE;
    break;
  case UDM_LEX_AMPERSAND:
    if (!UdmParseEqualityExpression(compiler, &expr2, NULL, NULL))
      return UDM_FALSE;
    break;
  case UDM_LEX_EQ_EQ:
  case UDM_LEX_NOT_EQ:
    if (!UdmParseRelationalExpression(compiler, &expr2, NULL, NULL))
      return UDM_FALSE;
    break; 
  case UDM_LEX_LE:
  case UDM_LEX_GE:
  case UDM_LEX_LT:
  case UDM_LEX_GT:
    if (!UdmParseShiftExpression(compiler, &expr2, NULL, NULL))
      return UDM_FALSE;
    break;
  case UDM_LEX_LSHIFT:
  case UDM_LEX_RSHIFT:
    if (!ParseAdditiveExpression(compiler, &expr2, NULL, NULL))
      return UDM_FALSE;
    break;
  case UDM_LEX_PLUS:
  case UDM_LEX_MINUS:
  case UDM_LEX_PERCENT:
    if (!ParseMultiplicativeExpression(compiler, &expr2, NULL, NULL))
      return UDM_FALSE;
    break;
  case UDM_LEX_ASTERISK:
  case UDM_LEX_SLASH:
    if (!UdmParseCastExpression(compiler, &expr2, NULL, NULL))
      return UDM_FALSE;
    break;
  default:
    UDM_ASSERT(0);
    return 0;
  }

  if (!UdmExprHandler(&expr2))
    return UdmProgCompilerVoidValueIsNotIgnored(compiler);

  if (!UdmGenerateExpressionToNativeReg0(compiler, &expr2, NULL))
    return UDM_FALSE;
  /* Restore the first operant to reg1 */
  if (!UdmGenerateVarToNativeReg1(compiler, &tmp))
    return UDM_FALSE;
  if (!UdmProgCompilerLeaveBlock(compiler))
    return UDM_FALSE;

  /* Now we have the left operant in reg1 and the right operand in reg0 */
  if (!UdmGenerateAggregateAndConvertReg0Reg1(compiler,
                                              UdmExprHandler(&expr2),
                                              UdmExprHandler(expr),
                                              &agg))
    return UDM_FALSE;

  switch (op)
  {
  case UDM_LEX_VBAR:
  case UDM_LEX_CARET:
  case UDM_LEX_AMPERSAND:
  case UDM_LEX_ASTERISK:
  case UDM_LEX_PLUS:
  case UDM_LEX_EQ_EQ:
  case UDM_LEX_NOT_EQ:
    break; /* Reverse operation, the order of arguments does not matter. */
  case UDM_LEX_SLASH:
  case UDM_LEX_PERCENT:
  case UDM_LEX_MINUS:
  case UDM_LEX_LSHIFT:
  case UDM_LEX_RSHIFT:
  case UDM_LEX_LE:
  case UDM_LEX_GE:
  case UDM_LEX_LT:
  case UDM_LEX_GT:
    if (!UdmGenerateSwapNativeReg0Reg1(compiler, agg))
      return UDM_FALSE;
    break;
  default:
    UDM_ASSERT(0);
    return UDM_FALSE;
  }
  if (!UdmGenerateDiadicOperationNativeReg0Reg1(compiler, &agg, op))
    return UDM_FALSE;
  UdmExprSetHandler(expr, agg);
  return UDM_TRUE;
}


/*
  <multiplicative-expression> ::= <cast-expression> 
                                  { <mul-op>   <cast-expression>  }*
  <cast-expression> ::= [ ( <type-name> )] <unary-expression>
*/
static udm_bool_t
ParseMultiplicativeExpression(UDM_PROG_COMPILER *compiler,
                              UDM_PROG_EXPRESSION *expr,
                              UDM_PROG_VAR *complex_func_result_on_stack,
                              const UDM_PROG_VAR *dstvar)
{
  udm_lex_t cmd;
  if (!UdmParseCastExpression(compiler, expr,
                              complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  for (; ParseTerm(compiler, (cmd= UDM_LEX_ASTERISK)) ||
         ParseTerm(compiler, (cmd= UDM_LEX_SLASH)) ||
         ParseTerm(compiler, (cmd= UDM_LEX_PERCENT)) ; )
  {
    if (!UdmGenerateDiadicOperationPart2(compiler, expr, cmd))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}



/*
  <additive-expression> ::= <multiplicative-expression>
                            { <add-op>   <multiplicative-expression> }*
*/
static udm_bool_t
ParseAdditiveExpression(UDM_PROG_COMPILER *compiler,
                        UDM_PROG_EXPRESSION *expr,
                        UDM_PROG_VAR *complex_func_result_on_stack,
                        const UDM_PROG_VAR *dstvar)
{
  udm_lex_t cmd;
  if (!ParseMultiplicativeExpression(compiler, expr,
                                     complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  for (; (ParseTerm(compiler, (cmd= UDM_LEX_PLUS)) ||
          ParseTerm(compiler, (cmd= UDM_LEX_MINUS))) ; )
  {
    if (!UdmGenerateDiadicOperationPart2(compiler, expr, cmd))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*
  <shift-expression> ::= <additive-expression> 
                         { <shift-op> <additive-expression>  }*
*/
static udm_bool_t
UdmParseShiftExpression(UDM_PROG_COMPILER *compiler,
                        UDM_PROG_EXPRESSION *expr,
                        UDM_PROG_VAR *complex_func_result_on_stack,
                        const UDM_PROG_VAR *dstvar)
{
  udm_lex_t cmd;
  if (!ParseAdditiveExpression(compiler, expr,
                               complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  for (; ParseTerm(compiler, (cmd= UDM_LEX_LSHIFT)) ||
         ParseTerm(compiler, (cmd= UDM_LEX_RSHIFT)); )
  {
    if (!UdmGenerateDiadicOperationPart2(compiler, expr, cmd))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*
  <relational-expression>     ::= <shift-expression>
                                  { <rel-op>   <shift-expression> 
*/
static udm_bool_t
UdmParseRelationalExpression(UDM_PROG_COMPILER *compiler,
                             UDM_PROG_EXPRESSION *expr,
                             UDM_PROG_VAR *complex_func_result_on_stack,
                             const UDM_PROG_VAR *dstvar)
{
  udm_lex_t cmd;
  if (!UdmParseShiftExpression(compiler, expr,
                               complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  for (; ParseTerm(compiler, (cmd= UDM_LEX_LT)) ||
         ParseTerm(compiler, (cmd= UDM_LEX_LE)) ||
         ParseTerm(compiler, (cmd= UDM_LEX_GT)) ||
         ParseTerm(compiler, (cmd= UDM_LEX_GE)); )
  {
    if (!UdmGenerateDiadicOperationPart2(compiler, expr, cmd))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*
  <equality-expression> ::= <relational-expression>
                            { <eq-op>    <relational-expression> }*
*/
static udm_bool_t
UdmParseEqualityExpression(UDM_PROG_COMPILER *compiler,
                           UDM_PROG_EXPRESSION *expr,
                           UDM_PROG_VAR *complex_func_result_on_stack,
                           const UDM_PROG_VAR *dstvar)
{
  udm_lex_t cmd;
  if (!UdmParseRelationalExpression(compiler, expr,
                                    complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  for (; ParseTerm(compiler, (cmd= UDM_LEX_EQ_EQ)) ||
         ParseTerm(compiler, (cmd= UDM_LEX_NOT_EQ)); )
  {
    if (!UdmGenerateDiadicOperationPart2(compiler, expr, cmd))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*
  <and-expression> ::= <equality-expression>
                       { &  <equality-expression> }*
*/
static udm_bool_t
UdmParseAndExpression(UDM_PROG_COMPILER *compiler,
                      UDM_PROG_EXPRESSION *expr,
                      UDM_PROG_VAR *complex_func_result_on_stack,
                      const UDM_PROG_VAR *dstvar)
{
  udm_lex_t cmd;
  if (!UdmParseEqualityExpression(compiler, expr,
                                  complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  for (; ParseTerm(compiler, (cmd= UDM_LEX_AMPERSAND)); )
  {
    if (!UdmGenerateDiadicOperationPart2(compiler, expr, cmd))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*
  <exclusive-or-expression> ::= <and-expression>  { ^  <and-expression> }*
*/
static udm_bool_t
UdmParseXorExpression(UDM_PROG_COMPILER *compiler,
                      UDM_PROG_EXPRESSION *expr,
                      UDM_PROG_VAR *complex_func_result_on_stack,
                      const UDM_PROG_VAR *dstvar)
{
  udm_lex_t cmd;
  if (!UdmParseAndExpression(compiler, expr,
                             complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  for (; ParseTerm(compiler, (cmd= UDM_LEX_CARET)); )
  {
    if (!UdmGenerateDiadicOperationPart2(compiler, expr, cmd))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*
  <inclusive-or-expression>   ::= <exclusive-or-expression>
                                 { |  <exclusive-or-expression>  }*
*/
static udm_bool_t
UdmParseOrExpression(UDM_PROG_COMPILER *compiler,
                      UDM_PROG_EXPRESSION *expr,
                      UDM_PROG_VAR *complex_func_result_on_stack,
                      const UDM_PROG_VAR *dstvar)
{
  udm_lex_t cmd;
  if (!UdmParseXorExpression(compiler, expr,
                             complex_func_result_on_stack, dstvar))
    return UDM_FALSE;
  for (; ParseTerm(compiler, (cmd= UDM_LEX_VBAR)); )
  {
    if (!UdmGenerateDiadicOperationPart2(compiler, expr, cmd))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateExpressionToBool(UDM_PROG_COMPILER *compiler,
                            UDM_PROG_EXPRESSION *expr)
{
  if (!UdmGenerateExpressionToNativeReg0(compiler, expr, NULL) ||
      !UdmGenerateIsTrueTest(compiler, UdmExprHandler(expr)))
    return UDM_FALSE;
  UdmExprSetHandler(expr, &UdmValueHandlerInt);
  return UDM_TRUE;
}


static udm_bool_t
UdmParseBoolAndExpression(UDM_PROG_COMPILER *compiler,
                          UDM_PROG_EXPRESSION *expr,
                          UDM_PROG_VAR *complex_func_result_on_stack,
                          const UDM_PROG_VAR *dst)
{
  size_t start= UdmProgCompilerGetProg(compiler)->nitems;
  if (!UdmParseOrExpression(compiler, expr, complex_func_result_on_stack, dst))
    return UDM_FALSE;
  if (!ParseTerm(compiler, UDM_LEX_BOOL_AND))
    return UDM_TRUE;
  if (!UdmGenerateExpressionToBool(compiler, expr))
    return UDM_FALSE;
  if (!UdmGenerateOp(compiler, UDM_PROG_JE))
    return UDM_FALSE;

  for ( ; ; )
  {
    UDM_PROG_EXPRESSION expr2;
    if (!UdmParseOrExpression(compiler, &expr2, NULL, dst))
      return UDM_FALSE;    
    if (!UdmGenerateExpressionToBool(compiler, &expr2))
      return UDM_FALSE;
    if (!UdmGenerateOp(compiler, UDM_PROG_JE))
      return UDM_FALSE;
    if (!ParseTerm(compiler, UDM_LEX_BOOL_AND))
      break;
  }
  UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                        start,
                        UdmProgCompilerGetProg(compiler)->nitems);
  return UDM_TRUE;
}


static udm_bool_t
UdmParseBoolOrExpression(UDM_PROG_COMPILER *compiler,
                         UDM_PROG_EXPRESSION *expr,
                         UDM_PROG_VAR *complex_func_result_on_stack,
                         const UDM_PROG_VAR *dst)
{
  size_t start= UdmProgCompilerGetProg(compiler)->nitems;
  if (!UdmParseBoolAndExpression(compiler, expr, complex_func_result_on_stack, dst))
    return UDM_FALSE;
  if (!ParseTerm(compiler, UDM_LEX_BOOL_OR))
    return UDM_TRUE;
  if (!UdmGenerateExpressionToBool(compiler, expr))
    return UDM_FALSE;
  if (!UdmGenerateOp(compiler, UDM_PROG_JNE))
    return UDM_FALSE;

  for ( ; ; )
  {
    UDM_PROG_EXPRESSION expr2;
    if (!UdmParseBoolAndExpression(compiler, &expr2, NULL, dst))
      return UDM_FALSE;    
    if (!UdmGenerateExpressionToBool(compiler, &expr2))
      return UDM_FALSE;
    if (!UdmGenerateOp(compiler, UDM_PROG_JNE))
      return UDM_FALSE;
    if (!ParseTerm(compiler, UDM_LEX_BOOL_OR))
      break;
  }
  UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                        start,
                        UdmProgCompilerGetProg(compiler)->nitems);
  return UDM_TRUE;
}


static udm_bool_t
UdmGenerateVarAssignmentFromExpression(UDM_PROG_COMPILER *compiler,
                                       const UDM_PROG_VAR *dst,
                                       const UDM_PROG_EXPRESSION *expr);
static udm_bool_t
UdmGenerateAssignmentOperator(UDM_PROG_COMPILER *compiler,
                              const UDM_PROG_VAR *Var,
                              const UDM_PROG_PRIMARY_EXPRESSION *Src,
                              udm_expression_t type);

static udm_bool_t
UdmParseAssignmentExpression(UDM_PROG_COMPILER *compiler,
                             UDM_PROG_EXPRESSION *expr,
                             UDM_PROG_VAR *complex_func_result_on_stack,
                             const UDM_PROG_VAR *dstvar);


/**
  <conditional-expression> ::= <logical-or-expression>
                               { ? <expression> : <conditional-expression> }*

  The conditional expression works like functions.
  - In case of a complex result type, it creates a temporary stack variable.
  - In case of a simple result type, the result is returned in registers.

  For example:
    string0= cond ? string1 : string2;

  The temporal stack variable is assigned to either string1 or string2,
  then copied to string 0.

  NOTE: Interesting: The right choice expression cannot be an assignment.
  Only the left can be:
     cond ? x= 10 : y;      // This is correct
     cond ? x     : y= 10;  // This is wrong
*/
static udm_bool_t
UdmParseConditionalExpression(UDM_PROG_COMPILER *compiler,
                              UDM_PROG_EXPRESSION *expr,
                              UDM_PROG_VAR *complex_func_result_on_stack,
                              const UDM_PROG_VAR *dst)
{
  size_t jmp0, jmp1;
  UDM_PROG_EXPRESSION expr2;
  if (!UdmParseBoolOrExpression(compiler, expr,
                                complex_func_result_on_stack, dst))
    return UDM_FALSE;
  if (!ParseTerm(compiler, UDM_LEX_QUESTION))
    return UDM_TRUE;

  if (!UdmGenerateExpressionToBool(compiler, expr))
    return UDM_FALSE;
  jmp0= UdmProgCompilerGetProg(compiler)->nitems;
  if (!UdmGenerateOp(compiler, UDM_PROG_JE)) /* To the second part */
    return UDM_FALSE;

  UDM_BZERO(expr, sizeof(*expr));
  if (!UdmParseExpression(compiler, expr,
                          complex_func_result_on_stack, dst))
    return UdmProgCompilerExpectedRule(compiler, "<expression>");

  if (UdmExprHandler(expr))
  {
    if (UdmValueHandlerIsComplex(UdmExprHandler(expr)))
    {
      if (expr->primary_expr.type == UDM_EXPRESSION_OBJECT)
      {
        /* The result variable was created on stack */
        UDM_ASSERT(UdmProgVarHandler(complex_func_result_on_stack) ==
                   UdmExprHandler(expr));
      }
      else
      {
        UDM_ASSERT(expr->primary_expr.type == UDM_EXPRESSION_IDENT ||
                   expr->primary_expr.type == UDM_EXPRESSION_CONST);
        /* Create the result variable on stack */
        if (!UdmGenerateReserveNonInitializedTemporaryVariableOnStack(compiler,
                                                          UdmExprHandler(expr),
                                                 complex_func_result_on_stack))
          return UDM_FALSE;      

        if (!UdmGenerateConstructor(compiler, complex_func_result_on_stack))
          return UDM_FALSE;
        if (!UdmGenerateAssignmentOperator(compiler,
                                           complex_func_result_on_stack,
                                           &expr->primary_expr,
                                           expr->primary_expr.type))
          return UDM_FALSE;
      }
    }
    else
    {
      if (!UdmGenerateExpressionToNativeReg0(compiler, expr,
                                             complex_func_result_on_stack))
        return UDM_FALSE;
    }
  }

  jmp1= UdmProgCompilerGetProg(compiler)->nitems;
  if (!UdmGenerateOp(compiler, UDM_PROG_JMP)) /* To the end */
    return UDM_FALSE;

  if (!ParseTermOrError(compiler, UDM_LEX_COLON))
    return UDM_TRUE;

  UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler), jmp0, jmp0 + 1);

  UDM_BZERO(&expr2, sizeof(expr2));
  /*
    If the left choice expression generated a new temporary variable
    (e.g. in case of a function or a method call), then the right
    choice expression should store the result into the same variable.
  */
  if (!UdmParseConditionalExpression(compiler, &expr2, NULL, complex_func_result_on_stack))
    return UdmProgCompilerExpectedRule(compiler, "<conditional-expression>");

  if (UdmExprHandler(expr) != UdmExprHandler(&expr2))
    return UdmProgCompilerNotSupported(compiler, "conditional expression with different argument types");

  if (UdmExprHandler(expr))
  {
    if (!UdmValueHandlerIsComplex(UdmExprHandler(expr)))
    {
      if (!UdmGenerateExpressionToNativeReg0(compiler, &expr2, NULL))
        return UDM_FALSE;
    }
    else
    {
      if (expr2.primary_expr.type != UDM_EXPRESSION_OBJECT)
      {
        /*
          Function call returns an initialized variable.
          For the other expression types call the costructor now.
        */
        if (!UdmGenerateConstructor(compiler, complex_func_result_on_stack))
          return UDM_FALSE;
        if (!UdmGenerateAssignmentOperator(compiler,
                                           complex_func_result_on_stack,
                                           &expr2.primary_expr,
                                           expr2.primary_expr.type))
        return UDM_FALSE;
      }
    }
  }

  UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler), jmp1, jmp1 + 1);
  expr->primary_expr.type= UDM_EXPRESSION_OBJECT;
  return UDM_TRUE;
}


/*
<assignment-operator> ::= =
                        | *=
                        | /=
                        | %=
                        | +=
                        | -=
                        | <<=
                        | >>=
                        | &=
                        | ^=
                        | |=
*/

static udm_bool_t
ParseEqAssignmentOperator(UDM_PROG_COMPILER *compiler)
{
  return ParseTerm(compiler, UDM_LEX_EQ);
}


static udm_bool_t
ParseSpecialAssignmentOperator(UDM_PROG_COMPILER *compiler,
                               udm_special_assignment_t *op,
                               udm_tmpl_cmd_t *cmd)
{
  *cmd= UdmProgCompilerLastTokenType(compiler);
  if (ParseTerm(compiler, UDM_LEX_MUL_EQ))
  {
    *op= UDM_ASSIGNMENT_MUL;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_DIV_EQ))
  {
    *op= UDM_ASSIGNMENT_DIV;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_REM_EQ))
  {
    *op= UDM_ASSIGNMENT_REM;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_INC_EQ))
  {
    *op= UDM_ASSIGNMENT_ADD;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_AND_EQ))
  {
    *op= UDM_ASSIGNMENT_AND;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_OR_EQ))
  {
    *op= UDM_ASSIGNMENT_OR;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_XOR_EQ))
  {
    *op= UDM_ASSIGNMENT_XOR;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_DEC_EQ))
  {
    *op= UDM_ASSIGNMENT_SUB;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_LSHIFT_EQ))
  {
    *op= UDM_ASSIGNMENT_LSHIFT;
    return UDM_TRUE;
  }
  if (ParseTerm(compiler, UDM_LEX_RSHIFT_EQ))
  {
    *op= UDM_ASSIGNMENT_RSHIFT;
    return UDM_TRUE;
  }
  return UDM_FALSE;
}


static udm_bool_t
UdmSpecialAssignmentOpDouble(UDM_PROG_COMPILER *c,
                             udm_tmpl_cmd_t *cmd, udm_special_assignment_t op,
                             const UDM_VALUE_HANDLER *dst,
                             const UDM_VALUE_HANDLER *src)
{
  switch (op)
  {
  case UDM_ASSIGNMENT_ADD: *cmd= UDM_PROG_ADD_DREG0_DREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_SUB: *cmd= UDM_PROG_SUB_DREG0_DREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_MUL: *cmd= UDM_PROG_MUL_DREG0_DREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_DIV: *cmd= UDM_PROG_DIV_DREG0_DREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_REM:
    return UdmProgCompilerBadDataTypeForOperator2(c, "%=", dst, src);
  case UDM_ASSIGNMENT_AND:
    return UdmProgCompilerBadDataTypeForOperator2(c, "&=", dst, src);
  case UDM_ASSIGNMENT_OR:
    return UdmProgCompilerBadDataTypeForOperator2(c, "|=", dst, src);
  case UDM_ASSIGNMENT_XOR:
    return UdmProgCompilerBadDataTypeForOperator2(c, "^=", dst, src);
  case UDM_ASSIGNMENT_LSHIFT:
    return UdmProgCompilerBadDataTypeForOperator2(c, "<<=", dst, src);
  case UDM_ASSIGNMENT_RSHIFT:
    return UdmProgCompilerBadDataTypeForOperator2(c, ">>=", dst, src);
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmSpecialAssignmentOpInt(UDM_PROG_COMPILER *c,
                          udm_tmpl_cmd_t *cmd, udm_special_assignment_t op,
                          const UDM_VALUE_HANDLER *dst,
                          const UDM_VALUE_HANDLER *src)
{
  switch (op)
  {
  case UDM_ASSIGNMENT_ADD: *cmd= UDM_PROG_ADD_IREG0_IREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_SUB: *cmd= UDM_PROG_SUB_IREG0_IREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_MUL: *cmd= UDM_PROG_MUL_IREG0_IREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_DIV: *cmd= UDM_PROG_DIV_IREG0_IREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_REM: *cmd= UDM_PROG_REM_IREG0_IREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_AND: *cmd= UDM_PROG_AND_IREG0_IREG1; return UDM_TRUE;
  case UDM_ASSIGNMENT_OR:  *cmd= UDM_PROG_OR_IREG0_IREG1;  return UDM_TRUE;
  case UDM_ASSIGNMENT_XOR: *cmd= UDM_PROG_XOR_IREG0_IREG1;  return UDM_TRUE;
  case UDM_ASSIGNMENT_LSHIFT: *cmd= UDM_PROG_SHL_IREG0_IREG1;  return UDM_TRUE;
  case UDM_ASSIGNMENT_RSHIFT: *cmd= UDM_PROG_SHR_IREG0_IREG1;  return UDM_TRUE;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmSpecialAssignmentOp(UDM_PROG_COMPILER *c,
                       udm_tmpl_cmd_t *cmd,
                       udm_special_assignment_t op,
                       const UDM_VALUE_HANDLER *dst,
                       const UDM_VALUE_HANDLER *src,
                       const UDM_VALUE_HANDLER *agg)
{
  UDM_ASSERT(!UdmValueHandlerIsComplex(agg));
  UDM_ASSERT(!UdmValueHandlerIsComplex(agg));
  return agg->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE ?
         UdmSpecialAssignmentOpDouble(c, cmd, op, dst, src):
         UdmSpecialAssignmentOpInt(c, cmd, op, dst, src);
}


static udm_bool_t
UdmGenerateSpecialAssignmentConvertAndPerform(UDM_PROG_COMPILER *compiler,
                                              const UDM_PROG_VAR *dst,
                                              const UDM_VALUE_HANDLER *src,
                                              udm_special_assignment_t op)
{
  UDM_VALUE_HANDLER *agg=  
    src->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE ||
    UdmProgVarHandler(dst)->native_reg_type == UDM_VALUE_DATA_TYPE_DOUBLE ?
    &UdmValueHandlerDouble : &UdmValueHandlerInt;
  UDM_PROG_CMD *cmdstore=  agg == &UdmValueHandlerDouble ?
                           &udm_tmpl_set_var_from_dreg0 :
                           &udm_tmpl_set_var_from_ireg0;
  udm_tmpl_cmd_t cmd;
  UDM_ASSERT(!UdmValueHandlerIsComplex(UdmProgVarHandler(dst)));
  UDM_ASSERT(!UdmValueHandlerIsComplex(src));
  return
    UdmSpecialAssignmentOp(compiler, &cmd, op,
                           UdmProgVarHandler(dst), src, agg) &&
    UdmGenerateNativeReg0TypeConversion(compiler, agg,
                                        UdmProgVarHandler(dst)) &&
    UdmGenerateNativeReg1TypeConversion(compiler, agg, src) &&
    UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg, cmd) &&
    UdmGenerateVariableOperation(compiler, dst, cmdstore);
}


/*
  Generate expression value in reg1,
  or move expression result from reg0 to reg1.
  Used to handle source expressions in special assignment operators,
  e.g. x<<=1.
*/
static udm_bool_t
UdmGenerateExpressionToNativeReg1(UDM_PROG_COMPILER *compiler,
                                 const UDM_PROG_VAR *Var,
                                 UDM_PROG_EXPRESSION *expr)
{
  switch (expr->primary_expr.type)
  {
  case UDM_EXPRESSION_CONST:
    expr->primary_expr.type= UDM_EXPRESSION_COMPUTED;
    return UdmGenerateLiteralToNativeRegN(compiler, &expr->primary_expr, 1);
  case UDM_EXPRESSION_IDENT:
    expr->primary_expr.type= UDM_EXPRESSION_COMPUTED;
    return UdmGenerateVarToNativeReg1(compiler, &expr->primary_expr.Var);
  case UDM_EXPRESSION_OBJECT:
    expr->primary_expr.type= UDM_EXPRESSION_COMPUTED;
    if (UdmValueHandlerIsComplex(UdmExprHandler(expr)))
      return UdmProgCompilerNotSupported(compiler, "Function of a complex return type in this context");
  case UDM_EXPRESSION_COMPUTED:
    return
      UdmGenerateMovNativeReg0ToReg1(compiler, UdmExprHandler(expr));
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
UdmGenerateExpressionSpecialAssignment(UDM_PROG_COMPILER *c,
                                       const UDM_PROG_VAR *Var,
                                       UDM_PROG_EXPRESSION *expr,
                                       udm_special_assignment_t op)
{

  return
    UdmGenerateExpressionToNativeReg1(c, Var, expr) &&
    UdmGenerateVarToNativeReg0(c, Var) &&
    UdmGenerateSpecialAssignmentConvertAndPerform(c, Var,
                                                  UdmExprHandler(expr), op);
}


static udm_bool_t
UdmGenerateVarFromNativeReg0(UDM_PROG_COMPILER *compiler,
                             const UDM_PROG_VAR *dst)
{
  switch (UdmProgVarHandler(dst)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
  case UDM_VALUE_DATA_TYPE_CHAR:
    return UdmGenerateVariableOperation(compiler, dst, &udm_tmpl_set_var_from_ireg0);
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmGenerateVariableOperation(compiler, dst, &udm_tmpl_set_var_from_dreg0);
  default:
    return UdmGenerateVariableOperation(compiler, dst, &udm_tmpl_set_var_from_cs0_z0);
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


/*
  @param name    - variable  name
  @param vname   - value name (e.g. function)
  @param handler - value handler
*/
static udm_bool_t
UdmGenerateVarAssignedFromFunctionCallResult(UDM_PROG_COMPILER *compiler,
                                             const UDM_PROG_VAR *dst,
                                             const UDM_VALUE_HANDLER *function_type_handler)
{
  switch (function_type_handler->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return
      UdmGenerateNativeReg0TypeConversion(compiler,
                                          UdmProgVarHandler(dst),
                                          function_type_handler) &&
      UdmGenerateVarFromNativeReg0(compiler, dst);
  case UDM_VALUE_DATA_TYPE_STR:
  {
    return
      UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_HREG0_0) &&
      UDM_OK == UdmProgAddArg0SimpleOp(compiler->prg, UDM_PROG_MOV_SREG0_0);
  }
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    break;
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
GenerateVarAssignedFromLiteral(UDM_PROG_COMPILER *compiler,
                               const UDM_PROG_VAR *dst,
                               const UDM_PROG_PRIMARY_EXPRESSION *expr)
{
  UDM_PROG_ITEM i;
  switch (UdmPrimaryExprHandler(expr)->datatype)
  {
  case UDM_VALUE_DATA_TYPE_INT:
  case UDM_VALUE_DATA_TYPE_CHAR:
  case UDM_VALUE_DATA_TYPE_DOUBLE:
    return UdmGenerateLiteralToNativeRegN(compiler, expr, 0) &&
           UdmGenerateNativeReg0TypeConversion(compiler,
                                               UdmProgVarHandler(dst),
                                               UdmPrimaryExprHandler(expr)) &&
           UdmGenerateVarFromNativeReg0(compiler, dst);
  case UDM_VALUE_DATA_TYPE_ENV:
  case UDM_VALUE_DATA_TYPE_RESULT:
  case UDM_VALUE_DATA_TYPE_DOCUMENT:
  case UDM_VALUE_DATA_TYPE_SQLRESULT:
  case UDM_VALUE_DATA_TYPE_EXCERPT_FRAGMENT:
    UDM_ASSERT(0);
    /* Pass through */
  case UDM_VALUE_DATA_TYPE_STR:
    break;
  }
  UdmProgItemInit(&i);
  UdmProgItemArgInitFromVar(&i.item_args, dst);
  UdmProgItemArgInitFromLiteral(&i.item_args, &expr->Var.name, &expr->textdata_addr);
  i.cmd= &udm_tmpl_set_var_from_text0;
  return UdmProgAdd(compiler->prg, &i) == UDM_OK;
}


static udm_bool_t
GenerateVarAssignedFromVar(UDM_PROG_COMPILER *compiler,
                           const UDM_PROG_VAR *dst,
                           const UDM_PROG_VAR *src)
{
  return UdmGenerateVarToNativeReg0(compiler, src) &&
         UdmGenerateNativeReg0TypeConversion(compiler,
                                             UdmProgVarHandler(dst),
                                             UdmProgVarHandler(src)) &&
         UdmGenerateVarFromNativeReg0(compiler, dst);
}


/**
  Generate an assignment:
    x= expr;
  The variable must be initialized
  (i.e. a contructor call was previously generated).
*/
static udm_bool_t
UdmGenerateVarAssignmentFromExpression(UDM_PROG_COMPILER *compiler,
                                       const UDM_PROG_VAR *dst,
                                       const UDM_PROG_EXPRESSION *expr)
{
  switch (expr->primary_expr.type)
  {
  case UDM_EXPRESSION_IDENT:
    return GenerateVarAssignedFromVar(compiler, dst, &expr->primary_expr.Var);
  case UDM_EXPRESSION_CONST:
    return GenerateVarAssignedFromLiteral(compiler, dst, &expr->primary_expr);
  case UDM_EXPRESSION_OBJECT:
    UDM_ASSERT(!UdmValueHandlerIsComplex(UdmExprHandler(expr)));
    return UdmGenerateVarAssignedFromFunctionCallResult(compiler, dst,
                                                        UdmExprHandler(expr));
  case UDM_EXPRESSION_COMPUTED:
    return UdmGenerateNativeReg0TypeConversion(compiler,
                                               UdmProgVarHandler(dst),
                                               UdmExprHandler(expr)) &&
           UdmGenerateVarFromNativeReg0(compiler, dst);
  }
  UDM_ASSERT(0);
  return UdmProgCompilerNotSupported(compiler, "This type of <postfix epxression> as a variable initializer");
}


static udm_bool_t
UdmProgAddAssignmentNone(UDM_PROG_COMPILER *compiler,
                         const UDM_PROG_EXPRESSION *expr)
{
  if (0 && !ExpressionHasEffect(expr))
  {
    char tmp[128];
    size_t tmplen;
    tmplen= udm_snprintf(tmp, sizeof(tmp),
                         "WARNING: statement has no effect: '%.*s'\n",
    (int) UdmConstTokenLength(&expr->primary_expr.Var.name),
            UdmConstTokenStr(&expr->primary_expr.Var.name));
    UdmProgGenerateCoutText(compiler, tmp, tmp + tmplen, UDM_TRUE);
    return UDM_TRUE;
  }
  return UDM_TRUE;
}


/**
  Parsee the right side of an assignment:

  int x= expr;
  x= expr;
  x+= expr;
*/
static udm_bool_t
UdmParseAssignmentSource(UDM_PROG_COMPILER *compiler,
                         UDM_PROG_EXPRESSION *expr,
                         UDM_PROG_VAR *complex_func_result_on_stack,
                         const UDM_PROG_VAR *dstvar)
{
  UDM_BZERO(expr, sizeof(*expr));
  /* Notice, it's <assignment-expression> here, not <expression> */
  if (!UdmParseAssignmentExpression(compiler, expr,
                                    complex_func_result_on_stack, dstvar))
    return UdmProgCompilerExpectedRule(compiler, "<assignment-expression>");
  if (!UdmExprHandler(expr))
    return UdmProgCompilerVoidValueIsNotIgnored(compiler);
  return UDM_TRUE;
}


static udm_bool_t
UdmExpressionIsGoodAssignmentTarget(UDM_PROG_COMPILER *compiler,
                                    UDM_PROG_EXPRESSION *expr)
{
  switch (expr->primary_expr.type)
  {
  case UDM_EXPRESSION_IDENT:
    return UDM_TRUE;
  case UDM_EXPRESSION_CONST:
  case UDM_EXPRESSION_OBJECT:
  case UDM_EXPRESSION_COMPUTED:
    break;
  }
  return UdmProgCompilerNotSupported(compiler,
                                     "This type of expression as an assignment target");
}


static udm_bool_t
UdmFindAssignmentOperatorForArgs(UDM_PROG_COMPILER *compiler,
                                 const UDM_VALUE_HANDLER *handler,
                                 const UDM_PROG_FUNCTION_CALL_ARGUMENTS *args,
                                 const UDM_FUNCTION **func)
{
  if (!(func[0]= UdmValueHandlerMethodFind(handler,
                                           UDM_CSTR_WITH_LEN("operator="),
                                           args)))
  {
    UDM_CONST_TOKEN name;
    UdmConstTokenSet(&name, UDM_CSTR_WITH_LEN("operator="));
    return UdmProgCompilerUnknownMethodError(compiler, handler, &name, args);
  }
  return UDM_TRUE;
}


static udm_bool_t
UdmFindAssignmentOperator(UDM_PROG_COMPILER *compiler,
                          const UDM_VALUE_HANDLER *dst,
                          const UDM_VALUE_HANDLER *src,
                          const UDM_FUNCTION **func)
{
  UDM_PROG_FUNCTION_CALL_ARGUMENTS args;
  UDM_BZERO(&args, sizeof(args));
  UdmProgVarSetHandler(&args.args[0].Var, src);
  args.args[0].type= UDM_EXPRESSION_IDENT;
  args.nargs= 1;
  return UdmFindAssignmentOperatorForArgs(compiler, dst, &args, func);
}


static udm_bool_t
UdmGenerateAssignmentOperator(UDM_PROG_COMPILER *compiler,
                              const UDM_PROG_VAR *Var,
                              const UDM_PROG_PRIMARY_EXPRESSION *Src,
                              udm_expression_t type)
{
  const UDM_FUNCTION *func;
  UDM_PROG_FUNCTION_CALL_ARGUMENTS args;
  UDM_BZERO(&args, sizeof(args));
  args.args[0].Var= Src[0].Var;
  args.args[0].textdata_addr= Src->textdata_addr;
  args.args[0].type= type;
  args.nargs= 1;
  return
    UdmFindAssignmentOperatorForArgs(compiler, UdmProgVarHandler(Var),
                                     &args, &func) &&
    UdmGenerateNewFunctionCall(compiler, func, Var, &args, NULL, Var);
}


/**
  Assign an expression to a variable.
  The variable must be initialized
  (i.e. a constructor call was previously generated).
*/
static udm_bool_t
UdmGenerateAssignment(UDM_PROG_COMPILER *compiler,
                      const UDM_PROG_VAR *Var,
                      const UDM_PROG_EXPRESSION *value,
                      const UDM_PROG_PRIMARY_EXPRESSION *tmp)
{
  if (UdmPrimaryExprHandler(tmp))
  {
    /*
      The expression was a function call returning a complex object.
      Copy temporary variable to the destination variable.
    */
    UDM_ASSERT(UdmValueHandlerIsComplex(UdmPrimaryExprHandler(tmp)));
    return UdmGenerateAssignmentOperator(compiler, Var, tmp,
                                         UDM_EXPRESSION_IDENT);
  }
  else if (!UdmValueHandlerIsComplex(UdmProgVarHandler(Var)))
  {
    if (!UdmDyadicOperatorFindWithError(compiler, assignment_operators,
                                        UdmProgVarHandler(Var),
                                        UdmExprHandler(value),
                                        "="))
      return UDM_FALSE;
    if (!UdmGenerateVarAssignmentFromExpression(compiler, Var, value))
      return UDM_FALSE;
    
  }
  else
  {
    UDM_ASSERT(UdmValueHandlerIsComplex(UdmProgVarHandler(Var)));
    switch (value->primary_expr.type)
    {
    case UDM_EXPRESSION_IDENT:
    case UDM_EXPRESSION_CONST:
      return UdmGenerateAssignmentOperator(compiler, Var,
                                           &value->primary_expr,
                                           value->primary_expr.type);
    case UDM_EXPRESSION_OBJECT:
      UDM_ASSERT(0);
      return UDM_FALSE;
    case UDM_EXPRESSION_COMPUTED:
      {
        /*
          TODO: get rid of this.
          <conditional-expression> currenly returns complex values in cs0/z0.
          This is limited to string type only and should be changed
          to support all other complex data types.
        */
        const UDM_FUNCTION *func;
        if (!UdmFindAssignmentOperator(compiler,
                                       UdmProgVarHandler(Var),
                                       UdmExprHandler(value), &func))
          return UDM_FALSE;
        if (!UdmGenerateVarAssignmentFromExpression(compiler, Var, value))
          return UDM_FALSE;
      }
      break;
    }
  }

  return UDM_TRUE;
}



/**
  Parse the expression in the right side of an assignment,
  and generate the assignment.
  TODO: this crashes:
    string x= "abcd";
    string y;
    y= x.substr(1,3).substr(0,2).length();     
*/
static udm_bool_t
UdmParseAssignmentPart2(UDM_PROG_COMPILER *compiler,
                        const UDM_PROG_VAR *dst)
{
  /*
    For an expression like this:
      string x= "test"; x= x.left();
    we create a temporary variable for the function/method result,
    which then copied to the final destination.
    If there are multiple temporary values in the same <assignment-expression>:    
      x= x.left(), y= y.left();
    then all temporary values destructed at once at the end
    (e.g. when the semicolon is read).
  */
  UDM_PROG_EXPRESSION rvalue;
  UDM_PROG_PRIMARY_EXPRESSION tmp;
  UDM_BZERO(&tmp, sizeof(tmp));
  tmp.type= UDM_EXPRESSION_IDENT; /*TODO: get rid if this, pass UDM_PROG_VAR*/
  return UdmParseAssignmentSource(compiler, &rvalue, &tmp.Var, NULL) &&
         UdmGenerateAssignment(compiler, dst, &rvalue, &tmp);
}


/**
  <assignment-expression> ::=
      <conditional-expression>
    | <unary-expression> <assignment-operator> <assignment-expression>
*/
static udm_bool_t
UdmParseAssignmentExpression(UDM_PROG_COMPILER *compiler,
                             UDM_PROG_EXPRESSION *expr,
                             UDM_PROG_VAR *complex_result_on_stack,
                             const UDM_PROG_VAR *dstvar)
{
  udm_special_assignment_t op;
  udm_tmpl_cmd_t cmd;
  if (!UdmParseConditionalExpression(compiler, expr,
                                     complex_result_on_stack, dstvar))
    return UDM_FALSE;
  if (ParseEqAssignmentOperator(compiler))  /* x= expression; */
  {
    return UdmExpressionIsGoodAssignmentTarget(compiler, expr) &&
           UdmParseAssignmentPart2(compiler, &expr->primary_expr.Var);
  }
  if (ParseSpecialAssignmentOperator(compiler, &op, &cmd)) /* e.g. x+= expression; */
  {
    UDM_PROG_EXPRESSION rvalue;
    UDM_PROG_VAR complex_result_on_stack2;
    UDM_BZERO(&complex_result_on_stack2, sizeof(complex_result_on_stack2));
    if (!UdmExpressionIsGoodAssignmentTarget(compiler, expr))
      return UDM_FALSE;
    if (!UdmParseAssignmentSource(compiler, &rvalue,
                                  &complex_result_on_stack2, NULL))
      return UDM_FALSE;
    if (!UdmDyadicOperatorFindWithError(compiler, assignment_operators,
                                        UdmExprHandler(expr),
                                        UdmExprHandler(&rvalue),
                                        UdmLex2str(cmd)))
      return UDM_FALSE;
    /* Special assignment operators for complex types are not supported yet */
    UDM_ASSERT(!UdmValueHandlerIsComplex(UdmExprHandler(&rvalue)));
    UDM_ASSERT(!UdmProgVarHandler(&complex_result_on_stack2));
    return UdmGenerateExpressionSpecialAssignment(compiler,
                                                  &expr->primary_expr.Var,
                                                  &rvalue, op);
  }
  if (!UdmProgAddAssignmentNone(compiler, expr))
    return UDM_FALSE;
  return UDM_TRUE;
}


/*
<expression> ::= <assignment-expression> {, <assignment-expression> }
*/
static udm_bool_t
UdmParseExpression(UDM_PROG_COMPILER *compiler,
                   UDM_PROG_EXPRESSION *expr,
                   UDM_PROG_VAR *complex_result_on_stack,
                   const UDM_PROG_VAR *dstvar)
{
  if (!UdmParseAssignmentExpression(compiler, expr,
                                    complex_result_on_stack, dstvar))
    return UDM_FALSE;
  for ( ; ; )
  {
    if (!ParseTerm(compiler, UDM_LEX_COMMA))
      return UDM_TRUE;
    UDM_BZERO(expr, sizeof(*expr));
    if (!UdmParseAssignmentExpression(compiler, expr,
                                      complex_result_on_stack, dstvar))
      return UdmProgCompilerExpectedRule(compiler, "<assignment-expression>");
  }
  return UDM_TRUE;
}


/**
  Parse an assignment expression in its own block.
  If the expression returns a value which is not used,
  it needs to be destructed immediately, e.g.:
  {
    string str= "xxx";
    str.left();
  }
  UdmProgCompilerLeaveBlock makes sure it's destructed.
  This effectively converts expressions like:
    str.left();
  into
    {
      string str= str.left(); // str is destructed here
    }
*/
static udm_bool_t
UdmParseExpressionIgnoreResult(UDM_PROG_COMPILER *compiler)
{
  udm_bool_t rc;
  UDM_PROG_EXPRESSION expr;
  UDM_PROG_VAR complex_result_on_stack;
  UDM_BZERO(&expr, sizeof(expr));
  UDM_BZERO(&complex_result_on_stack, sizeof(complex_result_on_stack));
  if (!UdmProgCompilerEnterBlock(compiler))
    return UDM_FALSE;  
  rc= UdmParseExpression(compiler, &expr,
                         &complex_result_on_stack, NULL);
  if (!UdmProgCompilerLeaveBlock(compiler))
    return UDM_FALSE;
  return rc;
}


static udm_bool_t
UdmGenerateCoutReg0(UDM_PROG_COMPILER *compiler,
                    const UDM_VALUE_HANDLER *handler)
{
  switch (handler->type)
  {
    case UDM_VALUE_HANDLER_TYPE_CHAR:
      return UDM_OK == UdmProgAddArg0Simple(compiler->prg, &udm_tmpl_cout_ireg0_char);
    case UDM_VALUE_HANDLER_TYPE_INT:
      return UDM_OK == UdmProgAddArg0Simple(compiler->prg, &udm_tmpl_cout_ireg0);
    case UDM_VALUE_HANDLER_TYPE_DOUBLE:
      return UDM_OK == UdmProgAddArg0Simple(compiler->prg, &udm_tmpl_cout_dreg0);
    case UDM_VALUE_HANDLER_TYPE_STR:
    case UDM_VALUE_HANDLER_TYPE_ENV:
    case UDM_VALUE_HANDLER_TYPE_RESULT:
    case UDM_VALUE_HANDLER_TYPE_DOCUMENT:
    case UDM_VALUE_HANDLER_TYPE_SQLRESULT:
    case UDM_VALUE_HANDLER_TYPE_EXCERPT_FRAGMENT:
      return UDM_OK == UdmProgAddArg0Simple(compiler->prg, &udm_tmpl_cout_cs0_z0);
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


/*
  Return UDM_TRUE on success, UDM_FALSE on error.
*/
static udm_bool_t
UdmGenerateCoutExpression(UDM_PROG_COMPILER *compiler,
                          UDM_PROG_EXPRESSION *expr,
                          const UDM_PROG_VAR *func_result)
{
  udm_expression_t orig_type= expr->primary_expr.type;
  if (!UdmGenerateExpressionToNativeReg0(compiler, expr, func_result))
    return UDM_FALSE;
  switch (orig_type)
  {
    case UDM_EXPRESSION_IDENT:
    case UDM_EXPRESSION_CONST:
    case UDM_EXPRESSION_COMPUTED:
      return UdmGenerateCoutReg0(compiler, UdmExprHandler(expr));
    case UDM_EXPRESSION_OBJECT:
      if (!UdmExprHandler(expr))
        return UdmProgCompilerVoidValueIsNotIgnored(compiler);
      if (UdmProgVarHandler(func_result))
      {
        UDM_ASSERT(UdmProgVarHandler(func_result) == UdmExprHandler(expr));
        UDM_ASSERT(UdmValueHandlerIsComplex(UdmExprHandler(expr)));
        return UdmGenerateCoutReg0(compiler, UdmProgVarHandler(func_result));
      }
      else
      {
        UDM_ASSERT(!UdmValueHandlerIsComplex(UdmExprHandler(expr)));
        return UdmGenerateCoutReg0(compiler, UdmExprHandler(expr));
      }
  }
  UDM_ASSERT(0);
  return UDM_FALSE;
}


static udm_bool_t
ParseEchoExpression(UDM_PROG_COMPILER *compiler)
{
  UDM_PROG_EXPRESSION expr;
  if (!ParseTerm(compiler, UDM_LEX_COUT))
    return UDM_FALSE;
  if (!ParseTermOrError(compiler, UDM_LEX_LSHIFT))
    return UDM_FALSE;
  for ( ; ; )
  {
    UDM_PROG_VAR func_result;
    bzero(&func_result, sizeof(UDM_PROG_VAR));
    if (!UdmProgCompilerEnterBlock(compiler)) /* In case if expr is a function call */
      return UDM_FALSE;
    if (!ParseAdditiveExpression(compiler, &expr, &func_result, NULL))
      return UDM_FALSE;
    if (!UdmGenerateCoutExpression(compiler, &expr, &func_result))
      return UDM_FALSE;
    if (!UdmProgCompilerLeaveBlock(compiler))
      return UDM_FALSE;
    if (!ParseTerm(compiler, UDM_LEX_LSHIFT))
      break;
  }
  return UDM_TRUE;
}


/*
<expression-statement> ::= {<expression>}? ;
*/
static udm_bool_t
ParseExpressionStatement(UDM_PROG_COMPILER *compiler)
{
  if (ParseEchoExpression(compiler))
    return ParseTermOrError(compiler, UDM_LEX_SEMICOLON);
  if (UdmParseExpressionIgnoreResult(compiler))
    return ParseTermOrError(compiler, UDM_LEX_SEMICOLON);
  return ParseTerm(compiler, UDM_LEX_SEMICOLON);
}


static udm_bool_t ParseStatement(UDM_PROG_COMPILER *compiler);


static udm_bool_t
ParseIfCondition(UDM_PROG_COMPILER *compiler)
{
  UDM_PROG_EXPRESSION expr;
  return 
    UdmParseExpression(compiler, &expr, NULL, NULL) &&
    UdmGenerateExpressionToBool(compiler, &expr);
}


static udm_bool_t
ParseParenthesizedCondition(UDM_PROG_COMPILER *compiler)
{
  if (!ParseTermOrError(compiler, UDM_LEX_LP))
    return UDM_FALSE;
  if (!ParseIfCondition(compiler))
    return UDM_FALSE;
  if (!ParseTermOrError(compiler, UDM_LEX_RP))
    return UDM_FALSE;
  return UDM_TRUE;
}


/*
<selection-statement> ::= if ( <expression> ) <statement>
                        | if ( <expression> ) <statement> else <statement>
                        | switch ( <expression> ) <statement>
*/
static udm_bool_t ParseSelectionStatement(UDM_PROG_COMPILER *compiler)
{
  size_t jne_to_end_of_if_block;
  if (!ParseTerm(compiler, UDM_LEX_IF))
    return UDM_FALSE;
  if (!ParseParenthesizedCondition(compiler))
    return UDM_FALSE;

  /* jne to end_of_if_block if the condition failed */
  jne_to_end_of_if_block= UdmProgCompilerGetProg(compiler)->nitems;
  if (UdmProgAddJmp(compiler->prg, UDM_PROG_JE, 0))
    return UDM_FALSE;

  if (!ParseStatement(compiler))
    return UdmProgCompilerExpectedRule(compiler, "<statement>");

  if (ParseTerm(compiler, UDM_LEX_ELSE))
  {
    /*
      We are at the end of the "if" block, and there is "else".
      Add unconditional JMP to the end of "else".
    */
    size_t start_of_else= UdmProgCompilerGetProg(compiler)->nitems;
    if (UdmProgAddJmp(compiler->prg, UDM_PROG_JMP, 0))
      return UDM_FALSE;

    /*
      Now we know the start address of the "else" block.
      Fix the jump address of the command pointed by jne_to_end_of_if_block
      to the current address, which is the beginning of the "else" block.
    */
    UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                          jne_to_end_of_if_block,
                          jne_to_end_of_if_block + 1);
    if (!ParseStatement(compiler))
      return UdmProgCompilerExpectedRule(compiler, "<statement>");

    UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                          start_of_else, start_of_else + 1);
  }
  else
  {
    /*
      No else, just fix the command pointed by
      jne_to_end_of_if_block to jump to the current address,
      which is the end of the entire "if" statement.
    */
    UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                          jne_to_end_of_if_block,
                          jne_to_end_of_if_block + 1);
  }
  return UDM_TRUE;
}


/*
  while ( <expression> ) <statement>
*/
static udm_bool_t
ParseWhileStatement(UDM_PROG_COMPILER *compiler)
{
  size_t backup_current_loop_continue;
  if (!ParseTerm(compiler, UDM_LEX_WHILE))
    return UDM_FALSE;
  backup_current_loop_continue= compiler->current_loop_continue;
  compiler->current_loop_continue= UdmProgCompilerGetProg(compiler)->nitems;
  if (!ParseParenthesizedCondition(compiler))
    return UDM_FALSE;

  /* Break the loop if the condition failed */
  if (UdmProgAddJmpWithComment(compiler->prg,
                               UDM_PROG_JE, 0, "while condition failed"))
    return UDM_FALSE;
 
  if (!ParseStatement(compiler))
    return UdmProgCompilerExpectedRule(compiler, "<statement>");

  /* Jump to the beginning */
  if (UdmProgAddJmpWithComment(compiler->prg, UDM_PROG_JMP,
                               compiler->current_loop_continue,
                               "implicit continue"))
    return UDM_FALSE;

  UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                        compiler->current_loop_continue,
                        UdmProgCompilerGetProg(compiler)->nitems - 1);
  compiler->current_loop_continue= backup_current_loop_continue;
  return UDM_TRUE;
}


static udm_bool_t
ParseDeclaration(UDM_PROG_COMPILER *compiler);


static udm_bool_t
UdmParseForStatementInitialization(UDM_PROG_COMPILER *compiler)
{
  return ParseDeclaration(compiler) ||
         UdmParseExpressionIgnoreResult(compiler);
}


/*
  The C grammar looks like:
    for ( {<expression>}? ; {<expression>}? ; {<expression>}? ) <statement>

  We allow declaraions:
    for ( {<declaration>|<expression>}? ;
          {<expression>}? ;
          {<expression>}? )
      <statement>
*/
static udm_bool_t
ParseForStatement(UDM_PROG_COMPILER *compiler)
{
  udm_bool_t has_finalize_expression;
  size_t condition, finalize_expression, backup_current_loop_continue;
  if (!ParseTerm(compiler, UDM_LEX_FOR))
    return UDM_FALSE;
  if (!ParseTermOrError(compiler, UDM_LEX_LP))
    return UDM_FALSE;
  /*
    Create its own level, in case the initialization part will
    declare variables.
  */
  if (!UdmProgCompilerEnterBlock(compiler))
    return UDM_FALSE;
  /* Optional for-init expression */
  if (!UdmParseForStatementInitialization(compiler) &&
      UdmProgCompilerError(compiler))
    return UDM_FALSE;
  if (!ParseTermOrError(compiler, UDM_LEX_SEMICOLON))
    return UDM_FALSE;

  /* Condition expression, currently non-optional */
  condition= UdmProgCompilerGetProg(compiler)->nitems;
  if (!ParseIfCondition(compiler))
    return UDM_FALSE;
  /* Break the loop if the condition <expression> failed */
  if (UdmProgAddJmpWithComment(compiler->prg,
                               UDM_PROG_JE, 0, "to the FOR end"))
    return UDM_FALSE;
  if (!ParseTermOrError(compiler, UDM_LEX_SEMICOLON))
    return UDM_FALSE;

  /* Jump to the body <statement> */
  if (UdmProgAddJmpWithComment(compiler->prg,
                               UDM_PROG_JMP, 0, "to the FOR body statement"))
    return UDM_FALSE;

  /* Optional for-finalize <expression> */
  finalize_expression= UdmProgCompilerGetProg(compiler)->nitems;
  if (!(has_finalize_expression= UdmParseExpressionIgnoreResult(compiler)) &&
      UdmProgCompilerError(compiler))
    return UDM_FALSE;
  /* Go to the condition <expression> after for-finalize <expression> is done */
  if (UdmProgAddJmpWithComment(compiler->prg,
                               UDM_PROG_JMP, condition,
                               "to the FOR condition"))
    return UDM_FALSE;
  if (!ParseTermOrError(compiler, UDM_LEX_RP))
    return UDM_FALSE;

  /*
    Fix JMP after condition and before finalize_expression
    to the address of the body <statement>.
  */
  UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                        finalize_expression - 1,
                        finalize_expression);
  backup_current_loop_continue= compiler->current_loop_continue;
  compiler->current_loop_continue= has_finalize_expression ?
                                      finalize_expression : condition;
  if (!ParseStatement(compiler))
    return UdmProgCompilerExpectedRule(compiler, "<statement>");

  /* Jump to for_finalize <expression> or to the condition <expression> */
  if (UdmProgAddJmpWithComment(compiler->prg, UDM_PROG_JMP,
                               has_finalize_expression ?
                               finalize_expression : condition,
                               has_finalize_expression ?
                               "to the FOR finalize expression" :
                               "to the FOR condition"))
    return UDM_FALSE;

  /* Fix the jump after the condition failure to the end of the loop */
  UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                        condition,
                        UdmProgCompilerGetProg(compiler)->nitems - 1);

  compiler->current_loop_continue= backup_current_loop_continue;
  return UdmProgCompilerLeaveBlock(compiler);
}


/*
  do <statement> while ( <expression> ) ;
*/
static udm_bool_t
ParseDoStatement(UDM_PROG_COMPILER *compiler)
{
  size_t backup_current_loop_continue;
  if (!ParseTerm(compiler, UDM_LEX_DO))
    return UDM_FALSE;

  backup_current_loop_continue= compiler->current_loop_continue;
  compiler->current_loop_continue= UdmProgCompilerGetProg(compiler)->nitems;

  if (!ParseStatement(compiler))
    return UdmProgCompilerExpectedRule(compiler, "<statement>");

  if (!ParseTermOrError(compiler, UDM_LEX_WHILE))
    return UDM_FALSE;

  if (!ParseParenthesizedCondition(compiler))
    return UDM_FALSE;

  /* Jump to the beginning if the condition is true */
  if (UdmProgAddJmp(compiler->prg, UDM_PROG_JNE, compiler->current_loop_continue))
    return UDM_FALSE;
  /* Fix jumps for "break" statements */
  UdmProgFixJumpInRange(UdmProgCompilerGetProg(compiler),
                        compiler->current_loop_continue,
                        UdmProgCompilerGetProg(compiler)->nitems - 1);
  compiler->current_loop_continue= backup_current_loop_continue; 
  return UDM_TRUE;
}


/*
<iteration-statement> ::= while ( <expression> ) <statement>
                        | do <statement> while ( <expression> ) ;
                        | for ( {<expression>}? ; {<expression>}? ; {<expression>}? ) <statement>
*/
static udm_bool_t
ParseIterationStatement(UDM_PROG_COMPILER *compiler)
{
  return
    ParseWhileStatement(compiler) ||
    ParseDoStatement(compiler) ||
    ParseForStatement(compiler);
}


static udm_bool_t
UdmProgCompilerNotWithin(UDM_PROG_COMPILER *compiler, const char *item, const char *where)
{
  udm_snprintf(compiler->errstr, sizeof(compiler->errstr),
               "'%s' not within a %s", item, where);
  return UDM_FALSE;
}


static udm_bool_t
ParseJumpStatement(UDM_PROG_COMPILER *compiler)
{
  if (ParseTerm(compiler, UDM_LEX_CONTINUE))
  {
    if (!compiler->current_loop_continue)
      return UdmProgCompilerNotWithin(compiler, "continue", "loop");    
    return UDM_OK == UdmProgAddJmpWithComment(compiler->prg,
                                              UDM_PROG_JMP, 
                                              compiler->current_loop_continue,
                                              "explicit continue");
  }
  if (ParseTerm(compiler, UDM_LEX_BREAK))
  {
    if (!compiler->current_loop_continue)
      return UdmProgCompilerNotWithin(compiler, "continue", "loop or switch");
    return UDM_OK == UdmProgAddJmpWithComment(compiler->prg,
                                              UDM_PROG_JMP, 0, "break");
  }
  return UDM_FALSE;
}


static udm_bool_t
UdmLexScannerScanUntilNativePIStart(UDM_PROG_COMPILER *compiler)
{
  const UDM_CONST_STR native= {UDM_CSTR_WITH_LEN("<?mnogosearch")};
  return UdmLexScannerScanUntil(&compiler->scanner, &native,
                                UDM_LEX_TEXT, UDM_LEX_EOF);
}


/*
  Embedded text:
  ?>some text<?mnogosearch
   
*/
static udm_bool_t
ParseEmbeddedTextStatement(UDM_PROG_COMPILER *compiler)
{
  const char *start, *end;
  if (UdmProgCompilerLastTokenType(compiler) != UDM_LEX_QUESTION)
    return UDM_FALSE;
  if (UdmLexScannerCur(&compiler->scanner)[0] != '>')
    return UDM_FALSE;
  start= UdmLexScannerCur(&compiler->scanner) + 1; /* Skip '>' */
  UdmLexScannerScanUntilNativePIStart(compiler);
  end= compiler->scanner.token.token.end;
  /* Skip the newline that immediately follows the closing '?>' */
  if (start < end && *start == '\r')
    start++;
  if (start < end && *start == '\n')
    start++;
  UdmProgGenerateCoutText(compiler, start, end, UDM_FALSE);
  UdmProgCompilerScan(compiler);
  return UDM_TRUE;
}


static udm_bool_t ParseCompoundStatement(UDM_PROG_COMPILER *compilerarser);
/*
<statement> ::= <labeled-statement>
              | <expression-statement>
              | <compound-statement>
              | <selection-statement>
              | <iteration-statement>
              | <jump-statement>
*/
static udm_bool_t
ParseStatement(UDM_PROG_COMPILER *compiler)
{
  return ParseExpressionStatement(compiler) ||
         ParseSelectionStatement(compiler) ||
         ParseIterationStatement(compiler) ||
         ParseJumpStatement(compiler) ||
         ParseCompoundStatement(compiler) ||
         ParseEmbeddedTextStatement(compiler);
}


static udm_bool_t
ParseStatementList(UDM_PROG_COMPILER *compiler)
{
  for ( ; ; )
  {
    if (!ParseStatement(compiler))
    {
      if (UdmProgCompilerError(compiler))
        return UDM_FALSE;
      break;
    }
  }
  return UDM_TRUE;
}


/*
<direct-declarator> ::= <identifier>
                      | ( <declarator> )
                      | <direct-declarator> [ {<constant-expression>}? ]
                      | <direct-declarator> ( <parameter-type-list> )
                      | <direct-declarator> ( {<identifier>}* )
*/
static udm_bool_t
ParseDirectDeclarator(UDM_PROG_COMPILER *compiler, UDM_CONST_TOKEN *Var)
{
  if (UdmProgCompilerLastTokenType(compiler) == UDM_LEX_IDENT)
    *Var= compiler->scanner.token.token;
  return ParseTerm(compiler, UDM_LEX_IDENT);
}


/*
<declarator> ::= {<pointer>}? <direct-declarator>
*/
static udm_bool_t
ParseDeclarator(UDM_PROG_COMPILER *compiler,  UDM_CONST_TOKEN *Var)
{
  return ParseDirectDeclarator(compiler, Var);
}


/*
  TODO:
  <initializer> ::= <assignment-expression>
                    | { <initializer-list> }
                    | { <initializer-list> , }                                
*/
static udm_bool_t
ParseVarInitializer(UDM_PROG_COMPILER *compiler, const UDM_PROG_VAR *Var)
{
  return UdmParseAssignmentPart2(compiler, Var);
}


/*
<init-declarator> ::= <declarator>
                    | <declarator> = <initializer>
*/
static udm_bool_t
ParseInitDeclarator(UDM_PROG_COMPILER *compiler, const UDM_VALUE_HANDLER *ha)
{
  UDM_PROG_VAR Var;
  if (!ParseDeclarator(compiler, &Var.name))
    return UDM_FALSE;

  UdmProgStackVarInitAttributes(&Var, compiler, ha);
  if (!UdmGenerateDeclareNonInitializedVariable(compiler, &Var))
    return UDM_FALSE;
  if (!UdmGenerateConstructor(compiler, &Var))
    return UDM_FALSE;
  if (!ParseTerm(compiler, UDM_LEX_EQ))
    return UDM_TRUE;
  if (!ParseVarInitializer(compiler, &Var))
    return UdmProgCompilerExpectedRule(compiler, "<initializer>");
  return UDM_TRUE;
}


static udm_bool_t
ParseOptInitDeclaratorList(UDM_PROG_COMPILER *compiler,
                           const UDM_VALUE_HANDLER *ha)
{
  /*
    Check for an empty declaration,
    type not followed by an identifier name, e.g.:
    { int; }
  */
  if (!ParseInitDeclarator(compiler, ha))
    return UdmProgCompilerGenericError(compiler,
                                       "declaration does not declare anything");
  while (ParseTerm(compiler, UDM_LEX_COMMA))
  {
    if (!ParseInitDeclarator(compiler, ha))
      return UdmProgCompilerExpectedRule(compiler, "<init-declarator>");
  }
  return UDM_TRUE;
}


/*
<declaration> ::=  {<declaration-specifier>}+ {<init-declarator>}*
*/
static udm_bool_t
ParseDeclaration(UDM_PROG_COMPILER *compiler)
{
  const UDM_VALUE_HANDLER *ha;
  if (!ParseDeclarationSpecifier(compiler, &ha))
    return UDM_FALSE;
  if (!ParseOptInitDeclaratorList(compiler, ha) &&
       UdmProgCompilerError(compiler))
    return UDM_FALSE;
  return UDM_TRUE;
}


static udm_bool_t
ParseOptDeclarationList(UDM_PROG_COMPILER *compiler)
{
  for ( ; ; )
  {
    if (!ParseDeclaration(compiler))
    {
      if (UdmProgCompilerError(compiler))
        return UDM_FALSE;
      break;
    }
    if (!ParseTermOrError(compiler, UDM_LEX_SEMICOLON))
      return UDM_FALSE;
  }
  return UDM_TRUE;
}


/*
<compound-statement> ::= { {<declaration>}* {<statement>}* }
*/
static udm_bool_t
ParseCompoundStatement(UDM_PROG_COMPILER *compiler)
{
  if (!ParseLCB(compiler))
    return UDM_FALSE;
  if (!UdmProgCompilerEnterBlock(compiler))
    return UDM_FALSE;
  if (!ParseOptDeclarationList(compiler))
    return UDM_FALSE;
  if (!ParseStatementList(compiler))
    return !UdmProgCompilerError(compiler);
  if (!UdmProgCompilerLeaveBlock(compiler))
    return UDM_FALSE;
  return ParseTermOrError(compiler, UDM_LEX_RCB);
}


static udm_bool_t
ParsePI(UDM_PROG_COMPILER *compiler)
{
  if (!UdmProgCompilerEnterBlock(compiler))
    return UDM_FALSE;
  if (!ParseOptDeclarationList(compiler))
    return UDM_FALSE;
  if (!ParseStatementList(compiler))
    return UDM_FALSE;
  if (!UdmProgCompilerLeaveBlock(compiler))
    return UDM_FALSE;
  return ParseTermOrError(compiler, UDM_LEX_EOF);
}


udm_rc_t
UdmCompilePI(UDM_PROG_COMPILER *compiler, const UDM_CONST_STR *src)
{
  udm_bool_t rc;
  UdmPIParserInit(compiler, src->str, src->length);
  UdmProgCompilerScan(compiler);
  if (!(rc= ParsePI(compiler)))
  {
    char tmp[124];
    size_t tmplen;
    tmplen= udm_snprintf(tmp, sizeof(tmp), "%s\n", compiler->errstr);
    UdmProgGenerateCoutText(compiler, tmp, tmp + tmplen, UDM_FALSE);
  }
  /*
  for (UdmProgCompilerScan(&parser);
       parser.scanner.token.type != UDM_LEX_UNKNOWN &&
       parser.scanner.token.type != UDM_LEX_EOF ;
       UdmProgCompilerScan(&parser))
  {
  }
  */
  UdmPIParserFree(compiler);
  return UDM_OK;
}


static udm_bool_t
UdmProgCompilerScanPrefixText(UDM_PROG_COMPILER *compiler)
{
  if (!UdmLexScannerScanUntilNativePIStart(compiler))
    return UDM_TRUE;
  UdmProgGenerateCoutText(compiler,
                          compiler->scanner.token.token.str,
                          compiler->scanner.token.token.end,
                          UDM_FALSE);
  compiler->pi_count++;
  UdmProgCompilerScan(compiler); /* For look-ahead */
  return UDM_TRUE;
}


static udm_bool_t
ParsePIProgram(UDM_PROG_COMPILER *compiler)
{
  if (!UdmProgCompilerEnterBlock(compiler))
    return UDM_FALSE;
  if (!UdmProgCompilerScanPrefixText(compiler))
    return UDM_FALSE;
  if (!ParseOptDeclarationList(compiler))
    return UDM_FALSE;
  if (!ParseStatementList(compiler))
    return UDM_FALSE;
  if (!UdmProgCompilerLeaveBlock(compiler))
    return UDM_FALSE;
  return ParseTermOrError(compiler, UDM_LEX_EOF);
}


udm_rc_t
UdmCompilePIProgram(UDM_PROG_COMPILER *compiler, const UDM_CONST_STR *src)
{
  udm_bool_t rc;
  UdmPIParserInit(compiler, src->str, src->length);
  if (!(rc= ParsePIProgram(compiler)))
  {
    char tmp[124];
    size_t tmplen;
    tmplen= udm_snprintf(tmp, sizeof(tmp), "%s\n", compiler->errstr);
    UdmProgGenerateCoutText(compiler, tmp, tmp + tmplen, UDM_FALSE);
  }
  UdmPIParserFree(compiler);
  return rc == UDM_TRUE ? UDM_OK : UDM_ERROR;
}
