/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>


#ifdef WIN32
#include <winsock.h>
#include <time.h>
#include <direct.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#ifdef HAVE_SELECT_H
#include <select.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef USE_HTTPS
#include <openssl/rand.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

#include "udm_dirent.h"
#include "udm_common.h"
#include "udm_proto.h"
#include "udm_utils.h"
#include "udm_socket.h"
#include "udm_http.h"
#include "udm_ftp.h"
#include "udm_xmalloc.h"
#include "udm_host.h"
#include "udm_execget.h"
#include "udm_id3.h"
#include "udm_log.h"
#include "udm_vars.h"
#include "udm_db.h"
#include "udm_match.h"
#include "udm_agent.h"
#include "udm_contentencoding.h"
#include "udm_url.h"
#include "udm_indexer.h" /* For UdmDocProcessContentResponseHeaders */

/************** connect with timeout **********/

static int
connect_tm(int s,
           const struct sockaddr *name,
           unsigned int namelen,
           unsigned int to)
{
#ifdef WIN32
  return connect(s, name, namelen);
#else
  int flags, res, s_err;
  socklen_t s_err_size= sizeof(int);
  fd_set sfds;
  struct timeval tv;

  if (!to)
    return connect(s,name, namelen);

  flags= fcntl(s, F_GETFL, 0);     /* Set socket to non-blocked */
#ifdef O_NONBLOCK
  fcntl(s, F_SETFL, flags | O_NONBLOCK);  /* and save the flags        */
#endif

  res= connect(s, name, namelen);
  s_err= errno;      /* Save errno */
  fcntl(s, F_SETFL, flags);
  if ((res != 0) && (s_err != EINPROGRESS))
  {
    errno= s_err;    /* Restore errno              */
    return -1;       /* Unknown error, not timeout */
  }
  if (!res)
    return 0;        /* Quickly connected */

  FD_ZERO(&sfds);
  FD_SET(s,&sfds);

  tv.tv_sec= (long) to;
  tv.tv_usec= 0;
  
  while(1)
  {
    res= select(s+1, NULL, &sfds, NULL, &tv);
    if(res == 0)
      return -1;          /* Timeout */
    if(res < 0)
    {
      if (errno == EINTR) /* Signal */
        continue;
      else
        return -1 ;       /* Error */
    }
    break;
  }

  s_err= 0;
  if (getsockopt(s, SOL_SOCKET, SO_ERROR, (char*) &s_err, &s_err_size) != 0)
    return -1;            /* Can't get sock options */

  if (s_err)
  {
    errno= s_err;
    return -1;
  }
  return 0;      /* We have a connection! */
#endif
}


/************** Open TCP Connect **************/


static int
open_host(UDM_AGENT *Agent, UDM_DOCUMENT *Doc)
{
  int net, res;
  
  UdmAgentSetTask(Agent, "Connecting");
  net= socket(AF_INET, SOCK_STREAM, 0);
  Doc->connp.sin.sin_family= AF_INET;
  res= connect_tm(net,
                  (struct sockaddr *)&Doc->connp.sin, sizeof (struct sockaddr_in),
                  (unsigned int)Doc->Spider.read_timeout);
  UdmAgentSetTask(Agent, "Downloading");
  if (res)
  {
    closesocket(net);
    return UDM_NET_CANT_CONNECT;
  }
  return net;
}


/**************** HTTP codes and messages ************/

UDM_API(const char *)
UdmHTTPErrMsg(int code)
{
  switch(code){
  case 0:   return("Not indexed yet");
  case 200: return("OK");
  case 204: return("No Content");
  case 206: return("Partial OK");
  case 301: return("Moved Permanently");
  case 302: return("Moved Temporarily");
  case 303: return("See Other");
  case 304: return("Not Modified");
  case 300: return("Multiple Choices");
  case 305: return("Use Proxy (proxy redirect)");
  case 307: return("Temporary Redirect");
  case 400: return("Bad Request");
  case 401: return("Unauthorized");
  case 402: return("Payment Required");
  case 403: return("Forbidden");
  case 404: return("Not found");
  case 405: return("Method Not Allowed");
  case 406: return("Not Acceptable");
  case 407: return("Proxy Authentication Required");
  case 408: return("Request Timeout");
  case 409: return("Conflict");
  case 410: return("Gone");
  case 411: return("Length Required");
  case 412: return("Precondition Failed");
  case 413: return("Request Entity Too Large");
  case 414: return("Request-URI Too Long");
  case 415: return("Unsupported Media Type");
  case 500: return("Internal Server Error");
  case 501: return("Not Implemented");
  case 502: return("Bad Gateway");
  case 505: return("Protocol Version Not Supported");
  case 503: return("Service Unavailable");
  case 504: return("Gateway Timeout");
  case 509: return("Bandwidth Limit Exceeded");
  default:  return("Unknown status");
  }
}


/*******************************************************/
#ifdef USE_HTTP

static int
UdmHTTPGet(UDM_AGENT *Agent,UDM_DOCUMENT *Doc)
{
  int fd,res= 0;
  fd_set sfds;
  struct timeval tv;

  /* Connect to HTTP or PROXY server */
  if ((fd= open_host(Agent,Doc)) < 0)
   return fd;

  /* Send HTTP request */
  if (UdmSend(fd, UdmHTTPBufPtr(&Doc->Buf), UdmHTTPBufSize(&Doc->Buf), 0) < 0)
    return -1;

  UdmHTTPBufReset(&Doc->Buf);

  /* Retrieve response */
  tv.tv_sec= (long) Doc->Spider.read_timeout;
  tv.tv_usec= 0;

  while (1)
  {
    time_t start_time= time(NULL);
    int status;

    if (UdmHTTPBufAvailableSize(&Doc->Buf) < 2) /* +1 for '\0' */
      break;

    FD_ZERO(&sfds);
    FD_SET(fd, &sfds);

    status= select(FD_SETSIZE, &sfds, NULL, NULL, &tv);
    if (status == -1)
    {
      res= UDM_NET_ERROR;
      break;
    }
    if (status == 0)
    {
      res= UDM_NET_TIMEOUT;
      break;
    }
    if (!FD_ISSET(fd, &sfds))
      break;

    if (!(status= UdmHTTPBufAppendRecv(&Doc->Buf, fd, UDM_NET_BUF_SIZE, 0)))
    {
      /*
        - The peer has performed an orderly shutdown.
        - The buffer is full
        - Timeout
      */
      if (Doc->Spider.doc_timeout < time(NULL) - start_time)
        res= UDM_NET_TIMEOUT;
      break;
    }
    if (status < 0) /* An error occured */
    {
      res= status;
      break;
    }
    /* Read some bytes, continue. */
  }
  closesocket(fd);
  return res;
}
#endif

#ifdef USE_HTTPS

#define sslcleanup closesocket(fd); SSL_free (ssl); SSL_CTX_free (ctx)

#ifndef OPENSSL_VERSION_NUMBER
#define OPENSSL_VERSION_NUMBER 0x0000
#endif

static int UdmHTTPSGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc)
{
  int fd;
  int res= 0, status;
  SSL_CTX* ctx;
  SSL*     ssl=NULL;
  const SSL_METHOD *meth;
  time_t start_time;
  int buf_size= UDM_NET_BUF_SIZE;
  pid_t pid;

#if OPENSSL_VERSION_NUMBER >= 0x00905100
  while (RAND_status() != 1)
  {
#endif
    /* seed in the current time and process id
     * these are normally 4 bytes each, which should be enough
     * for our necessary 128-bit minimum seed for the PRNG */
    start_time= time(NULL);
    RAND_seed((unsigned char *)&start_time, sizeof(time_t));
    pid= getpid();
    RAND_seed((unsigned char *)&pid, sizeof(pid_t));
#if OPENSSL_VERSION_NUMBER >= 0x00905100
  }
#endif

  /* Connect to HTTPS server */
  if((fd= open_host(Indexer,Doc)) < 0)
     return(fd);

  SSLeay_add_ssl_algorithms();
  meth= SSLv23_client_method();
  SSL_load_error_strings();

  if ((ctx= SSL_CTX_new (meth))==NULL)
  {
    sslcleanup;
    return UDM_NET_ERROR;
  }

  /* -------------------------------------------------- */
  /* Now we have TCP connection. Start SSL negotiation. */

  if ((ssl= SSL_new(ctx)) == NULL)
  {
    sslcleanup;
    closesocket(fd);
    return UDM_NET_ERROR;
  }

  SSL_set_fd (ssl, fd);

  if (SSL_connect(ssl) < 0)
  {
    sslcleanup;
    return UDM_NET_ERROR;
  }

  /* Send HTTP request */
  if ((SSL_write(ssl, UdmHTTPBufPtr(&Doc->Buf), (int) UdmHTTPBufSize(&Doc->Buf))) < 0)
  {
    sslcleanup;
    return UDM_NET_ERROR;
  }

  /* Get result */
  UdmHTTPBufReset(&Doc->Buf);
  start_time= time(NULL);
  while(1)
  {
    if (UdmHTTPBufSize(&Doc->Buf) + buf_size > UdmHTTPBufAllocedSize(&Doc->Buf))
      buf_size= UdmHTTPBufAllocedSize(&Doc->Buf) - UdmHTTPBufSize(&Doc->Buf);
    else
      buf_size= UDM_NET_BUF_SIZE;
    
    status= SSL_read(ssl, Doc->Buf.buf + UdmHTTPBufSize(&Doc->Buf), buf_size);
    if( status < 0 )
    {
      res= status;
      break;
    }
    else if( status == 0 )
    {
      break;
    }
    else
    {
      Doc->Buf.size+= status;
      if(Doc->Spider.doc_timeout<(time(NULL)-start_time))
      {
        res= UDM_NET_TIMEOUT;
        break;
      }
      if (UdmHTTPBufSize(&Doc->Buf) == UdmHTTPBufAllocedSize(&Doc->Buf))
        break;
    }
  }
  
/*    nread= SSL_read(ssl, Doc->buf, (int)(Doc->Buf.maxsize - 1));*/
/*
    if(nread <= 0)
    {
      sslcleanup;
      return UDM_NET_ERROR;
    }
*/    
  SSL_shutdown (ssl);  /* send SSL/TLS close_notify */

/*  Doc->Buf.size=nread;*/
  Doc->Buf.buf[Doc->Buf.size]= '\0';

  /* Clean up. */
  sslcleanup;
  return res;
}

#endif

#ifdef USE_FTP
static int UdmFTPGet(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc)
{
  int  last_mod_tm, code, res=0;
  char *full_path=NULL;
  size_t  len;
  time_t  last_mod_time=UdmHttpDate2Time_t(UdmVarListFindStr(&Doc->Sections,"Last-Modified",""));
  
  UdmHTTPBufReset(&Doc->Buf);
  
  /* Check for changing host */
  if (!Doc->connp.hostname || 
      strcmp(Doc->connp.hostname, UDM_NULL2EMPTY(Doc->CurURL.hostname)) || 
      (Doc->connp.connected == UDM_NET_NOTCONNECTED))
  {
    char *authstr= NULL;
    char * user=NULL;
    char * pass=NULL;
    
    if (Doc->CurURL.auth != NULL)
    {
      authstr= (char*)UdmStrdup(Doc->CurURL.auth);
      user=authstr;
      if((pass=strchr(authstr,':')))
      {
        *pass='\0';
        pass++;
      }
    }
    
    code=Udm_ftp_connect(Indexer, &Doc->connp,
                         UDM_NULL2EMPTY(Doc->CurURL.hostname),
                         Doc->CurURL.port ? Doc->CurURL.port : Doc->CurURL.default_port,
                         user, pass,  Doc->Spider.read_timeout);
    
    if (code == -1)
    {
      if (Doc->connp.err >0)
      {
        UdmHTTPBufPrintf(&Doc->Buf, "HTTP/1.1 401 OK\r\n\r\n  ");
      }
      else
      {
        res= Doc->connp.err;
      }
    }
    UDM_FREE(authstr);
  }
  
  if (Doc->connp.connected == UDM_NET_CONNECTED)
  {
    /* Make listing */
    if (Doc->CurURL.filename == NULL)
    {
      code= Udm_ftp_cwd(&Doc->connp, UDM_NULL2EMPTY(Doc->CurURL.path));
      if (code != -1)
      {
        code= Udm_ftp_list(Doc);
        if (code == -1)
        {
          if (Doc->connp.err >0)
          {
            UdmHTTPBufPrintf(&Doc->Buf, "HTTP/1.1 403 OK\r\n\r\n");
          }
          else
          {
            res= Doc->connp.err;
          }
        }
        else
        {
          UdmHTTPBufPrintf(&Doc->Buf,
                           "HTTP/1.1 200 OK\r\n"
                           "Content-Type: text/html\r\n\r\n"
                           "<html><body>%s</body></html>", 
                           Doc->connp.connp->buf);
        }
      }
      else
      {
        if (Doc->connp.err >0)
        {
          UdmHTTPBufPrintf(&Doc->Buf, "HTTP/1.1 403 OK\r\n\r\n");
        }
        else
        {
          res= Doc->connp.err;
        }
      }
    }
    else
    {
      len= strlen(UDM_NULL2EMPTY(Doc->CurURL.path)) + strlen(UDM_NULL2EMPTY(Doc->CurURL.filename));
      full_path= (char*) UdmMalloc(len+1);
      udm_snprintf(full_path, len + 1, "%s%s", UDM_NULL2EMPTY(Doc->CurURL.path), UDM_NULL2EMPTY(Doc->CurURL.filename));
      full_path[len]='\0';
      last_mod_tm= Udm_ftp_mdtm(&Doc->connp,full_path);
      if(last_mod_tm == -1 && Doc->connp.err)
      {
        if (Doc->connp.err >0)
        {
          UdmHTTPBufPrintf(&Doc->Buf, "HTTP/1.1 404 OK\r\n\r\n");
        }
        else
          res= Doc->connp.err;

      }
      else if (last_mod_tm == last_mod_time)
      {
        UdmHTTPBufPrintf(&Doc->Buf, "HTTP/1.1 304 OK\r\n\r\n");
      }
      else
      {
        char dbuf[UDM_MAXTIMESTRLEN];
        UdmTime_t2HttpStr(last_mod_tm, dbuf, sizeof(dbuf));
        if (Doc->method!=UDM_METHOD_HEAD)
        {
          int s, f= -1;
          size_t fsize= UdmHTTPBufAllocedSize(&Doc->Buf);
          sscanf(UdmVarListFindStr(&Doc->RequestHeaders, "Range", "bytes=0-0"), "bytes=%d-%d", &s, &f);
          if (f != 0)
          {
            if (s < 0)
            {
              s += Udm_ftp_size(&Doc->connp, full_path);
            }
            else if (f > 0)
            {
              fsize= f - s;
            }
            if (s > 0)
              Udm_ftp_rest(&Doc->connp, (size_t)s);
          }
          if (!Udm_ftp_get(&Doc->connp, Doc->connp.connp, full_path, fsize))
          {
            UdmHTTPBufPrintf(&Doc->Buf,
                             "HTTP/1.1 20%c OK\r\nLast-Modified: %s\r\n\r\n", 
                              (Doc->connp.connp->err == UDM_NET_FILE_TL) ? '6' : '0',
                              dbuf);
            UdmHTTPBufAppend(&Doc->Buf, Doc->connp.connp->buf, Doc->connp.connp->buf_len);
          }
          else
          {
            if (Doc->connp.err > 0)
            {
              UdmHTTPBufPrintf(&Doc->Buf, "HTTP/1.1 403 OK\r\n\r\n");
            }
            else
            {
              res= Doc->connp.err;
            }
          }
        }
        else
        {
          size_t size= Udm_ftp_size(&Doc->connp, full_path);
          UdmHTTPBufPrintf(&Doc->Buf,
                           "HTTP/1.1 200 OK\r\n"
                           "Content-Type: text/html\r\n"
                           "Content-Length: %d\r\n"
                           "Last-Modified: %s\r\n\r\n", size, dbuf);
        }
      }
    }
    Udm_ftp_close(&Doc->connp);
  }
  
  UDM_FREE(full_path);
  UDM_FREE(Doc->connp.buf);
  if (Doc->connp.connp && Doc->connp.connp->buf)
  {
    UDM_FREE(Doc->connp.connp->buf);
    Doc->connp.connp->buf=NULL;
  }
  return res;
}
#endif  

#ifdef USE_NEWS
static ssize_t fdgets(char * str,size_t size,int fd){
  size_t nbytes= 0;  /* Bytes read from fd and stored to str */
  int done=0;    /* Finish flag */

  while(!done)
  {
    if(nbytes+1>size)
    {
      break;
    }
    if(!recv(fd,str+nbytes,1,0))
    {
      break;
    }
    else
    {
      if(str[nbytes]=='\n')
      {
        done=1;
      }
      nbytes++;
    }
  }
  if(!nbytes)
  {
    return 0;
  }
  else
  {
    str[nbytes]='\0';
    return nbytes;
  }
}


/* #define DEBUG_NNTP */

static ssize_t
NNTPRecv(char * str,size_t sz, int fd)
{
  ssize_t res;
  res=fdgets(str,sz,fd);
#ifdef DEBUG_NNTP
  {
    char * s, ch;
    for(s=str; *s && *s!='\r' && *s!='\n' && (s-str<sz) ; s++);
    ch=*s;*s='\0';
    fprintf(stderr,"NNTPGet : NNTPRecv: '%s'\n",str);
    *s=ch;
  }
#endif
  return res;
}

static ssize_t
NNTPSend(char * str,size_t sz, int fd)
{
  ssize_t res;
  res= UdmSend(fd, str, sz, 0);
#ifdef DEBUG_NNTP
  {
    char * s, ch;
    for(s=str; *s && *s!='\r' && *s!='\n' && (s-str<sz) ; s++);
    ch=*s;*s='\0';
    fprintf(stderr,"NNTPGet : NNTPSend: '%s'\n",str);
    *s=ch;
  }
#endif
  return res;
}


#define UDM_NNTP_SEND    10
#define UDM_NNTP_RECV    11
#define UDM_NNTP_RECVCMD  12

#define UDM_NNTP_CONNECT  21
#define UDM_NNTP_HELLO    22
#define UDM_NNTP_READER    23
#define UDM_NNTP_BREAK    24
#define UDM_NNTP_QUIT    25

#define UDM_NNTP_USER    31
#define UDM_NNTP_USERRES  32
#define UDM_NNTP_PASS    33
#define UDM_NNTP_PASSRES  34

#define UDM_NNTP_LIST    41
#define UDM_NNTP_LISTRES  42
#define UDM_NNTP_LISTDAT  43

#define UDM_NNTP_STAT    51
#define UDM_NNTP_STATRES  52

#define UDM_NNTP_GROUP    61
#define UDM_NNTP_GROUPRES  62

#define UDM_NNTP_XOVER    71
#define UDM_NNTP_XOVERRES  72
#define UDM_NNTP_XOVERDAT  73

#define UDM_NNTP_ARTICLE  81
#define UDM_NNTP_ARTICLERES  82
#define UDM_NNTP_ARTICLEDAT  83



#define UDM_TERM(x)    (x)[sizeof(x)-1]='\0'

#define UDM_NNTP_MAXCMD    100

#ifdef DEBUG_NNTP
static const char * nntpcmd(int cmd)
{
  switch(cmd)
  {
    case UDM_NNTP_CONNECT:    return("connect");
    case UDM_NNTP_QUIT:    return("quit");
    case UDM_NNTP_BREAK:    return("break");
    case UDM_NNTP_HELLO:    return("hello");
    case UDM_NNTP_READER:    return("reader");
    case UDM_NNTP_RECV:    return("recv");
    case UDM_NNTP_RECVCMD:    return("recvcmd");
    case UDM_NNTP_SEND:    return("send");
    
    case UDM_NNTP_USER:    return("user");
    case UDM_NNTP_USERRES:    return("userres");
    
    case UDM_NNTP_PASS:    return("pass");
    case UDM_NNTP_PASSRES:    return("passres");
    
    case UDM_NNTP_LIST:    return("list");
    case UDM_NNTP_LISTRES:    return("listres");
    case UDM_NNTP_LISTDAT:    return("listdat");
    
    case UDM_NNTP_GROUP:    return("group");
    case UDM_NNTP_GROUPRES:    return("groupres");
    
    case UDM_NNTP_XOVER:    return("xover");
    case UDM_NNTP_XOVERRES:    return("xoverres");
    case UDM_NNTP_XOVERDAT:    return("xoverdat");
  
    case UDM_NNTP_STAT:    return("stat");
    case UDM_NNTP_STATRES:    return("statres");
  
    case UDM_NNTP_ARTICLE:    return("article");
    case UDM_NNTP_ARTICLERES:  return("articleres");
    case UDM_NNTP_ARTICLEDAT:  return("articledat");
  }
  return("UNKNOWN");
}
#endif
static void inscmd(int * commands,size_t * n,int command)
{
  if(((*n)+1)<UDM_NNTP_MAXCMD)
  {
    commands[*n]=command;
    (*n)++;
  }
}
static int delcmd(int * cmd,size_t * n)
{
  if(*n>0)
  {
    (*n)--;
    return cmd[*n];
  }
  else
  {
    return UDM_NNTP_BREAK;
  }
}


typedef struct
{
  time_t if_modified_since;
  char user[256];
  char pass[256];
} UDM_REQUEST_INFO;


static void
UdmRequestInfoInit(UDM_REQUEST_INFO *Request)
{
  Request->if_modified_since= 0;
  Request->user[0]= '\0';
  Request->pass[0]= '\0';
}


/* Parse request headers */
static void
UdmParseRequest(UDM_REQUEST_INFO *Request, const char *src, size_t srclen)
{
  char *tok, *lt, *tmp;
  if (!(tmp= udm_strndup(src, srclen)))
    return;

  for (tok= udm_strtok_r(tmp, "\r\n", &lt);
       tok;
       tok= udm_strtok_r(NULL,"\r\n",&lt))
  {
#ifdef DEBUG_NNTP
    fprintf(stderr,"HEADER: '%s'\n",tok);
#endif
    if (!strncasecmp(tok, UDM_CSTR_WITH_LEN("If-Modified-Since: ")))
    {
      Request->if_modified_since= UdmHttpDate2Time_t(tok + 19);
    }
    else if(!strncasecmp(tok, UDM_CSTR_WITH_LEN("Authorization: ")))
    {
      char *auth= UdmTrim(tok + 15," \t\r\n");
      char str[1024];
      strncpy(str, auth, sizeof(str) - 1);
      UDM_TERM(str);
      udm_base64_decode(Request->user, str, sizeof(Request->user) - 1);
      if ((auth= strchr(Request->user, ':')))
      {
        *auth++= '\0';
        strncpy(Request->pass, auth, sizeof(Request->pass));
        UDM_TERM(Request->pass);
      }
    }    
  }
  UdmFree(tmp);
}


static int UdmNNTPGet(UDM_AGENT * Indexer,UDM_DOCUMENT *Doc)
{
  UDM_REQUEST_INFO Request;
  char str[2048]="";
  char grp[256]="";
  char msg[256]="";
  char *tok,*lt;
  int  cmd[UDM_NNTP_MAXCMD];
  int  fd=0;
  int  status=0;
  int  lastcmd=0;
  int  a=0;
  int  b=0;
  int  from=0;
  int  to=0;
  int  headers=1;
  int  has_content=0;
  size_t ncmd=0;

  UdmRequestInfoInit(&Request);

  memset(cmd,0,sizeof(cmd));

  /* We'll send user/password given in URL first         */
  /* Only after that we'll use them from "header" field  */
  if (Doc->CurURL.auth != NULL)
  {
    strncpy(Request.user, Doc->CurURL.auth, sizeof(Request.user) - 1);
    UDM_TERM(Request.user);
    if ((tok= strchr(Request.user,':')))
    {
      *tok++= '\0';
      strncpy(Request.pass, tok, sizeof(Request.pass)-1);
      UDM_TERM(Request.pass);
    }
  }

  inscmd(cmd,&ncmd,UDM_NNTP_QUIT);

  /* Parse request headers */
  UdmParseRequest(&Request, UdmHTTPBufPtr(&Doc->Buf), UdmHTTPBufSize(&Doc->Buf));

  /* Find in which "url" field group and message are */
  /* It depends on protocol being used               */

  /* First bad case, make redirect */
  if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "news") &&
      strcmp(UDM_NULL2EMPTY(Doc->CurURL.path), "/"))
  {
    /* news://news.server.com/group/ */
    /* In news:// group should not   */
    /* have trailing slash !!!!!!!!  */
    /* Let's    redirect to the same */
    /* URL without slashes           */
    char *s;
    strncpy(grp,Doc->CurURL.path,sizeof(grp)-1);
    UDM_TERM(grp);
    /* Remove trailing slash */
    if((s=strrchr(grp+1,'/')))*s='\0';
    status=301;
    udm_snprintf(str, sizeof(str)-1, "%s://%s%s", Doc->CurURL.schema,Doc->CurURL.hostinfo,grp);
    UDM_TERM(str);
    UdmHTTPBufPrintf(&Doc->Buf,
                     "HTTP/1.0 %d %s\r\nLocation: %s\r\n\r\n",
                     status, UdmHTTPErrMsg(status), str);
    return 0;
  }
  
  
  /* Second bad case, make redirect */
  if(!strcasecmp(Doc->CurURL.schema,"nntp") &&
     !strcmp(Doc->CurURL.path,"/") &&
     Doc->CurURL.filename != NULL)
  {
    /* nntp://news.server.com/something        */
    /* This is wrong URL, assume that          */
    /* "something" is a group name, redirect   */
    /* to the same URL but with trailing slash */
    status=301;
    udm_snprintf(str,sizeof(str)-1,
                 "%s://%s/%s/",
                 Doc->CurURL.schema,
                 UDM_NULL2EMPTY(Doc->CurURL.hostinfo),
                 UDM_NULL2EMPTY(Doc->CurURL.filename));
    UDM_TERM(str);
    UdmHTTPBufPrintf(&Doc->Buf,
                     "HTTP/1.0 %d %s\r\nLocation: %s\r\n\r\n",
                     status, UdmHTTPErrMsg(status),str);
    return 0;
  }


  if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "news") &&
     (Doc->CurURL.filename != NULL))
  {
    /* url->path here is always "/"     */
    /* So there are two possible cases: */
    /* news://news.server.com/group     */
    /* news://news.server.com/messg     */
    /* url->filename contains either    */
    /* message ID  or group name        */
    /* and doesn't have any slashes     */
    /* Check whether @ does present     */
    if(strchr(Doc->CurURL.filename,'@'))
    {
      /* We have a message */
      strncpy(msg,Doc->CurURL.filename,sizeof(grp)-1);
      UDM_TERM(msg);
      if (Request.if_modified_since)
      {
        inscmd(cmd,&ncmd,UDM_NNTP_STAT);
      }
      else
      {
        inscmd(cmd,&ncmd,UDM_NNTP_ARTICLE);
      }
    }
    else
    {
      /* We have a news group */
      strncpy(grp,Doc->CurURL.filename,sizeof(grp)-1);
      UDM_TERM(grp);
      inscmd(cmd,&ncmd,UDM_NNTP_XOVER);
      inscmd(cmd,&ncmd,UDM_NNTP_GROUP);
    }
  }
  else if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "nntp") &&
          strcmp(UDM_NULL2EMPTY(Doc->CurURL.path), "/"))
  {
    /* There are two possible cases:         */
    /*   nntp://news.server.com/group/       */
    /*   nntp://news.server.com/group/msg    */
    /* Group is in url->path and contains    */
    /* both leading and trailing slashes     */ 
    char * s;
    
    /* Copy without leading slash */
    strncpy(grp,Doc->CurURL.path+1,sizeof(grp)-1);
    UDM_TERM(grp);
    
    /* Remove trailing slash */
    if((s=strchr(grp,'/')))*s='\0';
    if(Doc->CurURL.filename != NULL)
    {
      strncpy(msg,Doc->CurURL.filename,sizeof(msg)-1);
      UDM_TERM(msg);
      if (Request.if_modified_since)
      {
        inscmd(cmd,&ncmd,UDM_NNTP_STAT);
        inscmd(cmd,&ncmd,UDM_NNTP_GROUP);
      }
      else
      {
        inscmd(cmd,&ncmd,UDM_NNTP_ARTICLE);
        inscmd(cmd,&ncmd,UDM_NNTP_GROUP);
      }
    }
    else
    {
      inscmd(cmd,&ncmd,UDM_NNTP_XOVER);
      inscmd(cmd,&ncmd,UDM_NNTP_GROUP);
    }
  }
  else
  {
    inscmd(cmd,&ncmd,UDM_NNTP_LIST);
  }
  inscmd(cmd,&ncmd,UDM_NNTP_READER);
  inscmd(cmd,&ncmd,UDM_NNTP_CONNECT);


#ifdef DEBUG_NNTP
  fprintf(stderr,"NNTPGet : Enter with proto='%s' host='%s' port=%d path='%s' filename='%s'\n",url->schema,url->hostname,url->port?url->port:url->default_port,url->path,url->filename);
  fprintf(stderr,"NNTPGet : Assume grp='%s' msg='%s'\n",grp,msg);
#endif


  while(ncmd)
  {
#ifdef DEBUG_NNTP
    fprintf(stderr,"NCMD=%d\n",ncmd);
#endif
    lastcmd=delcmd(cmd,&ncmd);
#ifdef DEBUG_NNTP
    fprintf(stderr,"CMD=%s\n",nntpcmd(lastcmd));
#endif
    switch(lastcmd){
    case UDM_NNTP_CONNECT:
      /* Connect to NNTP server */
      fd=open_host(Indexer,Doc);
      if(fd<0)return(fd);    /* Could not connect */
      inscmd(cmd,&ncmd,UDM_NNTP_HELLO);
      inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
      break;
      
    case UDM_NNTP_RECV:
      /*nbytes=*/ NNTPRecv(str,sizeof(str),fd);
      break;
      
    case UDM_NNTP_RECVCMD:
      /*nbytes=*/ NNTPRecv(str,sizeof(str),fd);
      status=atoi(str);
      break;
      
    case UDM_NNTP_SEND:
      /*nbytes=*/NNTPSend(str,strlen(str),fd);
      break;
      
    case UDM_NNTP_READER:
      inscmd(cmd,&ncmd,UDM_NNTP_HELLO);
      inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
      inscmd(cmd,&ncmd,UDM_NNTP_SEND);
      sprintf(str,"mode reader\r\n");
      break;
      
    case UDM_NNTP_HELLO:
      if(status!=200)
      {
        status=500;
        UdmHTTPBufPrintf(&Doc->Buf,
                         "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                         status, UdmHTTPErrMsg(status), str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
      }
      break;
      
    case UDM_NNTP_USER:
      if (Request.user[0])
      {
        inscmd(cmd,&ncmd,UDM_NNTP_USERRES);
        inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
        inscmd(cmd,&ncmd,UDM_NNTP_SEND);
        sprintf(str,"authinfo user %s\r\n", Request.user);
      }
      else
      {
        status=401;
        UdmHTTPBufPrintf(&Doc->Buf,
                         "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                         status, UdmHTTPErrMsg(status), str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
      }
      break;
      
    case UDM_NNTP_USERRES:
      if(status==381)
      {
        inscmd(cmd,&ncmd,UDM_NNTP_PASS);
      }
      else
      {
        status=500;
        UdmHTTPBufPrintf(&Doc->Buf,
                         "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                         status, UdmHTTPErrMsg(status), str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
      }
      break;
      
    case UDM_NNTP_PASS:
      if(Request.pass[0])
      {
        inscmd(cmd,&ncmd,UDM_NNTP_PASSRES);
        inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
        inscmd(cmd,&ncmd,UDM_NNTP_SEND);
        sprintf(str, "authinfo pass %s\r\n", Request.pass);
      }
      else
      {
        status=401;
        UdmHTTPBufPrintf(&Doc->Buf,
                         "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                         status, UdmHTTPErrMsg(status), str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
      }
      break;
      
    case UDM_NNTP_PASSRES:
      if((status!=200) && (status!=281))
      {
        if(status==502)status=401; /* Unauthorzed */
        else  status=500;
        UdmHTTPBufPrintf(&Doc->Buf,
                         "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                         status, UdmHTTPErrMsg(status), str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
        break;
      }
      break;
      
    case UDM_NNTP_LIST:
      inscmd(cmd,&ncmd,UDM_NNTP_LISTRES);
      inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
      inscmd(cmd,&ncmd,UDM_NNTP_SEND);
      sprintf(str,"list\r\n");
      break;
      
    case UDM_NNTP_LISTRES:
      if(status==480){
        inscmd(cmd,&ncmd,UDM_NNTP_LIST);
        inscmd(cmd,&ncmd,UDM_NNTP_USER);
        break;
      }
      if(status!=215){
        status=500;
        UdmHTTPBufPrintf(&Doc->Buf,
                          "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                          status, UdmHTTPErrMsg(status), str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
        break;
      }
      UdmHTTPBufPrintf(&Doc->Buf,
                       "HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n<HTML><BODY>\n");
      inscmd(cmd,&ncmd,UDM_NNTP_LISTDAT);
      inscmd(cmd,&ncmd,UDM_NNTP_RECV);
      break;
      
    case UDM_NNTP_LISTDAT:
      if(str[0]=='.')
      {
        UdmHTTPBufAppendf(&Doc->Buf, "</BODY></HTML>\n");
        break;
      }
      if((tok=strchr(str,' ')))*tok=0;
      if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "nntp"))
        UdmHTTPBufAppendf(&Doc->Buf, "<A HREF=\"/%s/\"></A>\n", str);
      else
        UdmHTTPBufAppendf(&Doc->Buf, "<A HREF=\"/%s\"></A>\n", str);
      inscmd(cmd,&ncmd,UDM_NNTP_LISTDAT);
      inscmd(cmd,&ncmd,UDM_NNTP_RECV);
      break;
      
    case UDM_NNTP_XOVER:
      /* List of the articles in one news group */
      inscmd(cmd,&ncmd,UDM_NNTP_XOVERRES);
      inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
      inscmd(cmd,&ncmd,UDM_NNTP_SEND);
      sprintf(str,"xover %d-%d\r\n",from,to);
      break;
      
    case UDM_NNTP_XOVERRES:
      if(status==224)
      {/* 224 data follows */
        UdmHTTPBufPrintf(&Doc->Buf,
                         "HTTP/1.0 200 OK\r\nContent-Type: text/html\r\n\r\n"
                         "<HTML><BODY>\n");
        inscmd(cmd,&ncmd,UDM_NNTP_XOVERDAT);
        inscmd(cmd,&ncmd,UDM_NNTP_RECV);
      }
      else
      {
        UdmHTTPBufPrintf(&Doc->Buf,
                         "HTTP/1.0 404 Not found\r\n"
                         "NNTP-Server-Response: %s\r\n\r\n",
                         str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
      }
      break;
      
    case UDM_NNTP_XOVERDAT:
      if(str[0]=='.')
      {
        UdmHTTPBufAppendf(&Doc->Buf, "</BODY></HTML>\n");
        break;
      }
      else if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "nntp"))
      {
        char *e;
        if((e= udm_strtok_r(str,"\t\r\n",&lt)))
          UdmHTTPBufAppendf(&Doc->Buf,
                            "<A HREF=\"%s%s\"></A>\n",
                            UDM_NULL2EMPTY(Doc->CurURL.path), e);
        inscmd(cmd,&ncmd,UDM_NNTP_XOVERDAT);
        inscmd(cmd,&ncmd,UDM_NNTP_RECV);
      }
      else
      {
        int n=0;
        char *field[10];
        memset(field, 0, sizeof(field));
        
        tok= udm_strtok_r(str,"\t\r\n",&lt);
        while(tok)
        {
#ifdef DEBUG_NNTP
          fprintf(stderr,"NNTPGet : field[%d]: '%s'\n",n,tok);
#endif
          if(n<10)field[n]=tok;
          tok= udm_strtok_r(NULL,"\t\r\n",&lt);
          n++;
        }
        if(field[4])
        {
          char * msgb=field[4];
          char * msge;
          if(*msgb=='<')msgb++;
          for(msge=msgb ; *msge!='\0' && *msge!='>' ; msge++);
          if(*msge=='>')*msge='\0';
#ifdef DEBUG_NNTP
          fprintf(stderr,"NNTPGet : messg_id: '%s'\n",msgb);
#endif
          if (UdmHTTPBufSize(&Doc->Buf) + strlen(msgb) + 50 <
              UdmHTTPBufAllocedSize(&Doc->Buf))
          {
            UdmHTTPBufAppendf(&Doc->Buf, "<A HREF=\"/%s\"></A>\n", msgb);
          }
        }
        inscmd(cmd,&ncmd,UDM_NNTP_XOVERDAT);
        inscmd(cmd,&ncmd,UDM_NNTP_RECV);
      }
      break;
      
    case UDM_NNTP_GROUP:
      inscmd(cmd,&ncmd,UDM_NNTP_GROUPRES);
      inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
      inscmd(cmd,&ncmd,UDM_NNTP_SEND);
      sprintf(str, "group %s\r\n",grp);
      break;
      
    case UDM_NNTP_GROUPRES:
      if(status==480)
      { /* Authorization required */
        inscmd(cmd,&ncmd,UDM_NNTP_GROUP);
        inscmd(cmd,&ncmd,UDM_NNTP_USER);
      }
      else if(status==211)
      { /* Group selected */
        sscanf(str,"%d%d%d%d\n",&a,&b,&from,&to);
      }
      else
      {
        status=404;
        UdmHTTPBufPrintf(&Doc->Buf,
                          "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                          status, UdmHTTPErrMsg(status), str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
        break;
      }
      break;
      
    case UDM_NNTP_STAT:
      inscmd(cmd,&ncmd,UDM_NNTP_STATRES);
      inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
      inscmd(cmd,&ncmd,UDM_NNTP_SEND);
      if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "nntp"))
      {
        udm_snprintf(str,sizeof(str)-1,"stat %s\r\n",msg);
      }
      else
      {
        udm_snprintf(str,sizeof(str)-1,"stat <%s>\r\n",msg);
      }
      break;
      
    case UDM_NNTP_STATRES:
      if(status==480)
      { /* Authorization required */
        inscmd(cmd,&ncmd,UDM_NNTP_STAT);
        inscmd(cmd,&ncmd,UDM_NNTP_USER);
      }
      else if(status==223)
      {  /* Article exists             */
        status=304;  /* return 304 Not modified    */
      }
      else if(status==423)
      {  /* No such article            */
        status=404;  /* return 404 Not found       */
      }
      else
      {      /* Unknown status code        */
        status=500;  /* Return 505 Server error    */
      }
      UdmHTTPBufPrintf(&Doc->Buf,
                       "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                       status, UdmHTTPErrMsg(status), str);
      inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
      break;
      
    case UDM_NNTP_ARTICLE:
      inscmd(cmd,&ncmd,UDM_NNTP_ARTICLERES);
      inscmd(cmd,&ncmd,UDM_NNTP_RECVCMD);
      inscmd(cmd,&ncmd,UDM_NNTP_SEND);
      if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "nntp"))
      {
        udm_snprintf(str,sizeof(str),"article %s\r\n",msg);
      }
      else
      {
        udm_snprintf(str,sizeof(str),"article <%s>\r\n",msg);
      }
      break;
      
    case UDM_NNTP_ARTICLERES:
      if(status==480)
      { /* Authorization required */
        inscmd(cmd,&ncmd,UDM_NNTP_ARTICLE);
        inscmd(cmd,&ncmd,UDM_NNTP_USER);
      }
      else if(status==220){ /* Article follows */
        UdmHTTPBufPrintf(&Doc->Buf, "HTTP/1.0 200 OK\r\n");
        inscmd(cmd,&ncmd,UDM_NNTP_ARTICLEDAT);
        inscmd(cmd,&ncmd,UDM_NNTP_RECV);
      }
      else
      {
        status=404;
        UdmHTTPBufPrintf(&Doc->Buf,
                         "HTTP/1.0 %d %s\r\nNNTP-Response: %s\r\n\r\n",
                         status, UdmHTTPErrMsg(status), str);
        inscmd(cmd,&ncmd,UDM_NNTP_QUIT);
      }
      break;
      
    case UDM_NNTP_ARTICLEDAT:
      if(str[0]=='.')break;
      if(headers)
      {
        if(!strncasecmp(str,"Content-Type:",13))has_content=1;
        if((!strcmp(str,"\r\n"))||(!strcmp(str,"\n")))
        {
          headers=0;
          if(!has_content)
            UdmHTTPBufAppendf(&Doc->Buf, "Content-Type: text/plain\r\n");
        }
      }
      if(strlen(str) + 1 < UdmHTTPBufAvailableSize(&Doc->Buf))
        UdmHTTPBufAppendf(&Doc->Buf, "%s", str);
      inscmd(cmd,&ncmd,UDM_NNTP_ARTICLEDAT);
      inscmd(cmd,&ncmd,UDM_NNTP_RECV);
      break;
      
    case UDM_NNTP_QUIT:
      if(fd>0){
        sprintf(str,"quit\r\n");
        NNTPSend(str,strlen(str),fd);
        closesocket(fd);
        fd=0;
      }
      inscmd(cmd,&ncmd,UDM_NNTP_BREAK);
      break;
    
    case UDM_NNTP_BREAK:  
    default:
      ncmd=0;
      break;
    }
  }
  if(fd>0){
    /* This never should happen */
    /* Error in state machine   */
    closesocket(fd);
  }
  return 0;
}

#endif


#ifdef USE_FILE
static int
cmpdirent (char **a1, char **a2)
{
  return(strcmp(*a1, *a2));
}

static void
freedir(char **dir)
{
  char **p;

  if (! dir) return;
  for (p= dir; *p; p++) UdmFree(*p);
  UdmFree(dir);
}

static char**
loaddir(const char *path, int sort)
{
  struct dirent *buf;
  char **_, **tmp;
  size_t nrecs= 0, arecs= 32;
  DIR *dir;

  if (! path) return(NULL);

  dir= opendir(path);
  if (! dir) return(NULL);
  _= (char**) UdmMalloc(arecs * sizeof(char *));
  if (! _)
  {
    closedir(dir);
    return(NULL);
  }

  while ((buf= readdir(dir)))
  {
    if (arecs <= nrecs + 1)
    {
      arecs += 32;
      tmp= (char**) UdmRealloc(_, arecs * sizeof(char *));
      if (! tmp)
      {
        closedir(dir);
        _[nrecs]= NULL;
        freedir(_);
        return(NULL);
      }
      _= tmp;
    }
    _[nrecs++]= UdmStrdup(buf->d_name);
  }

  closedir(dir);

  if (sort) UdmSort(_, nrecs, sizeof(char *), (udm_qsort_cmp)cmpdirent);
  _[nrecs]= NULL;
  return(_);
}



static size_t
request_scan_filename(const char *src, size_t srclen, char *dst, size_t dstlen)
{
  const char *srcend= src + srclen;
  for ( ; src < srcend && *src != ' '; src++);

  if (src == srcend)
    return 0;

  if (*src == ' ')
  { 
    const char *e;
    src++;
    for (e= src; e < srcend && *e != ' '; e++);
    if (e < srcend && *e == ' ' && e - src < dstlen)
    {
      memcpy(dst, src, e - src);
      dst[e - src]= 0;
      return e - src;
    }
  }
  return 0;
}


static int
UdmFILEGet(UDM_AGENT *Indexer,UDM_DOCUMENT *Doc)
{
  int fd, status=0, is_dir;
  struct stat sb,sb1;
  char filename[4096]= "";
  char newfilename[4096];
  char openname[4096];
  char mystatname[4096];
  const char *hostname;
  int  cvs_ignore= Indexer->Conf->CVS_ignore;
  UDM_REQUEST_INFO Request;

  hostname= UdmVarListFindStr(&Doc->RequestHeaders,"Host",NULL);
  if (hostname && (hostname[0] == '\0' || !strcasecmp(hostname, "localhost")))
    hostname= NULL;
  
  if (!request_scan_filename(UdmHTTPBufPtr(&Doc->Buf), UdmHTTPBufSize(&Doc->Buf),
                             filename, sizeof(filename)))
  {
    UdmHTTPBufPrintf(&Doc->Buf,
                     "HTTP/1.0 500 %s\r\nX-Reason: bad file\r\n\r\n",
                     UdmHTTPErrMsg(500));
    return 0;
  }

  udm_snprintf(newfilename, sizeof(newfilename),
               "%s%s%s",
               hostname ? "//" : "",
               hostname ? hostname : "",
               filename);
  UdmUnescapeCGIQuery(openname,newfilename);

#ifdef WIN32
  {
    /* Remove leading "/" for windows path */
    char *s= openname;
    if (*s == '/' && (s[2] == ':' || s[2] == '\\'))
    {
      memmove(s, s + 1, strlen(s));
    }
  }
#endif

  /* Remember If-Modified-Since timestamp */
  UdmRequestInfoInit(&Request);
  UdmParseRequest(&Request, UdmHTTPBufPtr(&Doc->Buf), UdmHTTPBufSize(&Doc->Buf));

  strcpy(mystatname,openname);

#ifdef WIN32
  /*
    Remove traling slash sign stat("c:\a\") doesn't work
    under win. We'll pass stat("c:\a") instead
    In other hand when only drive name is checked it must
    have traling slash:  stat("c:\");
    For a network resouce we should execute WITH trailing
    slash, i.e. stat("\\servername\resoucename\");
  */
  {
    int len;    
    char *s= mystatname;
    len=strlen(s);
    while(*s)
    {
      if(*s=='/')*s='\\';
      s++;
    }
    if ((len>3) && (mystatname[len-1]=='\\') && !hostname)
      mystatname[len-1]='\0';
  }
#endif

  
  if(stat(mystatname,&sb))
  {
    switch(errno)
    {
      case ENOENT: status=404;break; /* Not found */
      case EACCES: status=403;break; /* Forbidden*/
      default: status=500;
    }
    
    UdmHTTPBufPrintf(&Doc->Buf,"HTTP/1.0 %d %s\r\n"
                               "X-Reason1: %d:%s\r\n"
                               "X-Statname: '%s'\r\n"
                               "\r\n",
                               status,UdmHTTPErrMsg(status),errno,
                               strerror(errno), mystatname);
    return 0;
  }


  /* If directory is given without ending "/"   */
  /* we must redirect to the same URL with "/"  */
  if((sb.st_mode&S_IFDIR)&&(filename[strlen(filename)-1]!='/'))
  {
    status=301;
    UdmHTTPBufPrintf(&Doc->Buf,
                     "HTTP/1.0 %d %s\r\nLocation: file:%s/\r\n\r\n",
                     status, UdmHTTPErrMsg(status), filename);
    return 0;
  }

  
  if(sb.st_mode&S_IFDIR)
  {
    char **recs= loaddir(openname, 1);
    size_t rpos;

    if (recs)
    {
      UdmHTTPBufPrintf(&Doc->Buf,
                       "HTTP/1.0 200 OK\r\nContent-type: text/html; generator=dirlist\r\n\r\n"
                       "<HTML><BODY>\n");
      
      for (rpos= 0; recs[rpos]; rpos++)
      {
        char escaped_name[1024]="";
        char *src, *e;

        /*
        This does not work on Solaris
        is_dir=(rec->d_type==DT_DIR);
        */
        if (cvs_ignore && !strcmp(recs[rpos],"CVS"))
          continue;
        
        sprintf(newfilename, "%s%s", openname, recs[rpos]);

#ifndef WIN32
        if (! lstat(newfilename, &sb1) &&
          S_ISLNK(sb1.st_mode) &&
          ! UdmVarListFindInt(&Doc->Sections, "FollowSymLinks", 0))
        {
          continue;
        }
#endif
        if(stat(newfilename,&sb1))
        {
          UdmLog(Indexer, UDM_LOG_EXTRA, "Can't stat '%s'", newfilename);
          is_dir= 0;
        }
        else
        {
          is_dir= ((sb1.st_mode&S_IFDIR) > 0);
        }

        e=escaped_name;
        for(src=recs[rpos];*src;src++)
        {
          if(strchr(" %&<>+[](){}/?#'\"\\;,`",*src))
          {
            sprintf(e,"%%%X",(int)*src);
            e+=3;
          }
          else
          {
            *e=*src;
            e++;
          }
        }
        *e=0;
        UdmHTTPBufAppendf(&Doc->Buf, "<A HREF=\"%s%s\">\n",
                          escaped_name, is_dir? "/" : "");
        if (UdmHTTPBufSize(&Doc->Buf) + 2048 > UdmHTTPBufAllocedSize(&Doc->Buf))
        {
          if (UDM_OK != UdmHTTPBufRealloc(&Doc->Buf,
                                          UdmHTTPBufAllocedSize(&Doc->Buf) + 64*1024))
            break;
        }
      }
      freedir(recs);
      UdmHTTPBufAppendf(&Doc->Buf,"</BODY><HTML>\n");
      return 0;
    }
    else
    {
      switch(errno)
      {
        case ENOENT: status=404;break; /* Not found */
        case EACCES: status=403;break; /* Forbidden*/
        default: status=500;
      }
      UdmHTTPBufPrintf(&Doc->Buf,
                       "HTTP/1.0 %d %s\r\n"
                       "X-Reason: Loaddir returned no results for '%s'\r\n"
                       "\r\n",
                       status, UdmHTTPErrMsg(status), openname);
      return 0;
    }
  }
  else
  {
    const UDM_REPLACE *M;
    UDM_MATCH_PART P[10];
    const char *fn= (Doc->CurURL.filename && Doc->CurURL.filename[0]) ?
                    Doc->CurURL.filename : "index.html";
    char content_type[80]= "";

    
    /* Lets compare last modification date */
    if (Request.if_modified_since >= sb.st_mtime)
    {
      /* Document seems to be unchanged */
      status=304;
      UdmHTTPBufPrintf(&Doc->Buf,
                       "HTTP/1.0 %d %s\r\n\r\n",
                       status, UdmHTTPErrMsg(status));
      return 0;
    }

    if((fd=open(openname,O_RDONLY|UDM_BINARY))<0)
    {
      switch(errno)
      {
        case ENOENT: status=404;break;
        case EACCES: status=403;break;
        default: status=1;
      }
      UdmHTTPBufPrintf(&Doc->Buf,
                       "HTTP/1.0 %d %s\r\nX-Reason: %d:%s\r\n\r\n",
                       status, UdmHTTPErrMsg(status),
                       errno, strerror(errno));
      return 0;
    }
    
    if ((M= UdmReplaceListFind(&Indexer->Conf->MimeTypes, fn, 10, P)))
      udm_snprintf(content_type, sizeof(content_type), "%s", M->Replace.str);
    
    if (strcasecmp(content_type, "application/http"))
    {
      UdmHTTPBufPrintf(&Doc->Buf, "HTTP/1.0 200 OK\r\n");
      UdmHTTPBufAppend(&Doc->Buf, UDM_CSTR_WITH_LEN("Last-Modified: "));
      UdmHTTPBufAppendHTTPTime(&Doc->Buf, sb.st_mtime);
      UdmHTTPBufAppend(&Doc->Buf, UDM_CSTR_WITH_LEN("\r\n\r\n"));
    }
    else
    {
      UdmHTTPBufReset(&Doc->Buf);
    }

    UdmHTTPBufAppendFromFile(&Doc->Buf, fd);
    close(fd);
    return 0;
  }
}
#endif



static udm_rc_t
UdmBuildHTTPRequest(UDM_DOCUMENT *Doc)
{
  const char  *method=(Doc->method==UDM_METHOD_HEAD)?"HEAD":"GET";
  udm_bool_t proxy= UDM_TEST(UdmVarListFindStr(&Doc->RequestHeaders,"Proxy",NULL) != NULL);
  size_t      i= strlen(UDM_NULL2EMPTY(Doc->CurURL.path)) + 
                 strlen(UDM_NULL2EMPTY(Doc->CurURL.filename)) + 1;
  char *url= (char*)UdmMalloc(i);
  char *eurl= (char*)UdmMalloc(3 * i);

  if (url == NULL || eurl == NULL) return UDM_ERROR;

  sprintf(url, "%s%s", UDM_NULL2EMPTY(Doc->CurURL.path),
                       UDM_NULL2EMPTY(Doc->CurURL.filename)); 
  UdmEscapeURI(eurl, url);
  
  UDM_ASSERT(UdmHTTPBufPtr(&Doc->Buf) != NULL);
  if(proxy && strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "file"))
  {
    UdmHTTPBufPrintf(&Doc->Buf,
                      "%s %s://%s%s HTTP/1.0\r\n",
                      method, UDM_NULL2EMPTY(Doc->CurURL.schema), 
                      UDM_NULL2EMPTY(Doc->CurURL.hostinfo), eurl);
  }
  else
  {
    UdmHTTPBufPrintf(&Doc->Buf, "%s %s HTTP/1.0\r\n", method, eurl);
  }
  UDM_FREE(eurl);UDM_FREE(url);
  
  /* Add user defined headers */
  for( i=0 ; i<Doc->RequestHeaders.nvars ; i++)
  {
    const UDM_VAR *Hdr= UdmVarListFindConstByIndex(&Doc->RequestHeaders, i);
    UDM_CONST_STR strbuf, *str= UdmVarGetConstStr(Hdr, &strbuf);
    UdmHTTPBufAppendf(&Doc->Buf, "%s: %.*s\r\n",
                      UdmVarName(Hdr), (int) str->length, str->str);
  }
  
  /* Empty line is the end of HTTP header */
  UdmHTTPBufAppendf(&Doc->Buf, "\r\n");
#ifdef DEBUG_REQUEST
  fprintf(stderr,"Request:'%.*s'\n",
          (int) UdmHTTPBufSize(&Doc->Buf), UdmHTTPBufPtr(&Doc->Buf));
#endif
  return UDM_OK;
}


static void
UdmCreateErrorResponse(UDM_DOCUMENT *Doc, int code)
{
  UdmHTTPBufPrintf(&Doc->Buf,
                    "HTTP/1.0 %d %s\r\n\r\n",
                    code, UdmHTTPErrMsg(code));
  UdmVarListReplaceInt(&Doc->Sections, "Status", code);
}


udm_rc_t
UdmGetURL(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc)
{
  udm_bool_t proxy;
  int mirror_period, found_in_mirror= 0, res= 0;
  udm_timer_t ticks= UdmStartTimer();
  udm_rc_t rc= UDM_OK;

  proxy= UDM_TEST(UdmVarListFindStr(&Doc->RequestHeaders,"Proxy",NULL) != NULL);
  mirror_period= UdmVarListFindInt(&Doc->Sections, "MirrorPeriod", -1);

  UdmBuildHTTPRequest(Doc);
  
  /* If mirroring is enabled */
  if (mirror_period >= 0)
  {
    /* 
      on u_m==0 it returned by mtime from mirrorget
      but we knew that it should be put in mirror
    */
    
    if (!UdmMirrorGET(Indexer, Doc, &Doc->CurURL))
    {
      UdmLog(Indexer, UDM_LOG_DEBUG, "Taken from mirror");
      found_in_mirror= 1;
      UdmVarListReplaceStr(&Doc->Sections, "Content-Source", "mirror");
    }
    else
    {
      UdmVarListReplaceStr(&Doc->Sections, "Content-Source", "original");
    }
  }

  if (!found_in_mirror)
  {
    if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "exec"))
    {
      rc= UdmExecGet(Indexer,Doc);
      goto ex;
    }
    else if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "cgi"))
    {
      rc= UdmExecGet(Indexer,Doc);
      goto ex;
    }
#ifdef HAVE_SQL /* TODO34 */
    else if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "htdb"))
    {
      rc= UdmHTDBGet(Indexer, Doc);
      goto ex;
    }
#endif
#ifdef USE_FILE
    else if(!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "file"))
    {
      res= UdmFILEGet(Indexer,Doc);
    }
#endif
#ifdef USE_NEWS
    else if((!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "news")))
    {
      res= UdmNNTPGet(Indexer,Doc);
    }
    else if((!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "nntp")))
    {
      res= UdmNNTPGet(Indexer,Doc);
    }
#endif
#ifdef USE_HTTPS
    else if((!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "https")))
    {
      res= UdmHTTPSGet(Indexer,Doc);
    }
#endif
#ifdef USE_HTTP
    else if((!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "http")) ||
            (!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "ftp") && (proxy)))
    {
      res= UdmHTTPGet(Indexer,Doc);
    }
#endif
#ifdef USE_FTP
    else if ((!strcasecmp(UDM_NULL2EMPTY(Doc->CurURL.schema), "ftp")) && (!proxy))
    {
      res= UdmFTPGet(Indexer,Doc);
    }
#endif
  }


  switch(res)
  {
    case UDM_NET_UNKNOWN:
      UdmLog(Indexer,UDM_LOG_WARN,"Protocol not supported");
      UdmCreateErrorResponse(Doc, UDM_HTTP_STATUS_NOT_SUPPORTED);
      break;
    case UDM_NET_ERROR:
      UdmLog(Indexer,UDM_LOG_WARN,"Network error");
      UdmCreateErrorResponse(Doc, UDM_HTTP_STATUS_SERVICE_UNAVAILABLE);
      break;
    case UDM_NET_TIMEOUT:
      UdmLog(Indexer,UDM_LOG_WARN,"Download timeout");
      UdmCreateErrorResponse(Doc, UDM_HTTP_STATUS_GATEWAY_TIMEOUT);
      break;
    case UDM_NET_CANT_CONNECT:
      UdmLog(Indexer,UDM_LOG_WARN,"Can't connect to host %s:%d",Doc->connp.hostname,Doc->connp.port);
      UdmCreateErrorResponse(Doc, UDM_HTTP_STATUS_SERVICE_UNAVAILABLE);
      break;
    case UDM_NET_CANT_RESOLVE:
      UdmLog(Indexer,UDM_LOG_WARN,"Unknown %shost '%s'",proxy?"proxy ":"",Doc->connp.hostname);
      UdmCreateErrorResponse(Doc, UDM_HTTP_STATUS_SERVICE_UNAVAILABLE);
      break;
    default:
      if(res<0)
      {  /* No connection */
        UdmLog(Indexer,UDM_LOG_WARN,"Can't connect to host %s:%d",Doc->connp.hostname,Doc->connp.port);
        UdmCreateErrorResponse(Doc, UDM_HTTP_STATUS_SERVICE_UNAVAILABLE);
        break;
      }
  }

ex:
  UdmVarListReplaceInt(&Doc->Sections, "ResponseTime",
                       (int) (UdmStartTimer() - ticks));

  /* Put into mirror if required */
  if ((mirror_period >= 0) && (!found_in_mirror))
  {
    if (UdmMirrorPUT(Indexer, Doc, &Doc->CurURL))
      return UDM_ERROR;
  }
  return rc;
}

/*
  UdmDocInit() must be called of Doc before calling this function
*/
udm_rc_t
UdmGetURLSimple(UDM_AGENT *Agent, UDM_DOCUMENT *Doc, const char *vurl)
{
  UDM_VARLIST *vars= &Agent->Conf->Vars;
  size_t max_doc_size= (size_t)UdmVarListFindInt(vars,"MaxDocSize",UDM_MAXDOCSIZE);
  if (!UdmHTTPBufPtr(&Doc->Buf) &&
      UDM_OK != UdmHTTPBufAlloc(&Doc->Buf, max_doc_size))
    return UDM_ERROR;

  UdmURLParse(&Doc->CurURL,vurl);
  UdmVarListReplaceStr(&Doc->RequestHeaders, "Host", UDM_NULL2EMPTY(Doc->CurURL.hostname));
  Doc->connp.hostname= (char*)UdmStrdup(UDM_NULL2EMPTY(Doc->CurURL.hostname));
  Doc->connp.port= Doc->CurURL.port ? Doc->CurURL.port : Doc->CurURL.default_port;
    
  if(UdmHostLookup2(Agent, &Agent->Conf->Hosts, &Doc->connp))
  {
  }

  if (UdmGetURL(Agent, Doc) != UDM_OK)
    return UDM_ERROR;

  UdmParseHTTPResponse(Agent, Doc);
  UdmDocProcessContentResponseHeaders(Agent, Doc);

#ifdef HAVE_ZLIB
  {
    const char *ce= UdmVarListFindStr(&Doc->Sections,"Content-Encoding", NULL);
    if (ce && ce[0])
    {
      udm_content_encoding_t encoding;
      if (UDM_CONTENT_ENCODING_UNKNOWN != (encoding= UdmContentEncodingID(ce)))
        return UdmHTTPBufContentEncodingDecode(&Doc->Buf, encoding);
    }
  }
#endif
  return UDM_OK;
}
