/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef   HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#include "udmsearch.h"

/* This should be last include */
#ifdef DMALLOC
#include "dmalloc.h"
#endif


#define POWERED_LOGO "<BR><a href=\"http://www.mnogosearch.org/\"><IMG BORDER=0 SRC=\"http://www.mnogosearch.org/img/mwin.gif\"></A>&nbsp<FONT SIZE=1><A HREF=\"http://www.mnogosearch.org/\">Powered by mnoGoSearch - free web search engine software</A></FONT><BR>\n"

#define UDM_HTTPD_CMDLINE 0
#define UDM_HTTPD_CGI     1
#define UDM_HTTPD_DAEMON  2

#define TEMPLATE_NAME_LENGTH 1024

typedef struct udm_self_template_query_st
{
  char template_name[TEMPLATE_NAME_LENGTH];
  char self[1024];
  char query_string[1024];
  int  httpd;                /* Execution type: httpd, inetd or command line */
} UDM_SELF_TEMPLATE_QUERY;


static void
UdmSelfTemplateQueryInit(UDM_SELF_TEMPLATE_QUERY *q)
{
  /*
    Some HTTPD servers do not pass QUERY_STRING if the query was empty,
    so additionally check REQUEST_METHOD to be safe.
  */
  q->httpd= getenv("QUERY_STRING") || getenv("REQUEST_METHOD") ?
            UDM_HTTPD_CGI : 0;
  q->template_name[0]= 0;
  q->self[0]= 0;
  q->query_string[0]= 0;
}


/******************** Variable functions *****************************/
/*
  Wrappers for UdmVarListFindXXX(),
  with a limit to command parameters only, ignoring query string values.
  Have two purposes:
  - generally avoid overwriting search.htm commands in query string
  - avoid cross-site scripting, in case of String variables
*/
static const char *
UdmVarListFindSafeStr(UDM_VARLIST *List, const char *name, const char *def)
{
  const UDM_VAR *Var= UdmVarListFind(List, name);
  if (!Var || UdmVarSecno(Var) == UDM_VARSRC_QSTRING || !UdmVarStr(Var))
    return def;
  return UdmVarStr(Var);
}

/* Avoid using directly */
#define UdmVarListFindStr(x)  { error;}

static int
UdmVarListFindSafeInt(UDM_VARLIST *List, const char *name, int def)
{
  const UDM_VAR *Var= UdmVarListFind(List, name);
  if (!Var || UdmVarSecno(Var) == UDM_VARSRC_QSTRING || !UdmVarStr(Var))
    return def;
  return atoi(UdmVarStr(Var));
}


/*****************************************************************/

#ifndef WIN32
#include <grp.h>

static udm_rc_t
UdmLoadGroups (UDM_ENV *E)
{
  struct group *grp;
  int cnt = 0;
  char name[32];
  char **p;
  const char *user= UdmVarListFindSafeStr(&E->Vars, "REMOTE_USER", NULL);

  if (! user) return(UDM_OK);

  setgrent();
  while ((grp = getgrent()))
  {
    for (p = grp->gr_mem; *p; p++)
    {
      if (! strcmp(user, *p))
      {
        udm_snprintf(name, sizeof(name), "REMOTE_GROUP%d", cnt);
        UdmVarListAddStr(&E->Vars, name, grp->gr_name);
        cnt++;
	break;
      }
    }
  }
  getgrent();

  return(UDM_OK);
}
#endif

/************************************************************/

/*
   We need to initialize WSA to process possible
   includes in template.
*/
static void UdmWSAStartup(void)
{
#ifdef WIN32
  WSADATA wsaData;
  if(WSAStartup(0x101,&wsaData)!=0)
  {
    fprintf(stderr,"WSAStartup() error %d\n",WSAGetLastError);
    exit(1);
  }
#endif
}

static void UdmWSACleanup(void)
{
#ifdef WIN32
  WSACleanup();
#endif
}
/************************************************************/

static udm_rc_t
UdmHTTPHeadersSend(UDM_SELF_TEMPLATE_QUERY *q,
                   const char *content_type, const char *bcs)
{
  if (q->httpd & UDM_HTTPD_DAEMON)
    printf("HTTP/1.0 200 OK\r\n");
  
  if (q->httpd && strcasecmp(content_type, "none"))
  {
    if (bcs)
      printf("Content-type: %s; charset=%s\r\n\r\n", content_type, bcs);
    else
      printf("Content-Type: %s\r\n\r\n", content_type);
  }
  return UDM_OK;
}


static void usage(void)
{
}


static udm_rc_t
UdmDetectSelfTemplateQueryCGI(UDM_ENV *Env,
                              UDM_SELF_TEMPLATE_QUERY *q,
                              int argc, char **argv)
{
  char *env;
  /* Detect self */
  if((env= getenv("UDMSEARCH_SELF")) || (env= getenv("REQUEST_URI")))
  {
    /* Use Apache's REQUEST_URI, which is not in CGI standard */
    char *end;
    udm_snprintf(q->self, sizeof(q->self), "%s", env);
    if ((end= strchr(q->self,'?')))
      end[0]='\0';
  }
  else if ((env= getenv("SCRIPT_NAME")))
  {
    /* Use standard CGI variables if not under Apache */
    const char *path_info= getenv("PATH_INFO");
    udm_snprintf(q->self, sizeof(q->self),"%s%s", env, path_info ? path_info:"");
  }
  else
  {
    udm_snprintf(q->self, sizeof(q->self), "search.cgi");
  }
  
  
  /* Detect template name */
  if (q->template_name[0])
  {
    /* Do nothing - already set in command line via -d option */
  }
  else if((env= getenv("UDMSEARCH_TEMPLATE")))
  {
    size_t envlen= strlen(env);
    if (envlen > TEMPLATE_NAME_LENGTH - 1)
      envlen= TEMPLATE_NAME_LENGTH - 1;
    UdmURLDecode(q->template_name, env, envlen);
  }
  else if(getenv("REDIRECT_STATUS") && (env= getenv("PATH_TRANSLATED")))
  {
    /* Check Apache internal redirect  */
    /* via   "AddHandler" and "Action" */
    udm_snprintf(q->template_name, sizeof(q->template_name), "%s", env);
  }
  else if ((env= getenv("PATH_INFO")) && env[0])
  {
    /* http://localhost/cgi-bin/search.cgi/path/to/search.htm */
    char name[128];
    
    strncpy(name,env,sizeof(name)-1);
    name[sizeof(name)-1]='\0';
    if (UDMSLASH !='/')
    {
      char *s;
      for (s=name; s[0]; s++)
        if (s[0]=='\\')s[0]='/';
    }
    
    /* Take from the config directory */
    UdmGetFileName(q->template_name, sizeof(q->template_name)-5, 
                   UDM_DIRTYPE_CONF, name);
  }
  else
  {
    /* CGI executed without Apache internal redirect */
    /* or started from command line                  */
    char *end;
    size_t length;
    
    /* Take from the config directory */
    length= UdmGetFileName(q->template_name, sizeof(q->template_name)-5,
                           UDM_DIRTYPE_CONF,
                           (end= strrchr(q->self,'/')) ? (end+1) : (q->self));
    end= q->template_name + length;
    if ( length > 3 && (!strcmp(end-4,".exe") || !strcmp(end-4,".cgi")))
      end-=4;
    strcpy(end,".htm");
  }
  
  if((env= getenv("QUERY_STRING")))
  {
    udm_snprintf(q->query_string, sizeof(q->query_string), "%s", env);
  }
  else if (!q->httpd && argv[0])
  {
    udm_snprintf(q->query_string, sizeof(q->query_string), "q=%s", argv[0]);
  }
  else
  {
    udm_snprintf(q->query_string, sizeof(q->query_string), "q=");
  }
  return UDM_OK;
}


static udm_rc_t
UdmDetectSelfTemplateQueryDaemon(UDM_ENV *Env,
                                 UDM_SELF_TEMPLATE_QUERY *q,
                                 FILE *stream)
{
  char str[1024];
  while (fgets(str, sizeof(str), stream))
  {
    UdmRTrim(str, "\r\n");
    if (!strncmp(str, "GET ", 4))
    {
      char *method, *uri, *lt;
      if ((method= udm_strtok_r(str, " ", &lt)) &&
          (uri= udm_strtok_r(NULL, " ", &lt)))
      {
        char *query_string= strchr(uri, '?');
        if (query_string)
        {
          *query_string= '\0';
          query_string++;
          udm_snprintf(q->query_string, sizeof(q->query_string), "%s", query_string);
        }
        UdmGetFileName(q->template_name, sizeof(q->template_name)-5, 
                       UDM_DIRTYPE_CONF, uri);
        udm_snprintf(q->self, sizeof(q->self), "%s",  uri);
      }
    }
    if (!strcmp(str,""))
      break;
  }
  return UDM_OK;
}


static UDM_SELF_TEMPLATE_QUERY q;


static int
UdmCmdLineHandleOption(void *user_data,
                       const UDM_CMDLINE_OPT *opt, const char *value)
{
  switch (opt->id)
  {
    case 'x':
      q.httpd|= UDM_HTTPD_DAEMON;
      break;
    case 'd':
      udm_snprintf(q.template_name, sizeof(q.template_name), "%s", value);
      break;
    case '?':
    case 'h':
    default:
      usage();
      exit(0);
  }
  return 0;
}


static UDM_CMDLINE_OPT udm_search_cgi_options[]=
{
  {'x', "", UDM_OPT_BOOL, NULL, "Act as itetd/xinetd service"},
  {'h', "", UDM_OPT_BOOL, NULL, "Print help page and exit; -hh print more help"},
  {'?', "", UDM_OPT_BOOL, NULL, "Print help page and exit; -?? print more help"},
  {'d', "", UDM_OPT_STR,  NULL, "Use the given template file instead of search.htm"},
  {0, NULL, 0, NULL, NULL},
};


#if ((defined WIN32) && (defined _DEBUG))
static void
WIN32_IIS_DEBUG_MSG(const char *name)
{
  char szMessage [256];
  wsprintf(szMessage, "Please attach a debugger to the process 0x%X (%s) and click OK",
           GetCurrentProcessId(), name);
  MessageBox(NULL, szMessage, "CGI Debug Time!", MB_OK|MB_SERVICE_NOTIFICATION);
}
#endif


int main(int argc, char ** argv)
{
  const char  *env;
  udm_rc_t    rc;
  UDM_ENV     Env;
  UDM_TEMPLATE tmpl;
  size_t noptions= 0;
  
  UdmSelfTemplateQueryInit(&q);

#if ((defined WIN32) && (defined _DEBUG))
  if(q.httpd)
  {
    WIN32_IIS_DEBUG_MSG(argv[0]);
    DebugBreak();
  }
#endif


  if (!q.httpd)
  {
    /*
      Don't parse command line options when under HTTPD,
      to avoid opening passing an arbitrary file as a template name:
      http://localhost/cgi-bin/search.cgi/search.htm?-d/path/to/arbitrary/file
    */
    if (UdmCmdLineOptionsGet(NULL, argc, argv, udm_search_cgi_options,
                             UdmCmdLineHandleOption, &noptions))
      return 0;
  }

  UdmWSAStartup();
  
  UdmInit();
  UdmEnvInit(&Env);
#if defined(HAVE_PTHREAD) && !defined(WIN32)
  UdmInitMutexes();
  UdmSetLockProc(&Env, UdmLockProc);
  Env.THDHandler.ThreadCreate= UdmThreadCreate;
  Env.THDHandler.ThreadJoin= UdmThreadJoin;
#endif
  UdmTemplateInit(&tmpl);
  UdmVarListAddEnviron(&Env.Vars,"ENV");
  
  if (q.httpd & UDM_HTTPD_DAEMON)
    UdmDetectSelfTemplateQueryDaemon(&Env, &q, stdin);
  else
    UdmDetectSelfTemplateQueryCGI(&Env, &q, argc - noptions , argv + noptions);
  
  
#ifndef WIN32
  env= getenv("REMOTE_USER");
  if (env)
  {
    UdmVarListAddStr(&Env.Vars, "REMOTE_USER", env);
    UdmLoadGroups(&Env);
  }
#endif
  
  if ((rc= UdmTemplateLoad(&tmpl, &Env, q.template_name)))
  {
    UdmHTTPHeadersSend(&q, "text/plain", NULL);
    printf("%s\n",Env.errstr);
    UdmTemplateFree(&tmpl);
    UdmEnvFree(&Env);
    return(0);
  }


#ifdef HAVE_LOCALE_H
  if ((env= UdmVarListFindSafeStr(&Env.Vars, "Locale", NULL)))
    setlocale(LC_ALL, env);
#endif

#ifdef HAVE_SETVBUF
  {
    /*
      Set stdout buffer size. It allows to avoid re-rendering
      in the browser. After setting StdoutBufferSize to a
      reasonably big value, search.cgi will send whole output
      at once in the end, rather than gradually.
    */
    size_t bsz= (size_t) UdmVarListFindSafeInt(&Env.Vars, "StdoutBufferSize", 0);
    if (bsz > 0)
      setvbuf(stdout, NULL, _IOFBF, bsz);
  }
#endif


  /*
    Prepare argv and argc to make the template name available as argv[0].
  */
  if (noptions)
  {
    /*
      Executed from command line.
      Put the template name into the last argv that was
      scanned by UdmCmdLineOptionsGet().
      The program will get the template name in argv[0],
      with additional parameters that were not scanned by
      UdmCmdLineOptionsGet().
    */
    noptions--;
  }
  else
  {
    /*
      Executed from HTTP server, UdmCmdLineOptionsGet() was not called.
      The program will get argc==1 with the template name in argv[0],
      without any additional arguments.
    */
    UDM_ASSERT(argc > 0);
    noptions= argc - 1;
  }
  argv[noptions]= q.template_name;
  if ((rc= UdmTemplatePrintToStreamWithArgs(argc - noptions,
                                            argv + noptions,
                                            &Env, stdout, &tmpl)))
  {
    UdmHTTPHeadersSend(&q, "text/plain", NULL);
    printf("%s\n", Env.errstr);
    rc= 0;
  }
  else
    rc= tmpl.rc;

  UdmTemplateFree(&tmpl);
  UdmEnvFree(&Env);
  if (q.httpd) 
    fflush(NULL); 
  else
    fclose(stdout);
  UdmWSACleanup();
#if defined(HAVE_PTHREAD) && !defined(WIN32)
  UdmDestroyMutexes();
#endif
  return rc;
}
