/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#if   (HAVE_MYSQL)
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "udm_common.h"
#include "udm_sqldbms.h"
#include "udm_utils.h"
#include "udm_vars.h"

#include "udm_xmalloc.h"
#ifdef WIN32
#include <process.h>
#endif
#include <mysql.h>

#define ER_DUP_ENTRY            1062
#define ER_DUP_KEY		1022
#define CR_SERVER_LOST          2013
#define CR_SERVER_GONE_ERROR    2006
#define ER_SERVER_SHUTDOWN      1053
#define ER_DUP_ENTRY_WITH_KEY_NAME 1582

#if (MYSQL_VERSION_ID > 40100)
#define HAVE_MYSQL_PS 1
#endif


/*
  Return MySQL version as number: 4.1.24 -> 40124
  This is a replacement for mysql_get_server_version()
  which appeared in 4.1.x
*/
static int
udm_mysql_version(MYSQL *mysql)
{
  int v1, v2, v3;
  char *pos= mysql->server_version, *end;
  v1=   strtol(pos, &end, 10); pos= end + 1;
  v2=   strtol(pos, &end, 10); pos= end + 1;
  v3= strtol(pos, &end, 10);
  return v1 * 10000 + v2 * 100 + v3;
}

#define UDM_MYSQL_MAX_BIND_PARAM UDM_SQL_MAX_BIND_PARAM

#ifdef HAVE_MYSQL_PS
typedef struct udm_mysql_stmt_st
{
  MYSQL_STMT *stmt;
  MYSQL_BIND bind[UDM_MYSQL_MAX_BIND_PARAM];
  unsigned long bind_length[UDM_MYSQL_MAX_BIND_PARAM];
} UDM_MYSQL_STMT;
#endif  


typedef struct udm_mysql_st
{
  MYSQL mysql;
#ifdef HAVE_MYSQL_PS
  UDM_MYSQL_STMT stmt;
#endif  
} UDM_MYSQL;


static UDM_MYSQL *
UdmMySQL(UDM_SQL *db)
{
  return (UDM_MYSQL*) db->specific;
}

static void
UdmMySQLSetError(UDM_SQL *db, const char *msg)
{
  UDM_MYSQL *mydb= UdmMySQL(db);
  db->errcode= 1;
  sprintf(db->errstr,"[MySQL driver]%s #%d: %s",
          msg, mysql_errno(&mydb->mysql), mysql_error(&mydb->mysql));
}


static udm_rc_t
UdmMySQLSetNames(UDM_SQL *db, UDM_MYSQL *mydb, const char *csname)
{
  int myrc;
#if  (MYSQL_VERSION_ID > 40100)
  myrc= mysql_set_character_set(&mydb->mysql, csname);
#else
  {
    char qbuf[64];
    udm_snprintf(qbuf, sizeof(qbuf), "SET NAMES %s", csname);
    myrc= mysql_query(&mydb->mysql, qbuf);
  }
#endif
  if (myrc)
  {
    UdmMySQLSetError(db, " SET NAMES:");
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
UdmMySQLConnect(UDM_SQL *db)
{
  const char* DBLog=  UdmVarListFindStr(&db->Vars,"sqllog",NULL);
  const char* DBLogBin= UdmVarListFindStr(&db->Vars,"sqllogbin",NULL);
  const char* DBSock= UdmVarListFindStr(&db->Vars,"socket",NULL);
  const char* DBUser= UdmVarListFindStr(&db->Vars,"DBUser",NULL);
  const char* DBPass= UdmVarListFindStr(&db->Vars,"DBPass",NULL);
  const char* DBHost= UdmVarListFindStr(&db->Vars, "DBHost", "localhost");
  const char* DBName= UdmVarListFindStr(&db->Vars, "DBName", "mnogosearch");
  const char* mycnfgroup= UdmVarListFindStr(&db->Vars, "mycnfgroup", "client");
  int DBPort= UdmVarListFindInt(&db->Vars, "DBPort", 0);
  const char* setnames=  UdmVarListFindStr(&db->Vars,"setnames", "");
  const char* storage_engine= UdmVarListFindStr(&db->Vars, "Engine", "MyISAM");
  int opt_compress= UdmVarListFindBool(&db->Vars, "compress", UDM_FALSE);
  int opt_client_multi_statements= UdmVarListFindBool(&db->Vars, "ClientMultiStatements", UDM_FALSE);
  UDM_MYSQL *mydb= (UDM_MYSQL*) UdmMalloc(sizeof(UDM_MYSQL));
  unsigned long client_flag= 0;
  
  if (!(db->specific= mydb))
  {
    db->errcode= 1;
    sprintf(db->errstr, "Can't alloc mydb. Not enough memory?");
    return UDM_ERROR;
  }
  
  bzero((void*) mydb, sizeof(UDM_MYSQL));
  
  if (!mysql_init(&mydb->mysql))
  {
    UdmMySQLSetError(db, "");
    return UDM_ERROR;
  }

  if (strcasecmp(mycnfgroup, "no"))
    mysql_options(&mydb->mysql, MYSQL_READ_DEFAULT_GROUP, mycnfgroup);
  
  if (opt_compress)
    mysql_options(&mydb->mysql, MYSQL_OPT_COMPRESS, 0);

#ifdef CLIENT_MULTI_STATEMENTS
  client_flag |= opt_client_multi_statements ? CLIENT_MULTI_STATEMENTS : 0;
#endif

  if(!(mysql_real_connect(&mydb->mysql, DBHost, DBUser, DBPass, DBName,
                          (unsigned) DBPort, DBSock, client_flag)))
  {
    UdmMySQLSetError(db, "");
    return UDM_ERROR;
  }
  db->connected= UDM_TRUE;
  db->version= udm_mysql_version(&mydb->mysql);
  if (db->version >= 40100)
  {
    db->flags|= UDM_SQL_HAVE_SUBSELECT;
    if (setnames[0])
    {
      UDM_CHARSET *cs= UdmGetCharSet(setnames);
      if (UdmMySQLSetNames(db, mydb, cs && cs->mysql_name ?
                                     cs->mysql_name :
                                     "binary"))
        return UDM_ERROR;
    }
  }
  else
  {
#ifdef HAVE_MYSQL_PS
    /* Switch from native to general PS API */
    if (db->handler.Prepare)
    {
      db->handler.Prepare=  UdmSQLPrepareGeneric;
      db->handler.Bind=     UdmSQLBindGeneric;
      db->handler.Exec=     UdmSQLExecGeneric;
      db->handler.StmtFree= UdmSQLStmtFreeGeneric;
    }
#endif
  }

  if (storage_engine)
  {
    int myrc;
    char qbuf[64];
    udm_snprintf(qbuf, sizeof(qbuf), "SET storage_engine=%s",
                 storage_engine[0] ?  storage_engine : "MyISAM");
    if ((myrc= mysql_query(&mydb->mysql, qbuf)))
    {
      UdmMySQLSetError(db, qbuf);
      return UDM_ERROR;
    }
    if (!strcasecmp(storage_engine, "InnoDB"))
      db->flags|= (UDM_SQL_HAVE_TRANSACT | UDM_SQL_HAVE_GOOD_COMMIT);
  }

  if (DBLog)
  {
    char qbuf[64];
    sprintf(qbuf, "SET SQL_LOG_OFF=%d", atoi(DBLog) ? 0 : 1);
    if (mysql_query(&mydb->mysql, qbuf))
    {
      UdmMySQLSetError(db, qbuf);
      return UDM_ERROR;
    }
  }
  if (DBLogBin)
  {
    char qbuf[64];
    sprintf(qbuf, "SET SQL_LOG_BIN=%d", atoi(DBLogBin) ? 1 : 0);
    if (mysql_query(&mydb->mysql, qbuf))
    {
      UdmMySQLSetError(db, qbuf);
      return UDM_ERROR;
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmMySQLStoreResult(UDM_SQL *db, UDM_SQLRES *R)
{
  size_t mitems= 0;
  MYSQL_ROW mysqlrow;
  
  while((mysqlrow=mysql_fetch_row((MYSQL_RES *)R->specific)))
  {
    size_t col;
    size_t coloffs= R->nRows * R->nCols;  
    unsigned long  *lengths= mysql_fetch_lengths((MYSQL_RES *)R->specific);
    
    if (coloffs + R->nCols >= mitems)
    {
      mitems= mitems ? mitems * 8 : 256;
      R->Items= (UDM_STR*) UdmRealloc(R->Items, mitems * sizeof(UDM_STR));
    }
    
    for(col= 0; col < R->nCols ; col++)
    {
      UDM_STR *I= &R->Items[coloffs + col];
      size_t len;
      len= I->length= lengths[col];
      I->str= (char*) UdmMalloc(len+1);
      memcpy(I->str, mysqlrow[col], len);
      I->str[len]='\0';
    }
    R->nRows++;
  }
  return UDM_OK;
}


static udm_rc_t
UdmMySQLStoreMetaData(UDM_SQL *db, UDM_SQLRES *R)
{
  if(R->specific)
  {
    MYSQL_FIELD  *field;
    size_t    nfields;
    
    R->nCols= mysql_num_fields((MYSQL_RES*)R->specific);
    R->nRows=0;
    R->Items=NULL;
    R->Fields=(UDM_SQLFIELD*)UdmMalloc(R->nCols*sizeof(UDM_SQLFIELD));
    bzero(R->Fields,R->nCols*sizeof(UDM_SQLFIELD));
    
    for(nfields=0; (field=mysql_fetch_field((MYSQL_RES *)R->specific)); nfields++)
    {
      R->Fields[nfields].sqlname = (char*)UdmStrdup(field->name);
      R->Fields[nfields].sqllen=field->length;
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmMySQLExecDirect(UDM_SQL *db,UDM_SQLRES *R,const char *query)
{
  size_t  i;
  UDM_MYSQL *mydb;
  MYSQL *mysql;
  
  db->errcode= 0;
  
  if (R)
  {
    bzero((void*) R, sizeof(UDM_SQLRES));
    R->db= db;
  }

  if (!db->connected)
  {
    udm_rc_t rc;
    if (UDM_OK != (rc= UdmMySQLConnect(db)))
      return rc;
  }

  mydb= UdmMySQL(db);
  mysql= &mydb->mysql;

  for(i=0;i<2;i++)
  {
    if((mysql_query(mysql,query)))
    {
      if((mysql_errno(mysql)==CR_SERVER_LOST)||
         (mysql_errno(mysql)==CR_SERVER_GONE_ERROR)||
         (mysql_errno(mysql)==ER_SERVER_SHUTDOWN))
      {
        UDMSLEEP(5);
      }
      else
      {
        sprintf(db->errstr,"MySQL driver: #%d: %s",
                mysql_errno(mysql),mysql_error(mysql));
        if((mysql_errno(mysql)!=ER_DUP_ENTRY) &&
           (mysql_errno(mysql)!=ER_DUP_KEY) &&
           (mysql_errno(mysql)!=ER_DUP_ENTRY_WITH_KEY_NAME))
        {
          db->errcode= 1;
          return UDM_ERROR;
        }
        db->errcode= 0;
        return UDM_OK;
      }
    }
    else
    {
      if (R)
      {
        R->specific= mysql_use_result((MYSQL*) db->specific);
        return UdmMySQLStoreMetaData(db, R);
      }
      else
        return UDM_OK;
    }
  }
  UdmMySQLSetError(db, "");
  return UDM_ERROR;
}


static udm_rc_t
UdmMySQLQuery(UDM_SQL *db,UDM_SQLRES *R,const char *query)
{
  udm_rc_t rc= UdmMySQLExecDirect(db, R, query);
  if (rc != UDM_OK)
    return rc;

  if (!R->specific)
    return UDM_OK;
  
  rc= UdmMySQLStoreResult(db, R);

#if (MYSQL_VERSION_ID > 40100)
  {
    UDM_MYSQL *mydb= (UDM_MYSQL*) db->specific;
    while (mysql_next_result(&mydb->mysql) == 0)
    {
      MYSQL_RES *res;
      if ((res= mysql_store_result(&mydb->mysql)))
      {
        mysql_free_result(res);
      }
      else
      {
        /* Non-select query */
      }
    }
  }
#endif

  return rc;
}


static udm_rc_t
UdmMySQLClose(UDM_SQL *db)
{
  UDM_MYSQL *mydb= (UDM_MYSQL*) db->specific;
  MYSQL *mysql= &mydb->mysql;
  mysql_close(mysql);
  UdmFree(db->specific);
  db->specific = NULL;
  return UDM_OK;
}


static udm_rc_t
UdmMySQLBegin(UDM_SQL *db)
{
  return UDM_OK;
}


static udm_rc_t
UdmMySQLCommit(UDM_SQL *db)
{
  return UDM_OK;
}


static size_t
UdmMySQLEscStr(UDM_SQL *db, char *to, const char *from, size_t len)
{
  return mysql_escape_string(to, from, len);
}


static udm_rc_t
UdmMySQLFreeResult(UDM_SQL *db, UDM_SQLRES *R)
{
  UdmSQLFreeResultSimple(db, R);
  if (R->specific)
  {
    mysql_free_result((MYSQL_RES *)R->specific);
    R->specific= NULL;
  }
  return UDM_OK;
}


static udm_rc_t
UdmMySQLFetchRow(UDM_SQL *db, UDM_SQLRES *R, UDM_STR *buf)
{
  size_t j;
  MYSQL_ROW row;
  unsigned long *lengths;
  if (!(row= mysql_fetch_row((MYSQL_RES *)R->specific)))
    return UDM_ERROR;
  lengths= mysql_fetch_lengths((MYSQL_RES *)R->specific);
  
  for (j = 0; j < R->nCols; j++)
  {
    buf[j].str= row[j];
    buf[j].length= lengths[j];
  }
  
  return(UDM_OK);
}


#ifdef HAVE_MYSQL_PS
static void
UdmMySQLPrintStmtError(UDM_SQL *db, const char *func)
{
  UDM_MYSQL *mydb= UdmMySQL(db);
  db->errcode= 1;
  udm_snprintf(db->errstr, sizeof(db->errstr), "%s failed: %s",
               func,  mysql_stmt_error(mydb->stmt.stmt));
}


static udm_rc_t
UdmMySQLPrepare(UDM_SQL *db, const char *query)
{
  UDM_MYSQL *mydb= UdmMySQL(db);
  db->errcode= 0;
  db->errstr[0]= '\0';
  if (!(mydb->stmt.stmt= mysql_stmt_init(&mydb->mysql)))
  {
    db->errcode= 1;
    sprintf(db->errstr, "mysql_stmt_init() failed. Out of memory?");
    return UDM_ERROR;
  }
  if (mysql_stmt_prepare(mydb->stmt.stmt, query, strlen(query)))
  {
    UdmMySQLPrintStmtError(db, "mysql_stmt_prepare()");
    return UDM_ERROR;
  }
  return UDM_OK;
}


static inline enum enum_field_types
UdmSQLType2MySQLType(udm_sqltype_t type)
{
  return type == UDM_SQLTYPE_INT32 ? MYSQL_TYPE_LONG : MYSQL_TYPE_STRING;
}


#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-qual"
static udm_rc_t
UdmMySQLBind(UDM_SQL *db, int position, const void *data, int size, udm_sqltype_t type)
{
  UDM_MYSQL *mydb= UdmMySQL(db);
  MYSQL_BIND *b= &mydb->stmt.bind[position - 1];
  b->buffer_type= UdmSQLType2MySQLType(type);
  b->buffer= (char*) data;
  b->is_null= 0;
  b->length= &mydb->stmt.bind_length[position - 1];
  b->length[0]= (unsigned long) size;
  return UDM_OK;
}
#pragma GCC diagnostic pop


static udm_rc_t
UdmMySQLStmtFree(UDM_SQL *db)
{
  UDM_MYSQL *mydb= UdmMySQL(db);
  if (mysql_stmt_close(mydb->stmt.stmt))
  {
    UdmMySQLPrintStmtError(db, "mysql_stmt_close()");
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
UdmMySQLExec(UDM_SQL *db)
{
  UDM_MYSQL *mydb= UdmMySQL(db);

  if (mysql_stmt_bind_param(mydb->stmt.stmt, mydb->stmt.bind))
  {
    UdmMySQLPrintStmtError(db, "mysql_stmt_bind_param()");
    return UDM_ERROR;
  }
  
  if (mysql_stmt_execute(mydb->stmt.stmt))
  {
    UdmMySQLPrintStmtError(db, "mysql_stmt_execute()");
    return UDM_ERROR;
  }

  return UDM_OK;
}
#endif


static udm_rc_t
UdmMySQLRenameTable(UDM_SQL *db, const char *from, const char *to)
{
  char buf[256];
  udm_snprintf(buf, sizeof(buf), "ALTER TABLE %s RENAME TO %s", from, to);
  return UdmSQLExecDirect(db, NULL, buf);
}


static udm_rc_t
UdmMySQLCopyStructure(UDM_SQL *db, const char *from, const char *to)
{
  char buf[256];
  udm_snprintf(buf, sizeof(buf),
              "CREATE TABLE %s ENGINE=MyISAM MAX_ROWS=300000000 AVG_ROW_LENGTH=512 "
              "SELECT * FROM %s LIMIT 0", to, from);
  return UdmSQLExecDirect(db, NULL, buf);
}


static udm_rc_t
UdmMySQLLockOrBegin(UDM_SQL *db, const char *param)
{
  char buf[128];
  udm_snprintf(buf, sizeof(buf), "LOCK TABLE %s", param);
  return UdmSQLExecDirect(db, NULL, buf);
}


static udm_rc_t
UdmMySQLUnlockOrCommit(UDM_SQL *db)
{
  return UdmSQLExecDirect(db, NULL, "UNLOCK TABLES");
}


const UDM_SQLDB_HANDLER udm_sqldb_mysql_handler =
{
  UdmMySQLEscStr,
  UdmMySQLQuery,
  UdmMySQLConnect,
  UdmMySQLClose,
  UdmMySQLBegin,
  UdmMySQLCommit,
#ifdef HAVE_MYSQL_PS
  UdmMySQLPrepare,
  UdmMySQLBind,
  UdmMySQLExec,
  UdmMySQLStmtFree,
#else
  UdmSQLPrepareGeneric,
  UdmSQLBindGeneric,
  UdmSQLExecGeneric,
  UdmSQLStmtFreeGeneric,
#endif
  UdmMySQLFetchRow,
  UdmSQLStoreResultSimple,
  UdmMySQLFreeResult,
  UdmMySQLExecDirect,
  UdmMySQLRenameTable,
  UdmMySQLCopyStructure,
  UdmMySQLLockOrBegin,
  UdmMySQLUnlockOrCommit,
};
#endif
