/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#if (HAVE_ODBC)


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "udm_common.h"
#include "udm_sqldbms.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_xmalloc.h"

#if HAVE_IODBC
#include <sql.h>
#include <sqlext.h>
#endif

#if HAVE_EASYSOFT
#include <sql.h>
#include <sqlext.h>
#endif

#if HAVE_VIRT
#include <iodbc.h>
#include <isql.h>
#include <isqlext.h>
#endif

#if HAVE_UNIXODBC
#include <sql.h>
#include <sqlext.h>
#endif

#if HAVE_SAPDB
#include <WINDOWS.H>
#include <sql.h>
#include <sqlext.h>
#endif

#if HAVE_SOLID
#include <cli0cli.h>
#endif

#if HAVE_DB2
#include <sqlcli1.h>
#endif

#ifdef WIN32
#include <process.h>
#endif


/************************** ODBC ***********************************/
#if HAVE_DB2
typedef SQLHANDLE HENV;
typedef SQLHANDLE HDBC;
typedef SQLHANDLE HSTMT;
#endif

typedef struct udm_odbc_bind_st
{
  SQLLEN size;
  SQLPOINTER data;
#if defined(__x86_64__)
  udm_uint8 int8buf;
#endif
} UDM_ODBC_BINDBUF;


typedef struct udm_odbc_stmt_st
{
  HSTMT hstmt;
  UDM_ODBC_BINDBUF BindBuf[UDM_SQL_MAX_BIND_PARAM];
} UDM_ODBC_STMT;


typedef struct udm_odbc_conn_st
{
  HDBC  hDbc;
  HENV  hEnv;
  UDM_ODBC_STMT stmt;
} UDM_ODBC_CONN;


typedef struct udm_odbc_field_st
{
  SQLCHAR     Name[32];
  SQLSMALLINT NameLen;
  SQLSMALLINT Type;
  SQLULEN     Size;
  SQLSMALLINT Scale;
  SQLSMALLINT Nullable;
} UDM_ODBC_FIELD;



static inline int
SQL_OK(int rc)
{
  return rc == SQL_SUCCESS || rc == SQL_SUCCESS_WITH_INFO;
}


static UDM_ODBC_CONN*
UdmOB(UDM_SQL *db)
{
  return (UDM_ODBC_CONN*) db->specific;
}


static void
UdmSetErrorCode(UDM_SQL *db, int errcode)
{
  db->errcode= errcode;
}


static void
UdmODBCDisplayError(UDM_SQL *db, HSTMT *hstmt, const char *prefix)
{
  UCHAR szSqlState[10];
  UCHAR szErrMsg[400];
  SDWORD naterr;
  SWORD length;
  RETCODE rc=SQL_SUCCESS;
  size_t len= strlen(prefix);
  UDM_ODBC_CONN *sdb= UdmOB(db);

  strcpy(db->errstr, prefix);
  
  while (1)
  {   
    rc= SQLError(sdb->hEnv, sdb->hDbc, hstmt, szSqlState,
                 &naterr,szErrMsg,sizeof(szErrMsg)-1,&length);
    if ((SQL_SUCCESS != rc) && (SQL_SUCCESS_WITH_INFO != rc))
      break;
    szErrMsg[sizeof(szErrMsg)-1]='\0';
    len+= sprintf(db->errstr+len, "[SQLSTATE:%s]%s",szSqlState,szErrMsg);
  }
}


static int
UdmODBCTypeIsBinary(int Type)
{
  return  Type == SQL_BINARY ||
          Type == SQL_VARBINARY ||
          Type == SQL_LONGVARBINARY ||
          Type == -98/* DB2 SQL_BLOB */;
}


static int
IsHexEncodedBinary(UDM_SQL *db, int SourceType, int TargetType)
{
  /*
    MonetDB does not encode binary values to HEX  when bind type is SQL_C_CHAR
  */
  return UdmODBCTypeIsBinary(SourceType) &&
         TargetType != SQL_C_BINARY &&
         db->DBType != UDM_DB_MONETDB;
}


static SQLSMALLINT
TargetFetchType(UDM_SQL *db, SQLSMALLINT Type)
{
  if (Type == SQL_FLOAT)
    return SQL_C_FLOAT;

  if (Type == SQL_DOUBLE)
    return SQL_C_DOUBLE;

  if (UdmODBCTypeIsBinary(Type))
  {
    switch (db->DBType)
    {
    case UDM_DB_PGSQL:       /* bytea */
    case UDM_DB_MSSQL:       /* blob  */
    case UDM_DB_IBASE:       /* blob  */
    case UDM_DB_ORACLE8:     /* blob  */
    case UDM_DB_SYBASE:      /* image/varbinary */
    case UDM_DB_DB2:
    case UDM_DB_MYSQL:
      return SQL_C_BINARY;
    case UDM_DB_SQLITE:
    case UDM_DB_SQLITE3:
    case UDM_DB_SOLID:
    case UDM_DB_VIRT:
    case UDM_DB_SAPDB:
    case UDM_DB_ACCESS:
    case UDM_DB_MIMER:
    case UDM_DB_CACHE:
    case UDM_DB_MONETDB:
      /*
        SQLGetData as SQL_C_BINARY does not work with MonetDB:
        "Restricted data type attribute violation" error is returned
      */
      return SQL_CHAR;
    }
  }

#ifdef HAVE_UNIXODBC
  /*
     For some weird reasons, Sybase native (i.e. not FreeTDS)
     ODBC driver in conjunction with unixODBC always returns 4
     in pcbValue if get_type is SQL_CHAR. Let's fetch as SQL_C_SLONG.
  */
  if (Type == SQL_INTEGER && db->DBType == UDM_DB_SYBASE)
    return SQL_C_SLONG;
#endif

  if (Type  == SQL_WLONGVARCHAR && db->DBType == UDM_DB_MYSQL)
  {
    /*
      Workaround for a bug in MyODBC driver.
      It reports "TEXT BINARY" as SQL_WLONGVARCHAR.
    */
    return SQL_C_BINARY;
  }
  return SQL_CHAR;
}


static udm_rc_t
execDB(UDM_SQL *db, UDM_SQLRES *result, const char *sqlstr)
{
  RETCODE dbrc;
  SWORD iResColumns;
  SQLLEN iRowCount;
  static char  bindbuf[(int)(64L * 1024L - 16L)];
  UDM_ODBC_CONN *sdb= UdmOB(db);
  HSTMT hstmt;
  udm_rc_t rc= UDM_OK;

  if(!strcmp(sqlstr,"COMMIT"))
  {
    if (!SQL_OK(dbrc= SQLTransact(sdb->hEnv, sdb->hDbc, SQL_COMMIT)))
    {
      UdmSetErrorCode(db, 1);
      UdmODBCDisplayError(db, SQL_NULL_HSTMT, "SQLTransact:");
      return UDM_ERROR;
    }
    else
    {
      UdmSetErrorCode(db, 0);
      return UDM_OK;
    }
  }

  if (!SQL_OK(dbrc= SQLAllocStmt(sdb->hDbc, &hstmt)))
  {
    UdmSetErrorCode(db, 1);
    UdmODBCDisplayError(db, SQL_NULL_HSTMT, "SQLAllocStmt:");
    return UDM_ERROR;
  }

#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-qual"
#endif
  if (!SQL_OK(dbrc= SQLExecDirect(hstmt, (SQLCHAR *)sqlstr, SQL_NTS)))
  {
    if (dbrc == SQL_NO_DATA)
      goto ex;
    UdmODBCDisplayError(db, hstmt, "SQLExecDirect:");
    rc= UDM_ERROR;
    goto ex;
  }
#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic pop
#endif

  if (!SQL_OK(dbrc= SQLNumResultCols(hstmt, &iResColumns)))
  {
    UdmODBCDisplayError(db, hstmt, "SQLNumResultCols:");
    rc= UDM_ERROR;
    goto ex;
  }
  
  if (!iResColumns)
  {
    if (!SQL_OK(dbrc= SQLRowCount(hstmt, &iRowCount)))
    {
      UdmODBCDisplayError(db, hstmt, "SQLRowCount:");
      rc= UDM_ERROR;
      goto ex;
    }
  }
  else
  {
    size_t res_count;
    int col;
    size_t mRows;

    UDM_ODBC_FIELD Field[128];

    result->nRows= 0;
    result->nCols= iResColumns;
    result->Items= NULL;

    result->Fields=(UDM_SQLFIELD*)UdmMalloc(result->nCols*sizeof(UDM_SQLFIELD));
    bzero(result->Fields, result->nCols*sizeof(UDM_SQLFIELD));

    for(col=0; col < iResColumns; col++)
    {
      SQLDescribeCol(hstmt, col+1,
                     Field[col].Name, 
                     (SQLSMALLINT) sizeof(Field[col].Name) - 1,
                     &Field[col].NameLen,
                     &Field[col].Type,
                     &Field[col].Size,
                     &Field[col].Scale,
                     &Field[col].Nullable);
      result->Fields[col].sqlname= (char*)UdmStrdup((const char*)Field[col].Name);
    }

    dbrc= SQL_NO_DATA_FOUND;
    for (mRows=0, res_count=0;
         (db->res_limit ? (res_count < db->res_limit) : 1);
         res_count++)
    {
      int i;
      
      if (!SQL_OK(dbrc= SQLFetch(hstmt)))
      {
        if (dbrc != SQL_NO_DATA_FOUND)
        {
          UdmODBCDisplayError(db, hstmt, "SQLFetch:");
          if (db->DBType == UDM_DB_DB2) /* FIXME: check with all databases */
          {            
            rc= UDM_ERROR;
            goto ex;
          }
        }
        break;
      }
      if (mRows <= (result->nRows+1) * iResColumns)
      {
        mRows+= 1024 * iResColumns;
        result->Items=(UDM_STR*)UdmXrealloc(result->Items,mRows*sizeof(UDM_STR));
      }

      for (i= 0; i < iResColumns; i++)
      {
        size_t offs=result->nRows*iResColumns+i;
        const char *p= "";
        char *tmp= NULL;
        SQLLEN  pcbValue= 0; /* DB2 Linux driver needs 0 */
        SQLSMALLINT get_type= TargetFetchType(db, Field[i].Type);

        dbrc= SQLGetData(hstmt,i+1,get_type,bindbuf,sizeof(bindbuf),&pcbValue);
        if (dbrc == SQL_SUCCESS_WITH_INFO)
        {
          if (pcbValue <= 0)
          {
            udm_snprintf(db->errstr, sizeof(db->errstr),
                         "execDB: Unexpected pcbValue %d", (int) pcbValue);
            return UDM_ERROR;
          }

          if (pcbValue > (SQLLEN) sizeof(bindbuf))
          {
            char *dst;
            SQLLEN tmp_pcbValue, remainingLength;
            SQLLEN copiedLength= sizeof(bindbuf);
            int reserved_for_null= 0;

            if (IsHexEncodedBinary(db, Field[i].Type, get_type))
            {
              int null_length= bindbuf[copiedLength - 2] == '\0' ? 2 :
                               bindbuf[copiedLength - 1] == '\0' ? 1 : 0;
              copiedLength-= null_length;
              reserved_for_null= null_length > 0 ? 1 : 0;
            }

            /*
               When fetching hex encoded binary data,
               need extra byte for the null terminator.
            */
            if (!(p= tmp= (char*) UdmMalloc((pcbValue + reserved_for_null))))
            {
              p= "";
              pcbValue= 0;
              goto decode;
            }
            memcpy(tmp, bindbuf, copiedLength);
            dst= tmp + copiedLength;
            remainingLength= pcbValue - copiedLength;

            dbrc= SQLGetData(hstmt, i + 1, get_type, dst,
                             remainingLength + reserved_for_null,
                             &tmp_pcbValue);
            if (dbrc != SQL_SUCCESS || tmp_pcbValue <= 0)
            {
              udm_snprintf(db->errstr, sizeof(db->errstr),
                           "execDB: SQLGetData returned unexpected code (%d) or pcbValue (%d)",
                           dbrc, (int) tmp_pcbValue);
              return UDM_ERROR;
            }
          }
          else
          {
            p= "";
            pcbValue= 0;
          }
        }
        else if (dbrc == SQL_SUCCESS)
        {
          if (pcbValue == SQL_NULL_DATA)
          {
            bindbuf[0]= '\0';
            pcbValue= 0;
          }
          else if (get_type == SQL_C_FLOAT)
          {
            float num;
            memcpy(&num, bindbuf, sizeof(num));
            pcbValue= sprintf(bindbuf, "%f", num);
          }
          else if (get_type == SQL_C_DOUBLE)
          {
            double num;
            memcpy(&num, bindbuf, sizeof(num));
            pcbValue= sprintf(bindbuf, "%f", num);
          }
          else if (get_type == SQL_C_SLONG)
          {
            long int num;
            memcpy(&num, bindbuf, sizeof(num));
            pcbValue= sprintf(bindbuf, "%ld", num);
          }
          p= bindbuf;
        }
        else if (dbrc == SQL_NO_DATA)
        {
          /* MSSQL returns this in case of an empty IMAGE value */
          /* Do nothing */
        }
        else if (UdmODBCTypeIsBinary(Field[i].Type) &&
                 dbrc == SQL_ERROR &&
                 db->DBType == UDM_DB_MONETDB)
        {
          /* SQLGetData returns SQL_ERROR on empty BLOB for MonetDB */
          bindbuf[0]= '\0';
          pcbValue= 0;
        }
        else
        {
          UdmODBCDisplayError(db, hstmt, "Unknown data type:");
          rc= UDM_ERROR;
          goto ex;
        }

decode:
        if (IsHexEncodedBinary(db, Field[i].Type, get_type))
        {
          DecodeHexStr(p, &result->Items[offs], pcbValue);
        }
        else
        {
          result->Items[offs].str= (char*)malloc(pcbValue+1);
          memcpy(result->Items[offs].str, p, pcbValue);
          result->Items[offs].str[pcbValue]='\0';
          result->Items[offs].length= pcbValue;
        }
        if (tmp)
          UdmFree(tmp);
      }
      result->nRows++;
    }
  }

ex:
  SQLFreeStmt(hstmt, SQL_DROP);
  db->res_limit=0;
  db->errcode= (rc == UDM_OK) ? 0 : 1;
  return rc;
}


static udm_rc_t
UdmODBCGetVersion(UDM_SQL *db)
{
  UDM_ODBC_CONN *sdb= UdmOB(db);
  char version[128];
  SQLSMALLINT DummyLen;
  SQLRETURN rc= SQLGetInfo(sdb->hDbc, SQL_DBMS_VER,
                           version, sizeof(version), &DummyLen);
  if (SQL_OK(rc))
  {
    version[sizeof(version) - 1]= '\0';
    if (!strncmp(version, "PostgreSQL ", 11))
    {
      int a, b, c;
      if (3 == sscanf(version + 11, "%d.%d.%d", &a, &b, &c))
        db->version= a * 10000 + b * 100 + c;
    }
    return UDM_OK;
  }
  db->version= 0;
  return UDM_ERROR;
}


static udm_rc_t
UdmODBCConnect(UDM_SQL *db)
{
  char DSN[512]="";
  const char* DBUser= UdmVarListFindStr(&db->Vars,"DBUser",NULL);
  const char* DBPass= UdmVarListFindStr(&db->Vars,"DBPass",NULL);
  const char* DBName= UdmVarListFindStr(&db->Vars, "DBName", "mnogosearch");
#if HAVE_SOLID || HAVE_SAPDB
  const char* DBHost= UdmVarListFindStr(&db->Vars, "DBHost", "localhost");
  int DBPort= UdmVarListFindInt(&db->Vars, "DBPort", 0);
#endif
  UDM_SQLRES SQLRes;
  UDM_ODBC_CONN *sdb= (UDM_ODBC_CONN*) UdmMalloc(sizeof(UDM_ODBC_CONN));
  db->specific= (void*) sdb;

  sdb->hDbc=  SQL_NULL_HDBC;
  sdb->hEnv=  SQL_NULL_HENV;
  sdb->stmt.hstmt= SQL_NULL_HSTMT;

#if (HAVE_SOLID)
  udm_snprintf(DSN, sizeof(DSN)-1, "tcp %s %d", DBHost, DBPort?DBPort:1313);
#elif (HAVE_SAPDB)
  udm_snprintf(DSN, sizeof(DSN)-1, "%s:%s", DBHost, DBName ? DBName : "");
#else
  strncpy(DSN, DBName ? DBName : "", sizeof(DSN)-1);
#endif

#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-qual"  
#endif
  if (SQL_SUCCESS != (db->errcode= SQLAllocEnv(&sdb->hEnv)))
    return UDM_ERROR;
  if (SQL_SUCCESS != (db->errcode= SQLAllocConnect(sdb->hEnv, &sdb->hDbc)))
    return UDM_ERROR;
  if (SQL_SUCCESS != (db->errcode= SQLSetConnectOption(sdb->hDbc, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_ON)))
    return UDM_ERROR;
  if (!SQL_OK(db->errcode = SQLConnect(sdb->hDbc,
                                       (SQLCHAR *) DSN, SQL_NTS,
                                       (SQLCHAR *) DBUser, SQL_NTS,
                                       (SQLCHAR*) DBPass, SQL_NTS)))
    return UDM_ERROR;
#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic pop
#endif

  UdmSetErrorCode(db, 0);
  db->connected= UDM_TRUE;

  /*
  if (db->DBType == UDM_DB_DB2)
  {
    1253 = SQL_ATTR_LONGDATA_COMPAT
    int rc= SQLSetConnectAttr(sdb->hDbc, 1253, (SQLPOINTER) 1, 0);
  }
  */
  if (db->DBType == UDM_DB_ORACLE8)
    execDB(db, &SQLRes, "ALTER SESSION SET NLS_NUMERIC_CHARACTERS='. '");

  /*
    MySQL ODBC Driver does not allow to do SET NAMES
  */
  if (/*db->DBType == UDM_DB_MYSQL || */db->DBType == UDM_DB_PGSQL)
  {
    const char* setnames= UdmVarListFindStr(&db->Vars, "setnames", NULL);
    if (setnames)
    {
      char qbuf[64];
      udm_snprintf(qbuf, sizeof(qbuf), "SET NAMES '%s'", setnames);
      if (UDM_OK != execDB(db, &SQLRes, qbuf))
        return UDM_ERROR;;
    }
  }
  
  UdmODBCGetVersion(db);
  return UDM_OK;
}


static udm_rc_t
UdmODBCClose(UDM_SQL *db)
{
  if (db->connected)
  {
    UDM_ODBC_CONN *sdb= UdmOB(db);
    db->errcode= SQLTransact(sdb->hEnv, sdb->hDbc, SQL_COMMIT);
    if(SQL_SUCCESS != db->errcode)
      goto ret;
    db->errcode= SQLDisconnect(sdb->hDbc );
    if(SQL_SUCCESS != db->errcode)
      goto ret;
    db->errcode= SQLFreeConnect(sdb->hDbc);
    if(SQL_SUCCESS != db->errcode)
      goto ret;
    db->errcode= SQLFreeEnv(sdb->hEnv);
    if(SQL_SUCCESS != db->errcode)
      goto ret;
ret:
    db->connected= UDM_FALSE;
    UDM_FREE(db->specific);
  }
  return UDM_OK;
}


static udm_rc_t
UdmODBCExecDirect(UDM_SQL *db,UDM_SQLRES *res, const char *qbuf)
{
  udm_rc_t rc;
   
  if (res)
  {
    bzero((void*) res, sizeof(UDM_SQLRES));
    res->db= db;
  }

  if (!db->connected)
  {
    UdmODBCConnect(db);
    if (db->errcode)
    {
      if (!db->errstr[0])
        UdmODBCDisplayError(db, SQL_NULL_HSTMT, "");
      return UDM_ERROR;
    }
    else
    {
      db->connected= UDM_TRUE;
    }
  }

  rc= execDB(db,res,qbuf);
  if ((db->errcode))
  {
    if(strstr(db->errstr,"[SQLSTATE:23000]")) /* MS Access */
    {
      UdmSetErrorCode(db, 0);
      rc= UDM_OK;
    }
#if 0
    /* Cannot use S1000: it's returned by unixODBC in many cases */
    else if(strstr(db->errstr,"[SQLSTATE:S1000]"))
    { /* Oracle 8i */
      UdmSetErrorCode(db, 0);
      rc= UDM_OK;
    }
#endif
    else if(strstr(db->errstr,"uplicat")) /* PgSQL,MySQL*/
    {
      UdmSetErrorCode(db, 0);
      rc= UDM_OK;
    }
    else if(strstr(db->errstr,"nique")) /* Solid, Virtuoso */
    {
      UdmSetErrorCode(db, 0);
      rc= UDM_OK;
    }
    else if(strstr(db->errstr,"UNIQUE")) /* Mimer */
    {
      UdmSetErrorCode(db, 0);
      rc= UDM_OK;
    }
    else if (strstr(db->errstr, "PRIMARY KEY")) /* MonetDB */
    {
      UdmSetErrorCode(db, 0);
      rc= UDM_OK;
    }
    else if (db->DBType == UDM_DB_MONETDB &&
             strstr(db->errstr, "General error"))
    {
      /*
        MonetDB returns "General error" in case of primary key
        violation in some cases.
      */
      fprintf(stderr, "MONETDB FIXME: %s\n%s\n", db->errstr, qbuf);
      UdmSetErrorCode(db, 0);
      rc= UDM_OK;
    }
    else
    {
      UdmSetErrorCode(db, 1);
    }
  }
  return rc;
}


static udm_rc_t
UdmODBCBegin(UDM_SQL *db)
{
  UDM_ODBC_CONN *sdb;
  if (!db->connected)
  {
    UdmODBCConnect(db);
    if(db->errcode)
    {
      UdmODBCDisplayError(db, SQL_NULL_HSTMT, "");
      return UDM_ERROR;
    }
    else
    {
      db->connected= UDM_TRUE;
    }
  }
  sdb= UdmOB(db);
  db->errcode= SQLSetConnectOption(sdb->hDbc, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF);
  if (SQL_SUCCESS != db->errcode)
  {
    UdmODBCDisplayError(db, SQL_NULL_HSTMT, "SQLSetConnectOption(SQL_AUTOCOMMIT_OFF) failed");
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
UdmODBCCommit(UDM_SQL *db)
{
  UDM_ODBC_CONN *sdb= UdmOB(db);
  udm_rc_t rc;
  if (UDM_OK != (rc= UdmODBCExecDirect(db,NULL,"COMMIT")))
    return rc;
  if (!SQL_OK(db->errcode= SQLSetConnectOption(sdb->hDbc,
                                               SQL_AUTOCOMMIT,
                                               SQL_AUTOCOMMIT_ON)))
  {
    UdmODBCDisplayError(db, SQL_NULL_HSTMT, "SQLSetConnectOption(SQL_AUTOCOMMIT_ON) failed");
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
UdmODBCPrepare(UDM_SQL *db, const char *query)
{
  UDM_ODBC_CONN *sdb= UdmOB(db);
  UDM_ODBC_STMT *stmt= &sdb->stmt;
  SQLRETURN dbrc;
  UdmSetErrorCode(db, 0);
  db->errstr[0] = 0;
  dbrc= SQLAllocStmt(sdb->hDbc, &stmt->hstmt);

  if (!SQL_OK(dbrc))
  {
    UdmSetErrorCode(db, dbrc);
    UdmODBCDisplayError(db, SQL_NULL_HSTMT, "SQLAllocStmt: ");
    return(UDM_ERROR);
  }
#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-qual"  
#endif
  dbrc= SQLPrepare(stmt->hstmt, (SQLCHAR*) query, SQL_NTS);
#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic pop
#endif
  if (!SQL_OK(dbrc))
  {
    UdmSetErrorCode(db, dbrc);
    UdmODBCDisplayError(db, stmt->hstmt, "SQLPrepare: ");
    return(UDM_ERROR);
  }
  return(UDM_OK);
}


static int
UdmSQLType2ODBCType(udm_sqltype_t udm_type)
{
  switch (udm_type)
  {
    case UDM_SQLTYPE_LONGVARBINARY: return SQL_LONGVARBINARY;
    case UDM_SQLTYPE_LONGVARCHAR  : return SQL_LONGVARCHAR;
    case UDM_SQLTYPE_VARCHAR      : return SQL_VARCHAR;
    case UDM_SQLTYPE_INT32        : return SQL_INTEGER;
    default                       : break;
  }
  return SQL_UNKNOWN_TYPE;
}


static int
UdmSQLParamLen2ODBC(int udm_len)
{
  switch (udm_len)
  {
    case UDM_SQL_NULL_DATA    : return SQL_NULL_DATA;
    case UDM_SQL_DATA_AT_EXEC : return SQL_DATA_AT_EXEC;
    case UDM_SQL_NTS          : return SQL_NTS;
  }
  return udm_len;
}


static udm_rc_t
UdmODBCBind(UDM_SQL *db, int position, const void *data, int size, udm_sqltype_t type)
{
  UDM_ODBC_CONN *sdb= UdmOB(db);
  UDM_ODBC_STMT *stmt= &sdb->stmt;
  SQLRETURN dbrc;
  int ctype= SQL_C_DEFAULT;
  int odbctype= UdmSQLType2ODBCType(type);
  stmt->BindBuf[position].size= UdmSQLParamLen2ODBC(size);
#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-qual"
#endif
  stmt->BindBuf[position].data= (SQLPOINTER) data;
#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic pop
#endif
  UdmSetErrorCode(db, 0);
  db->errstr[0] = 0;

#if defined(__x86_64__)
  /* Virtuoso ODBC driver wants a 64-bit number when binding SQL_INTEGER */
  if (type == UDM_SQLTYPE_INT32 && db->DBType == UDM_DB_VIRT)
  {
    stmt->BindBuf[position].int8buf= *((const int *) data);
    stmt->BindBuf[position].data= (SQLPOINTER) &stmt->BindBuf[position].int8buf;
    data= (SQLPOINTER) &stmt->BindBuf[position].int8buf;
  }
#endif

#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-qual" 
#endif
  dbrc= SQLBindParameter(stmt->hstmt, position, SQL_PARAM_INPUT, ctype,
                         odbctype, 0 /* Column size */, 0 /* Decimal digits */,
                         /*&BindBuf[position]*/ (char*) data, size < 0 ? 0 : size,
                         &stmt->BindBuf[position].size);
#ifdef HAVE_GCC_PRAGMA_PUSH
#pragma GCC diagnostic pop
#endif

#ifdef WIN32  
/* s_size= SQL_DATA_AT_EXEC; */
#endif

  if (!SQL_OK(dbrc))
  {
    UdmSetErrorCode(db, dbrc);
    UdmODBCDisplayError(db, stmt->hstmt, "UdmSQLBindParameter: ");
    return(UDM_ERROR);
  }
  return(UDM_OK);
}


static udm_rc_t
UdmODBCStmtFree(UDM_SQL *db)
{
  UDM_ODBC_CONN *sdb= UdmOB(db);
  UDM_ODBC_STMT *stmt= &sdb->stmt;
  int rc= SQLFreeStmt(stmt->hstmt, SQL_DROP);
  return SQL_OK(rc) ? UDM_OK : UDM_ERROR;
}


static udm_rc_t
UdmODBCExec(UDM_SQL *db)
{
  UDM_ODBC_CONN *sdb= UdmOB(db);
  UDM_ODBC_STMT *stmt= &sdb->stmt;
  SQLRETURN dbrc;
  int put_data= 0;
  UDM_ODBC_BINDBUF *Buf;

  UdmSetErrorCode(db, 0);
  db->errstr[0]= 0;
  dbrc= SQLExecute(stmt->hstmt);

#ifdef WIN32
  put_data= 1;
#endif

  if (!put_data || dbrc != SQL_NEED_DATA)
    goto check_err;

  for (dbrc= SQLParamData(stmt->hstmt, (SQLPOINTER*) &Buf);
       dbrc == SQL_NEED_DATA;
       dbrc= SQLParamData(stmt->hstmt, (SQLPOINTER*) &Buf))
  {
    if (SQL_ERROR == (dbrc= SQLPutData(stmt->hstmt, Buf->data, Buf->size)))
      goto check_err;
  }

check_err:

  if (!SQL_OK(dbrc))
  {
    UdmODBCDisplayError(db, stmt->hstmt, "UdmODBCExec: ");
    UdmSetErrorCode(db, dbrc);
    return UDM_ERROR;
  }
  return UDM_OK;
}


static udm_rc_t
UdmODBCRenameTable(UDM_SQL *db, const char *from, const char *to)
{
  if (db->DBType == UDM_DB_MYSQL ||
      db->DBType == UDM_DB_PGSQL ||
      db->DBType == UDM_DB_ORACLE8)
  {
    char buf[256];
    udm_snprintf(buf, sizeof(buf), "ALTER TABLE %s RENAME TO %s", from, to);
    return UdmSQLExecDirect(db, NULL, buf);
  }
  else if (db->DBType == UDM_DB_DB2)
  {
    char buf[256];
    udm_snprintf(buf, sizeof(buf), "RENAME TABLE %s TO %s", from, to);
    return UdmSQLExecDirect(db, NULL, buf);
  }
  else if (db->DBType == UDM_DB_MSSQL)
  {
    char buf[256];
    udm_snprintf(buf, sizeof(buf), "sp_rename %s, %s", from, to);
    return UdmSQLExecDirect(db, NULL, buf);
  }
  sprintf(db->errstr, "This database type does not support RENAME TABLE");
  return UDM_ERROR;
}


/*
  Copy table structure from another table, without indexes.
*/
static udm_rc_t
UdmODBCCopyStructure(UDM_SQL *db, const char *from, const char *to)
{
  char buf[256];
  switch (db->DBType)
  {
    case UDM_DB_MYSQL:
      udm_snprintf(buf, sizeof(buf),
                   "CREATE TABLE %s MAX_ROWS=300000000 AVG_ROW_LENGTH=512 "
                   "SELECT * FROM %s LIMIT 0", to, from);
      return UdmSQLExecDirect(db, NULL, buf);

    case UDM_DB_PGSQL:
      udm_snprintf(buf, sizeof(buf), "CREATE TABLE %s (LIKE %s)", to, from);
      return UdmSQLExecDirect(db, NULL, buf);

    case UDM_DB_ORACLE8:
      udm_snprintf(buf, sizeof(buf),
                   "CREATE TABLE %s AS SELECT * FROM %s WHERE 1=0", to, from);
      return UdmSQLExecDirect(db, NULL, buf);

    case UDM_DB_DB2:
      udm_snprintf(buf, sizeof(buf), "CREATE TABLE %s LIKE %s", to, from);
      return UdmSQLExecDirect(db, NULL, buf);

    case UDM_DB_MSSQL:
      udm_snprintf(buf, sizeof(buf), "SELECT TOP 0 * INTO %s FROM %s", to, from);
      return UdmSQLExecDirect(db, NULL, buf);
    default: break;
  }
  sprintf(db->errstr, "This database type does not support"
          "<CREATE TABLE name LIKE> or <CREATE TABLE name AS SELECT>");
  return UDM_ERROR;
}


static udm_rc_t
UdmODBCLockOrBegin(UDM_SQL *db, const char *param)
{
  if (db->DBType == UDM_DB_MYSQL)
  {
    char buf[128];
    udm_snprintf(buf, sizeof(buf), "LOCK TABLE %s", param);
    return UdmSQLExecDirect(db, NULL, buf);
  }
  return UdmSQLBegin(db);
}


static udm_rc_t
UdmODBCUnlockOrCommit(UDM_SQL *db)
{
  if (db->DBType == UDM_DB_MYSQL)
    return UdmSQLExecDirect(db, NULL, "UNLOCK TABLES");
  else
    return UdmSQLCommit(db);
}


const UDM_SQLDB_HANDLER udm_sqldb_odbc_handler =
{
  UdmSQLEscStrGeneric,
  UdmODBCExecDirect,
  UdmODBCConnect,
  UdmODBCClose,
  UdmODBCBegin,
  UdmODBCCommit,
  UdmODBCPrepare,
  UdmODBCBind,
  UdmODBCExec,
  UdmODBCStmtFree,
  UdmSQLFetchRowSimple,
  UdmSQLStoreResultSimple,
  UdmSQLFreeResultSimple,
  UdmODBCExecDirect,
  UdmODBCRenameTable,
  UdmODBCCopyStructure,
  UdmODBCLockOrBegin,
  UdmODBCUnlockOrCommit,
};

#endif /* HAVE_ODBC */

