/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#if defined(HAVE_SQLITE) || defined(HAVE_SQLITE3)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "udm_common.h"
#include "udm_sqldbms.h"
#include "udm_utils.h"
#include "udm_vars.h"
#include "udm_xmalloc.h"


#if HAVE_SQLITE
#include <sqlite.h>
#endif


#if HAVE_SQLITE3
#ifdef WIN32
#define HAVE_SQLITE3_AMALGAMATION
#endif
#ifdef HAVE_SQLITE3_AMALGAMATION
#define SQLITE_OMIT_LOAD_EXTENSION
#ifndef HAVE_DEBUG
#define NDEBUG 1
#endif
#include "sqlite3.c"
#else
#include <sqlite3.h>
#endif
#endif


#ifdef WIN32
#include <process.h>
#endif

/*********************************** SQLITE *********************/

#if (HAVE_SQLITE)

static udm_rc_t
UdmSQLiteConnect(UDM_SQL *db)
{
  char edbname[1024];
  char dbname[1024], *e, *errmsg;
  const char* DBName= UdmVarListFindStr(UdmSQLDBVars(db), "DBName", "mnogosearch");
  strncpy(edbname, DBName, sizeof(dbname));
  dbname[sizeof(edbname)-1]='\0';
  
  UdmUnescapeCGIQuery(dbname,edbname);
  /* Remove possible trailing slash */
  e= dbname+strlen(dbname);
  if (e>dbname && e[-1]=='/')
    e[-1]='\0';
  
  if (!(db->specific= (void*) sqlite_open(dbname, 0, &errmsg)))
  {
    sprintf(db->errstr, "sqlite driver: %s", errmsg ? errmsg : "<NOERROR>");
    UDM_FREE(errmsg);
    db->errcode=1;
    return UDM_ERROR;
  }
  db->connected= UDM_TRUE;
  sqlite_busy_timeout((struct sqlite*)(db->specific),30*1000);
  return UDM_OK;
}


static int
xCallBack(void *pArg, int argc, char **argv, char **name)
{
  UDM_SQLRES *res= (UDM_SQLRES*) pArg;
  int i;

  if (res->nCols == 0)
  {
    res->nCols= argc;
    for (i=0 ; i<argc; i++)
    {
      res->Fields=(UDM_SQLFIELD*)UdmMalloc(res->nCols*sizeof(UDM_SQLFIELD));
      bzero(res->Fields,res->nCols*sizeof(UDM_SQLFIELD));
      for (i=0 ; i < res->nCols; i++)
      {
        res->Fields[i].sqlname = (char*)UdmStrdup(name[i]);
        res->Fields[i].sqllen= 0;
        res->Fields[i].sqltype= 0;
      }
    }
  }
  res->nRows++;
  res->Items=(UDM_STR*)UdmRealloc(res->Items,res->nRows*res->nCols*sizeof(UDM_STR));
  
  for (i = 0; i < argc; i++)
  {
    size_t offs= (res->nRows-1)*res->nCols+(size_t)i;
    size_t len= strlen(argv[i]?argv[i]:"");
    res->Items[offs].len= len;
    res->Items[offs].val= (char*)UdmMalloc(len+1);
    memcpy(res->Items[offs].val,argv[i]?argv[i]:"",len+1);
  } 
  return 0;
}


static udm_rc_t
UdmSQLiteQuery(UDM_SQL *db, UDM_SQLRES *res, const char *q)
{
  char *errmsg;
  int dbrc;
  
  if (res)
  {
    bzero((void*) res, sizeof(UDM_SQLRES));
    res->db= db;
  }
  
#ifdef WIN32
  {
    const char* DBName= UdmVarListFindStr(UdmSQLDBVars(db), "DBName", "mnogosearch");
    if (DBName[0] != '\0' && DBName[1] != '\0' && DBName[2] == ':')
      UdmVarListReplaceStr(UdmSQLDBVars(db), "DBName", DBName + 1);
  }
#endif /* #ifdef WIN32 */

  db->errcode=0;
  db->errstr[0]='\0';
  if(!db->connected && UDM_OK != UdmSQLiteConnect(db))
    return UDM_ERROR;
  
  if ((dbrc= sqlite_exec((struct sqlite *)db->specific, q, xCallBack, (void*)res, &errmsg)))
  {
    sprintf(db->errstr,"sqlite driver: %s",errmsg ? errmsg : "<NOERROR>");
    sqlite_freemem(errmsg);
    if (!strstr(db->errstr,"unique"))
    {
      db->errcode=1;
      return UDM_ERROR;
    }
  }
  return UDM_OK;
}


static udm_rc_t
UdmSQLiteClose(UDM_SQL *db)
{
  sqlite_close((struct sqlite*)(db->specific));
  db->specific = NULL;
  return UDM_OK;
}


static udm_rc_t
UdmSQLiteBegin(UDM_SQL *db)
{
  return UdmSQLiteQuery(db,NULL,"BEGIN TRANSACTION");
}


static udm_rc_t
UdmSQLiteCommit(UDM_SQL *db)
{
  return UdmSQLiteQuery(db,NULL,"COMMIT");
}


UDM_SQLDB_HANDLER udm_sqldb_sqlite_handler =
{
  UdmSQLEscStrGeneric,
  UdmSQLiteQuery,
  UdmSQLiteConnect,
  UdmSQLiteClose,
  UdmSQLiteBegin,
  UdmSQLiteCommit,
  UdmSQLPrepareGeneric,
  UdmSQLBindGeneric,
  UdmSQLExecGeneric,
  UdmSQLStmtFreeGeneric,
  UdmSQLFetchRowSimple,
  UdmSQLStoreResultSimple,
  UdmSQLFreeResultSimple,
  UdmSQLiteQuery,
  NULL, /* RenameTable */
  UdmSQLLockOrBeginGeneric,
  UdmSQLUnlockOrCommitGeneric,
};

#endif  /* HAVE_SQLITE */


#if (HAVE_SQLITE3)


static void
UdmSetErrorCode(UDM_SQL *db, int errcode)
{
   db->errcode= errcode;
}


static sqlite3*
UdmSQLite3Conn(UDM_SQL *db)
{
  return (sqlite3*) db->specific;
}

static udm_rc_t
UdmSQLite3Connect(UDM_SQL *db)
{
  sqlite3 *ppDb;
  char edbname[1024];
  char dbname[1024], *e;
  const char* DBName= UdmVarListFindStr(&db->Vars, "DBName", "mnogosearch");
  strncpy(edbname,DBName,sizeof(dbname));
  dbname[sizeof(edbname)-1]='\0';
  
  UdmUnescapeCGIQuery(dbname,edbname);
  /* Remove possible trailing slash */
  e= dbname+strlen(dbname);
  if (e>dbname && e[-1]=='/')
    e[-1]='\0';
  
  if (SQLITE_OK != sqlite3_open(dbname, &ppDb))
  {
    const char *errmsg= sqlite3_errmsg(ppDb);
    udm_snprintf(db->errstr, sizeof(db->errstr),
                "sqlite3 driver: %s", errmsg ? errmsg : "<NOERROR>");
    UdmSetErrorCode(db, 1);
    sqlite3_close(ppDb);
    return UDM_ERROR;
  }
  db->specific= (void*) ppDb;
  db->connected= UDM_TRUE;
  sqlite3_busy_timeout(UdmSQLite3Conn(db), 30*1000);
  return UDM_OK;
}


#if 0
static int SQLite3CallBack(void *pArg, int argc, char **argv, char **name)
{
  UDM_SQLRES *res= (UDM_SQLRES*) pArg;
  int i;

  if (res->nCols == 0)
  {
    res->nCols= argc;
    for (i=0 ; i<argc; i++)
    {
      res->Fields=(UDM_SQLFIELD*)UdmMalloc(res->nCols*sizeof(UDM_SQLFIELD));
      bzero(res->Fields,res->nCols*sizeof(UDM_SQLFIELD));
      for (i=0 ; i < res->nCols; i++)
      {
        res->Fields[i].sqlname = (char*)UdmStrdup(name[i]);
        res->Fields[i].sqllen= 0;
        res->Fields[i].sqltype= 0;
      }
    }
  }

  res->nRows++;
  res->Items=(UDM_STR*)UdmRealloc(res->Items,res->nRows*res->nCols*sizeof(UDM_STR));
  
  for (i = 0; i < argc; i++)
  {
    size_t offs= (res->nRows-1)*res->nCols+(size_t)i;
    size_t len= strlen(argv[i]?argv[i]:"");
    res->Items[offs].len= len;
    res->Items[offs].val= (char*)UdmMalloc(len+1);
    memcpy(res->Items[offs].val,argv[i]?argv[i]:"",len+1);
  } 
  return 0;
}
#endif


static udm_rc_t
UdmSQLite3Query(UDM_SQL *db, UDM_SQLRES *res, const char *q)
{
  int dbrc;
  sqlite3_stmt *pStmt;
  const char *Tail;
  int columns= 0;
  
  if (res)
  {
    bzero((void*) res, sizeof(UDM_SQLRES));
    res->db= db;
  }
  
#ifdef WIN32
  /** Remove heading "/" from DBName (in case of absolute path) */
  {
    const char* DBName= UdmVarListFindStr(&db->Vars, "DBName", "mnogosearch");
    if (DBName[0] != '\0' && DBName[1] != '\0' && DBName[2] == ':')
      UdmVarListReplaceStr(&db->Vars, "DBName", DBName + 1);
  }
#endif /* #ifdef WIN32 */

  UdmSetErrorCode(db, 0);
  db->errstr[0]= '\0';

  if (!db->connected && UDM_OK != UdmSQLite3Connect(db))
    return UDM_ERROR;

  if ((dbrc= sqlite3_prepare(UdmSQLite3Conn(db), q, -1, &pStmt, &Tail)))
  {
    udm_snprintf(db->errstr, sizeof(db->errstr),
                 "sqlite3 driver: (%d) %s", 
                 sqlite3_errcode((sqlite3*) db->specific),
                 sqlite3_errmsg((sqlite3*) db->specific));
    UdmSetErrorCode(db, 1);
    return UDM_ERROR;
  }
  
  for (dbrc= sqlite3_step(pStmt) ; ; dbrc= sqlite3_step(pStmt))
  {
    int i;
    switch (dbrc)
    {
      case SQLITE_DONE:
        goto fin;

      case SQLITE_ROW:
        if (!columns)
        {
          columns= sqlite3_column_count(pStmt);
          res->nCols= columns;
          res->Fields=(UDM_SQLFIELD*)UdmMalloc(res->nCols*sizeof(UDM_SQLFIELD));
          bzero(res->Fields,res->nCols*sizeof(UDM_SQLFIELD));
          for (i=0 ; i < (int) res->nCols; i++)
          {
            const char *name= sqlite3_column_name(pStmt, i);
            res->Fields[i].sqlname= (char*)UdmStrdup(name);
            res->Fields[i].sqllen= 0;
            res->Fields[i].sqltype= 0;
          }
        }
        res->nRows++;
        res->Items=(UDM_STR*)UdmRealloc(res->Items,res->nRows*res->nCols*sizeof(UDM_STR));

        for (i= 0 ; i < columns; i++)
        {
          const unsigned char *value;
          int type, bytes;
          size_t offs= (res->nRows-1)*res->nCols+(size_t)i;
          UDM_STR *Item= &res->Items[offs];
          type= sqlite3_column_type(pStmt, i);
          if (type == SQLITE_BLOB)
            value= (const unsigned char*) sqlite3_column_blob(pStmt, i);
          else
            value= sqlite3_column_text(pStmt, i);
          bytes= sqlite3_column_bytes(pStmt, i);
          Item->length= bytes;
          Item->str= (char*)UdmMalloc(bytes + 1);
          if (bytes)
            memcpy(Item->str, value, bytes);
          res->Items[offs].str[bytes]= '\0';
        }
        break;

      case SQLITE_ERROR:
        sqlite3_finalize(pStmt);
        udm_snprintf(db->errstr, sizeof(db->errstr),
                     "sqlite3 driver: (%d) %s",
                     sqlite3_errcode(UdmSQLite3Conn(db)),
                     sqlite3_errmsg(UdmSQLite3Conn(db)));
        if (!strstr(db->errstr,"unique"))
        {
          UdmSetErrorCode(db, 1);
          return UDM_ERROR;
        }
        return UDM_OK;
        break;
    
      case SQLITE_MISUSE:
      case SQLITE_BUSY:
      default:
        udm_snprintf(db->errstr, sizeof(db->errstr),
                     "sqlite3_step() returned MISUSE or BUSY");
        UdmSetErrorCode(db, 1);
        sqlite3_finalize(pStmt);
        return UDM_ERROR;
        break;
      }
    }

fin:
  if ((dbrc= sqlite3_finalize(pStmt)))
    return UDM_ERROR;

  return UDM_OK;
}


static udm_rc_t
UdmSQLite3Close(UDM_SQL *db)
{
  sqlite3_close(UdmSQLite3Conn(db));
  db->specific = NULL;
  return UDM_OK;
}


static udm_rc_t
UdmSQLite3Begin(UDM_SQL *db)
{
  return UdmSQLite3Query(db,NULL,"BEGIN TRANSACTION");
}


static udm_rc_t
UdmSQLite3Commit(UDM_SQL *db)
{
  return UdmSQLite3Query(db,NULL,"COMMIT");
}


static udm_rc_t
UdmSQLite3RenameTable(UDM_SQL *db, const char *from, const char *to)
{
  char buf[256];
  udm_snprintf(buf, sizeof(buf), "ALTER TABLE %s RENAME TO %s", from, to);
  return UdmSQLExecDirect(db, NULL, buf);
}


static udm_rc_t
UdmSQLite3CopyStructure(UDM_SQL *db, const char *from, const char *to)
{
  char buf[256];
  /*
    CREATE TABLE t1 (LIKE t2); -- works, but creates a table with 
    one column name "LIKE".
  */
  udm_snprintf(buf, sizeof(buf),
               "CREATE TABLE %s AS SELECT * FROM %s WHERE 0=1", to, from);
  return UdmSQLExecDirect(db, NULL, buf);
}


const UDM_SQLDB_HANDLER udm_sqldb_sqlite3_handler =
{
  UdmSQLEscStrGeneric,
  UdmSQLite3Query,
  UdmSQLite3Connect,
  UdmSQLite3Close,
  UdmSQLite3Begin,
  UdmSQLite3Commit,
  UdmSQLPrepareGeneric,
  UdmSQLBindGeneric,
  UdmSQLExecGeneric,
  UdmSQLStmtFreeGeneric,
  UdmSQLFetchRowSimple,
  UdmSQLStoreResultSimple,
  UdmSQLFreeResultSimple,
  UdmSQLite3Query,
  UdmSQLite3RenameTable,
  UdmSQLite3CopyStructure,
  UdmSQLLockOrBeginGeneric,
  UdmSQLUnlockOrCommitGeneric,
};

#endif  /* HAVE_SQLITE3 */

#endif  /* defined(HAVE_SQLITE) || defined(HAVE_SQLITE3) */
