/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include "udm_common.h"
#include "udm_vars.h"
#include "udm_log.h"
#include "udm_sgml.h"

#include "udm_prog.h"


typedef struct
{
  UDM_VALUE generic;
  UDM_SECTION Section;
} UDM_VALUE_SECTION;


typedef struct
{
  UDM_VAR_HEADER Header;
  UDM_VALUE_SECTION Section;
} UDM_VAR_SECTION;


static udm_rc_t
UdmSectionCopyVal(UDM_SECTION *Section,
                  const char *val, size_t length)
{
  if (val == NULL)
  {
    UdmSectionSetNullStr(Section);
    return UDM_OK;
  }
  UdmDSTRReset(&Section->Value);
  UdmDSTRAppend(&Section->Value, val, length);
  return UDM_OK;
}


static udm_rc_t
UdmSectionSetConv(UDM_SECTION *Section, UDM_CONV *conv, int cnvflags,
                  const char *src, size_t srclen)
{
  UdmDSTRReset(&Section->Value);
  return UdmDSTRAppendConv(&Section->Value, conv, src, srclen, cnvflags);
}


static udm_rc_t
UdmSectionConvertCharset(UDM_SECTION *Section, UDM_CONV *conv, int flags)
{
  if (!(Section->Param.flags & UDM_VARFLAG_RAW))
  {
    udm_rc_t rc;
    UDM_DSTR dstr;
    UdmDSTRInit(&dstr, 256);
    rc= UdmDSTRAppendConv(&dstr, conv, UdmSectionPtr(Section),
                                       UdmSectionLength(Section), flags);
    UdmDSTRFree(&Section->Value);
    Section->Value= dstr;
    return rc;
  }
  return UDM_OK;
}


static size_t
UdmSectionRemainingSpace(const UDM_SECTION *Section, size_t maxlen)
{
  return UDM_TEST(Section->Param.flags & UDM_VARFLAG_READONLY) ||
         maxlen < UdmSectionLength(Section) ? 0 :
         maxlen - UdmSectionLength(Section);
}


static udm_rc_t
UdmSectionAppendConvWithoutLimit(UDM_SECTION *Section,
                                 UDM_CONV *cnv, int cnvflags,
                                 const char *src, size_t srclen)
{
  UdmDSTRAppendConv(&Section->Value, cnv, src, srclen, cnvflags);
  return UDM_OK;
}


/*
  Append with conversion, but not more than maxlen.
  Mark the variable as readonly if it would get longer than maxlen
  if converted without limits.
*/
static udm_rc_t
UdmSectionAppendConvWithLimit(UDM_SECTION *Section,
                              size_t maxlen,
                              UDM_CONV *cnv, int cnvflags,
                              const char *src, size_t length)
{
  int cnvres;
  size_t dstlen= UdmSectionRemainingSpace(Section, maxlen);
  UDM_DSTR *dstr= &Section->Value;
  if (UdmDSTRReserve(dstr, maxlen))
    return UDM_ERROR;
  cnvres= UdmConv(cnv,
                  dstr->Val.str + dstr->Val.length, dstlen,
                  src, length, cnvflags);
  dstr->Val.length+= cnvres;
  dstr->Val.str[dstr->Val.length]= '\0'; 
  if (cnvres < 0) /* TODO34: is this correct? */
    Section->Param.flags= (udm_var_flag_t) (Section->Param.flags | UDM_VARFLAG_READONLY);
  return UDM_OK;
}


/**************************************************************/

static udm_rc_t
UdmValueSimpleDump(const char *Data, const char *name, FILE *f)
{
  const UDM_SECTION *Section= (const UDM_SECTION *) Data;
  char str[128];
  udm_snprintf(str, sizeof(str),
               "Section %s %d", /*TODO34: add maxlen */
               name, Section->Param.secno);
  fprintf(f, "%s\n", str);
  return UDM_OK;
}


static udm_rc_t
UdmValueSimpleConstructor(char *Data,
                          const UDM_VALUE **args, size_t nargs)
{
  UDM_SECTION *Section= (UDM_SECTION *) Data;
  UDM_ASSERT(nargs == 0);
  UDM_BZERO(Data, sizeof(UDM_SECTION));
  UdmDSTRInit(&Section->Value, 64);
  return UDM_OK;
}


static void
UdmValueSimpleDestructor(char *Data)
{
  UDM_SECTION *Section= (UDM_SECTION *) Data;
  UdmDSTRFree(&Section->Value);
}


static size_t
UdmValueSimpleVarSize(void)
{
  return sizeof(UDM_VAR_SECTION);
}


static size_t
UdmValueSimpleValueSize(void)
{
  return sizeof(UDM_VALUE_SECTION);
}


static size_t
UdmValueSimpleDataOffset(void)
{
  return offsetof(UDM_VALUE_SECTION, Section);
}


static size_t
UdmValueSimpleDataSize(void)
{
  return sizeof(UDM_SECTION);
}


static size_t
UdmValueSimpleDataAlignment(void)
{
  return sizeof(size_t);
}



static udm_var_flag_t
UdmValueSimpleFlags(const char *Data)
{
  const UDM_SECTION *Section= (const UDM_SECTION *) Data;
  return Section->Param.flags;
}


static udm_secno_t
UdmValueSimpleSecno(const char *Data)
{
  const UDM_SECTION *Section= (const UDM_SECTION *) Data;
  return Section->Param.secno;
}


static size_t
UdmValueSimpleMemUsed(const char *Data)
{
  const UDM_SECTION *Section= (const UDM_SECTION *) Data;
  return Section->Value.size_alloced;
}


static void
UdmValueSimpleGetConstStr(const char *Data, UDM_CONST_STR *Val)
{
  const UDM_SECTION *Sec= (const UDM_SECTION *) Data;
  UdmDSTRToConstStr(&Sec->Value, Val);
}


static void
UdmValueSimpleGetBool(const char *Data, udm_bool_t *val)
{
  const UDM_SECTION *Sec= (const UDM_SECTION *) Data;
  const UDM_STR *str= &Sec->Value.Val;
  *val= str->str &&
         ((str->length == 3 && UDM_TEST(!strncasecmp(str->str, "yes", 3))) ||
          atoi(str->str) == 1);
}


static void
UdmValueSimpleGetInt(const char *Data, int *val)
{
  const UDM_SECTION *Sec= (const UDM_SECTION *) Data;
  const UDM_STR *str= &Sec->Value.Val;
  *val= str->str ? atoi(str->str) : 0;
}


static void
UdmValueSimpleGetUInt(const char *Data, unsigned int *val)
{
  const UDM_SECTION *Sec= (const UDM_SECTION *) Data;
  const UDM_STR *str= &Sec->Value.Val;
  *val= str->str ? strtoul(str->str, (char**) NULL, 10) : 0;
}


static void
UdmValueSimpleGetDouble(const char *Data, double *val)
{
  const UDM_SECTION *Sec= (const UDM_SECTION *) Data;
  const UDM_STR *str= &Sec->Value.Val;
  *val= str->str ? strtod(str->str, (char**) NULL) : 0;
}


static udm_rc_t
UdmValueSimpleSetStrn(char *Data, const char *str, size_t length)
{
  UDM_SECTION *Sec= (UDM_SECTION *) Data;
  UDM_ASSERT(str || !length);
  return UdmSectionCopyVal(Sec, str, length);
}


static udm_rc_t
UdmValueSimpleSetInt(char *Data, int value)
{
  char str[32];
  size_t length= udm_snprintf(str, sizeof(str), "%d", value);
  return UdmValueSimpleSetStrn(Data, str, length);
}


static udm_rc_t
UdmValueSimpleSetDouble(char *Data, double value)
{
  char str[128];
  size_t length= udm_snprintf(str, sizeof(str), "%f", value);
  return UdmValueSimpleSetStrn(Data, str, length);
  return UDM_OK;
}


static udm_rc_t
UdmValueSimpleCopy(char *D, const char *S)
{
  UDM_SECTION *DSec= (UDM_SECTION *) D;
  const UDM_SECTION *SSec= (const UDM_SECTION *) S;
  DSec->Param= SSec->Param;
  return UdmSectionCopyVal(DSec, UdmDSTRPtr(&SSec->Value),
                                 UdmDSTRLength(&SSec->Value));
}


static udm_rc_t
UdmValueSimpleSetConv(char *Data, UDM_CONV *conv, int cnvflags,
                      const char *src, size_t srclen)
{
  UDM_SECTION *Section= (UDM_SECTION *) Data;
  return UdmSectionSetConv(Section, conv, cnvflags, src, srclen);
}


static udm_rc_t
UdmValueSimpleConvertCharset(char *Data, UDM_CONV *conv, int flags)
{
  UDM_SECTION *Section= (UDM_SECTION *) Data;
  return UdmSectionConvertCharset(Section, conv, flags);
}


static udm_rc_t
UdmValueSimpleAppendStrn(char *Data, const char *str, size_t length)
{
  UDM_SECTION *Section= (UDM_SECTION *) Data;
  return UdmSectionAppendStrn(Section, str, length);
}


static udm_rc_t
UdmValueSimpleAppendConv(char *Data,
                         size_t maxlen,
                         UDM_CONV *cnv, int cnvflags,
                         const char *src, size_t srclen)
{
  UDM_SECTION *Section= (UDM_SECTION *) Data;
  return maxlen ?
         UdmSectionAppendConvWithLimit(Section, maxlen, cnv, cnvflags, src, srclen) :
         UdmSectionAppendConvWithoutLimit(Section, cnv, cnvflags, src, srclen);
}



static udm_rc_t
UdmValueSimplePrintToFile(const char *Data, FILE *f)
{
  const UDM_SECTION *Sec= (const UDM_SECTION *) Data;
  fwrite(UdmDSTRPtr(&Sec->Value), UdmDSTRLength(&Sec->Value), 1,f);
  return UDM_OK;
}


static udm_rc_t
UdmValueSimplePrintToDSTR(const char *Data, UDM_DSTR *dstr)
{
  const UDM_SECTION *Sec= (const UDM_SECTION *) Data;
  return UdmDSTRAppend(dstr, UdmDSTRPtr(&Sec->Value), UdmDSTRLength(&Sec->Value));
}


static void
UdmValueSimpleMethodClear(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data; /*this*/
  UdmDSTRReset(&Section->Value);
}


static void
UdmValueSimpleMethodLength(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data; /*this*/
  UDM_FUNC_RETURN_INT(state, (int) UdmDSTRLength(&Section->Value));
}


static void
UdmValueSimpleMethod_stoi(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data; /*this*/
  UDM_FUNC_RETURN_INT(state, (int) atoi(UdmDSTRPtr(&Section->Value)));
}


static void
UdmValueSimpleMethod_stod(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data; /*this*/
  UDM_FUNC_RETURN_DOUBLE(state, atof(UdmDSTRPtr(&Section->Value)));
}


static void
UdmValueSimpleMethodCompare(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section1= (UDM_SECTION *) state->vreg[0].Data; /*this*/
  UDM_SECTION *Section2= (UDM_SECTION *) state->vreg[1].Data;
  size_t minlen= UDM_MIN(Section1->Value.Val.length, Section2->Value.Val.length);
  if (!UDM_FUNC_RETURN_INT(state, memcmp(Section1->Value.Val.str,
                                         Section2->Value.Val.str, minlen)))
  {
    if (Section1->Value.Val.length < Section2->Value.Val.length)
      UDM_FUNC_RETURN_INT(state, -1);
    else if (Section1->Value.Val.length > Section2->Value.Val.length)
      UDM_FUNC_RETURN_INT(state, 1);
  }
}


static void
UdmValueSimpleMethod_find(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section1= (UDM_SECTION *) state->vreg[0].Data; /*this*/
  UDM_SECTION *Section2= (UDM_SECTION *) state->vreg[1].Data;
  const char *pos= udm_memmem(UdmDSTRPtr(&Section1->Value),
                              UdmDSTRLength(&Section1->Value),
                              UdmDSTRPtr(&Section2->Value),
                              UdmDSTRLength(&Section2->Value));
  UDM_FUNC_RETURN_INT(state,
                      pos ? pos - UdmDSTRPtr(&Section1->Value) : 0x7FFFFFFF);
}


/**
  TODO: return address of the left operand.
*/
static void
UdmValueSimpleOperatorCopy(UDM_PROG_EXECUTOR_STATE *state)
{
  char *dst= state->vreg[0].Data;
  const char *src= state->vreg[1].Data;
  if (dst != src)
    state->fatal_error= UdmValueSimpleCopy(dst, src);
}


static void
UdmValueSimpleMethodLeft(UDM_PROG_EXECUTOR_STATE *state)
{
  size_t length;
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  length= UDM_MIN(UdmDSTRLength(&Section->Value), 1);
  state->fatal_error= UdmDSTRAppend(&ReturnValue->Value, UdmDSTRPtr(&Section->Value), length);
}


static void
UdmValueSimpleMethodSubstr(UDM_PROG_EXECUTOR_STATE *state)
{
  size_t length;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  int pos= ((int *) state->vreg[1].Data)[0];
  if (pos < 0 || (size_t) pos > UdmDSTRLength(&Section->Value))
    return;
  length= UdmDSTRLength(&Section->Value) - pos;
  state->fatal_error= UdmDSTRAppend(&ReturnValue->Value,
                                    UdmDSTRPtr(&Section->Value) + pos,
                                    length);
}


static void
UdmValueSimpleMethodSubstr2(UDM_PROG_EXECUTOR_STATE *state)
{
  size_t length;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  int pos= ((int *) state->vreg[1].Data)[0];
  int npos= ((int *) state->vreg[2].Data)[0];
  
  if (pos < 0 || (size_t) pos > UdmDSTRLength(&Section->Value) ||
      npos < 0)
    return;
  length= UdmDSTRLength(&Section->Value) - pos;
  if (length > (size_t) npos)
    length= (size_t) npos;
  state->fatal_error= UdmDSTRAppend(&ReturnValue->Value,
                                    UdmDSTRPtr(&Section->Value) + pos,
                                    length);
}


static void
UdmValueSimpleMethodPCase(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UdmDSTRAppend(&ReturnValue->Value,
                UdmDSTRPtr(&Section->Value), 
                UdmDSTRLength(&Section->Value));
  UdmDSTRPCase(&ReturnValue->Value);
}


static void
UdmValueSimpleMethodLower(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UdmDSTRAppend(&ReturnValue->Value,
                UdmDSTRPtr(&Section->Value),
                UdmDSTRLength(&Section->Value));
  UdmDSTRLCase(&ReturnValue->Value);
}


static void
UdmValueSimpleMethodUpper(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  UdmDSTRAppend(&ReturnValue->Value,
                UdmDSTRPtr(&Section->Value),
                UdmDSTRLength(&Section->Value));
  UdmDSTRUCase(&ReturnValue->Value);
}


static void
UdmValueSimpleMethodURLDecode(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UdmDSTRAppendURLDecode(&ReturnValue->Value,
                                             UdmDSTRPtr(&Section->Value),
                                             UdmDSTRLength(&Section->Value));
}


static void
UdmValueSimpleMethodHTMLEncode(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UdmDSTRAppendHTMLEncode(&ReturnValue->Value,
                                              UdmDSTRPtr(&Section->Value),
                                              UdmDSTRLength(&Section->Value));
}


static void
UdmValueSimpleMethodURLEncode(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UdmDSTRAppendURLEncode(&ReturnValue->Value,
                                             UdmDSTRPtr(&Section->Value),
                                             UdmDSTRLength(&Section->Value));
}


static void
UdmValueSimpleMethodBase64Encode(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *ReturnValue= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UdmDSTRAppendBase64Encode(&ReturnValue->Value,
                                                UdmDSTRPtr(&Section->Value),
                                                UdmDSTRLength(&Section->Value));
}


static void
UdmValueSimpleMethodAppend(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *Section= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *Other= (UDM_SECTION *) state->vreg[1].Data;
  state->fatal_error= UdmDSTRAppend(&Section->Value, UdmDSTRPtr(&Other->Value), UdmDSTRLength(&Other->Value));
  UDM_FUNC_RETURN_INT(state, 0);
}


static void
UdmValueSimpleMethodRegexReplace(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *src= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *pattern= (UDM_SECTION *) state->vreg[1].Data;
  UDM_SECTION *replacement= (UDM_SECTION *) state->vreg[2].Data;
  UDM_SECTION *dst= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UdmDSTRAppendRegexReplace(&dst->Value,
                                                UdmSectionPtr(src),
                                                UdmSectionLength(src),
                                                UdmSectionPtr(pattern),
                                                UdmSectionLength(pattern),
                                                UdmSectionPtr(replacement),
                                                UdmSectionLength(replacement));
}


static void
UdmValueSimpleMethodRegexCut(UDM_PROG_EXECUTOR_STATE *state)
{
  UDM_SECTION *src= (UDM_SECTION *) state->vreg[0].Data;
  UDM_SECTION *pattern= (UDM_SECTION *) state->vreg[1].Data;
  UDM_SECTION *dst= UdmProgExecutorStateSetupReturnValue(state, &UdmValueHandlerSimple);
  state->fatal_error= UdmDSTRAppendRegexCut(&dst->Value,
                                            UdmSectionPtr(src),
                                            UdmSectionLength(src),
                                            UdmSectionPtr(pattern),
                                            UdmSectionLength(pattern));
}


static UDM_VALUE_PROTOTYPE UdmValuePrototypeSimpleThis[]=
{{&UdmValueHandlerSimple}};

static UDM_VALUE_PROTOTYPE UdmValuePrototypeSimpleThisAndOther[]=
{
  {&UdmValueHandlerSimple},
  {&UdmValueHandlerSimple}
};

static UDM_VALUE_PROTOTYPE UdmValuePrototypeSimpleIntInt[]=
{
  {&UdmValueHandlerSimple},
  {&UdmValueHandlerInt},
  {&UdmValueHandlerInt}
};

static UDM_VALUE_PROTOTYPE UdmValuePrototypeSimple3[]=
{
  {&UdmValueHandlerSimple},
  {&UdmValueHandlerSimple},
  {&UdmValueHandlerSimple}
};


static const UDM_FUNCTION StrMethods[]=
{
  {{UDM_CSTR_WITH_LEN("operator=")},
    {{NULL}, 2, UdmValuePrototypeSimpleThisAndOther},
    UdmValueSimpleOperatorCopy},

  {{UDM_CSTR_WITH_LEN("clear")},
    {{NULL/* void*/}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodClear},
  {{UDM_CSTR_WITH_LEN("length")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodLength},
  {{UDM_CSTR_WITH_LEN("stoi")},
    {{&UdmValueHandlerInt}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethod_stoi},
  {{UDM_CSTR_WITH_LEN("stod")},
    {{&UdmValueHandlerDouble}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethod_stod},
  {{UDM_CSTR_WITH_LEN("compare")},
    {{&UdmValueHandlerInt}, 2, UdmValuePrototypeSimpleThisAndOther},
    UdmValueSimpleMethodCompare},
  {{UDM_CSTR_WITH_LEN("find")},
    {{&UdmValueHandlerInt}, 2, UdmValuePrototypeSimpleThisAndOther},
    UdmValueSimpleMethod_find},
  {{UDM_CSTR_WITH_LEN("append")},
    {{&UdmValueHandlerInt}, 2, UdmValuePrototypeSimpleThisAndOther},
    UdmValueSimpleMethodAppend},
  {{UDM_CSTR_WITH_LEN("regex_substr")},
    {{&UdmValueHandlerSimple}, 3, UdmValuePrototypeSimple3},
    UdmValueSimpleMethodRegexReplace},
  {{UDM_CSTR_WITH_LEN("regex_cut")},
    {{&UdmValueHandlerSimple}, 2, UdmValuePrototypeSimple3},
    UdmValueSimpleMethodRegexCut},
  {{UDM_CSTR_WITH_LEN("left")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodLeft},
  {{UDM_CSTR_WITH_LEN("substr")},
    {{&UdmValueHandlerSimple}, 2, UdmValuePrototypeSimpleIntInt},
    UdmValueSimpleMethodSubstr},
  {{UDM_CSTR_WITH_LEN("substr")},
    {{&UdmValueHandlerSimple}, 3, UdmValuePrototypeSimpleIntInt},
    UdmValueSimpleMethodSubstr2},
  {{UDM_CSTR_WITH_LEN("lower")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodLower},
  {{UDM_CSTR_WITH_LEN("upper")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodUpper},
  {{UDM_CSTR_WITH_LEN("pcase")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodPCase},
  {{UDM_CSTR_WITH_LEN("urldecode")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodURLDecode},
  {{UDM_CSTR_WITH_LEN("htmlencode")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodHTMLEncode},
  {{UDM_CSTR_WITH_LEN("urlencode")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodURLEncode},
  {{UDM_CSTR_WITH_LEN("base64encode")},
    {{&UdmValueHandlerSimple}, 1, UdmValuePrototypeSimpleThis},
    UdmValueSimpleMethodBase64Encode},
  {{0,0}, {{NULL}, 0, NULL}, NULL}
};


UDM_VALUE_HANDLER UdmValueHandlerSimple=
{
  UDM_VALUE_HANDLER_TYPE_STR,
  UDM_VALUE_DATA_TYPE_STR,
  UDM_VALUE_DATA_TYPE_STR,
  "string",
  UdmValueSimpleConstructor,
  UdmValueSimpleDestructor,
  UdmValueSimpleCopy,
  UdmValueSimpleVarSize,
  UdmValueSimpleValueSize,
  UdmValueSimpleDataOffset,
  UdmValueSimpleDataSize,
  UdmValueSimpleDataAlignment,
  UdmValueSimpleDump,
  UdmValueSimpleFlags,
  UdmValueSimpleSecno,
  UdmValueSimpleMemUsed,
  UdmValueSimpleGetConstStr,
  UdmValueSimpleGetBool,
  UdmValueSimpleGetInt,
  UdmValueSimpleGetUInt,
  UdmValueSimpleGetDouble,
  UdmValueSimpleSetInt,
  UdmValueSimpleSetDouble,
  UdmValueSimpleSetStrn,
  UdmValueSimplePrintToFile,
  UdmValueSimplePrintToDSTR,
  UdmValueSimpleConvertCharset,
  UdmValueSimpleSetConv,
  UdmValueSimpleAppendStrn,
  UdmValueSimpleAppendConv,
  StrMethods,              /* Methods */
};
