/* Copyright (C) 2000-2015 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_unicode.h"
#include "udm_word.h"
#include "udm_searchtool.h"


#define WSIZE    1024
#define BSIZE    10

/*
  offs=0 means normal word
  offs=1 means seclen marker
*/
udm_rc_t
UdmWordListAddEx(UDM_WORDLIST *Words,
                 const char *word,
                 size_t secno,
                 size_t wordpos,
                 size_t offs)
{
  UDM_WORD *W;
  if (wordpos > 0x1FFFFF)
    return(UDM_OK);
  /* Realloc memory when required  */
  if(Words->nwords >= Words->mwords)
  {
    Words->mwords+= WSIZE;
    Words->Word= (UDM_WORD *) UdmRealloc(Words->Word, Words->mwords * sizeof(UDM_WORD));
  }
  /* Add new word */
  W= &Words->Word[Words->nwords];
  W->word= (char*) UdmStrdup(word);
  W->coord.pos= wordpos /*+ offs*/;
  W->coord.secno= secno;
  W->hash= 0;
  W->seclen_marker= offs;
  Words->nwords++;
  return UDM_OK;
}

/* This function adds a normalized word form(s) into list using Ispell */
udm_rc_t
UdmWordListAdd(UDM_WORDLIST *List, char *word, int secno)
{
  return UdmWordListAddEx(List, word, secno, ++List->wordpos[secno],0);
}


void
UdmWordListReset(UDM_WORDLIST *List)
{
  size_t i;
  for (i= 0; i < List->nwords; i++)
    UDM_FREE(List->Word[i].word);
  List->nwords= 0;
}


void
UdmWordListFree(UDM_WORDLIST * List)
{
  size_t i;
  for(i=0;i<List->nwords;i++)
    UDM_FREE(List->Word[i].word);
  List->nwords=0;
  UDM_FREE(List->Word);
}


void
UdmWordListListInit(UDM_WORDLISTLIST *WL)
{
  bzero((void*) WL, sizeof(*WL));
}


void
UdmWordListListFree(UDM_WORDLISTLIST *WL)
{
  size_t i;
  for (i= 0; i < 255; i ++)
  {
    UdmWordListFree(&WL->Item[i]);
  }
}


void
UdmWordListListReset(UDM_WORDLISTLIST *WL)
{
  size_t i;
  for (i= 0; i < 255; i ++)
  {
    UdmWordListReset(&WL->Item[i]);
  }
}

/***************** ConstWordList *************************/
void
UdmConstWordListInit(UDM_CONSTWORDLIST *L)
{
  bzero((void*)L, sizeof(*L));
}


void
UdmConstWordListFree(UDM_CONSTWORDLIST *L)
{
  UdmFree(L->Item);
}


udm_rc_t
UdmConstWordListAdd(UDM_CONSTWORDLIST *L, UDM_CONSTWORD *W)
{
  if (L->nitems >= L->mitems)
  {
    L->mitems+= 8*1024;
    L->Item= (UDM_CONSTWORD *) UdmRealloc(L->Item, L->mitems * sizeof(UDM_CONSTWORD));
    if (!L->Item)
    {
      L->mitems= L->nitems= 0;
      return UDM_ERROR;
    }
  }
  L->Item[L->nitems]= W[0];
  L->nitems++;
  return UDM_OK;
}


static int cwcmp_sort(UDM_CONSTWORD *w1, UDM_CONSTWORD *w2)
{
  /*register int _;*/
  if (w1->crc != w2->crc)
    return w1->crc < w2->crc ? -1 : 1;
/*  if (w1->length != w2->length)
    return w1->length < w2->length ? -1 : 1;*/
  /* TODO34: UdmStrCaseCmp2 */
/*  if ((_= memcmp(w1->word.str, w2->word.str, w1->word.length)))
    return _;*/
  if (w1->coord.secno != w2->coord.secno)
    return w1->coord.secno < w2->coord.secno ? -1 : 1;
  if (w1->coord.pos != w2->coord.pos)
    return w1->coord.pos < w2->coord.pos ? -1 : 1;
  return 0;
}


void
UdmConstWordListSort(UDM_CONSTWORDLIST *WL)
{
  if (WL->nitems)
    UdmSort(WL->Item, WL->nitems, sizeof(UDM_CONSTWORD), (udm_qsort_cmp) cwcmp_sort);
}


udm_rc_t
UdmConstWordListAddString(UDM_WORD_SCANNER *scanner, int cnvflags,
                          UDM_CONSTWORDLIST *CWL,
                          udm_secno_t secno, const char *src, size_t srclen)
{
  udm_rc_t rc= UDM_OK;
  int len;
  int (*getword)(UDM_WORD_SCANNER *, UDM_WORD_SCANNER_TOKEN *);
  UDM_WORD_SCANNER_TOKEN word;

  /*fprintf(stderr, "===========cs=%s\n[%d:fl=%d]%.*s\n", cs->name, secno, cnvflags, (int) srclen, src);*/
  getword= scanner->cs->cset->getword;
  UDM_ASSERT(getword != NULL);

  UdmWordScannerSetSource(scanner, cnvflags, src, srclen);

  for (len= getword(scanner, &word); len; len= getword(scanner, &word))
  {
    UDM_CONSTWORD W;
    if (len > 255) /* TODO34: limit in UDM_CONSTWORD */
    {
      CWL->wordpos[secno]++;
      continue;
    }
    W.str= word.str;
    W.length= len;
    W.crc= word.crc;
    W.coord.pos= ++(CWL->wordpos[secno]); /* TODO34: check overflow */         
    /*fprintf(stderr, "[%d:%d] %08X '%.*s'\n", secno, W.pos, W.crc, (int) W.length, W.str);*/
    W.coord.secno= secno;
    if (UDM_OK != (rc= UdmConstWordListAdd(CWL, &W)))
      break;
  }
  return rc;
}


/*********************************************************/

void UdmWideWordParamInit(UDM_WIDEWORD_PARAM *P)
{
  bzero((void*) P, sizeof(UDM_WIDEWORD_PARAM));
  P->user_weight= UDM_DEFAULT_USER_WORD_WEIGHT;
}


void UdmWideWordInit(UDM_WIDEWORD *W)
{
  bzero((void*)W, sizeof(UDM_WIDEWORD));
  W->Param.user_weight= UDM_DEFAULT_USER_WORD_WEIGHT;
}


void UdmWideWordFree(UDM_WIDEWORD *W)
{
  UDM_FREE(W->Word.str);
}


void
UdmWideWordParamCopySpell(UDM_WIDEWORD_PARAM *Dst, const UDM_WIDEWORD_PARAM *Src)
{
  Dst->order=  Src->order;
  Dst->phrpos= Src->phrpos;
  Dst->phrlen= Src->phrlen;
  Dst->origin= UDM_WORD_ORIGIN_SPELL;
}


void
UdmWideWordParamCopySynonym(UDM_WIDEWORD_PARAM *Dst, const UDM_WIDEWORD_PARAM *Src)
{
  Dst->order=  Src->order;
  Dst->phrpos= Src->phrpos;
  Dst->phrlen= Src->phrlen;
  Dst->match_mode=  Src->match_mode;
  Dst->origin= UDM_WORD_ORIGIN_SYNONYM;
}


static void
UdmWideWordCopyParam(UDM_WIDEWORD_PARAM *Dst, const UDM_WIDEWORD_PARAM *Src)
{
  UDM_ASSERT(Src->origin >= UDM_WORD_ORIGIN_QUERY &&
             Src->origin <= UDM_WORD_ORIGIN_COLLATION);
  /* TODO34: Dst->Param= Src->Param */
  Dst->order=  Src->order;
  Dst->order_extra_width= Src->order_extra_width;
  Dst->count=  Src->count;
  /* doccount? */
  Dst->origin= Src->origin;
  /* weight? */
  Dst->user_weight= Src->user_weight;
  Dst->match_mode=  Src->match_mode;
  Dst->secno=  Src->secno;
  Dst->phrpos= Src->phrpos;
  Dst->phrlen= Src->phrlen;
  Dst->phrwidth= Src->phrwidth;
}


static void
UdmWideWordCopyWord(UDM_WIDEWORD *Dst, const char *str, size_t length)
{
  Dst->Word.length= length;
  Dst->Word.str= str ? UdmStrdup(str) : NULL;
}


static void
UdmWideWordCopy(UDM_WIDEWORD *Dst, UDM_WIDEWORD *Src)
{
  UdmWideWordCopyParam(&Dst->Param, &Src->Param);
  UdmWideWordCopyWord(Dst, Src->Word.str, Src->Word.length);
}


/*
  Replace the last three words in the list to range designator
*/
udm_rc_t
UdmWideWordListMakeRange(UDM_WIDEWORDLIST *WWL, int beg, int end)
{
  UDM_WIDEWORD *W= &WWL->Word[WWL->nwords - 3];
  char *word;
  size_t len;
  UDM_ASSERT(WWL->nwords >= 3);
  UDM_ASSERT(WWL->nuniq >= 3);
  len= 1 + W[0].Word.length + 4 + W[2].Word.length + 1;
  if (!(word= (char*) UdmMalloc(len + 1)))
    return UDM_ERROR;
  udm_snprintf(word, len + 1, "%c%s TO %s%c",
               beg, W[0].Word.str, W[2].Word.str, end);
  UdmWideWordFree(&W[0]);
  UdmWideWordFree(&W[1]);
  UdmWideWordFree(&W[2]);
  WWL->nwords-= 2;
  WWL->nuniq-= 2;
  W->Word.str= word;
  W->Word.length= len;
  W->Param.match_mode= UDM_MATCH_RANGE;
  return UDM_OK;
}



UDM_WORDLIST * UdmWordListInit(UDM_WORDLIST * List)
{
  bzero((void*)List, sizeof(*List));
  return(List);
}


UDM_WIDEWORDLIST * UdmWideWordListInit(UDM_WIDEWORDLIST * List)
{
  bzero((void*)List, sizeof(*List));
  return(List);
}

static size_t
UdmWideWordListAddInternal(UDM_WIDEWORDLIST * List,
                           UDM_WIDEWORD_PARAM *Param,
                           UDM_CONST_STR w,
                           int for_stat)
{
  size_t i;

  for (i = 0; i < List->nwords; i++)
  {
    UDM_WIDEWORD *ListWord= &List->Word[i];
    if (ListWord->Word.length == w.length &&
        strcmp(ListWord->Word.str, w.str) == 0)
    {
      if (for_stat)
      {
        ListWord->Param.count+= Param->count;
        return List->nwords;
      }
      else if (ListWord->Param.order == Param->order)
      {
        ListWord->Param.count+= Param->count;
        if (ListWord->Param.phrpos != Param->phrpos ||
            ListWord->Param.phrlen != Param->phrlen)
        {
          ListWord->Param.phrpos= 0; /* No certain in-phrase position */
          ListWord->Param.phrlen= 0;
        }
        switch (Param->origin)
        {
          case UDM_WORD_ORIGIN_QUERY:
            if (ListWord->Param.origin == UDM_WORD_ORIGIN_STOP)
              break;
          case UDM_WORD_ORIGIN_STOP:
            ListWord->Param.origin= Param->origin;
            break;
          case UDM_WORD_ORIGIN_SPELL:
          case UDM_WORD_ORIGIN_SYNONYM:
          case UDM_WORD_ORIGIN_SYNONYM_FINAL:
          case UDM_WORD_ORIGIN_SUGGEST:
          case UDM_WORD_ORIGIN_COLLATION:
            break;
        }
        ListWord->Param.order= Param->order;
        return List->nwords;
      }
    }
  }
  
  /* Realloc memory */
  List->Word= (UDM_WIDEWORD*)UdmRealloc(List->Word,sizeof(*(List->Word))*(List->nwords+1));
  UdmWideWordInit(&List->Word[List->nwords]);
  
  /* Copy data */
  UdmWideWordCopyParam(&List->Word[List->nwords].Param, Param);
  UdmWideWordCopyWord(&List->Word[List->nwords], w.str, w.length);
  
  List->nwords++;
  return(List->nwords);
}


size_t UdmWideWordListAdd(UDM_WIDEWORDLIST * List, UDM_WIDEWORD * Word)
{
  UDM_CONST_STR cstr;
  cstr.str= Word->Word.str;
  cstr.length= Word->Word.length;
  return UdmWideWordListAddInternal(List, &Word->Param, cstr, 0);
}


size_t
UdmWideWordListAddLike(UDM_WIDEWORDLIST *WWList,
                       UDM_WIDEWORD_PARAM *param, const char *word)
{
  UDM_CONST_STR cstr;
  cstr.str= word;
  cstr.length= strlen(word);
  return UdmWideWordListAddInternal(WWList, param, cstr, 0);
}


size_t
UdmWideWordListAddLikeConstStr(UDM_WIDEWORDLIST *WWList,
                               UDM_WIDEWORD_PARAM *param,
                               UDM_CONST_STR str)
{
  return UdmWideWordListAddInternal(WWList, param, str, 0);
}


size_t UdmWideWordListAddForStat(UDM_WIDEWORDLIST * List, UDM_WIDEWORD * Word)
{
  UDM_CONST_STR cstr;
  cstr.str= Word->Word.str;
  cstr.length= Word->Word.length;
  return UdmWideWordListAddInternal(List, &Word->Param, cstr, 1);
}


void UdmWideWordListFree(UDM_WIDEWORDLIST * List)
{
  size_t i;
  for(i=0;i<List->nwords;i++)
    UdmWideWordFree(&List->Word[i]);
  UDM_FREE(List->Word);
  UdmWideWordListInit(List);
}


udm_rc_t
UdmWideWordListCopy(UDM_WIDEWORDLIST *Dst, UDM_WIDEWORDLIST *Src)
{
  size_t i;
  *Dst= *Src;
  Dst->Word= (UDM_WIDEWORD*) UdmMalloc(sizeof(*(Src->Word))*(Src->nwords));
  for (i= 0; i < Src->nwords; i++)
    UdmWideWordCopy(&Dst->Word[i], &Src->Word[i]);
  return UDM_OK;
}


static int wwcmp(const UDM_WIDEWORD *w1, const UDM_WIDEWORD *w2)
{
  int rc;
  if ((rc= strcmp(w1->Word.str, w2->Word.str)))
    return rc;
  return (int) w1->Param.secno - (int) w2->Param.secno;
}


void UdmWideWordListSort(UDM_WIDEWORDLIST *L)
{
  if (L->nwords)
    UdmSort(L->Word, L->nwords, sizeof(UDM_WIDEWORD), (udm_qsort_cmp) wwcmp);
}


static int wlcmp(UDM_WORD *w1, UDM_WORD *w2)
{
  register int _;
  if ((_= strcmp(w1->word, w2->word)))
    return _;
  return (int) w1->coord.secno - (int) w2->coord.secno;
}


udm_rc_t
UdmWordListSaveSectionSize(UDM_WORDLIST *Words)
{
  size_t i= Words->nwords;
  int prev_sec= 0;
  const char *prev_word= "#non-existing";
  if (i)
    UdmSort(Words->Word, i, sizeof(UDM_WORD), (udm_qsort_cmp)wlcmp);
  while (i--)
  {
    /* 
       This assignement must be inside the loop, since Word could be
       realloced by AddOneWord
    */
    UDM_WORD *W= &Words->Word[i];
    if (W->coord.secno != prev_sec || strcmp(W->word, prev_word))
    {
      udm_rc_t rc;
      prev_word= W->word;
      prev_sec= W->coord.secno;
      if (UDM_OK != (rc= UdmWordListAddEx(Words, prev_word, prev_sec,
                                          Words->wordpos[prev_sec] + 1, 1)))
        return rc;
    }
  }
  return UDM_OK;
}
